; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ARITH.ASM						!
; 									!
; Description:	Definitions for playing with integer arithmetic.	!
; 									!
; List of FROTH definitions defined in this module:			!
;	NEGATE		NEGATE						!
;	MIN		MIN						!
;	MAX		MAX						!
;	UMIN		UMIN						!
;	UMAX		UMAX						!
;	+		PLUS						!
;	-		MINUS						!
;	*		STAR						!
;	/		SLASH						!
;	/MOD		SLASHMOD					!
;	MOD		IMOD						!
;	*/		STARSLASH					!
;	*/MOD		STARSLASHMOD					!
;	ABS		IABS						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = NEGATE							=
; =                                                 		=
; = Code for the forth word:	negate				=
; =                     	                            	=
; = Stack:	( n --- -n )					=
; ===============================================================

		align	4
NEGATE		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	1
		neg	PARAM0					; not n
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack1				; CALL   VStack1
@@NoOpt2:	NEG_xESIi8					; NEG    dword ptr [ESI]
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		neg	eax
		COMP_NB

@@exit:		ret

E_NEGATE:
NEGATE		endp


; ===============================================================
; = MIN								=
; =                                                 		=
; = Code for the forth word:	MIN				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ===============================================================

		align	4
MIN		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <n2>
		cmp	PARAM0, eax
		jl	@@ll
		mov	PARAM0, eax
@@ll:		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CMP_xESIi8_EAX					;     CMP    [ESI], EAX
		JL_rel8		2				;     JL     ll
		MOV_xESIi8_EAX					;     MOV    PARAM0, EAX
		jmp	@@exit					; ll:


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	cmp	eax, ebx
		jge	@@TwoNb2
		mov	eax, ebx
@@TwoNb2:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		CMP_xESIi8_i8	0				;     CMP   [ESI], 
		jmp	@@i_done
@@i32:		CMP_xESIi8_i32	0
@@i_done:	JL_rel8		6				;     JL     ll
		MOV_xESIi8_i32					;     MOV    PARAM0, EAX
		mov	eax, ebx
		call	CodeComma	             		;     Code,
								; ll:
@@exit:		ret

E_MIN:
MIN		endp


; ===============================================================
; = MAX								=
; =                                                 		=
; = Code for the forth word:	MAX				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ===============================================================

		align	4
MAX		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <n2>
		cmp	PARAM0, eax
		jg	@@ll
		mov	PARAM0, eax
@@ll:		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CMP_xESIi8_EAX					;     CMP    [ESI], EAX
		JG_rel8		2				;     JG     ll
		MOV_xESIi8_EAX					;     MOV    PARAM0, EAX
		jmp	@@exit					; ll:


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	cmp	eax, ebx
		jle	@@TwoNb2
		mov	eax, ebx
@@TwoNb2:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		CMP_xESIi8_i8	0				;     CMP   [ESI], 
		jmp	@@i_done
@@i32:		CMP_xESIi8_i32	0
@@i_done:	JG_rel8		6				;     JG     ll
		MOV_xESIi8_i32					;     MOV    PARAM0, EAX
		mov	eax, ebx
		call	CodeComma	             		;     Code,
								; ll:
@@exit:		ret

E_MAX:
MAX		endp


; ===============================================================
; = UMIN							=
; =                                                 		=
; = Code for the forth word:	UMIN				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ===============================================================

		align	4
UMIN		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <n2>
		cmp	PARAM0, eax
		jb	@@ll
		mov	PARAM0, eax
@@ll:		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CMP_xESIi8_EAX					;     CMP    [ESI], EAX
		JB_rel8		2				;     JB     ll
		MOV_xESIi8_EAX					;     MOV    PARAM0, EAX
		jmp	@@exit					; ll:


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	cmp	eax, ebx
		jae	@@TwoNb2
		mov	eax, ebx
@@TwoNb2:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		CMP_xESIi8_i8	0				;     CMP   [ESI], 
		jmp	@@i_done
@@i32:		CMP_xESIi8_i32	0
@@i_done:	JB_rel8		6				;     JB     ll
		MOV_xESIi8_i32					;     MOV    PARAM0, EAX
		mov	eax, ebx
		call	CodeComma	             		;     Code,
								; ll:
@@exit:		ret

E_UMIN:
UMIN		endp


; ===============================================================
; = UMAX							=
; =                                                 		=
; = Code for the forth word:	UMAX				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; ==============================================================

		align	4
UMAX		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <n2>
		cmp	PARAM0, eax
		ja	@@ll
		mov	PARAM0, eax
@@ll:		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CMP_xESIi8_EAX					;     CMP    [ESI], EAX
		JG_rel8		2				;     JG     ll
		MOV_xESIi8_EAX					;     MOV    PARAM0, EAX
		jmp	@@exit					; ll:


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	cmp	eax, ebx
		jbe	@@TwoNb2
		mov	eax, ebx
@@TwoNb2:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		CMP_xESIi8_i8	0				;     CMP   [ESI], 
		jmp	@@i_done
@@i32:		CMP_xESIi8_i32	0
@@i_done:	JA_rel8		6				;     JA     ll
		MOV_xESIi8_i32					;     MOV    PARAM0, EAX
		mov	eax, ebx
		call	CodeComma	             		;     Code,
								; ll:
@@exit:		ret

E_UMAX:
UMAX		endp


; ===============================================================
; = PLUS							=
; =                                                 		=
; = Code for the forth word:	+				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1+n2)				=
; ==============================================================

		align	4
PLUS		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; get <n2>
		add	PARAM0, eax				; n1 += n2
		into						; Integer arithmetic overflow test
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack1
@@NoOpt2:	_LODSD_						; LODSD
		ADD_xESIi8_EAX					; ADD    [ESI], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	add	eax, ebx
		sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit2

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, 1
		jne	@@NotInc
@@Inc:		INC_xESIi8
		jmp	@@exit
@@NotInc:	cmp	eax, -1
		jne	@@NotDec
@@Dec:		DEC_xESIi8
		jmp	@@exit
@@NotDec:	cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		ADD_xESIi8_i8	0
		jmp	@@exit
@@i32:		ADD_xESIi8_i32	0

@@exit:		call	isQIArithOverfl				; ?iarithoverflow
		jnc	@@exit2
		_INTO_						; Integer arithmetic overflow test
@@exit2:	ret

E_PLUS:
PLUS		endp


; ===============================================================
; = MINUS							=
; =                                                 		=
; = Code for the forth word:	-				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1-n2)				=
; ==============================================================

		align	4
MINUS		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; get <n2>
		sub	PARAM0, eax				; n1 -= n2
		into						; Integer arithmetic overflow test
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack1
@@NoOpt2:	_LODSD_						; LODSD
		SUB_xESIi8_EAX					; SUB    [ESI], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	sub	eax, ebx
		sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit2

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -1
		jne	@@NotInc
@@Inc:		INC_xESIi8
		jmp	@@exit
@@NotInc:	cmp	eax, 1
		jne	@@NotDec
@@Dec:		DEC_xESIi8
		jmp	@@exit
@@NotDec:	cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		SUB_xESIi8_i8	0
		jmp	@@exit
@@i32:		SUB_xESIi8_i32	0

@@exit:		call	isQIArithOverfl				; ?iarithoverflow
		jnc	@@exit2
		_INTO_						; Integer arithmetic overflow test
@@exit2:	ret

E_MINUS:
MINUS		endp


; ===============================================================
; = STAR							=
; =                                                 		=
; = Code for the forth word:	*				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1*n2)				=
; =                     	                            	=
; = Usage:	Multiply the two values on top stack.		=
; ===============================================================

		align	4
STAR		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; get <n2>
		imul	PARAM0					; n1 * n2
		mov	PARAM0, eax				; Store result
		into						; Integer arithmetic overflow test
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						; LODSD
		IMUL_xESIi8					; IMUL  DWORD PTR [ESI]
		MOV_xESIi8_EAX					; MOV   [ESI], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		cdq						; EDX:EAX
		imul	ebx
		COMP_NB
		jmp	@@exit2

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, 2
		jne	@@Not2
@@2:		SAL_xESIi8_n	0, 1				; SAL DWORD PTR [esi], 1
		jmp	@@exit
@@Not2:		cmp	eax, 4
		jne	@@Not4
@@4:		SAL_xESIi8_n	0, 2				; SAL DWORD PTR [esi], 2
		jmp	@@exit
@@Not4:		MOV_EAX_xESIi8					; MOV EAX, [ESI]
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		IMUL_EAX_i8					; IMUL EAX, imm8
		jmp	@@idone
@@i32:		IMUL_EAX_i32					; IMUL EAX, imm32
@@idone:	MOV_xESIi8_EAX					; MOV [ESI], EAX

@@exit:		call	isQIArithOverfl				; ?iarithoverflow
		jnc	@@exit2
		_INTO_						; Integer arithmetic overflow test
@@exit2:	ret

E_STAR:
STAR		endp


; ===============================================================
; = SLASH							=
; =                                                 		=
; = Code for the forth word:	/				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n1/n2)				=
; =                     	                            	=
; = Usage:	Divide the two values on top stack.		=
; ===============================================================

		align	4
SLASH		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Get <n2>
		xchg	eax, PARAM0				; <n1> <==> <n2>
		cdq						; EDX:EAX
		idiv	PARAM0					; n1 / n2
		mov	PARAM0, eax
		into						; Integer arithmetic overflow test
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						; LODSD
		XCHG_EAX_xESIi8					; XCHG  EAX, PARAM0
		_CDQ_						; CDQ
		IDIV_xESIi8					; IDIV  PARAM0
		MOV_xESIi8_EAX					; MOV   [ESI], EAX
		jmp	@@exit2


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		cdq						; EDX:EAX
		idiv	ebx
		COMP_NB
		jmp	@@exit2

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, 2
		jne	@@Not2
@@2:		SAR_xESIi8_n	0, 1				; SAR DWORD PTR [esi], 1
		jmp	@@exit
@@Not2:		cmp	eax, 4
		jne	@@Not4
@@4:		SAR_xESIi8_n	0, 2				; SAR DWORD PTR [esi], 2
		jmp	@@exit
@@Not4:		MOV_EAX_xESIi8					; MOV  EAX, [ESI]
		_CDQ_						; CDQ
		MOV_EBX_i32					; MOV  EBX, i32
		IDIV_EAX_EBX					; IDIV EBX
		MOV_xESIi8_EAX					; MOV [ESI], EAX
		jmp	@@exit2

@@exit:		call	isQIArithOverfl				; ?iarithoverflow
		jnc	@@exit2
		_INTO_						; Integer arithmetic overflow test
@@exit2:	ret

E_SLASH:
SLASH		endp


; ===============================================================
; = SLASHMOD							=
; =                                                 		=
; = Code for the forth word:	/MOD				=
; =                     	                            	=
; = Stack:	( n1 n2 --- remainder quotient )		=
; =                     	                            	=
; = Usage:	Divide the two values on top stack.		=
; ===============================================================

		align	4
SLASHMOD		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the operation
		; ------------------
		
		mov	eax, PARAM1				; Get n1
		cdq						; EDX:EAX
		idiv	PARAM0					; n1 / n2
		mov	PARAM1, edx				; remainder
		mov	PARAM0, eax				; quotient
		

		; Return to caller
		; ----------------
		
		ret

E_SLASHMOD:
SLASHMOD	endp


; ===============================================================
; = IMOD							=
; =                                                 		=
; = Code for the forth word:	MOD				=
; =                     	                            	=
; = Stack:	( n1 n2 --- remainder )				=
; =                     	                            	=
; = Usage:	Divide the two values on top stack.		=
; ===============================================================

		align	4
IMOD		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the operation
		; ------------------
		
		mov	eax, PARAM1				; Get n1
		cdq						; EDX:EAX
		idiv	PARAM0					; EDX:EAX / [esi]
		DISCARD	1
		mov	PARAM0, edx

		; Return to caller
		; ----------------
		
		ret

E_IMOD:
IMOD		endp


; ===============================================================
; = STARSLASH							=
; =                                                 		=
; = Code for the forth word:	*/				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- n1*n2/n3)			=
; ===============================================================

		align	4
STARSLASH	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3

		
		; Perform operation
		; -----------------

		mov	eax, PARAM2				; Get n1
		imul	PARAM1					; EDX:EAX = n1 * n2
		mov	ebx, PARAM0
		DISCARD	2					; drop drop
		idiv	ebx					; EDX:EAX / n3
		mov	PARAM0, eax


		; Return to caller
		; ----------------
		
		ret

E_STARSLASH:
STARSLASH	endp


; ===============================================================
; = STARSLASHMOD						=
; =                                                 		=
; = Code for the forth word:	*/MOD				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- mod n1*n2/n3 )			=
; ===============================================================

		align	4
STARSLASHMOD	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Perform operation
		; -----------------

		mov	eax, PARAM2				; Get n1
		mov	ebx, PARAM0
		DISCARD	1
		imul	PARAM0					; EDX:EAX = n1 * n2
		idiv	ebx					; EDX:EAX / n3
		mov	PARAM0, eax
		mov	PARAM1, edx


		; Return to caller
		; ----------------
		
		ret

E_STARSLASHMOD:
STARSLASHMOD	endp


; ===============================================================
; = IABS							=
; =                                                 		=
; = Code for the forth word:	ABS				=
; =                     	                            	=
; = Stack:	( n1 --- n2 )					=
; ===============================================================

		align	4
IABS		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Make the operation
		; ------------------
		
		cmp	PARAM0, 0
		jge	@@exit
		neg	PARAM0


		; Return to caller
		; ----------------
		
@@exit:		ret

E_IABS:
IABS		endp


code32		ends
		end

