; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	BIO.ASM							!
; 									!
; Description:	Definitions for playing with control structurs.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	EXIT		EXIT						!
;	BEGIN		BEGIN						!
;	AGAIN		AGAIN						!
;	UNTIL		UNTIL						!
;	FIF		IF						!
;	FELSE		ELSE						!
;	FENDIF		FENDIF						!
;	doTimes		rt_times					!
;	doEndTimes	rt_EndTimes					!
;	TIMES		TIMES						!
;	ENDTMES		ENDTIMES					!
;	I J K L M	I J K L M					!
;	doDo		rt_do						!
;	doEndDo		rt_enddo					!
;	doPEndDo	rt_+enddo					!
;	doPUEndDo	rt_+uenddo					!
;	doMake		rt_make						!
;	MAKE		MAKE						!
;	doPlusMake	rt_+make					!
;	PLUSMAKE	+MAKE						!
;	doEndMake	+endmake					!
;	ENDMAKE		ENDMAKE						!
;	DO		DO						!
;	doWhen		rt_when						!
;	WHEN		WHEN						!
;	doNotWhen	rt_notwhen					!
;	NOTWHEN		NOTWHEN						!
;	ENDDO		ENDDO						!
;	PENDDO		+ENDDO						!
;	PUENDDO		+UENDDO						!
;	WHILE		WHILE						!
;	_REPEAT		REPEAT						!
;	doCase		rt_case						!
;	doOf		rt_of						!
;	doEndCase	rt_endcase					!
;	CASE		CASE						!
;	OF		OF						!
;	ENDOF		ENDOF						!
;	TIPOF		TIPOF						!
;	ENDCASE		ENDCASE						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions

		; -----------------------------------
		; Put 'Operand-size prefix' operation
		; -----------------------------------
		
_OPSIZ_OVERR_	MACRO
		StCodOp	066h
		ENDM


; ===============================================================
; = EXIT							=
; =								=
; = Function for the forth word:	EXIT			=
; ===============================================================

		align	4
EXIT		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; Too many EXIT in one definition ?
		; ---------------------------------

		mov	edx, [ExitStackPtr]
		cmp	edx, offset ExitStack0
		jne	@@StackOk
		mov	ah, ERR_TOOMANYEXIT			; "Too many exit into a colon definition"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@StackOk:	

		; Compile 'JMP rel32'
		; ------------------

		; Compile 'JMP rel32'
		JMP_rel32					; Put 'JMP rel32' operation

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax					; Offset will be resolved later
		call	CodeComma				; Code,


		; Save rel16 cell's address
		; -------------------------

		mov	eax, [CDA_SIZE]
		sub	eax, @WordSize				; Return to rel32 address
		sub	edx, @WordSize
		mov	[edx], eax				; HERE
		mov	[ExitStackPtr], edx


		; Return to caller
		; ----------------
		
		ret

E_EXIT:
EXIT		endp


; ===============================================================
; = BEGIN							=
; =								=
; = Function for the forth word:	BEGIN			=
; ===============================================================

		align	4
BEGIN		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Leave the current code dictionnary pointer
		; ------------------------------------------

		mov	edx, [ControlStackPtr]		
		cmp	edx, offset ControlStack0
		jne	@@StackOk
		mov	ah, ERR_TOOMANYCTRL			; "Too many nested control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@StackOk:	mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adress of last Begin/If (do not optimize literal integer number ?)
		sub	edx, @WordSize
		mov	[edx], eax				; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, @WordSize
		mov	dword ptr [edx], 'Begi'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_BEGIN:
BEGIN		endp


; ===============================================================
; = AGAIN							=
; =								=
; = Function for the forth word:	AGAIN			=
; ===============================================================

		align	4
AGAIN		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'Begin' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]		
		mov	eax, [edx]
		add	edx, @WordSize
		cmp	eax, 'Begi'
		je	@@BeginCalled
		
		; No, it's an error
		mov	ah, ERR_BEGINNC				; "'Begin' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; Put 'CALL Branch' offset
		; ------------------------
		
@@BeginCalled:	

		; Put 'JMP rel32' code op
		JMP_rel32
		
		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [edx]				; Address to jump
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 is the size of 'rel32'
		call	CodeComma				; Code,
		add	edx, @WordSize
		mov	[ControlStackPtr], edx


		; Return to caller
		; ----------------
		
		ret

E_AGAIN:
AGAIN		endp


; ===============================================================
; = UNTIL							=
; =								=
; = Function for the forth word:	UNTIL			=
; ===============================================================

		align	4
UNTIL		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; Verify that 'Begin' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]		
		mov	eax, [edx]
		add	edx, @WordSize
		cmp	eax, 'Begi'
		je	@@BeginCalled
		
		; No, it's an error
		mov	ah, ERR_BEGINNC				; "'Begin' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; Put 'CALL ZBranch' offset
		; -------------------------
		
@@BeginCalled:	

		; ?safe
		call	isSafe
		jnc	@@safe_done
		compCALL_rel32	VStack1				; CALL VStack1
@@safe_done:

		; 0branch
		_LODSD_						; LODSD
		OR_EAX_EAX					; OR     EAX, EAX

		; Relative address to jump
		mov	ebx, [edx]				; Address to jump
		add	edx, @WordSize
		mov	[ControlStackPtr], edx

		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, ebx
		sub	eax, 1					; Size of the codop 'JE rel8'
		sub	eax, [CDA_SIZE]				; HERE
		sub	eax, 1					; 1 for REL8
		cmp	eax, -128
		jl	@@LongJe
		cmp	eax, 127
		jg	@@LongJe
		JE_rel8						; Put 'JE rel8'		operation
		call	CodeCComma				; CodeC,
		jmp	@@ShortLongJe
@@LongJe:	JE_rel32					; Put 'JE rel32'	operation
		mov	eax, ebx
		sub	eax, [CDA_SIZE]				; HERE
		sub	eax, 4					; 1 for REL32
		call	CodeComma				; Code,		
@@ShortLongJe:


		; Return to caller
		; ----------------
		
		ret

E_UNTIL:
UNTIL		endp


; ===============================================================
; = FIF								=
; =								=
; = Function for the forth word:	IF			=
; ===============================================================

		align	4
FIF		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Put 'CALL ZBranch' offset (offset will be resolved by 'else' or 'endif'
		; -----------------------------------------------------------------------

		; ?safe
		call	isSafe
		jnc	@@safe_done
		compCALL_rel32	VStack1				; CALL VStack1
@@safe_done:

		; 0branch
		_LODSD_						; LODSD
		OR_EAX_EAX					; OR     EAX, EAX
		JE_rel32					; JE     rel32

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax					; Offset whil be resolved later
		call	CodeComma				; Code,


		; Leave the current code dictionnary pointer
		; ------------------------------------------

		mov	edx, [ControlStackPtr]		
		cmp	edx, offset ControlStack0
		jne	@@StackOk
		mov	ah, ERR_TOOMANYCTRL			; "Too many nested control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@StackOk:	sub	edx, @WordSize
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adress of last Begin/If (do not optimize literal integer number ?)
		sub	eax, @WordSize
		mov	[edx], eax				; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, @WordSize
		mov	dword ptr [edx], 'IfSi'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_FIF:
FIF		endp


; ===============================================================
; = FELSE							=
; =								=
; = Function for the forth word:	ELSE			=
; ===============================================================

		align	4
FELSE		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'If' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]		
		mov	eax, [edx]
		add	edx, @WordSize
		cmp	eax, 'IfSi'
		je	@@IfCalled
		
		; No, it's an error
		mov	ah, ERR_IFNC				; "'If' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; -----------------------------
		; Resolve 'CALL ZBranch' offset
		; -----------------------------
		
@@IfCalled:	

		; ----------------------------------------------------------------------
		; Put 'CALL Branch' offset (offset will be resolved by 'else' or 'endif'
		; ----------------------------------------------------------------------
		
		JMP_rel32					; Put 'JMP rel32' code operation		

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax					; Offset whil be resolved lated
		call	CodeComma				; Code,


		; Put the 32-bits offset into the Code Dictionnary
		; ------------------------------------------------
		
		mov	eax, [CDA_SIZE]				; HERE
		mov	ebx, [edx]				; Relative address to store offset
		sub	eax, ebx				; Address to jump
		sub	eax, 4					; 4: offset 32rel
		add	edx, @WordSize				; DROP
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax


		; Leave the current code dictionnary pointer
		; ------------------------------------------
		
		sub	edx, @WordSize
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, @WordSize
		mov	[edx], eax				; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, @WordSize
		mov	dword ptr [edx], 'SiIf'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_FELSE:
FELSE		endp


; ===============================================================
; = FENDIF							=
; =								=
; = Function for the forth word:	ENDIF			=
; ===============================================================

		align	4
FENDIF		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		enter	4, 0
t_else		equ	byte ptr [ebp - 4]
		call	QCOMPILE
		
		
		; Verify that 'If' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]		
		mov	eax, [edx]
		add	edx, @WordSize
		mov	[t_else], 0
		cmp	eax, 'IfSi'
		je	@@IfCalled
		cmp	eax, 'SiIf'
		je	@@ElseCalled
		
		; No, it's an error
		mov	ah, ERR_IFNC				; "'If' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; Resolve 'CALL ZBranch' offset
		; -----------------------------
		
@@ElseCalled:	mov	[t_else], 1
@@IfCalled:	
		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		mov	ebx, [edx]				; Relative address to store offset
		sub	eax, ebx				; Address to jump
		add	ebx, [CDA_START]			; EBX = address to store offset
		cmp	[t_else], 0
		je	@@long
		cmp	eax, -128 + 1
		jl	@@long
		cmp	eax, 127 + 1
		jg	@@long

@@Short:	dec	eax
		mov	byte ptr [ebx - 1], JMPrel8_CODOP
		mov	byte ptr [ebx], al
		mov	byte ptr [ebx + 1], NOP_CODOP
		mov	byte ptr [ebx + 2], NOP_CODOP
		mov	byte ptr [ebx + 3], NOP_CODOP
		jmp	@@rel_done

@@long:		sub	eax, 4					; 4 for rel32
		mov	[ebx], eax
@@rel_done:	add	edx, @WordSize				; DROP
		mov	[ControlStackPtr], edx


		; Reset pointer to last word compiled
		;	to avoid to transform last CALL word to a JMP word
		; ---------------------------------------------------------

		mov	[LastWrdCompiled], 0


		; Return to caller
		; ----------------

		leave		
		ret

E_FENDIF:
FENDIF		endp


; ===============================================================
; = doTimes							=
; =								=
; = Function for the run-time forth word:	RT_TIMES	=
; ===============================================================

		align	4
doTimes		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; The offset address
		; ------------------
		
		mov	ecx, [esp]


		; Do the branch if top stack is zero
		; ----------------------------------
		
		mov	eax, PARAM0
		DISCARD	1
		cmpz	eax, 0
		jne	@@NotNull


		; -------------
		; Do the branch
		; -------------
		
		add	ecx, [ecx]
		mov	[esp], ecx
		ret						; jmp	@@Exit
		

		; -----------
		; Do the loop
		; -----------

@@NotNull:	

		; Skip the offset: move the IP after the offset
	      	add	ecx, @WordSize				; 4 is the size of a 32-bits number
		mov	[esp], ecx

		; Initialize the loop
		mov	ebx, [CurrActThread]			; Current active thread
		mov	edx, [LoopStackNb + ebx * 4]
		cmp	dl, LOOPSTACKMAXNB
		jl	@@MaxLoopOk

		mov	ah, ERR_TOOMANYLOOPS			; "Too many nexted loops control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@MaxLoopOk:	inc	[LoopStackNb + ebx * 4]
		imul	edx, size loopentry
		add	edx, offset LoopStack
		imul	ebx, LOOPSTACKMAXNB * size loopentry
		add	edx, ebx
		mov	[edx.loop_index], 0			; index <- 0
		mov	[edx.loop_end], eax			; end value


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doTimes:
doTimes		endp


; ===============================================================
; = doEndTimes							=
; =								=
; = Function for the run-time forth word:	(ENDTIMES)	=
; ===============================================================

		align	4
doEndTimes	proc

		; Update the loop index
		mov	eax, [CurrActThread]				; Current active thread
		mov	edx, [LoopStackNb + eax * 4]
		dec	edx
		imul	edx, size loopentry
		add	edx, offset LoopStack
		imul	eax, LOOPSTACKMAXNB * size loopentry
		add	edx, eax
		inc	[edx.loop_index]				; I++
		mov	eax, [edx.loop_index]				; I
		cmp	eax, [edx.loop_end]				; end value
		je	@@LoopEnded
		

		; Loop not ended: do the branch
		; -----------------------------
		
		mov	eax, [esp]
		add	eax, [eax]
		mov	[esp], eax
		ret							; jmp	@@Exit


		; Loop ended: skip the offset
		; ---------------------------

@@LoopEnded:	add	dword ptr [esp], @WordSize			; Skip next cell

		; Update the loop number
		mov	eax, [CurrActThread]				; Current active thread
		dec	[LoopStackNb + eax * 4]


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doEndTimes:
doEndTimes	endp


; ===============================================================
; = TIMES							=
; =								=
; = Function for the forth word:	TIMES			=
; ===============================================================

		align	4
TIMES		proc


		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Put 'CALL doTimes' offset (offset will be resolved later by ENDTIMES
		; --------------------------------------------------------------------
		
		; Put 'SUMMON rt_times' code operation
		compCALL_rel32	doTimes

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax						; Offset whil be resolved lated
		call	CodeComma					; Code,


		; Leave the current code dictionnary pointer
		; ------------------------------------------

		mov	edx, [ControlStackPtr]		
		cmp	edx, offset ControlStack0
		jne	@@StackOk

		mov	ah, ERR_TOOMANYCTRL			; "Too many nested control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@StackOk:	sub	edx, 4
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, 4
		mov	[edx], eax					; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, 4
		mov	dword ptr [edx], 'Time'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_TIMES:
TIMES		endp


; ===============================================================
; = ENDTIMES							=
; =								=
; = Function for the forth word:	ENDTIMES		=
; ===============================================================

		align	4
ENDTIMES	proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'TIMES' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'Time'
		je	@@TimesCalled
		
		; No, it's an error
		mov	ah, ERR_TIMESNC				; "'TIMES' not called beforehand."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

@@TimesCalled:	

		; Put 'CALL doEndTimes' offset (offset will be resolved later by ENDTIMES
		; -----------------------------------------------------------------------
		
		; Put 'SUMMON rt_endtimes' code operation
		compCALL_rel32	doEndTimes

		; Put the know 32-bits jump offset into the Code Dictionnary
		mov	eax, [edx]
		add	eax, 4
		sub	eax, [CDA_SIZE]					; Offset
		call	CodeComma					; Code,


		; Resolve the offset in the 'doTimes' field
		; -----------------------------------------
		
		mov	eax, [CDA_SIZE]
		mov	ebx, [edx]
		sub	eax, ebx
		add	ebx, [CDA_START]
		mov	[ebx], eax
		add	edx, 4						; DROP
		mov	[ControlStackPtr], edx


		; Return to caller
		; ----------------
		
		ret

E_ENDTIMES:
ENDTIMES	endp


; ===============================================================
; = I-J-K-L-M							=
; =								=
; = Function for the forth words:	I-J-K-L-M		=
; ===============================================================

IJKLM		MACRO	name, level

		align	4
name		proc

		mov	ebx, [CurrActThread]			; Current active thread
		mov	eax, [LoopStackNb + ebx * 4]
		cmp	al, level + 1
		jge	@@Ok

		mov	ah, ERR_BADLOOPIND			; "Bad loop indice"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok:
		if	level eq 0
		elseif	level eq 1
			dec	al
		else
			sub	al, level
		endif


		; Push the current loop index
		; ---------------------------

		imul 	ebx, LOOPSTACKMAXNB * size loopentry		
		dec	eax
		imul	eax, size loopentry
		add	eax, offset LoopStack
		add	eax, ebx
		sub	esi, @WordSize
		mov	eax, [eax.loop_index]
		mov	[esi], eax


		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK
		ret

E_&name:
name		endp

		ENDM
	
		IJKLM	I, 0
		IJKLM	J, 1
		IJKLM	K, 2
		IJKLM	L, 3
		IJKLM	M, 4
		
		purge	IJKLM


; ===============================================================
; = doDo							=
; =								=
; = Function for the run-time forth word:	RT_DO		=
; ===============================================================

		align	4
doDo		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -------------------
		; Initialize the loop
		; -------------------
		
		mov	ebx, [CurrActThread]			; Current active thread
		mov	edx, [LoopStackNb + ebx * 4]
		cmp	dl, LOOPSTACKMAXNB
		jl	@@MaxLoopOk

		mov	ah, ERR_TOOMANYLOOPS			; "Too many nexted loops control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@MaxLoopOk:	inc	[LoopStackNb + ebx * 4]
		imul	edx, size loopentry
		add	edx, offset LoopStack
		imul	ebx, LOOPSTACKMAXNB * size loopentry
		add	edx, ebx
		mov	eax, PARAM0				; Start value
		mov	ebx, PARAM1				; End value
		mov	[edx.loop_index], eax			; index <- start value
		mov	[edx.loop_start], eax			; Start value
		mov	[edx.loop_end], ebx			; end value
		DISCARD	2


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doDo:
doDo		endp


; ===============================================================
; = doEndDo							=
; =								=
; = Function for the run-time forth word:	rt_enddo	=
; ===============================================================

		align	4
doEndDo		proc

		; Update the loop index
		; ---------------------
		
		mov	eax, [CurrActThread]			; Current active thread
		mov	edx, [LoopStackNb + eax * 4]
		dec	edx
		imul	edx, size loopentry
		add	edx, offset LoopStack
		imul	eax, LOOPSTACKMAXNB * size loopentry
		add	edx, eax
		inc	[edx.loop_index]			; I++
		mov	eax, [edx.loop_index]			; I
		cmp	eax, [edx.loop_end]			; end value
		jge	@@LoopEnded
		

		; Loop not ended: do the branch
		; -----------------------------

		mov	eax, [esp]
		add	eax, [eax]
		mov	[esp], eax
		ret						; jmp	@@Exit


		; Loop ended: skip the offset
		; ---------------------------

@@LoopEnded:	add	dword ptr [esp], @WordSize		; Skip next cell

		; Update the loop number
		mov	eax, [CurrActThread]			; Current active thread
		dec	[LoopStackNb + eax * 4]


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doEndDo:
doEndDo		endp


; ===============================================================
; = doPEndDo							=
; =								=
; = Function for the run-time forth word:	rt_+enddo	=
; ===============================================================

		align	4
doPEndDo	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Store into DI the offset address
		mov	ecx, [esp]

		; Update the loop index
		mov	edx, [LoopStackNb]
		cmp	dl, LOOPSTACKMAXNB
		jl	@@MaxLoopOk

		mov	ah, ERR_TOOMANYLOOPS				; "Too many nexted loops control structures"
		clr	edx						; No additionnal msg
		clr	ecx						; No line #
		call	doError

@@MaxLoopOk:	dec	edx
		imul	edx, size loopentry
		add	edx, offset LoopStack
		mov	eax, PARAM0					; Pop loop increment
		DISCARD	1						; DROP
		cmpz	eax, 0
		jl	@@IncNeg

		; Loop increment is positive
		add	[edx.loop_index], eax				; I += step
		mov	eax, [edx.loop_index]				; I
		cmp	eax, [edx.loop_end]				; end value
		jge	@@LoopEnded
		jmp	@@LoopNotEnded					; Stop if I >= end value
		
		; Loop increment is negative
@@IncNeg:	add	[edx.loop_index], eax				; I += step
		mov	eax, [edx.loop_index]				; I
		cmp	eax, [edx.loop_end]				; end value
		jle	@@LoopEnded					; Stop if I <= end value
		

		; Loop not ended: do the branch
		; -----------------------------

@@LoopNotEnded:	add	ecx, [ecx]
		mov	[esp], ecx
		jmp	@@Exit


		; Loop ended: skip the offset
		; ---------------------------

@@LoopEnded:	add	ecx, 4						; 4 is the size of a 32-bits number
		mov	[esp], ecx

		; Update the loop number
		mov	eax, [CurrActThread]				; Current active thread
		dec	[LoopStackNb + eax * 4]


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doPEndDo:
doPEndDo	endp


; ===============================================================
; = doPUEndDo							=
; =								=
; = Function for the run-time forth word:	rt_+uenddo	=
; ===============================================================

		align	4
doPUEndDo	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Store into DI the offset address
		mov	ebx, esp
		mov	ecx, [ebx]

		; Update the loop index
		mov	edx, [LoopStackNb]
		cmp	dl, LOOPSTACKMAXNB
		jb	@@MaxLoopOk

		mov	ah, ERR_TOOMANYLOOPS				; "Too many nexted loops control structures"
		clr	edx						; No additionnal msg
		clr	ecx						; No line #
		call	doError

@@MaxLoopOk:	dec	edx
		imul	edx, size loopentry
		add	edx, offset LoopStack
		lodsd							; Pop loop increment
		cmpz	eax, 0
		jl	@@IncNeg

		; Loop increment is positive
		add	[edx.loop_index], eax				; I += step
		mov	eax, [edx.loop_index]				; I
		cmp	eax, [edx.loop_end]				; end value
		jge	@@LoopEnded
		jmp	@@LoopNotEnded					; Stop if I >= end value
		
		; Loop increment is positive
@@IncNeg:	add	[edx.loop_index], eax				; I += step
		mov	eax, [edx.loop_index]				; I
		cmp	eax, [edx.loop_end]				; end value
		jle	@@LoopEnded					; Stop if I <= end value
		

		; Loop not ended: do the branch
		; -----------------------------

@@LoopNotEnded:	add	ecx, [ecx]
		mov	[ebx], ecx
		jmp	@@Exit


		; Loop ended: skip the offset
		; ---------------------------

@@LoopEnded:	add	ecx, 4						; 4 is the size of a 32-bits number
		mov	[ebx], ecx

		; Update the loop number
		mov	eax, [CurrActThread]				; Current active thread
		dec	[LoopStackNb + eax * 4]


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doPUEndDo:
doPUEndDo	endp


; ===============================================================
; = doMake							=
; =								=
; = Function for the run-time forth word:	RT_MAKE		=
; ===============================================================

		align	4
doMake		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; -------------------
		; Initialize the loop
		; -------------------
		
		mov	edx, [CurrActThread]			; Current active thread
		mov	edx, [LoopStackNb + edx * 4]
		cmp	dl, LOOPSTACKMAXNB
		jl	@@MaxLoopOk

		mov	ah, ERR_TOOMANYLOOPS			; "Too many nexted loops control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@MaxLoopOk:	
		; Pop parameters
		; --------------
		
		mov	eax, PARAM1				; Start value
		mov	ebx, PARAM0				; End value
		mov	ecx, 1					; Step value
		DISCARD	2


		; Do not make the loop ?
		; ----------------------
		
		cmp	ecx, 0
		je	@@NotLoop
		jl	@@Neg
@@Pos:		cmp	eax, ebx
		jle	@@Loop
		jmp	@@NotLoop
@@Neg:		cmp	eax, ebx
		jge	@@Loop
		
@@NotLoop:	; Skip loop
		mov	ecx, [esp]
		add	ecx, [ecx]
		mov	[esp], ecx
		jmp	@@Exit

		
		; Make the loop
		; -------------

@@Loop:		imul	edx, size loopentry
		add	edx, offset LoopStack
		push	ebx
		mov	ebx, [CurrActThread]
		imul	ebx, LOOPSTACKMAXNB * size loopentry
		add	edx, ebx
		pop	ebx
		mov	[edx.loop_index], eax			; index <- start value
		mov	[edx.loop_start], eax			; Start value
		mov	[edx.loop_step], ecx			; Step value
		mov	[edx.loop_end], ebx			; end value
		mov	edx, [CurrActThread]			; Current active thread
		inc	[LoopStackNb + edx * 4]
		
		; Skip next cell
		add	dword ptr [esp], @WordSize


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doMake:	
doMake		endp


; ===============================================================
; = MAKE							=
; =								=
; = Function for the forth word:	MAKE			=
; ===============================================================

		align	4
MAKE		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE

		
		; Put 'CALL doMake'
		; -----------------
		
		; Put code operation		
		compCALL_rel32	doMake

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax						; Offset whil be resolved later
		call	CodeComma					; Code,



		; Leave the current code dictionnary pointer
		; ------------------------------------------
		
		mov	edx, [ControlStackPtr]
		cmp	edx, offset ControlStack0
		jne	@@StackOk

		mov	ah, ERR_TOOMANYCTRL				; "Too many nested control structures"
		clr	edx						; No additionnal msg
		clr	ecx						; No line #
		call	doError

@@StackOk:	sub	edx, 4
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, 4
		mov	[edx], eax					; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, 4
		mov	dword ptr [edx], 'Make'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_MAKE:		
MAKE		endp


; ===============================================================
; = doPlusMake							=
; =								=
; = Function for the run-time forth word:	RT_PLUSMAKE	=
; ===============================================================

		align	4
doPlusMake	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; -------------------
		; Initialize the loop
		; -------------------
		
		mov	edx, [CurrActThread]			; Current active thread
		mov	edx, [LoopStackNb + edx * 4]
		cmp	dl, LOOPSTACKMAXNB
		jl	@@MaxLoopOk

		mov	ah, ERR_TOOMANYLOOPS			; "Too many nexted loops control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@MaxLoopOk:	
		; Pop parameters
		; --------------
		
		mov	eax, PARAM2				; Start value
		mov	ebx, PARAM1				; End value
		mov	ecx, PARAM0				; Step value
		DISCARD	3


		; Do not make the loop ?
		; ----------------------
		
		cmp	ecx, 0
		je	@@NotLoop
		jl	@@Neg
@@Pos:		cmp	eax, ebx
		jle	@@Loop
		jmp	@@NotLoop
@@Neg:		cmp	eax, ebx
		jge	@@Loop
		
@@NotLoop:	; Skip loop
		mov	ecx, [esp]
		add	ecx, [ecx]
		mov	[esp], ecx
		jmp	@@Exit

		
		; Make the loop
		; -------------

@@Loop:		imul	edx, size loopentry
		add	edx, offset LoopStack
		push	ebx
		mov	ebx, [CurrActThread]
		imul	ebx, LOOPSTACKMAXNB * size loopentry
		add	edx, ebx
		pop	ebx
		mov	[edx.loop_index], eax			; index <- start value
		mov	[edx.loop_start], eax			; Start value
		mov	[edx.loop_step], ecx			; Step value
		mov	[edx.loop_end], ebx			; end value
		mov	edx, [CurrActThread]			; Current active thread
		inc	[LoopStackNb + edx * 4]
		
		; Skip next cell
		add	dword ptr [esp], @WordSize


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doPlusMake:	
doPlusMake	endp


; ===============================================================
; = PLUSMAKE							=
; =								=
; = Function for the forth word:	PLUSMAKE		=
; ===============================================================

		align	4
PLUSMAKE	proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE

		
		; Put 'CALL +Make'
		; ----------------
		
		; Put code operation		
		compCALL_rel32	doPlusMake

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax						; Offset whil be resolved later
		call	CodeComma					; Code,



		; Leave the current code dictionnary pointer
		; ------------------------------------------
		
		mov	edx, [ControlStackPtr]
		cmp	edx, offset ControlStack0
		jne	@@StackOk

		mov	ah, ERR_TOOMANYCTRL				; "Too many nested control structures"
		clr	edx						; No additionnal msg
		clr	ecx						; No line #
		call	doError

@@StackOk:	sub	edx, 4
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, 4
		mov	[edx], eax					; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, 4
		mov	dword ptr [edx], 'Make'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_PLUSMAKE:		
PLUSMAKE	endp


; ===============================================================
; = doEndMake							=
; =								=
; = Function for the run-time forth word:	RT_ENDMAKE	=
; ===============================================================

		align	4
doEndMake	proc

		; Update the loop index
		; ---------------------
		
		mov	eax, [CurrActThread]			; Current active thread
		mov	edx, [LoopStackNb + eax * 4]
		dec	edx
		imul	edx, size loopentry
		add	edx, offset LoopStack
		imul	eax, LOOPSTACKMAXNB * size loopentry
		add	edx, eax
		mov	eax, [edx.loop_step]
		add	[edx.loop_index], eax			; I += step
		mov	ebx, [edx.loop_index]			; I
		cmp	eax, 0
		jl	@@Neg
@@Pos:		cmp	ebx, [edx.loop_end]			; end value
		jg	@@LoopEnded
		jmp	@@LoopNotEnded
@@Neg:		cmp	ebx, [edx.loop_end]			; end value
		jl	@@LoopEnded
		

		; Loop not ended: do the branch
		; -----------------------------

@@LoopNotEnded:	mov	eax, [esp]
		add	eax, [eax]
		mov	[esp], eax
		ret						; jmp	@@Exit


		; Loop ended: skip the offset
		; ---------------------------

@@LoopEnded:	add	dword ptr [esp], @WordSize		; Skip next cell

		; Update the loop number
		mov	eax, [CurrActThread]			; Current active thread
		dec	[LoopStackNb + eax * 4]


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doEndMake:	
doEndMake	endp


; ===============================================================
; = ENDMAKE							=
; =								=
; = Function for the forth word:	ENDMAKE			=
; ===============================================================

		align	4
ENDMAKE		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'MAKE' was called before
		; ------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'Make'
		je	@@DoCalled
		cmp	eax, 'When'
		je	@@WhenCalled
		
		; No, it's an error
		mov	ah, ERR_MAKENC				; "''MAKE' not called beforehand."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; -----------------------------
		; Resolve 'CALL ZBranch' offset
		; -----------------------------
		
@@DoCalled:	

		; Put 'CALL doEndloop' offset
		; -------------------------
		
		; Put 'CALL rel32' code operation		
		compCALL_rel32	doEndMake

		; Put the know 32-bits jump offset into the Code Dictionnary
		mov	ecx, [edx]
		mov	eax, ecx
		add	edx, 4						; DROP
		add	eax, 4
		sub	eax, [CDA_SIZE]					; Offset
		call	CodeComma					; Code,
		
		jmp	@@Resolv


		; -----------------------------
		; WHEN (or NOTWHENN) was called
		; -----------------------------
		
@@WhenCalled:	
		
		; Resolve 'CALL ZBranch' offset
		; -----------------------------

		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		add	eax, 9					; 9: call-rel32 (5) + offset (4)
		mov	ebx, [edx]				; Relative address to store offset
		sub	eax, ebx				; Address to jump
		add	edx, 4					; DROP
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax

		; Continue normally
		add	edx, 4
		jmp	@@DoCalled


		; -----------------------------------------
		; Resolve the offset in the 'doTimes' field
		; -----------------------------------------
		
@@Resolv:	
		; Resolve the offset in the 'doMake' field
		; -----------------------------------------

		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, ecx
		add	ecx, [CDA_START]
		mov	[ecx], eax


		; ----------------
		; Return to caller
		; ----------------
		
		mov	[ControlStackPtr], edx
		ret

E_ENDMAKE:	
ENDMAKE		endp


; ===============================================================
; = DO								=
; =								=
; = Function for the forth word:	DO			=
; ===============================================================

		align	4
DO		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE

		
		; Put 'CALL doDo'
		; ---------------
		
		; Put 'SUMMON rt_do' code operation		
		compCALL_rel32	doDo


		; Leave the current code dictionnary pointer
		; ------------------------------------------
		
		mov	edx, [ControlStackPtr]
		cmp	edx, offset ControlStack0
		jne	@@StackOk

		mov	ah, ERR_TOOMANYCTRL				; "Too many nested control structures"
		clr	edx						; No additionnal msg
		clr	ecx						; No line #
		call	doError

@@StackOk:	sub	edx, 4
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, 4
		mov	[edx], eax					; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, 4
		mov	dword ptr [edx], 'DoLo'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_DO:
DO		endp


; ===============================================================
; = doWhen							=
; =								=
; = Function for the run-time forth word	rt_when		=
; =								=
; = Regs modified:	none					=
; =								=
; ===============================================================

		align	4
doWhen		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Get next cells address
		mov	ecx, [esp]

		; Pop flag from the top of the stack
		cmp	dword ptr [esi], 0
		je	@@Zero


		; Not zero : we do not make the branch
		; ------------------------------------
		
		; Move the IP after the jump offset
	      	add	ecx, 4					; 4 is the size of a 32-bits number

		; Done
		jmp	@@Next
		
		
		; Zero; we do the branch
		; ----------------------
		
@@Zero:		; Move the IP to the new position
		add	ecx, [ecx]

		; Update the loop number
		mov	eax, [CurrActThread]			; Current active thread
		dec	[LoopStackNb + eax * 4]


@@Next:		mov	[esp], ecx


		; Return to caller
		; ----------------
		
		add	esi, @WordSize				; DROP
		ret

E_doWhen:
doWhen		endp


; ===============================================================
; = WHEN							=
; =								=
; = Function for the forth word:	WHEN			=
; ===============================================================

		align	4
WHEN		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'DO' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		cmp	eax, 'DoLo'
		je	@@DoCalled
		cmp	eax, 'Make'
		je	@@DoCalled
		
		; No, it's an error
		mov	ah, ERR_DONC				; "'DO' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; Put 'CALL ZBranch' offset (offset will be resolved by 'ENDDO' or '+ENDDO')
		; ------------------------------------------------------------------------
		
		; Put 'SUMMON doWhen' code operation		
@@DoCalled:	compCALL_rel32	doWhen

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax						; Offset whil be resolved lated
		call	CodeComma					; Code,


		; Leave the current code dictionnary pointer
		; ------------------------------------------
		
		sub	edx, 4
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, 4
		mov	[edx], eax					; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, 4
		mov	dword ptr [edx], 'When'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_WHEN:
WHEN		endp


; ===============================================================
; = doNotWhen							=
; =								=
; = Function for the run-time forth word (NotWhen)		=
; =								=
; = Regs modified:	none					=
; ===============================================================

		align	4
doNotWhen	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; Get next cells address
		mov	ecx, [esp]

		; Pop flag from the top of the stack
		cmp	dword ptr [esi], 0
		je	@@Zero


		; Not Zero: we do the branch
		; ---------------------------
		
		; Move the IP to the new position
		add	ecx, [ecx]
		
		; Update the loop number
		mov	eax, [CurrActThread]			; Current active thread
		dec	[LoopStackNb + eax * 4]

		; Done
		jmp	@@Next
		
		
		; Zero : we do not make the branch
		; --------------------------------
		
@@Zero:		
		; Move the IP after the jump offset
	      	add	ecx, 4					; 4 is the size of a 32-bits number


@@Next:		mov	[esp], ecx


		; Return to caller
		; ----------------
		
		add	esi, @WordSize				; DROP
		ret

E_doNotWhen:
doNotWhen	endp


; ===============================================================
; = NOTWHEN							=
; =								=
; = Function for the forth word:	NOTWHEN			=
; ===============================================================

		align	4
NOTWHEN		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'DO' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		cmp	dword ptr [edx], 'DoLo'
		je	@@DoCalled
		cmp	dword ptr [edx], 'Make'
		je	@@DoCalled
		
		; No, it's an error
		mov	ah, ERR_DONC				; "'DO' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; Put 'CALL ZBranch' offset (offset will be resolved by 'ENDDO' or '+ENDDO')
		; ------------------------------------------------------------------------
		
		; Put 'CALL rel32' code operation		
@@DoCalled:	compCALL_rel32	doNotWhen

		; Put the unknow 32-bits jump offset into the Code Dictionnary
		clr	eax						; Offset whil be resolved later
		call	CodeComma					; Code,


		; Leave the current code dictionnary pointer
		; ------------------------------------------
		
		sub	edx, 4
		mov	eax, [CDA_SIZE]
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, 4
		mov	[edx], eax					; HERE


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, 4
		mov	dword ptr [edx], 'When'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E_NOTWHEN:
NOTWHEN		endp


; ===============================================================
; = ENDDO							=
; =								=
; = Function for the forth word:	ENDDO			=
; ===============================================================

		align	4
ENDDO		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'DO' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'DoLo'
		je	@@DoCalled
		cmp	eax, 'When'
		je	@@WhenCalled
		
		; No, it's an error
		mov	ah, ERR_DONC				; "'DO' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; -----------------------------
		; Resolve 'CALL ZBranch' offset
		; -----------------------------
		
@@DoCalled:	

		; Put 'CALL doEndDo' offset
		; -------------------------
		
		; Put 'CALL rel32' code operation		
		compCALL_rel32	doEndDo

		; Put the know 32-bits jump offset into the Code Dictionnary
		mov	eax, [edx]
		add	edx, 4						; DROP
		add	eax, 4
		sub	eax, [CDA_SIZE]					; Offset
		call	CodeComma					; Code,
		
		jmp	@@Exit


		; -----------------------------
		; WHEN (or NOTWHENN) was called
		; -----------------------------
		
@@WhenCalled:	
		
		; Resolve 'CALL ZBranch' offset
		; -----------------------------

		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		add	eax, 9					; 9: call-rel32 (5) + offset (4)
		mov	ebx, [edx]				; Relative address to store offset
		sub	eax, ebx				; Address to jump
		add	edx, 4					; DROP
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax

		; Continue normally
		add	edx, 4
		jmp	@@DoCalled


		; Return to caller
		; ----------------
		
@@Exit:		mov	[ControlStackPtr], edx
		mov	eax, [CDA_SIZE]				; HERE
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		ret

E_ENDDO:
ENDDO		endp


; ===============================================================
; = PENDDO							=
; =								=
; = Function for the forth word:	+ENDDO			=
; ===============================================================

		align	4
PENDDO		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'DO' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'DoLo'
		je	@@DoCalled
		cmp	eax, 'When'
		je	@@WhenCalled
		
		; No, it's an error
		mov	ah, ERR_DONC				; "'DO' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; -----------------------------
		; Resolve 'CALL ZBranch' offset
		; -----------------------------
		
@@DoCalled:	

		; Put 'SUMMON doPEndDo'
		compCALL_rel32	doPEndDo
		
		; Put the know 32-bits jump offset into the Code Dictionnary
		mov	eax, [edx]
		add	edx, 4						; DROP
		add	eax, 4
		sub	eax, [CDA_SIZE]					; Offset
		call	CodeComma					; Code,

		jmp	@@Exit


		; -----------------------------
		; WHEN (or NOTWHENN) was called
		; -----------------------------
		
@@WhenCalled:	
		
		; Resolve 'CALL ZBranch' offset
		; -----------------------------

		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		add	eax, 9					; 9: call-rel32 (5) + offset (4)
		mov	ebx, [edx]				; Relative address to store offset
		sub	eax, ebx				; Address to jump
		add	edx, 4					; DROP
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax

		; Continue normally
		jmp	@@DoCalled


		; Return to caller
		; ----------------
		
@@Exit:		mov	[ControlStackPtr], edx
		mov	eax, [CDA_SIZE]				; HERE
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		ret

E_PENDDO:
PENDDO		endp


; ===============================================================
; = PUENDDO							=
; =								=
; = Function for the forth word:	+UENDDO			=
; ===============================================================

		align	4
PUENDDO		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'DO' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'DoLo'
		je	@@DoCalled
		cmp	eax, 'When'
		je	@@WhenCalled
		
		; No, it's an error
		mov	ah, ERR_DONC				; "'DO' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; -----------------------------
		; Resolve 'CALL ZBranch' offset
		; -----------------------------
		
@@DoCalled:	

		; Put 'SUMMON rel32 doPUEndDo'
		compCALL_rel32	doPUEndDo

		; Put the know 32-bits jump offset into the Code Dictionnary
		mov	eax, [edx]
		add	edx, 4					; DROP
		add	eax, 4
		sub	eax, [CDA_SIZE]				; Offset
		call	CodeComma				; Code,

		jmp	@@Exit


		; -----------------------------
		; WHEN (or NOTWHENN) was called
		; -----------------------------
		
@@WhenCalled:	
		
		; Resolve 'CALL ZBranch' offset
		; -----------------------------

		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		add	eax, 10					; 9: call-rel32 (6) + offset (4)
		mov	ebx, [edx]				; Relative address to store offset
		sub	eax, ebx				; Address to jump
		add	edx, 4					; DROP
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax

		; Continue normally
		jmp	@@DoCalled


		; Return to caller
		; ----------------
		
@@Exit:		mov	[ControlStackPtr], edx
		mov	eax, [CDA_SIZE]				; HERE
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		ret

E_PUENDDO:
PUENDDO		endp


; ===============================================================
; = WHILE							=
; =								=
; = Function for the forth word:	WHILE			=
; ===============================================================

		align	4
_WHILE		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'Begin' was called before
		; -------------------------------------

		mov	edx, [ControlStackPtr]		
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'Begi'
		je	@@BeginCalled
		
		; No, it's an error
		mov	ah, ERR_BEGINNC				; "'Begin' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

		; Put 'CALL ZBranch' offset
		; -------------------------
		
@@BeginCalled:	

		_LODSD_						; LODSD
		OR_EAX_EAX					; OR     EAX, EAX
		JE_rel32					; JE     rel32

		; Push the current address to store offset (HERE)
		sub	edx, @WordSize
		mov	eax, [CDA_SIZE]				; HERE
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		mov	[edx], eax

		; Put the 32-bits offset (still unknon) into the Code Dictionnary
		clr	eax
		call	CodeComma				; Code,		


		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, @WordSize
		mov	dword ptr [edx], 'Whil'
		mov	[ControlStackPtr], edx
		
		
		; Return to caller
		; ----------------
		
		ret

E__WHILE:
_WHILE		endp


; ===============================================================
; = _REPEAT							=
; =								=
; = Function for the forth word:	REPEAT			=
; ===============================================================

		align	4
_REPEAT		proc

		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; Verify that 'While' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'Whil'
		je	@@WhileCalled
		
		; No, it's an error
		mov	ah, ERR_WHILENC				; "'WHILE' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

@@WhileCalled:	

		; ------------------------
		; Put 'CALL Branch' offset
		; ------------------------
		
		; Relative address to jump
		mov	ebx, [edx + @WordSize]			; Address to jump

		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, ebx
		sub	eax, 1					; Size of the codop 'JMP rel8'
		sub	eax, [CDA_SIZE]				; HERE
		sub	eax, 1					; 1 for REL8
		cmp	eax, -128
		jl	@@LongJe
		cmp	eax, 127
		jg	@@LongJe
		JMP_rel8					; Put 'JMP rel8'		operation
		call	CodeCComma				; CodeC,
		jmp	@@ShortLongJe
@@LongJe:	JMP_rel32					; Put 'JMP rel32'		operation
		mov	eax, ebx
		sub	eax, [CDA_SIZE]				; HERE
		sub	eax, 4					; 1 for REL32
		call	CodeComma				; Code,		
@@ShortLongJe:


		; -----------------------------
		; Resolve 'CALL ZBranch' offset
		; -----------------------------
		
		; Put the 32-bits offset into the Code Dictionnary
		mov	ebx, [edx]				; Relative address to store offset
		add	edx, 8					; DROP DROP
		mov	[ControlStackPtr], edx
		mov	eax, [CDA_SIZE]				; HERE
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		sub	eax, ebx				; Address to jump
		sub	eax, 4					; 4 for: REL32 (JE rel32)
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax


		; Return to caller
		; ----------------
		
		ret

E__REPEAT:
_REPEAT		endp


; ===============================================================
; = doCase							=
; =								=
; = Function for the tun-time forth word:	RT_CASE		=
; ===============================================================

		align	4
doCase		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ----------------------------------------------------------------------------
		; Pop number to test in the case..endcase struct, then push it onto case stack
		; ----------------------------------------------------------------------------
		
		mov	ebx, [CurrActThread]			; Current active thread
		mov	edx, [CaseStackNb + ebx * 4]
		cmp	dl, CASESTACKMAXNB
		jl	@@MaxCasesOk

		; "Too many nested cases control structures"
		mov	ah, ERR_TOOMANYCASES			; "Too many nested cases control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@MaxCasesOk:	
		; Update the number of case..endcase nested
		inc	[CaseStackNb + ebx * 4]
		
		imul	edx, @WordSize
		add	edx, offset CaseStack
		imul	ebx, CASESTACKMAXNB * @WordSize
		add	edx, ebx
		lodsd
		mov	[edx], eax
		

		; ----------------
		; Return to caller
		; ----------------

		ret

E_doCase:
doCase		endp


; ===============================================================
; = doOf							=
; =								=
; = Function for the run-time forth word:	RT_OF		=
; ===============================================================

		align	4
doOf		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; --------------------------------
		; Store into DI the offset address
		; --------------------------------
		
		mov	ecx, [esp]


		; -------------------------------------------------
		; Do the branch if case value is equal to top stack
		; -------------------------------------------------

		lodsd

		mov	ebx, [CurrActThread]				; Current active thread
		mov	edx, [CaseStackNb + ebx * 4]
		dec	edx
		shl	edx, 2						; * 4
		add	edx, offset CaseStack
		imul	ebx, CASESTACKMAXNB * @WordSize
		add	edx, ebx

		cmp	eax, [edx]
		je	@@Equal


		; -------------------------
		; Not equal : Do the branch
		; -------------------------
		
		add	ecx, [ecx]
		mov	[esp], ecx
		ret							; jmp	@@Exit
		

		; ---------------------
		; Equal : Do not branch
		; ---------------------

@@Equal:	

		; Skip the offset: move the IP after the offset
		add	dword ptr [esp], 4				; Skip next cell


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_doOf:
doOf		endp


; ===============================================================
; = doEndCase							=
; =								=
; = Function for the tun-time forth word:	(ENDCASE)	=
; ===============================================================

		align	4
doEndCase	proc

		; -----------------------------------------
		; Update the number of case..endcase nested
		; -----------------------------------------
		
		mov	ebx, [CurrActThread]			; Current active thread
		dec	[CaseStackNb + ebx * 4]
		

		; ----------------
		; Return to caller
		; ----------------

		ret

E_doEndCase:
doEndCase	endp


; ===============================================================
; = CASE							=
; =								=
; = Function for the forth word:	CASE			=
; ===============================================================

		align	4
CASE		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; -----------------
		; Put 'CALL (CASE)'
		; -----------------
		
		; Put 'SUMMON doCase'
		compCALL_rel32	doCase
		

		; ----------------------------------------------------------------
		; Push the initial number of forward jumps to resolve (by ENDCASE)
		; ----------------------------------------------------------------
		
		mov	edx, [ControlStackPtr]
		cmp	edx, offset ControlStack0
		jne	@@StackOk
		mov	ah, ERR_TOOMANYCTRL			; "Too many nested control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@StackOk:	sub	edx, @WordSize
		mov	dword ptr [edx], 0
				
		
		; ---------------------------------------
		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, @WordSize
		mov	dword ptr [edx], 'Case'
		mov	[ControlStackPtr], edx


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_CASE:
CASE		endp


; ===============================================================
; = OF								=
; =								=
; = Function for the forth word:	OF			=
; ===============================================================

		align	4
OF		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; -------------------------------------
		; Verify that 'Case' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]
		cmp	edx, offset ControlStack0
		jne	@@StackOk
		mov	ah, ERR_TOOMANYCTRL			; "Too many nested control structures"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@StackOk:	mov	eax, [edx]
		add	edx, @WordSize
		cmp	eax, 'Case'
		je	@@CaseCalled
		cmp	eax, 'CASE'
		je	@@TipofCalled

		; It's an error
		mov	ah, ERR_CASENC				; "'CASE' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@CaseCalled:	
		; ----------------------
		; Put 'CALL (OF)' offset
		; ----------------------
		
		; Put 'SUMMON rt_of'
		compCALL_rel32	doOf
		
		; Push the current address to store offset (HERE)
		sub	edx, @WordSize
		mov	eax, [CDA_SIZE]				; HERE
		mov	[edx], eax

		; Put the 32-bits offset into the Code Dictionnary
		clr	eax					; Where to jump is now unknow (will be resolved by ENDOF)
		call	CodeComma				; Code,		

		; Done
		jmp	@@Exit
		
		
@@TipofCalled:	
		; ----------------------
		; Put 'CALL (OF)' offset
		; ----------------------

		mov	ebx, [edx + @WordSize]
		add	ebx, [CDA_START]
		mov	eax, [CDA_SIZE]
		add	eax, [CDA_START]
		add	eax, 5+4-4				; Destination address
		sub	eax, ebx
		mov	[ebx], eax
		mov	dword ptr [edx + @WordSize], 0		; Skip this address in ENDCASE

		
		; Put 'SUMMON rt_of'
		compCALL_rel32	doOf
		
		; Push the current address to store offset (HERE)
		sub	edx, 4
		mov	eax, [CDA_SIZE]				; HERE
		mov	[edx], eax

		; Put the 32-bits offset into the Code Dictionnary
		clr	eax					; Where to jump is now unknow (will be resolved by ENDOF)
		call	CodeComma				; Code,		


		; ---------------------------------------
		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
@@Exit:		sub	edx, @WordSize
		mov	dword ptr [edx], '-OF-'
		mov	[ControlStackPtr], edx

		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_OF:
OF		endp


; ===============================================================
; = ENDOF							=
; =								=
; = Function for the forth word:	ENDOF			=
; ===============================================================

		align	4
ENDOF		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; ----------------------------------
		; Verify that 'OF' was called before
		; ----------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, @WordSize
		cmp	eax, '-OF-'
		je	@@OfCalled
		
		; No, it's an error
		mov	ah, ERR_OFNC				; "'OF' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

@@OfCalled:	

		; --------------------------------------------
		; Resolve 'CALL ZBranch' offset (stored by OF)
		; --------------------------------------------
		
		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		sub	eax, 4					; 4 for REL32 ('JMP rel32' stored by OF)
		add	eax, 1+4				; 1: size codop JMP rel32  4: REL32
		add	eax, 4					; 4 for size cell after 'CALL (OF)'
		mov	ebx, [edx]				; Relative address to store offset
		add	edx, @WordSize				; DROP
		sub	eax, ebx				; Address to jump
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax


		; ------------------------
		; Put 'CALL Branch' offset
		; ------------------------

		JMP_rel32

		; Push the current address to store offset (HERE) and update the number of forward jumps to resolve later
		mov	eax, [edx]				; Actual nb of forward jumps
		mov	ebx, [CDA_SIZE]				; HERE
		mov	[edx], ebx
		sub	edx, @WordSize
		inc	eax
		mov	[edx], eax

		; Put the 32-bits offset into the Code Dictionnary
		clr	eax					; Where to jump (still unknow, will be resolved by ENDCASE)
		call	CodeComma				; Code,		


		; ---------------------------------------
		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, @WordSize
		mov	dword ptr [edx], 'Case'
		mov	[ControlStackPtr], edx


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ENDOF:
ENDOF		endp


; ===============================================================
; = TIPOF							=
; =								=
; = Function for the forth word:	TIPOF			=
; ===============================================================

		align	4
TIPOF		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; ----------------------------------
		; Verify that 'OF' was called before
		; ----------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, '-OF-'
		je	@@OfCalled
		
		; No, it's an error
		mov	ah, ERR_OFNC				; "'OF' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

@@OfCalled:	

		; --------------------------------------------
		; Resolve 'CALL ZBranch' offset (stored by OF)
		; --------------------------------------------
		
		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		sub	eax, 4					; 4 for REL32 ('JMP rel32' stored by OF)
		add	eax, 1+4				; 1: size codop JMP rel32  4: REL32
		add	eax, 4					; 4 for size cell after 'CALL (OF)'
		mov	ebx, [edx]				; Relative address to store offset
		add	edx, 4					; DROP
		sub	eax, ebx				; Address to jump
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax


		; ------------------------
		; Put 'CALL Branch' offset
		; ------------------------

		JMP_rel32

		; Push the current address to store offset (HERE) and update the number of forward jumps to resolve later
		mov	eax, [edx]				; Actual nb of forward jumps
		mov	ebx, [CDA_SIZE]				; HERE
		mov	[edx], ebx
		sub	edx, 4
		inc	eax
		mov	[edx], eax

		; Put the 32-bits offset into the Code Dictionnary
		clr	eax					; Where to jump (still unknow, will be resolved by ENDCASE)
		call	CodeComma				; Code,		


		; ---------------------------------------
		; Push a 'mark' value (compiler security)
		; ---------------------------------------
		
		sub	edx, 4
		mov	dword ptr [edx], 'CASE'
		mov	[ControlStackPtr], edx


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_TIPOF:
TIPOF		endp


; ===============================================================
; = ENDCASE							=
; =								=
; = Function for the forth word:	ENDCASE			=
; ===============================================================

		align	4
ENDCASE		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; ------------------------------------
		; Verify that 'CASE' was called before
		; ------------------------------------
		
		mov	edx, [ControlStackPtr]
		mov	eax, [edx]
		add	edx, 4
		cmp	eax, 'Case'
		je	@@CaseCalled
		
		; No, it's an error
		mov	ah, ERR_CASENC				; "'CASE' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		

@@CaseCalled:	

		; --------------------------------------------------
		; Resolve the 'CALL Branch' offset (stored by ENDOF)
		; -------------------------------------------------

		; How many forward jump to resolve ?
		mov	ecx, [edx]
		add	edx, @WordSize				; DROP
		
		; Next forward jump to resolve
@@LL10:		cmpz	ecx, 0
		je	@@LL1
		
		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, [CDA_SIZE]				; HERE
		mov	ebx, [edx]				; Relative address to store offset
		cmpz	ebx, 0
		je	@@LL20
		sub	eax, ebx				; Address to jump
		sub	eax, 4					; 4 for REL32
		add	ebx, [CDA_START]			; EBX = address to store offset
		mov	[ebx], eax
@@LL20:		add	edx, @WordSize				; DROP

		; Next
		dec	ecx
		jmp	@@LL10


		; ---------------------------
		; Put 'CALL (ENDCASE)' offset
		; ---------------------------

@@LL1:
		
		; Put 'SUMMON do_endcase'
		compCALL_rel32	doEndCase


		; ----------------
		; Return to caller
		; ----------------
		
		mov	[ControlStackPtr], edx
		mov	eax, [CDA_SIZE]				; HERE
		mov	[AddrLastBeginIf],eax 		; Adrees of last Begin/If (do not optimize literal integer number ?)
		ret

E_ENDCASE:
ENDCASE		endp


code32		ends
		end
