; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	CREF.ASM						!
; 									!
; Description:	Definitions for playing with cross-ref.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	UPDATE		UPDATE						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32 		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

if TURNKEYED eq 0

include		Ed.inc                      		; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = CRef_add							=
; =								=
; = Usage:							=
; = 	Called every time a new word is compiled in the		=
; =	current word being defined: update words cross ref.	=
; =								=
; = I) ECX : NIA pointer to caller word				=
; =    EDX : NIA pointer to called word				= 
; =								=
; = O) none							=
; =								=
; = M) none							=
; ===============================================================

		align	4
CRef_add	proc

		; --------------
		; Save used regs
		; --------------
		
		push	eax ebx ecx edx ebp
		
		
		; -----------------------------------------
		; Update the nb of uses for the called word
		; -----------------------------------------

		push	edx		
		add	edx, [NIA_START]
		inc	[edx.word_NbUses]
		pop	edx
		
		
		; ---------------------------
		; If words ref table is empty
		; ---------------------------
		
		cmp	[CrossRefTotalNb], 0
		jne	@@NotEmpty
		
		
		; -------------------
		; Create a new record
		; -------------------

		; Create the 1st record in the table
		mov	ebp, [CrossRef]				; Compute address on new record
		mov	[ebp.wref_caller], ecx			; Pointer to caller word
		mov	[ebp.wref_called], edx			; Pointer to called word
		mov	[ebp.wref_calls], 1			; Nb of uses
		
		; Create the 1st index relative to callers
		mov	ebx, [CRefNdxRefRD]
		mov	dword ptr [ebx], 0			; Pointer to record #0

		; Create the 1st index relative to calleds
		mov	ebx, [CRefNdxRefDR]			; Pointer to record #0
		mov	dword ptr [ebx], 0
		
		; Update the nb of records
		mov	[CrossRefNb], 1
		mov	[CrossRefTotalNb], 1
		
		; Done
		jmp	@@Exit

@@NotEmpty:

		; ----------------------------------------------------
		; Search if the record (Caller, Called) already exists
		; ----------------------------------------------------
	
		; in ECX: pointer to caller,   in EDX: pointer to called		
		call	CRefSrch_RD
		jc	@@RecExists
		

		; --------------------------------
		; Record does not exist, create it
		; --------------------------------
		
		; Compute address on new record
		mov	ebp, [CrossRefFFreed]
		cmp	ebp, -1
		jne	@@Freed
		inc	[CrossRefTotalNb]
		mov	ebp, [CrossRefNb]
		jmp	@@Freed2
@@Freed:	mov	eax, ebp
		imul	eax, size wrefentry
		add	eax, [CrossRef]
		mov	eax, [eax.wref_caller]			; Link to next freed
		mov	[CrossRefFFreed], eax
@@Freed2:	mov	eax, ebp
		imul	eax, size wrefentry
		add	eax, [CrossRef]

		; Create a new record in the table
		mov	[eax.wref_caller], ecx			; Pointer to caller word
		mov	[eax.wref_called], edx			; Pointer to called word
		mov	[eax.wref_calls], 1			; Nb of uses
		
		; Update the 1st index relative to callers
		cmp	ebx, [CrossRefNb]
		je	@@NotMoveIdx1
		push	ecx esi edi
		mov	ecx, [CrossRefNb]
		mov	edi, ecx
		shl	edi, 2
		add	edi, [CRefNdxRefRD]			; Destination address
		mov	esi, edi
		sub	esi, 4					; Source address
		sub	ecx, ebx
		std						; Dec ESI and EDI
		rep	movsd  		          	    	; Move the block
		cld
		pop	edi esi ecx
@@NotMoveIdx1:	mov	eax, ebx
		shl	eax, 2
		add	eax, [CRefNdxRefRD]			; Index record address
		mov	[eax], ebp

		; Update the 2nd index relative to calleds
		; in ECX: pointer to caller,   in EDX: pointer to called		
		call	CRefSrch_DR
		cmp	ebx, [CrossRefNb]
		je	@@NotMoveNdx1
		push	ecx esi edi
		mov	ecx, [CrossRefNb]
		mov	edi, ecx
		shl	edi, 2
		add	edi, [CRefNdxRefDR]			; Destination address
		mov	esi, edi
		sub	esi, 4					; Source address
		sub	ecx, ebx
		std						; Dec ESI and EDI
		rep	movsd  		          	    	; Move the block
		cld
		pop	edi esi ecx
@@NotMoveNdx1:	mov	eax, ebx
		shl	eax, 2
		add	eax, [CRefNdxRefDR]			; Index record address
		mov	[eax], ebp
		
		; Update the nb of records
		inc	[CrossRefNb]
		
		; Done
		jmp	@@Exit
		
		
		; -------------------------------------
		; Record already exists: just update it
		; -------------------------------------
		
		; Locate record address (in EBP)
@@RecExists:	mov	ebp, ebx
		shl	ebp, 2
		add	ebp, [CRefNdxRefRD]
		mov	ebp, [ebp]
		imul	ebp, size wrefentry
		add	ebp, [CrossRef]
		
		; Update the number of uses
		inc	[ebp.wref_calls]


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ebp edx ecx ebx eax
		ret

CRef_add	endp


; ===============================================================
; = CRef_del							=
; =								=
; = Usage:							=
; = 	Delete a cross reference record.			=
; =								=
; = I) ECX : NIA pointer to caller word				=
; =    EDX : NIA pointer to called word				= 
; =								=
; = O) none							=
; =								=
; = M) none							=
; ===============================================================

		align	4
CRef_del	proc

		; --------------
		; Save used regs
		; --------------
		
		push	eax ebx ecx edx ebp
		
		
		; -----------------
		; Locate the record
		; -----------------

		; in ECX: pointer to caller,   in EDX: pointer to called		
		; Found index #1 record #
		call	CRefSrch_DR
		jc	@@RecExists1

		; Internal error !
		; !!!
		add	esp, 5 * @WordSize			; Clean stack
		mov	ah, ERR_NULLWORD			; Can't locate a word to delete in cross ref table (called, caller)
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError
@@RecExists1:
	
		; Found index #0 record #
		mov	ebp, ebx
		call	CRefSrch_RD
		jc	@@RecExists0

		; Internal error !
		; !!!
		add	esp, 5 * @WordSize			; Clean stack
		mov	ah, ERR_NULLWORD			; Can't locate a word to delete in cross ref table (caller, called)
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError

		; Record nb, in EAX
@@RecExists0:	mov	eax, ebx
		shl	eax, 2
		add	eax, [CRefNdxRefRD]
		mov	eax, [eax]
		
		; Delete the 1st index record (index on Caller, Called)
		push	esi edi ecx
		mov	ecx, [CrossRefNb]
		sub	ecx, ebx
		dec	ecx					; Nb of records
		cmpz	ecx, 0
		je	@@NotMove1
		mov	edi, ebx
		shl	edi, 2
		add	edi, [CRefNdxRefRD]			; Destination address
		mov	esi, edi
		add	esi, 4					; Source address
		cld						; Inc ESI and EDI
		rep	movsd  		          	    	; Move the block
@@NotMove1:	pop	ecx edi esi
		
		; Delete the 2nd index record (index on Called, Caller)
		mov	ebx, ebp
		push	esi edi ecx
		mov	ecx, [CrossRefNb]
		sub	ecx, ebx
		dec	ecx					; Nb of records
		cmpz	ecx, 0
		je	@@NotMove2
		mov	edi, ebx
		shl	edi, 2
		add	edi, [CRefNdxRefDR]			; Destination address
		mov	esi, edi
		add	esi, 4					; Source address
		cld						; Inc ESI and EDI
		rep	movsd  		          	    	; Move the block
@@NotMove2:	pop	ecx edi esi
		
		; Delete this record
		push	esi
		mov	esi, [CrossRefFFreed]
		mov	ebx, eax
		imul	ebx, size wrefentry
		add	ebx, [CrossRef]				; Destination address
		mov	[ebx.wref_caller], esi
		mov	[ebx.wref_called], -1
		mov	[ebx.wref_calls], -1
		mov	[CrossRefFFreed], eax
		cmp	esi, -1
		jne	@@NotDec
;		dec	[CrossRefTotalNb]			; !!!
@@NotDec:	pop	esi
		
		; Update the number of records
		dec	[CrossRefNb]
		
		
		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	ebp edx ecx ebx eax
		ret

CRef_del	endp


; ===============================================================
; = CRefSrch_RD							=
; =								=
; = Usage:							=
; = 	Search a word reference, (CALLER, CALLED).		=
; =								=
; = I) ECX : NIA pointer to caller word				=
; =    EDX : NIA pointer to called word				= 
; =								=
; = O) CF = 1, ref is found, so address is in EBX		=
; =    CF = 0, ref not found					=
; =								=
; = M) CF, EBX							=
; ===============================================================

		align	4
CRefSrch_RD	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebp esi edi
		
		
		; ---------------
		; Initializations
		; ---------------
		
		; Left limit
		clr	esi
		
		; Right limit
		mov	edi, [CrossRefNb]

		
		; ------------------
		; Search (dichotomy)
		; ------------------

		; Stop the search ?
@@DichoSearch:	cmp	esi, edi
		ja	@@NotFound
		cmp	edi, -1
		je	@@NotFound
		
		; Compute middle point
		mov	ebx, edi
		add	ebx, esi
		shr	ebx, 1					; Middle = (Left + Right) / 2
		
		cmp	ebx, [CrossRefNb]
		je	@@NotFound

		; Compare this point with the asked record
		mov	ebp, ebx
		shl	ebp, 2
		add	ebp, [CRefNdxRefRD]
		mov	ebp, [ebp]
		imul	ebp, size wrefentry
		add	ebp, [CrossRef]
		cmp	[ebp.wref_caller], ecx
		ja	@@GoLeft
		jb	@@GoRight
		cmp	[ebp.wref_called], edx
		ja	@@GoLeft
		jb	@@GoRight		
		
		; Found
		mov	esi, ebx
		jmp	@@Found
		
		; Go left : search between [left, middle - 1]
@@GoLeft:	mov	edi, ebx
		dec	edi
		jmp	@@DichoSearch

		; Go right : search between [middle + 1, right]
@@GoRight:	mov	esi, ebx
		inc	esi
		jmp	@@DichoSearch
		
		
		; ----------------
		; Return to caller
		; ----------------
	
@@NotFound:	clc						; CF = 0
		jmp	@@Exit

@@Found:	stc						; CF = 1
		
@@Exit:		mov	ebx, esi
		pop	edi esi ebp eax
		ret

CRefSrch_RD	endp


; ===============================================================
; = CRefSrch_DR							=
; =								=
; = Usage:							=
; = 	Search a word reference, (CALLED, CALLER).		=
; =								=
; = I) ECX : NIA pointer to caller word				=
; =    EDX : NIA pointer to called word				= 
; =								=
; = O) CF = 1, ref is found, so address is in EBX		=
; =    CF = 0, ref not found					=
; =								=
; = M) CF, EBX							=
; ===============================================================

		align	4
CRefSrch_DR	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebp esi edi
		
		
		; ---------------
		; Initializations
		; ---------------
		
		; Left limit
		clr	esi
		
		; Right limit
		mov	edi, [CrossRefNb]

		
		; ------------------
		; Search (dichotomy)
		; ------------------

		; Stop the search ?
@@DichoSearch:	cmp	esi, edi
		ja	@@NotFound
		cmp	edi, -1
		je	@@NotFound
		
		; Compute middle point
		mov	ebx, edi
		add	ebx, esi
		shr	ebx, 1					; Middle = (Left + Right) / 2
		
		cmp	ebx, [CrossRefNb]
		je	@@NotFound

		; Compare this point with the asked record
		mov	ebp, ebx
		shl	ebp, 2
		add	ebp, [CRefNdxRefDR]
		mov	ebp, [ebp]
		imul	ebp, size wrefentry
		add	ebp, [CrossRef]
		cmp	[ebp.wref_called], edx
		ja	@@GoLeft
		jb	@@GoRight
		cmp	[ebp.wref_caller], ecx
		ja	@@GoLeft
		jb	@@GoRight		
		
		; Found
		mov	esi, ebx
		jmp	@@Found
		
		; Go left : search between [left, middle - 1]
@@GoLeft:	mov	edi, ebx
		dec	edi
		jmp	@@DichoSearch

		; Go right : search between [middle + 1, right]
@@GoRight:	mov	esi, ebx
		inc	esi
		jmp	@@DichoSearch
		
		
		; ----------------
		; Return to caller
		; ----------------
	
@@NotFound:	clc						; CF = 0
		jmp	@@Exit

@@Found:	stc						; CF = 1
		
@@Exit:		mov	ebx, esi
		pop	edi esi ebp eax
		ret

CRefSrch_DR	endp


; ===============================================================
; = Create_Line							=
; =								=
; = Usage:							=
; =	Create a new lineto store more references.		=
; =								=
; = I)	EDI : pointer to start of text				=
; =	EBP : pointer to buffer description (buffedentry)	=
; =	ECX : pointer on the word name				=
; =	EAX : nb of calls					=
; =								=
; = O)	EDI : is updated to the start of next line		=
; =	ESI : line #						=
; =								=
; = M)	EDX							=
; ===============================================================

		align	4
CreateLine	proc

		; --------------
		; Registers used
		; --------------
		
		push	ebx ecx esi
		

		; ------------------------------
		; Move to the start of next line
		; ------------------------------

		mov	ecx, [ebp.binf_CurLinAddr]		; Current line
		call	EdGetCurLinBuff
		mov	[ebp.binf_CurLinAddr], ecx		; Current line


		; --------------------
		; Choose 80 or 132 col
		; --------------------
		
		cmp	[EdColsNb], 80
		jne	@@L132
		lea	edx, @@Header_80
		mov	ebx, @@End80 - @@Header_80
		jmp	@@80or132
@@L132:		lea	edx, @@Header_132
		mov	ebx, @@End132 - @@Header_132
@@80or132:


		; -----------------
		; Ask for new space
		; -----------------
		
		mov	eax, ebx
		call	EdRequestMem
		jc	@@Exit
; !!!


		; --------------------------
		; Update the number of lines
		; --------------------------
		
		inc	[ebp.binf_LinesNb]
		

		; --------------------------------------------------
		; Move the text, make a 'hole' where header will fit
		; --------------------------------------------------
		
		; Source address
		call	EdCompTextSize					; Size of the text into EAX
		mov	esi, eax
		add	esi, [ebp.binf_BuffAddr]
		dec	esi						; ESI : last char of the text
		
		; Destination address
		mov	edi, esi
		add	edi, ebx
		
		; The nb of bytes to move
		mov	ecx, esi
		sub	ecx, [ebp.binf_CurLinAddr]
		inc	ecx

		; Move block
		std							; Dec ESI and EDI
		rep	movsb  			                	; Move/Copy the block
		cld
		

		; ------------------------------------
		; Now insert the header into the 'hole'
		; ------------------------------------

		; Source address
		mov	esi, edx
		
		; Destination address
		mov	edi, [ebp.binf_CurLinAddr]
		
		; The nb of bytes to move
		mov	ecx, ebx

		; Move block
		cld
		rep	movsb


		; -----------------------------------------------
		; Update the current position to store references 
		; -----------------------------------------------

		; In term of address		
		mov	edi, [ebp.binf_CurLinAddr]
		add	edi, 16 - 1
		
		; In term of line #
		inc	[ebp.binf_trow]


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	esi ecx ebx
		ret


		align	4
@@Header_80	db	"\                                                                            \", 13, 10
@@End80:

		align	4
@@Header_132	db	"\                                                                                                                               \", 13, 10
@@End132:

CreateLine	endp


; ===============================================================
; = DeleteLines							=
; =								=
; = Usage:							=
; =	Create a new lineto store more references.		=
; =								=
; = I)	EDI : pointer to start of text				=
; =	EBP : pointer to buffer description (buffedentry)	=
; =	ECX : pointer on the word name				=
; =	EAX : nb of calls					=
; =								=
; = O)	EDI : is updated to the start of next line		=
; =								=
; = M)	EDX							=
; ===============================================================

		align	4
DeleteLines	proc

		; --------------
		; Registers used
		; --------------
		
		push	eax ebx ecx edx edi esi
		
		
		; --------------------------
		; Delete lines for 'Callers'
		; --------------------------

@@DelLine:

		; Delete current line
		; -------------------

		; Destination address
		mov	edi, [ebp.binf_CurLinAddr]		; Current line addr
		
		; Source address
		mov	ecx, edi
		call	EdGetCurLinBuff				; Bring the current line into editing buffer
		mov	esi, ecx
		
		; Nb of characters to move
		mov	ecx, [ebp.binf_BuffAddr]		; Buffer addr
		call	EdCompTextSize				; Text size in EAX		
		add	ecx, eax
		sub	ecx, esi
		
		; Move text
		cld						; Inc ESI and EDI
		rep	movsb  			                ; Move/Copy the block

		
		; Get current line address
		; ------------------------
		
		mov	ecx, [ebp.binf_CurLinAddr]
		call	EdGetCurLinBuff
		
		
		; Update the nb of lines of the text
		; ----------------------------------
		
		dec	[ebp.binf_LinesNb]
		
		
		; Other line to delete ?
		; ----------------------
		
		mov	ecx, [ebp.binf_CurLinAddr]
		cmp	[ecx], '  \'
		je	@@DelLine
		

		; ---------------------------
		; Now store again empty lines
		; ---------------------------

			; --------------------
			; Choose 80 or 132 col
			; --------------------
		
		cmp	[EdColsNb], 80
		jne	@@L132
		lea	edx, @@Header_80
		mov	ebx, @@End80 - @@Header_80
		jmp	@@80or132
@@L132:		lea	edx, @@Header_132
		mov	ebx, @@End132 - @@Header_132
@@80or132:


			; -----------------
			; Ask for new space
			; -----------------
		
		mov	eax, ebx
		call	EdRequestMem
		jc	@@Exit
; !!!


			; --------------------------
			; Update the number of lines
			; --------------------------
		
		add	[ebp.binf_LinesNb], 2
		

			; --------------------------------------------------
			; Move the text, make a 'hole' where header will fit
			; --------------------------------------------------
		
		; Source address
		call	EdCompTextSize					; Size of the text into EAX
		mov	esi, eax
		add	esi, [ebp.binf_BuffAddr]
		dec	esi						; ESI : last char of the text
		
		; Destination address
		mov	edi, esi
		add	edi, ebx
		
		; The nb of bytes to move
		mov	ecx, esi
		sub	ecx, [ebp.binf_CurLinAddr]
		inc	ecx

		; Move block
		std							; Dec ESI and EDI
		rep	movsb  			                	; Move/Copy the block
		cld	
		

			; ------------------------------------
			; Now insert the header into the 'hole'
			; ------------------------------------

		; Source address
		mov	esi, edx
		
		; Destination address
		mov	edi, [ebp.binf_CurLinAddr]
		
		; The nb of bytes to move
		mov	ecx, ebx

		; Move block
		cld
		rep	movsb


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	esi edi edx ecx ebx eax
		ret


		align	4
@@Header_80	db	"\    Callers :                                                               \", 13, 10
		db	"\      Calls :                                                               \", 13, 10
@@End80:

		align	4
@@Header_132	db	"\    Callers :                                                                                                                  \", 13, 10
		db	"\      Calls :                                                                                                                  \", 13, 10
@@End132:

DeleteLines	endp
		

; ===============================================================
; = Update_Line							=
; =								=
; = Usage:							=
; =	Store a word into the line. Called only by 		=
; =	'Update_Callers' and 'Update_Calleds'.			=
; =								=
; = I)	EDI : pointer to start of text				=
; =	EBP : pointer to buffer description (buffedentry)	=
; =	ECX : pointer on the word name				=
; =	EAX : nb of calls					=
; =								=
; = O)	EDX : is updated to the start of next line		=
; =								=
; = M)	EDX							=
; ===============================================================

		align	4
Update_Line	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx


		; ---------------------------
		; Length of the word to store
		; ---------------------------
		
		clr	ebx					; Length
@@Length:	cmp	byte ptr [ecx + ebx], 0
		je	@@Length2
		inc	ebx
		jmp	@@Length		
@@Length2:

		
		; ------------------------
		; Copy word to work buffer
		; ------------------------

		; Word name
		push	ax
		clr	ebx
		mov	byte ptr [@@Ref], ' '
@@CopyWord:	mov	al, [ecx + ebx]
		cmpz	al, 0
		je	@@CopyWord2
		mov	[@@Ref + 1 + ebx], al
		inc	ebx
		jmp	@@CopyWord
@@CopyWord2:	pop	ax

		; Nb of references
		cmp	ax, 1
		ja	@@More1Ref
		
		; Only one reference
		jmp	@@CopyDone
		
		; At least, two references
@@More1Ref:	mov	byte ptr [@@Ref + 1 + ebx], ':'
		inc	ebx
		cmp	ax, 10
		jae	@@Not1Digit
		add	al, '0'
		mov	[@@Ref + 1 + ebx], al
		inc	ebx
		jmp	@@CopyDone

		; Two digits (10-99)
@@Not1Digit:	mov	cl, 10
		div	cl				; ax / cl   al=quo ah=rem
		add	al, '0'
		mov	[@@Ref + 1 + ebx], al
		inc	ebx
		add	ah, '0'
		mov	[@@Ref + 1 + ebx], ah
		inc	ebx
		
@@CopyDone:	mov	[@@Ref + 1 + ebx], 0
	

		; --------------
		; Line to long ?
		; --------------
		
		mov	ecx, edi
		sub	ecx, [ebp.binf_CurLinAddr]
		add	ecx, ebx
		inc	ecx
		cmp	[EdColsNb], 80
		jne	@@132
		cmp	cx, 79 - 3
		jbe	@@LengthOk
		jmp	@@LengthNOk
@@132:		cmp	cx, 130 - 3
		jbe	@@LengthOk

		
		; -----------------------------------
		; Line is too long, create a new line
		; -----------------------------------

@@LengthNOk:	call	CreateLine


		; --------------------
		; Store word into line
		; --------------------

@@LengthOk:

		clr	ebx		
@@StoreRef:	mov	al, [@@Ref + ebx]
		cmpz	al, 0
		je	@@StoreRef2
		mov	[edi + ebx], al
		inc	ebx
		jmp	@@StoreRef
@@StoreRef2:

		; Update position
		add	edi, ebx
		

		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	ebx eax
		ret

		align	4
@@Ref		db	WORDMAXLEN+1+4 dup (?)

Update_Line	endp
		

; ===============================================================
; = Update_Callers						=
; =								=
; = Usage:							=
; =	Update callers. Called only by Update_Text.		=
; =								=
; = I)	EDX : pointer to start of text				=
; =	EBP : pointer to buffer description (buffedentry)	=
; =	ECX : pointer on the word name				=
; =								=
; = O)	EDX : is updated to the start of next line		=
; =								=
; = M)	EDX							=
; ===============================================================

		align	4
Update_Callers	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx edi
		mov	edi, edx


		; -------------------------------------------
		; Is this word defined into the dictionnary ?
		; -------------------------------------------
		
		push	esi
		mov	esi, [PStackStart]				; Stack pointer for thread #0
		sub	esi, @WordSize
		mov	[esi], ecx
		call	PFind
		mov	eax, [esi]					; Get flag (word found ?)
		pop	esi
		cmpz	eax, 0
		jne	@@WordDef
		
		; Error: word not defined
		add	esp, (1 + 6 + 1 + 3) * @WordSize		; Clean stack (and cancel a call level)
		mov	ah, ERR_WORDNDEF				; Word not defined
		mov	edx, ecx					; Msg is the word name
		mov	ecx, [ebp.binf_trow]				; Line #
		call	EdDoMoveBeginTxt				; Restart on top document
		jmp	doError
		
@@WordDef:


		; -----------------------------------------------------
		; Find the 1st record of (Called, Caller) for this word
		; -----------------------------------------------------

		clr	ecx						; 1st word caller 
		mov	edx, ebx					; NIA pointer of the word called (PFIND stored in EBX)
		call	CRefSrch_DR

@@Call1:	cmp	ebx, [CrossRefNb]
		je	@@Call2
		mov	ecx, ebx
		shl	ecx, 2
		add	ecx, [CRefNdxRefDR]
		mov	ecx, [ecx]
		imul	ecx, size wrefentry
		add	ecx, [CrossRef]
		cmp	[ecx.wref_called], edx
		jne	@@Call2

		; Nb of calls
		movzx	eax, [ecx.wref_calls]

		; Pointer to caller word name
		mov	ecx, [ecx.wref_caller]
		add	ecx, [NIA_START]
		add	ecx, word_name

		; Update the text
		call	Update_Line

		; Next word caller, if any
		inc	ebx
		jmp	@@Call1
		
		; Done
@@Call2:

		; ----------------
		; Return to caller
		; ----------------

		pop	edi ebx eax
		ret
		
Update_Callers	endp
		

; ===============================================================
; = Update_Calleds						=
; =								=
; = Usage:							=
; =	Update calleds. Called only by Update_Text.		=
; =								=
; = I)	EDX : pointer to start of text				=
; =	EBP : pointer to buffer description (buffedentry)	=
; =	ECX : pointer on the word name				=
; =								=
; = O)	EDX : is updated to the start of next line		=
; =								=
; = M)	EDX							=
; ===============================================================

		align	4
Update_Calleds	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx edi
		mov	edi, edx


		; -------------------------------------------
		; Is this word defined into the dictionnary ?
		; -------------------------------------------
		
		push	esi
		mov	esi, [PStackStart]			; Stack pointer for thread #0
		sub	esi, @WordSize
		mov	[esi], ecx
		call	PFind
		mov	eax, [esi]				; Get flag (word found ?)
		pop	esi
		cmpz	eax, 0
		jne	@@WordDef
		
		; Error: word not defined
		add	esp, (1 + 6 + 1 + 3) * @WordSize	; Clean stack (and cancel a call level)
		mov	ah, ERR_WORDNDEF			; Word not defined
		mov	edx, ecx				; Msg is the word name
		mov	ecx, [ebp.binf_trow]			; Line #
		call	EdDoMoveBeginTxt			; Restart on top document
		jmp	doError
		
@@WordDef:


		; -----------------------------------------------------
		; Find the 1st record of (Caller, Called) for this word
		; -----------------------------------------------------

		mov	ecx, ebx				; NIA pointer of the caller word (PFIND stored in EBX)
		clr	edx					; 1st word called
		call	CRefSrch_RD

@@Call1:	cmp	ebx, [CrossRefNb]
		je	@@Call2
		mov	edx, ebx
		shl	edx, 2
		add	edx, [CRefNdxRefRD]
		mov	edx, [edx]
		imul	edx, size wrefentry
		add	edx, [CrossRef]
		cmp	[edx.wref_caller], ecx
		jne	@@Call2

		; Nb of calls
		movzx	eax, [edx.wref_calls]

		; Pointer to called word name
		mov	edx, [edx.wref_called]
		add	edx, [NIA_START]
		add	edx, word_name

		; Update the text
		push	ecx
		mov	ecx, edx
		call	Update_Line
		pop	ecx

		; Next word caller, if any
		inc	ebx
		jmp	@@Call1
		
		; Done
@@Call2:

		; ----------------
		; Return to caller
		; ----------------

		pop	edi ebx eax
		ret
		
Update_Calleds	endp
		

; ===============================================================
; = Update_Text							=
; =								=
; = Usage:							=
; =	Update a given text. Called only by UPDATE.		=
; =								=
; = I)	EDX : pointer to buffer description (buffedentry)	=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
Update_Text	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx ebp
		

		; --------
		; Pointers
		; --------

		mov	ebp, edx				; Pointer on buffer infos
		mov	edx, [ebp.binf_BuffAddr]		; Pointer to start of line


		; ---------------------------------------------
		; Take each line to search:  "\       Word : "
		; ---------------------------------------------
		
		mov	ecx, edx				; Start address
		mov	[ebp.binf_trow], 0			; Start at line #0
		
		; Get current line into work buffer
@@NextLine:	inc	[ebp.binf_trow]				; Next line #
		mov	ebx, ecx				; Save in EBX the starting addr of the line
		mov	edx, ecx				; Save in EDX the starting addr of the line
		call	EdGetCurLinBuff

		
		; End of text ?
		cmp	byte ptr [ecx], K_EOF
		je	@@EndText
		
		; Compute line length
		sub	ebx, ecx
		neg	ebx
		dec	ebx
		dec	ebx
		mov	[@@Len], bx

		; Line length not ok ?
		cmp	bx, 79
		je	@@LengthOk
		cmp	bx, 130
		jne	@@NextLine
@@LengthOk:

		; Header found ?
		clr	ebx					; Char counter
@@FindHeader:	mov	al, [@@StrStart + ebx]
		cmpz	al, 0
		je	@@HeaderFound
		cmp	al, [edx + ebx]
		jne	@@NextLine
		inc	ebx
		jmp	@@FindHeader
		
		; Header is found
@@HeaderFound:

		; Copy the word name
		clr	ebx
@@CopyWordName:	mov	al, [edx + ebx + 16]
		cmp	al, ' '
		je	@@CopyWordN2
		mov	[@@WordName + ebx], al
		inc	ebx
		jmp	@@CopyWordName
@@CopyWordN2:	mov	[@@WordName + ebx], 0

		; Update pointer to start of line "\    Callers : "
		mov	ecx, edx				; Start addr
		call	EdGetCurLinBuff				; +1 line	"  Word:
@@Stack:	call	EdGetCurLinBuff				; +2 line	" Stack:
		cmp	dword ptr [ecx], '  \'
		jne	@@Descrip
		cmp	dword ptr [ecx + 4], '    '
		je	@@Stack
@@Descrip:	call	EdGetCurLinBuff				; +3 line
		cmp	dword ptr [ecx], '  \'
		jne	@@Callers
		cmp	dword ptr [ecx + 4], '    '
		je	@@Descrip
@@Callers:	mov	edx, ecx				; New start addr (next line)
		mov	[ebp.binf_CurLinAddr], ecx
		
		; Is-it the good begin of line ?
		clr	ebx					; Char counter
@@FindRHdr:	mov	al, [@@StrCallers + ebx]
		cmpz	al, 0
		je	@@HdrRFound
		cmp	al, [edx + ebx]
		jne	@@HdrRNFound
		inc	ebx
		jmp	@@FindRHdr

		; Header not found: it's an error
@@HdrRNFound:	add	esp, (6) * @WordSize			; Clean stack
		mov	ah, ERR_BADWORDHEADER			; Invalid word header.
		lea	edx, @@WordName				; Additionnal msg is the word name
		mov	ecx, [ebp.binf_trow]			; Line #
		call	EdDoMoveBeginTxt			; Restart on top document
		jmp	doError
		
		; Header is found: clean lines where to store references
@@HdrRFound:	call	DeleteLines

		; Update the line: callers for this words
		inc	[ebp.binf_trow]
		mov	[@@SaveECX], ecx
		lea	ecx, @@WordName				; Pointer on word name
		mov	edx, [ebp.binf_CurLinAddr]
		add	edx, 16 - 1				; Store 1st ref here
		call	Update_Callers
		mov	ecx, [@@SaveECX]
		mov	ecx, [ebp.binf_CurLinAddr]		; Current line
		call	EdGetCurLinBuff
		mov	[ebp.binf_CurLinAddr], ecx		; Current line

		; Update the line: calls for this words
		inc	[ebp.binf_trow]
		mov	[@@SaveECX], ecx
		lea	ecx, @@WordName				; Pointer on word name
		mov	edx, [ebp.binf_CurLinAddr]
		add	edx, 16 - 1				; Store 1st ref here
		call	Update_Calleds
		mov	ecx, [@@SaveECX]

		; Print a dot (.) for each header processed
		mov	al, '.'
		call	COUT

		; Analyse now next line
		jmp	@@NextLine

		
		; -----------
		; End of text
		; -----------
		
@@EndText:


		; ----------------
		; Return to caller
		; ----------------

		pop	ebp edx ecx ebx eax
		ret

		align	4
@@Len		dw	?		
		align	4
@@StrStart	db	"\       Word : ", 0
		align	4
@@StrCallers	db	"\    Callers : ", 0
		align	4
@@WordName	db	WORDMAXLEN+1 dup (?)
		align	4
@@SaveECX	dd	?

Update_Text	endp


; ===============================================================
; = UPDATE                            				=
; =								=
; = Code for the forth word:    UPDATE                		=
; =								=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
UPDATE		proc

		; -------------------
		; Set 'Editor' active
		; -------------------

		mov	[EditorActive], 2


		; -------------------------------------
		; For each text into FROSTED buffer, do
		; -------------------------------------

		lea	edx, EdBuffInfos
		clr	cx

@@ForEachBuff:	cmp	byte ptr [edx.binf_FileName], 0
		je	@@NextBuff

		; Is this file swapped to disk ? If yes, we can't process it....
		cmp	[edx.binf_InMemory], FALSE
		je	@@NextBuff
		
		
		; -----------------
		; Analyse this text
		; -----------------

		; Print a msg		
		call	PDOTQ					; rt."
		db	13, 10
		db	"Analyse "
		db	0
		sub	esi, @WordSize
		mov	[esi], edx				; Filename address
		call	_TYPE
		
		; Analyse this text
		mov	[edx.binf_NotSaved], TRUE		; Modified and not saved ?
		call	Update_Text

		; Put current position on top of text
		push	ebp
		mov	ebp, edx
		call	EdDoMoveBeginTxt
		pop	ebp
		
		; Print a msg when done
		call	PDOTQ					; rt."
		db	"   done"
		db	0
		
		
		; ----------------------------------
		; Next text to analyse, if not empty
		; ----------------------------------
		
@@NextBuff:	add	edx, size buffedentry
		inc	cx
		cmp	cx, EDMAXNBBUFF
		jne	@@ForEachBuff


		; ----------------
		; Return to caller
		; ----------------
		
		mov	[EditorActive], 0
		ret

E_UPDATE:
UPDATE		endp

endif

code32		ends
		end
