; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	DICTMANG.ASM						!
; 									!
; Description:	Definitions for playing with the dictionnary.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	QCOMPILE		?COMPILE				!
;	QEXECUTE		?EXECUTE				!
;	BUILD			BUILD>					!
;	COLON			:					!
;	SEMICOLON		;					!
;	IMMEDIATE		IMMEDIATE				!
;	LITERAL			LITERAL					!
;	LBRACKET		[					!
;	RBRACKET		]					!
;	COMPILE			COMPILE					!
;	LPARENT			(					!
;	RPARENT			)					!
;	SLASHSLASH		//			\		!
;	WORDS			WORDS					!
;	XWORDS			XWORDS					!
;	VOCS			VOCS					!
;	VOCABULARY		VOCABULARY				!
;	CURRENT			CURRENT					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = isOptimize							=
; = 	*** This is a service function ***			=
; =                                                 		=
; = Usage:							=
; =    Return if ?OPTIMIZE is TRUE (cf=1), else cf=0.		=
; =								=
; = I) none							=
; = O) CF							=
; = M) none							=
; ===============================================================

		align	4
isOptimize	proc

		; Test flag
		push	eax
		mov	eax, [edi + ddir_blkaddr + D_QOPTIMIZE]
		cmp	byte ptr [eax], FALSE
		jne	@@On
		
@@Off:		pop	eax
		clc						; CF = 0
		ret						; jmp	@@Exit

@@On:		pop	eax
		stc						; CF = 1
		ret

isOptimize	endp


; ===============================================================
; = isQIArithOverfl						=
; = 	*** This is a service function ***			=
; =                                                 		=
; = Usage:							=
; =    Return if ?IARITHOVERFLOW is TRUE (cf=1), else cf=0.	=
; =								=
; = I) none							=
; = O) CF							=
; = M) none							=
; ===============================================================

		align	4
isQIArithOverfl	proc

		; Test flag
		push	eax
		mov	eax, [edi + ddir_blkaddr + D_QIARITHOVERFLOW]	; ?iarithoverflow
		cmp	byte ptr [eax], FALSE
		jne	@@On
		
@@Off:		pop	eax
		clc						; CF = 0
		ret						; jmp	@@Exit

@@On:		pop	eax
		stc						; CF = 1
		ret

isQIArithOverfl	endp


; ===============================================================
; = QPrev_WasNb							=
; = 	*** This is a service function ***			=
; =                                                 		=
; = Usage:							=
; =    Return (CF=1) if previous compiled is a literal number.	=
; =								=
; = I) EDX : current dictionnary pointer			=
; = O) CF							=
; =    EDX : is updated (backward) is was a number.		=
; =    EAX : the number.					=
; = M) none							=
; ===============================================================

		align	4
QPrev_WasNb	proc

		; Pointer to last word defined with :
		push	ecx
		mov	ecx, [VOC_CURRENT]
		add	ecx, [VOC_START]
		mov	ecx, [ecx.voc_lastword]
		add	ecx, [NIA_START]
		mov	ecx, [ecx.word_realaddr]
		sub	edx, 3+6				; 3 for 'sub esi, 4' /// 4 for 'mov dword ptr [esi], i32'
		cmp	edx, ecx
		jae	@@CanBe

@@NotANb:	pop	ecx
		add	edx, 3+6
		clc
		ret

@@CanBe:
		cmp	dword ptr [edx], 0c704ee83h		; 83 EE 04 ==> sub esi, 4
		jne	@@NotANb
		cmp	byte ptr [edx + 4], 06h			; C7 06 i32 ==> mov dword ptr [esi], i32
		jne	@@NotANb
		mov	eax, edx
		sub	eax, [CDA_START]
		cmp	eax, [AddrLastBeginIf]		; Adress of last Begin/If (do not optimize literal integer number ?)
		jb	@@NotANb
		mov	eax, [edx + 5]				; Get the 32-bits literal
		pop	ecx
		stc						; CF = 1
		ret

QPrev_WasNb	endp


; ===============================================================
; = QPrev_WasCall						=
; = 	*** This is a service function ***			=
; =                                                 		=
; = Usage:							=
; =    Return (CF=1) if previous compiled is a CALL word.	=
; =								=
; = I) EDX : current dictionnary pointer			=
; = O) CF							=
; =    EDX : is updated (backward) is was a number.		=
; =    EAX : the address (absolute).				=
; = M) none							=
; ===============================================================

		align	4
QPrev_WasCall	proc

		; Pointer to last word defined with :
		push	ecx
		mov	ecx, [VOC_CURRENT]
		add	ecx, [VOC_START]
		mov	ecx, [ecx.voc_lastword]
		add	ecx, [NIA_START]
		mov	ecx, [ecx.word_realaddr]
		sub	edx, 5				; 5 for 'CALL rel32'
		cmp	edx, ecx
		jae	@@CanBe

@@NotACall:	pop	ecx
		add	edx, 5
		clc
		ret

@@CanBe:
		cmp	byte ptr [edx], CALLr32_CODOP		; CALL rel32 ?
		jne	@@NotACall
		mov	eax, [edx + 1]				; Get the 32-bits relative branch
		add	eax, edx
		add	eax, 5					; Aboslute address
		cmp	eax, [LastWrdCompiled]			; Pointer to last word compiled
		jne	@@NotACall
		pop	ecx
		stc						; CF = 1
		ret

QPrev_WasCall	endp


; ===============================================================
; = QCOMPILE							=
; =                                                 		=
; = Code for the forth word:	?COMPILE			=
; =                     	                            	=
; = Stack:	( --- )						=
; =                     	                            	=
; = Usage:	Issue an error is system is not in 		=
; =		compile state.					=
; ===============================================================

		align	4
QCOMPILE	proc

		; Error is system is not in compile state
		; ---------------------------------------
		
		cmp	[CompileMode], TRUE
		je	@@Ok

		mov	ah, ERR_NOTINCOMP
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
@@Ok:		ret

E_QCOMPILE:
QCOMPILE	endp


; ===============================================================
; = QEXECUTE							=
; =                                                 		=
; = Code for the forth word:	?EXECUTE			=
; =                     	                            	=
; = Stack:	( --- )						=
; =                     	                            	=
; = Usage:	Issue an error is system is not in 		=
; =		execute state.					=
; ===============================================================

		align	4
QEXECUTE	proc

		; Error is system is not in execute state
		; ---------------------------------------
		
		cmp	[CompileMode], FALSE
		je	@@Ok

		mov	ah, ERR_NOTINEXEC
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError


		; Return to caller
		; ----------------
		
@@Ok:		ret

E_QEXECUTE:
QEXECUTE	endp


; ===============================================================
; = PFind							=
; =                                                 		=
; = Code for the forth word (FIND)				=
; =								=
; = found	: ( addr --- cfa tf )				=
; = not found	: ( addr --- ff )				=
; =                                                 		=
; = adr		: address of zstring				=
; = cfa		: code address, if found			=
; = f		: flag (0 or 1)					=
; =                                                 		=
; = O) EBX points to the word_entry of the found word.		=
; ===============================================================

		align	4
PFind		proc

		; Save regs used
		push	eax ecx edx edi

		; Pop address of the word
		mov	edi, PARAM0				; Word address in EDI

		; Pointer to the CONTEXT vocabulary
		mov	ecx, [VOC_CONTEXT]
		
		; Pointer to the last word of the vocab
@@LookVoc:	add	ecx, [VOC_START]
		mov	edx, [ecx.voc_lastword]
		cmp	edx, -1					; Vocab empty ?
		je	@@NextVoc
		
		; Init counter inside words names
@@ScanWord:	add	edx, [NIA_START]
		test	[edx.word_flags], FSMUDGE		; Smudge ?
		jnz	@@FF3					; Definition not endded
		clr	ebx
@@CmpNames:	mov	al, [edx.word_name + ebx]
		mov	ah, [edi + ebx]
		TOUPPER	al
		TOUPPER	ah
		cmp	al, ah
		jne	@@FF3
		or	al, ah
		jz	@@WordFound
		inc	ebx
		jmp	@@CmpNames
		
@@FF3:		; Strings not equal : goto previous word
		mov	edx, [edx.word_prevword]		; Address of the previous word
		cmp	edx, -1
		jne	@@ScanWord

		; Previous vocab, if any
@@NextVoc:	mov	ecx, [ecx.voc_fatherptr]
		cmp	ecx, -1
		jne	@@LookVoc				; Look the father of this vocab
		jmp	@@WordNotFound
		
		; Strings are equal -> word is found
@@WordFound:	mov	ecx, [edx.word_realaddr]
		mov	PARAM0, ecx				; Push CFA
		IPUSH	TRUE					; 1 (tf)
		mov	ebx, edx				; Output in EBX
		sub	ebx, [NIA_START]
		jmp	@@Exit

		; Word is not found
@@WordNotFound:	mov	PARAM0, FALSE				; 0 (ff)

		; Restore regs used
@@Exit:		pop	edi edx ecx eax

		; Return to caller
		ret

PFind		endp


; ===============================================================
; = BUILD							=
; =                                                 		=
; = Code for the forth word:	BUILD>				=
; =                     	                            	=
; = Stack:		( --- )					=
; =                           					=
; = Input stream:	build> name				=
; ===============================================================

		align	4
BUILD		proc

		; Pointer on the word NIA address		
		; -------------------------------
		
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		add	eax, [NIA_START]

		
		; 'BUILD>' already called ?
		; -------------------------
		
		test 	[eax.word_flags], FBUILDDOES
		jz	@@Ok

		; 'BUILD>' already called.
		mov	ah, ERR_BUILDALCALLED			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok:


		; 'BUILD>' the 1st word called for this word ?
		; --------------------------------------------
		
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		cmp	ebx, [eax.word_realaddr]
		je	@@OK2
		
		; "'BUILD>' must be the 1st definition compiled."
		mov	ah, ERR_BUILDNOTFIRST			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@OK2:


		; Set the bit 'BuildDoes' for the current word under definition
		; -------------------------------------------------------------

		or	[eax.word_flags], FBUILDDOES
		mov	[CDA_BUILD_CALLED], TRUE


		; Return to caller
		; ----------------
		
		ret

E_BUILD:
BUILD		endp


; ===============================================================
; = DynVarNotAll						=
; =                                                 		=
; = Compiled by doBuild only.					=
; =                     	                            	=
; = I) none							=
; = O) none							=
; ===============================================================

		align	4
DynVarNotAll	proc

		; Allocate the variable ?
		cmp	[AllocFlag], TRUE
		je	@@AllocVarOk

		
		; Error: "Dynamic variable not allocated"
		; ---------------------------------------
		
		; If not turnkey application, search the name if the variable
		clr	edx				; No additionnal msg
		cmp	[ExeFile], TRUE
		je	@@error

		mov	ebx, [esp]
		mov	ecx, [NIA_START]
		mov	eax, ecx
		add	eax, [NIA_SIZE]
@@SearchName:	cmp	ebx, [ecx.word_realaddr]
		jb	@@NextName
		cmp	ebx, [ecx.word_EndCode]
		jbe	@@FoundName
@@NextName:	add	ecx, size wordentry
		cmp	ecx, eax
		jb	@@SearchName
@@FoundName:	lea	edx, [ecx.word_name]

		; "Dynamic variable not allocated."
@@error:	mov	ah, ERR_DYNVARNOTALL		; Error #
		clr	ecx				; No line #
		jmp	doError


@@AllocVarOk:
		
		; Update address of the beginning of the dynamic variable
		mov	edx, [esp]
		mov	eax, edx
		sub	eax, 5+4+5			; 5:call_rel32  4:rel32  5:call_rel32
		mov	[DynVarAddr], eax

		; Get address of compile-time part
		mov	eax, [edx]
		
		; Now give control to compile-time part
		add	eax, [CDA_START]
		call	eax

		; Skip return level (call $+0)		
		add	esp, 4

		; Parse next word into INTERPRET
		add	esp, 4				; Skip a level return -> fall to: InterpretAddr
		ret

E_DynVarNotAll:
DynVarNotAll	endp


; ===============================================================
; = IsolateWordEx						=
; =                                                 		=
; = Usage:							=
; =	Isolate next word from input stream.			=
; =                     	                            	=
; = Note:							=
; =	Parsed word must be defined, else an error occurs.	=
; =                     	                            	=
; = I)	none							=
; = O)	EDX	 	ptr to word (NIA)			=
; = M)	all regs						=
; ===============================================================

		align	4
IsolateWordEx	proc

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; Can't create or parse a null word
		mov	ah, ERR_NULLWORD
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		je	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		jmp	@@LL4

		; Found the end of the word
@@LL5:		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx			; Current pointer into the buffer


		; -------------------------------------------
		; Is this word defined into the dictionnary ?
		; -------------------------------------------

		IPUSH	edx		
		call	PFind
		mov	eax, PARAM0				; Get flag (word found ?)
		DISCARD	1					; drop
		cmpz	eax, 0
		jne	@@WordDef
		
		; Error: word not defined
		mov	ah, ERR_WORDNDEF			; Word not defined
		clr	ecx					; No line #
		jmp	doError
		
@@WordDef:

		; Get word NIA address (in EDX)
		DISCARD	1					; drop
		mov	edx, ebx
		add	edx, [NIA_START]


		; ----------------
		; Return to caller
		; ----------------
		
		ret

IsolateWordEx	endp


; ===============================================================
; = doBuild							=
; =                                                 		=
; = Code for the forth word:	RT_BUILD>			=
; =                     	                            	=
; = Stack:		( --- )					=
; =                     	                            	=
; = Input stream:	create name				=
; =                     	                            	=
; = REGISTERS ARE PRESERVED.					=
; ===============================================================

		align	4
doBuild		proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx
		

		; ------------------------------
		; The last word is not smudged ?
		; ------------------------------

		mov	edx, [LastWordDefWColon]
		add	edx, [NIA_START]
		test	[edx.word_flags], FSMUDGE		; Smudge ?
		jz	@@Ok					; Definition endded

		; "Bad use of [ ... ]."
		mov	ah, ERR_BADUSELRBRACKET			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@Ok:		


		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; Can't create a null word
		add	esp, 4 * @WordSize			; Clean stack before error
		mov	ah, ERR_NULLWORD
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		je	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		cmp	al, 020h
		jb	@@InvWordName
		cmp	al, 07Fh
		jbe	@@LL4
		
		; "Invalide word name."
@@InvWordName:	add	esp, 4 * @WordSize		; Clean stack before error
		mov	ah, ERR_BADWORDNAME		; "Invalide word name."
		clr	ecx				; No line #
		call	doError

		; Found the end of the word
@@LL5:		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx		; Current pointer into the buffer


		; Is this word name too long ?
		; ----------------------------
		
		mov	eax, ecx
		sub	eax, edx				; EAX = length of the word name
		cmp	eax, WORDMAXLEN
		jae	@@InvWordName


		; Is this word already into the dictionnary ?
		; -------------------------------------------
		
		sub	esi, @WordSize
		mov	[esi], edx
		call	PFind
		mov	eax, [esi]				; Get flag (word found ?)
		add	esi, @WordSize
		cmpz	eax, FALSE
		je	@@WordNotDef
		call	QREDEF					; ?redefinitions
		call	BOOL_QTRUE				; true?
		lodsd
		cmpz	eax, FALSE
		je	@@WordAD
		DISCARD	1
		jmp	@@WordNotDef
		
		; Error: word already defined
@@WordAD:	add	esp, 4 * @WordSize			; Clean stack before error
		mov	ah, ERR_NOTREDEFWORD
		clr	ecx					; No line #
		call	doError
		
@@WordNotDef:	

		; Is this word defined as a vocabulary ?
		; --------------------------------------

		sub	esi, @WordSize
		mov	PARAM0, edx
		call	VocFind
		mov	eax, PARAM0				; Get flag (voc found ?)
		DISCARD	1
		cmpz	eax, FALSE
		je	@@VocNotDef
		
		; Error: vocabulary already defined
		add	esp, 4 * @WordSize			; Clean stack before error
		mov	ah, ERR_VOCALRDEFV			; Already defined as vocabulary.
		clr	ecx					; No line #
		call	doError
		
@@VocNotDef:	


		; Verify that it can't be a number
		; --------------------------------

		sub	esi, @WordSize
		mov	PARAM0, edx
		call	NUMBER
		lodsd
		cmpz 	eax, FALSE
		je	@@NotANumber
		
		; "Cannot redefine a number."
		add	esp, 4 * @WordSize			; Clean stack before error
		mov	ah, ERR_NOTREDEFNUMBER			; Error #
		clr	ecx					; No line #
		call	doError

@@NotANumber:	

		; ----------------------------------------------------------------
		; If ?CHAT is true, print a dot (.) (only if compiled from a file)
		; ----------------------------------------------------------------

		; ?chat
		call	QCHAT					; ?char
		call	BOOL_QTRUE				; true?		
		lodsd
		cmpz	eax, FALSE
		je	@@NotChat
		
		; Loading from a file ?
		cmp	[LoadFilesNb], 0		; Level of files which are loaded (0 means input stream is terminal)
		je	@@NotChat

		; Print '.'
		mov	al, '.'
		call	COUT

@@NotChat:	



		; -------------------------------------------
		; Create a new entry in NIA (Names Info Area)
		; -------------------------------------------
		
		; Add a new record
		mov	ebx, [NIA_START]
		add	ebx, [NIA_SIZE]				; EBX points to the new record
		
		; Always ends the list with -3
		mov	dword ptr [ebx + size wordentry], -3
		
		; Update the actual size
		add	[NIA_SIZE], size wordentry

		; Zero the record
		clr	eax
@@ZZ1:		mov	byte ptr [ebx + eax], 0
		inc	eax
		cmp	ax, size wordentry
		jl	@@ZZ1
		
		; Copy the word name
		clr	ecx					; Counter
@@CWN:		mov	al, [edx + ecx]
		mov	[ebx.word_name + ecx], al
		inc	ecx
		cmpz	al, 0
		jne	@@CWN
		

		; The word is not immediate, not system
		; -------------------------------------
		
		mov	[ebx.word_flags], 0


		; --------------------------------------------------------
		; Point to the last word defined in the current vocabulary
		; --------------------------------------------------------
		
		mov	ecx, [VOC_CURRENT]
		add	ecx, [VOC_START]
		cmp	[ecx.voc_lastword], -1
		jne	@@LW1
		
		; The vocabulary is empty of words
		; --------------------------------
		
		mov	[ecx.voc_lastword], ebx			; The last word of the current voc is the being defined word
		mov	[ebx.word_prevword], -1			; -1, because it's the 1st word defined for the voc
		jmp	@@LW2


		; The vocabulary contains already words
		; -------------------------------------
		
@@LW1:		mov	eax, [ecx.voc_lastword]			; Points to the actual last word of the voc
		add	eax, [NIA_START]

		; Chain to the next word is the new word (in creation)
		mov	edx, ebx
		sub	edx, [NIA_START]
		mov	[eax.word_nextword], edx

		; Point the previous word (next to last word)
		mov	edx, eax
		sub	edx, [NIA_START]
		mov	[ebx.word_prevword], edx
		
@@LW2:

		; Pointer to the next word
		mov	[ebx.word_nextword], -1			; -1, because it's the last word defined

		; Update in the vocab field, the number of last word if this vocab
		mov	edx, ebx
		sub	edx, [NIA_START]
		mov	[ecx.voc_lastword], edx
		
		; Update the number of words defined for the current vocabulary
		inc	[ecx.voc_wordsnb]
		

		; The vocabulary which 'owns' this word
		; -------------------------------------
		
		mov	eax, [VOC_CURRENT]
		mov	[ebx.word_vocptr], ax
		

		; Put the real code address into the NIA field
		; --------------------------------------------

		; Real code address		
		mov	ecx, [CDA_START]
		add	ecx, [CDA_SIZE]
		mov	[ebx.word_realaddr], ecx
		
		; Initialize the code size
		mov	[ebx.word_CodeSize], 0
		
		; Initialize the number of using this word
		mov	[ebx.word_NbUses], 0


		; ----------------------------------------------------------------------------
		; Update the filename/linenb for the file where this word is defined (if case)
		; ----------------------------------------------------------------------------

		; Loading from a file or from terminal ?
		; --------------------------------------
		
		cmp	[LoadFilesNb], 0
		jne	@@FromFile

		
		; Loading from terminal
		; ---------------------
		
		mov	byte ptr [ebx.word_FileName], 0
		mov	[ebx.word_LineNb], -1			; No info available
		jmp	@@FileDone
	
		
		; Loading from file		
		; -----------------
		
@@FromFile:
		; Pointer to file infos
		mov	edx, [LoadFilesNb]
		dec	edx
		imul	edx, size loadfileentry
		add	edx, [LoadFilesInfos]			; EDX : pointer to file infos
		
		; Line nb
		mov	eax, [edx.loadfile_linenb]
		dec	eax
		mov	[ebx.word_LineNb], eax
		
		; Find the end of the filename
		clr	ecx
@@FindEndFN:	cmp	byte ptr [edx.loadfile_name + ecx], 0
		je	@@FindEndFN2
		inc	ecx
		jmp	@@FindEndFN
@@FindEndFN2:

		; Find the beginning of the primary name ('/' or begin)
@@FindEndFN3:	cmpz	ecx, 0
		je	@@FindEndFN5
		cmp	byte ptr [edx.loadfile_name + ecx], '/'
		je	@@FindEndFN4
		cmp	byte ptr [edx.loadfile_name + ecx], '\'
		je	@@FindEndFN4
		dec	ecx
		jmp	@@FindEndFN3
@@FindEndFN4:	inc	ecx
@@FindEndFN5:	

		; Copy the primary filename
		push	esi
		mov	esi, ecx
		clr	ecx
		dec	ecx
		dec	esi
@@CopyFN:	inc	ecx
		inc	esi
		mov	al, [edx.loadfile_name + esi]
		mov	[ebx.word_FileName + ecx], al
		cmpz	al, 0
		jne	@@CopyFN
		pop	esi
		
@@FileDone:


ifdef NEVER
		; ---------------------------
		; Create a DYNAMIC variable ?
		; ---------------------------
		
		cmp	[DynamicFlag], FALSE
		je	@@NoDynamic

		; Put: 'CALL DynVarNotAll' (5 bytes)
		compCALL_rel32	DynVarNotAll

@@NoDynamic:
endif


		; ----------------
		; Return to caller
		; ----------------

		pop	edx ecx ebx eax		
		ret

E_doBuild:
doBuild		endp


; ===============================================================
; = doLocalBuild						=
; =                                                 		=
; = Usage:							=
; =	Create a definition into local dictionnary area (LDA).	=
; ===============================================================

		align	4
doLocalBuild	proc

		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; Can't create a null word
		mov	ah, ERR_NULLWORD
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		je	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		jmp	@@LL4

		; Found the end of the word
@@LL5:		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx			; Current pointer into the buffer


		; Is this word already into the dictionnary ?
		; -------------------------------------------

		IPUSH	edx		
		call	PFind
		mov	eax, PARAM0				; Get flag (word found ?)
		DISCARD	1
		cmpz	eax, 0
		je	@@WordNotDef
		call	QREDEF					; ?redefinitions
		call	BOOL_QTRUE				; true?		
		lodsd
		cmpz	eax, FALSE
		je	@@WordAD
		DISCARD	1
		jmp	@@WordNotDef
		
		; Error: word already defined
@@WordAD:	mov	ah, ERR_NOTREDEFWORD
		clr	ecx					; No line #
		call	doError
		
@@WordNotDef:	

		; Is this word defined as a vocabulary ?
		; --------------------------------------

		sub	esi, 4
		mov	[esi], edx
		call	VocFind
		mov	eax, [esi]				; Get flag (voc found ?)
		add	esi, 4
		cmpz	eax, 0
		je	@@VocNotDef
		
		; Error: vocabulary already defined
		mov	ah, ERR_VOCALRDEFV			; Already defined as vocabulary.
		clr	ecx					; No line #
		call	doError
		
@@VocNotDef:	

		; -----------------------------------------------------------
		; Is this word already defined into Local Dictionnary (LDA) ?
		; -----------------------------------------------------------
		
		; Search the word into local words table
		call	AutoFind
		jc	@@LWordNFound				; If word is not found

		; Automatic word already defined
		mov	ebx, ecx
		jmp	@@UpdateRec				; Just update the record


		; Ok, local word not already defined
@@LWordNFound:


		; ---------------------------------
		; Not too munch local definitions ?
		; ---------------------------------
		
		mov	eax, ecx
		mov	bx, size lwordentry
		idiv	bl
		cmp	al, LDA_MAXNB
		jne	@@MaxNbOk
		
		; Error: too many local words defined
		mov	ah, ERR_TOOMANYLWORDS			; "Too many local words defined."
		clr	ecx					; No line #
		call	doError

		; Max nb of local definitions is OK
@@MaxNbOk:
		

		; -------------------------
		; Create a new entry in LDA
		; -------------------------
		
		; Found a free record (in EBX)
		; ----------------------------
		
		; Update the actual size
		mov	ebx, [LDA_SIZE]
		add	ebx, size lwordentry
		mov	[LDA_SIZE], ebx

		; Record address is:
		sub	ebx, size lwordentry
		add	ebx, [LDA_START]

		; Zero the record
		clr	eax
@@ZZ1:		mov	byte ptr [ebx + eax], 0
		inc	eax
		cmp	eax, size lwordentry
		jl	@@ZZ1
		
		; Copy the word name
		clr	ecx					; Counter
@@CWN:		mov	al, [edx + ecx]
		mov	[ebx.lword_name + ecx], al
		inc	ecx
		cmpz	al, 0
		jne	@@CWN
		
		; Code nb for the run-time procedure
@@UpdateRec:	mov	eax, [CDA_CURR_WORD]
		add 	eax, word_runtaddr
		mov	[ebx.lword_rtaddr], eax
		

		; Put 'CALL rel32' to compile part of the definig word
		; ----------------------------------------------------
		
		CALL_rel32					; CALL rel32

		mov	eax, [CDA_CURR_WORD]
		add 	eax, [NIA_START]
		mov	eax, [eax.word_realaddr]
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 for REL32
		call	CodeComma				; Code,


		; Put the block nb address
		; ------------------------

		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		mov	[ebx.lword_blocknbad], eax


		; Allocate space for later Block Nb Addr storage
		; ----------------------------------------------
		
		mov	eax, -1			; IMPORTANT: -1, if memory not allocated, not free by FreeDynVar !
		call	CodeComma		; Code,


		; Return to caller
		; ----------------
		
		ret

E_doLocalBuild:
doLocalBuild	endp


; ===============================================================
; = COLON							=
; =								=
; = Code for the forth word:	:				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
COLON		proc

		; Verify if the system is is execute state
		; ----------------------------------------

		call	QEXECUTE
		
		
		; CTRL-BREAK not allowed
		; ----------------------
		
		inc	[NoCtrlBreak]


		; Create a new entry in the code dictionnary
		; ------------------------------------------

		call	doBuild		

		
		; Pointer to last word defined with :
		; -----------------------------------

		mov	edx, [VOC_CURRENT]
		add	edx, [VOC_START]
		mov	edx, [edx.voc_lastword]
		mov	[LastWordDefWColon], edx
	
		
		; Pointer to last word compiled
		; -----------------------------

		mov	[LastWrdCompiled], 0


		; Put the system in compilation mode
		; ----------------------------------
		
		mov	[CompileMode], TRUE
		

		; This is not a word built by a defining-word
		; -------------------------------------------

		add	edx, [NIA_START]
		mov	[edx.word_DefiningWord], 0		


		; Flags for 'BUILD> ... DOES>' construction
		; -----------------------------------------

		mov	[CDA_BUILD_CALLED], FALSE		; Is 'BUILD>' called ?		
		mov	[CDA_ALLOT_CALLED], FALSE		; Is 'ALLOT'  called ?
		mov	[CDA_DOES_CALLED], FALSE		; Is 'DOES>'  called ?
		mov	[REF_BLK_ALLCTED], -1			; Last block allocated by ALLOT
		
		
		; This is a colon-definition
		; --------------------------
		
		mov	edx, [LastWordDefWColon]
		add	edx, [NIA_START]
		or	[edx.word_flags], FCOLONDEF		; Colon definition
		or	[edx.word_flags], FSMUDGE		; Not endded


		; Local variables
		; ---------------

		mov	[LocVar], 0				; 0, 1 (1LOC), 2 (2LOC), 3 (3LOC), 4 (4LOC)


		; Adress of last Begin/If (do not optimize literal integer number ?)
		; ------------------------------------------------------------------

		mov	[AddrLastBeginIf], 0		; Adress of last Begin/If (do not optimize literal integer number ?)


		; Return to caller
		; ----------------
		
		ret

E_COLON:
COLON		endp


; ===============================================================
; = VrfCtrlStrPending						=
; =                                                 		=
; = Usage:							=
; =   Verify that there is no pending control structure.	=
; ===============================================================

		align	4
VrfCtrlStrPending proc

		mov	edx, [ControlStackPtr]
		cmp	edx, offset ControlStack
		je	@@ControlOk
		
		; Begin...Until ?
		cmp	dword ptr [edx], 'Begi'
		jne	@@NotBegin
		mov	ah, ERR_BEGINNEND			; BEGIN... structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotBegin:
		
		; Begin...While...Repeat ?
		cmp	dword ptr [edx], 'Whil'
		jne	@@NotWhile
		mov	ah, ERR_WHILENEND			; BEGIN...WHILE...REPEAT structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotWhile:
		
		; If...else...endif ?
		cmp	dword ptr [edx], 'IfSi'
		jne	@@NotIf
		mov	ah, ERR_IFNEND				; IF...ELSE...ENDIF structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotIf:
		
		; Times...Endtimes ?
		cmp	dword ptr [edx], 'Time'
		jne	@@NotTimes
		mov	ah, ERR_TIMESNEND			; TIMES...ENDTIMES structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotTimes:
		
		; Do...Enddo ?
		cmp	dword ptr [edx], 'DoLo'
		jne	@@NotDo
		mov	ah, ERR_DONEND				; DO...ENDDO structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotDo:
		
		; Make...EndMake ?
		cmp	dword ptr [edx], 'Make'
		jne	@@NotMake
		mov	ah, ERR_MAKENEND			; MAKE...ENDMAKE structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotMake:
		
		; Do...(not)When...Enddo ?
		cmp	dword ptr [edx], 'When'
		jne	@@NotDoWhen
		mov	ah, ERR_DOWHENNEND			; DO...(NOT)WHEN...ENDDO structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotDoWhen:
		
		; Case...Endcase ?
		cmp	dword ptr [edx], 'Case'
		jne	@@NotCase
		mov	ah, ERR_CASENEND			; CASE...ENDCASE structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotCase:
		
		; Case...Of...Endof...Endcase ?
		cmp	dword ptr [edx], '-OF-'
		jne	@@NotOf
		mov	ah, ERR_OFNEND				; CASE...OF...ENDOF...ENDCASE structure not ended
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError		
@@NotOf:
		
@@ControlOk:

		; return to caller
		ret

VrfCtrlStrPending endp


; ===============================================================
; = PurgeAutoVars						=
; =								=
; = Usage:							=
; =   Purge any automatic variables, when ending a definition.	=
; ===============================================================

		align	4
PurgeAutoVars	proc

		; At least one local definition called ?
		cmp	[LDA_SIZE], 0
		je	@@LocDefDone
		
		; Put 'CALL_rel32 FreeLocalVar'
		compCALL_rel32	FreeLocalVar
		
		; Store the nb of local definitions
		mov	eax, [LDA_SIZE]
		clr	edx
		mov	ecx, size lwordentry
		idiv	ecx
		call	CodeComma				; Code,
		
		; Store each memory ref nb
		mov	ebx, [LDA_SIZE]
@@StMemRN:	cmpz	bl, 0
		je	@@StMemRN2
		sub	ebx, size lwordentry
		mov	eax, ebx
		add	eax, [LDA_START]
		mov	eax, [eax.lword_blocknbad]
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		call	CodeComma				; CodeC
		jmp	@@StMemRN
@@StMemRN2:

		; Reset the nb of local definitions
		mov	[LDA_SIZE], 0

@@LocDefDone:

		; ----------------
		; Return to caller
		; ----------------
		
		ret

PurgeAutoVars	endp


; ===============================================================
; = SEMICOLON							=
; =                                                 		=
; = Code for the forth word:	;				=
; =                     	                            	=
; = Stack:	( --- )						=
; ===============================================================

		align	4
SEMICOLON	proc


		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; --------------------------------------------------------------		
		; Verify if there is not a control structure pending (not ended)
		; --------------------------------------------------------------		
		
		call	VrfCtrlStrPending


		; --------------------------------------------------------------
		; Verify is construction 'BUILD> ... ALLOT ... DOES>' is correct
		; --------------------------------------------------------------
		
		cmp	[CDA_BUILD_CALLED], FALSE
		je	@@NotBuildDoes

		; 'ALLOT' not called ?
		cmp	[CDA_ALLOT_CALLED], FALSE
		je	@@ErrBuildDoes
		
		; 'DOES>' not called ?
		cmp	[CDA_DOES_CALLED], FALSE
		jne	@@BuildDoesOk
		
		; Error: "'BUILD> ... ALLOT ... DOES>' construction is incorrect."
@@ErrBuildDoes:	mov	ah, ERR_BUILDDOESINC			; "'BUILD> ... ALLOT ... DOES>' construction is incorrect."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError	

@@BuildDoesOk:	
@@NotBuildDoes:


		; ---------------------------
		; Is the [] mode not closed ?
		; ---------------------------
		
		cmp	[CDA_BRACKET_MODE], FALSE
		je	@@BracketOk
		
		mov	ah, ERR_BRACKETNOTCLOSED		; [ mode not closed.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@BracketOk:


		; --------------------
		; Resolve pending EXIT
		; --------------------

@@ExitLoop:	mov	edx, [ExitStackPtr]
		cmp	edx, offset ExitStack
		je	@@ExitDone

		mov	ecx, [edx]
		mov	eax, [CDA_SIZE]
		sub	eax, ecx
		sub	eax, @WordSize				; 4 for rel32
		add	ecx, [CDA_START]

		; Change 'JMP rel32' to 'JMP rel8' ?
		cmp	eax, 128-3
		jae	@@ExitCont
		mov	byte ptr [ecx - 1], JMPrel8_CODOP
		add	al, 3					; 3 for the difference between rel32 and rel8
		or	eax, (NOP_CODOP shl 24) + (NOP_CODOP shl 16) + (NOP_CODOP shl 8)
@@ExitCont:	mov	[ecx], eax

		add	edx, @WordSize
		mov	[ExitStackPtr], edx
		jmp	@@ExitLoop
@@ExitDone:		


		; ------------------------------------------------
		; Local Dictionnary: purge any automatic variables
		; ------------------------------------------------

		call	PurgeAutoVars


		; ----------------------------------------------
		; Purge local variables if any (1LOC, 2LOC, ...)
		; ----------------------------------------------

		cmp	[LocVar], 0
		je	@@NoLocVar
		_LEAVE_						; Store 386 code op: LEAVE
		mov	[LocVar], 0				; 0, 1 (1LOC), 2 (2LOC), 3 (3LOC), 4 (4LOC)
@@NoLocVar:	


		; ------------------------
		; Put 'RET' code operation
		; ------------------------

		; Is last operation CALL word ?
		mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasCall				; CALL rel32 ?
		jne	@@put_ret
		mov	byte ptr [edx], JMPr32_CODOP
		jmp	@@ret_done
@@put_ret:	_RET_		
@@ret_done:

		; align 4
@@Align4:	mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		and	al, 3
		jz	@@Align4Done
		_NOP_
		jmp	@@Align4
@@Align4Done:


		; -----------------------------------------------------
		; Store the code size of the word which is endded there
		; -----------------------------------------------------

		; Pointer on the word NIA address		
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		add	eax, [NIA_START]

		; Compute the size of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		sub	ebx, [eax.word_realaddr]
		
		; Store the code size
		mov	[eax.word_CodeSize], bx
		
		; Update the end of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		dec	ebx
		mov	[eax.word_EndCode], ebx

		; Definition endded
		and	[eax.word_flags], NOT FSMUDGE
		mov	[LastWordDefWColon], 0


		; --------------------------------------------------
		; Put the system in immediate mode (not compilation)
		; --------------------------------------------------
		
		mov	[CompileMode], FALSE


		; ------------------
		; CTRL-BREAK allowed
		; ------------------
		
		dec	[NoCtrlBreak]


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SEMICOLON:
SEMICOLON	endp


; ===============================================================
; = IMMEDIATE							=
; =                                                 		=
; = Code for the forth word:	IMMEDIATE			=
; =                     	                            	=
; = Stack:	( --- )						=
; ===============================================================

		align	4
IMMEDIATE	proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; -----------------------------------------------
		; Set the last word defined (with :) as immediate
		; -----------------------------------------------
		
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		add	eax, [NIA_START]
		or	[eax.word_flags], FIMMEDIATE
		
		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_IMMEDIATE:
IMMEDIATE	endp


; ===============================================================
; = LITERAL							=
; =                                                 		=
; = Code for the forth word:	LITERAL				=
; =                     	                            	=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
LITERAL		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------------
		; Compile the number as literal
		; -----------------------------

		; Put 'SUB ESI, 4'
		SUB_ESI_i8	@WordSize
		
		; Put 'MOV DWORD PTR [ESI], IMM32'
		MOV_xESIi8_i32
		mov	eax, PARAM0
		DISCARD	1
		call	CodeComma               			; Code,


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_LITERAL:
LITERAL		endp


; ===============================================================
; = LBRACKET							=
; =                                                 		=
; = Code for the forth word:	[				=
; =                     	                            	=
; = Stack:	( --- )						=
; ===============================================================

		align	4
LBRACKET	proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; --------------------------------------------------
		; Put the system in immediate mode (not compilation)
		; --------------------------------------------------
		
		mov	[CompileMode], FALSE


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_LBRACKET:
LBRACKET	endp


; ===============================================================
; = RBRACKET							=
; =								=
; = Code for the forth word:	]				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
RBRACKET	proc

		; Verify if the system is is execute state
		; ----------------------------------------

		call	QEXECUTE
		
		
		; The last word is not smudged ?
		; ------------------------------

		mov	edx, [LastWordDefWColon]
		add	edx, [NIA_START]
		test	[edx.word_flags], FSMUDGE		; Smudge ?
		jnz	@@Ok					; Definition endded

		; "Bad use of [ ... ]."
		mov	ah, ERR_BADUSELRBRACKET			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@Ok:		

		
		; Put the system in compilation mode
		; ----------------------------------
		
		mov	[CompileMode], TRUE
		

		; Return to caller
		; ----------------
		
		ret

E_RBRACKET:
RBRACKET	endp


; ===============================================================
; = COMPILE							=
; =                                                 		=
; = Code for the forth word:	COMPILE				=
; =                     	                            	=
; = Stack:	( --- )						=
; ===============================================================

		align	4
COMPILE		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; Can't create a null word
		mov	ah, ERR_NULLWORD
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		je	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		jmp	@@LL4

		; Found the end of the word
@@LL5:		
		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx			; Current pointer into the buffer


		; -------------------------------------------
		; Is this word defined into the dictionnary ?
		; -------------------------------------------

		IPUSH	edx		
		call	PFind
		mov	eax, [esi]				; Get flag (word found ?)
		DISCARD	1
		cmpz	eax, FALSE
		jne	@@WordDef
		
		
		; Error: word not defined
		mov	ah, ERR_WORDNDEF			; Word not defined
		clr	ecx					; No line #
		call	doError
		
@@WordDef:

		; -----------------------------
		; Get word NIA address (in EDX)
		; -----------------------------
		
		add	esi, 4
		mov	edx, ebx
		add	edx, [NIA_START]

		
		; --------------------
		; Word not immediate ?
		; --------------------

		test	byte ptr [edx.word_flags], FIMMEDIATE
		jnz	@@Imm
		
		; "'COMPILE' can be used only with immediate words."
		mov	ah, ERR_COMPILENOTIMMW			; Error "
		add	edx, word_name
		clr	ecx					; No line #
		call	doError
		
@@Imm:
		; -----------------------
		; Is this word 'system' ?
		; -----------------------

		; Word 'system' ?
		test	byte ptr [edx.word_flags], FSYSTEM
		jz	@@NotSystem
		
		; "This system word can't be compiled inside user definitions."
		mov	ah, ERR_CANTBECOMPILED			; Error #
		add	edx, word_name
		clr	ecx					; No line #
		call	doError
		
@@NotSystem:

		; ----------------------------
		; Compile now the word address
		; ----------------------------
		
		CALL_rel32
		mov	eax, [edx.word_realaddr]
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 is for REL32
		call	CodeComma               		; Code,
	

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_COMPILE:
COMPILE		endp


; ===============================================================
; = LPARENT							=
; =                                                 		=
; = Code for the forth word:	(				=
; =                     	                            	=
; = Stack:		( --- )					=
; =                     	                            	=
; = Input stream:	( ... )					=
; ===============================================================

		align	4
LPARENT		proc

		; Just increment the comment level
		inc	[CDA_COMMENT]
		
		; Return to caller
		ret

E_LPARENT:
LPARENT		endp


; ===============================================================
; = RPARENT							=
; =                                                 		=
; = Code for the forth word:	)				=
; =                     	                            	=
; = Stack:		( --- )					=
; =                     	                            	=
; = Input stream:	( ... )					=
; ===============================================================

		align	4
RPARENT		proc

		; Decrement the comment level
		cmp	[CDA_COMMENT], 0
		jg	@@Ok

		; Bad right comment sign (lacks left signs: ')')
		mov	ah, ERR_BADCOMM				; Code # error
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
		; Decrement the comment level
@@Ok:		dec	[CDA_COMMENT]
		
		; Return to caller
		ret

E_RPARENT:
RPARENT		endp


; ===============================================================
; = SLASHSLASH							=
; =                                                 		=
; = Code for the forth word:	//				=
; =                     	                            	=
; = Stack:		( --- )					=
; =                     	                            	=
; = Input stream:	// ... 					=
; ===============================================================

		align	4
SLASHSLASH	proc

		mov	[CDA_SLASHSLASH], TRUE

		; Return to caller
		ret

E_SLASHSLASH:
SLASHSLASH	endp


; ===============================================================
; = XWORDS							=
; =                                                 		=
; = Code for the forth word:	XWORDS				=
; =                     	                            	=
; = Stack:		( --- )					=
; ===============================================================

		align	4
XWORDS		proc

		; ---------------------------------
		; Pointer to the CONTEXT vocabulary
		; ---------------------------------
		
		mov	ecx, [VOC_CONTEXT]
		

		; -----------------------
		; For each vocabulary met
		; -----------------------

@@PrVoc:
		
		; Print header
		; ------------
		
		; Vocabulary name
		add	ecx, [VOC_START]
		call	CR					; CR
		call	CR					; CR
		call	PDOTQ
		db	"Vocabulary: ", 0
		mov	eax, ecx
		add	eax, voc_name				; Voc name
		IPUSH	WORDMAXLEN
		DISPOSE	1
		mov	PARAM0, eax
		call	TYPED					; TYPE-
		
		; Nb of words
		call	PDOTQ
		db	"  Number of words: ", 0
		mov	eax, [ecx.voc_wordsnb]			; Nb of words
		IPUSH	eax
		IPUSH	4
		call	DOTR					; .R
		call	CR					; CR
		

		; Print each word for this vocabulary
		; -----------------------------------

		; Vocab empty ?		
		mov	edx, [ecx.voc_lastword]
		cmp	edx, -1					; Vocab empty ?
		jne	@@NotEmpty
		call	PDOTQ					; ."
		db	"<empty>", 0
		jmp	@@NextVoc
		
		; Vocab is not empty
@@NotEmpty:	clr	eax					; Char counter on the line

		; Print words
@@PrWordName:	add	edx, [NIA_START]

		; Print this word name only if not system
		push	edx
		add	edx, word_flags
		movzx	edx, byte ptr [edx]
		bt	dx, BSYSTEM
		pop	edx
		jc	@@PrevWord

		; Print this word name		
		push	ecx edx
		add	edx, word_name
		sub	esi, 4
		mov	[esi], edx
@@WordLen0:	inc	eax					; Count the 'space' bewteen each word
@@WordLen:	cmp	byte ptr [edx], 0
		je	@@WordLen2
		inc	eax
		inc	edx
		jmp	@@WordLen
@@WordLen2:	
		push	eax
		call	NBCOLS					; #COLS
		pop	eax
		mov	ecx, [esi]
		add	esi, 4
		cmp	eax, ecx
		jl	@@WordLen3
		call	CR					; CR
		clr	eax					; Reset counter on next line
		pop	edx
		push	edx					; Restore EDX
		jmp	@@WordLen0
@@WordLen3:	call	_TYPE					; TYPE
		call	SPACE					; SPACE
		pop	edx ecx
		
		; Previous word, if any
@@PrevWord:	mov	edx, [edx.word_prevword]		; Address of the previous word
		cmp	edx, -1
		jne	@@PrWordName
		

		; Next vocabulary, if any
		; -----------------------
		
@@NextVoc:	mov	ecx, [ecx.voc_fatherptr]
		cmp	ecx, -1
		jne	@@PrVoc					; Look the father of this vocab
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_XWORDS:
XWORDS		endp


; ===============================================================
; = WORDS							=
; =                                                 		=
; = Code for the forth word:	WORDS				=
; =                     	                            	=
; = Stack:		( --- )					=
; ===============================================================

		align	4
WORDS		proc

		; ---------------------------------
		; Pointer to the CONTEXT vocabulary
		; ---------------------------------
		
		mov	ecx, [VOC_CONTEXT]
		

		; -----------------------
		; For each vocabulary met
		; -----------------------

@@PrVoc:
		
		; Print header
		; ------------
		
		; Vocabulary name
		add	ecx, [VOC_START]
		call	CR					; CR
		call	CR					; CR
		call	PDOTQ
		db	"Vocabulary: ", 0
		mov	eax, ecx
		add	eax, voc_name				; Voc name
		IPUSH	WORDMAXLEN
		DISPOSE	1
		mov	PARAM0, eax
		call	TYPED					; TYPE-
		
		; Nb of words
		call	PDOTQ
		db	"  Number of words: ", 0
		mov	eax, [ecx.voc_wordsnb]			; Nb of words
		IPUSH	eax
		IPUSH	4
		call	DOTR					; .R
		call	CR					; CR
		

		; Print each word for this vocabulary
		; -----------------------------------

		; Vocab empty ?		
		mov	edx, [ecx.voc_lastword]
		cmp	edx, -1					; Vocab empty ?
		jne	@@NotEmpty
		call	PDOTQ					; ."
		db	"<empty>", 0
		jmp	@@NextVoc
		
		; Vocab is not empty
@@NotEmpty:	clr	eax					; Char counter on the line

		; Print words
@@PrWordName:	add	edx, [NIA_START]

		; Do no show run-time words
		test	[edx.word_flags], FRUNTIME
		jnz	@@PrevWord

		; Print this word name		
		push	ecx edx
		add	edx, word_name
		sub	esi, 4
		mov	[esi], edx
@@WordLen0:	inc	eax					; Count the 'space' bewteen each word
@@WordLen:	cmp	byte ptr [edx], 0
		je	@@WordLen2
		inc	eax
		inc	edx
		jmp	@@WordLen
@@WordLen2:	
		push	eax
		call	NBCOLS					; #COLS
		pop	eax
		mov	ecx, [esi]
		add	esi, 4
		cmp	eax, ecx
		jl	@@WordLen3
		call	CR					; CR
		clr	eax					; Reset counter on next line
		pop	edx
		push	edx					; Restore EDX
		jmp	@@WordLen0
@@WordLen3:	call	_TYPE					; TYPE
		call	SPACE					; SPACE
		pop	edx ecx
		
		; Previous word, if any
@@PrevWord:	mov	edx, [edx.word_prevword]		; Address of the previous word
		cmp	edx, -1
		jne	@@PrWordName
		

		; Next vocabulary, if any
		; -----------------------
		
@@NextVoc:	mov	ecx, [ecx.voc_fatherptr]
		cmp	ecx, -1
		jne	@@PrVoc					; Look the father of this vocab
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_WORDS:
WORDS		endp


if TURNKEYED eq 0
; ===============================================================
; = VOCS							=
; =                                                 		=
; = Code for the forth word:	VOCS				=
; =                     	                            	=
; = Stack:		( --- )					=
; ===============================================================

		align	4
VOCS		proc

		; ---------------
		; Initializations
		; ---------------

		; Print the nb of vocabularies defined
		call	PDOTQ
		db	13, 10
		db	"Nb of vocabulary defined: ", 0
		mov	eax, [VOC_NB]
		sub	esi, @WordSize
		mov	[esi], eax
		call	DOT					; .

		; Print the context vocabulary
		call	PDOTQ
		db	13, 10
		db	"Context vocabulary is: ", 0
		mov	eax, [VOC_CONTEXT]
		add	eax, [VOC_START]
		add	eax, voc_name				; Pointer to voc name
		sub	esi, @WordSize
		mov	[esi], eax				; CONTEXT @
		call	_TYPE					; TYPE
		
		; Print the current vocabulary
		call	PDOTQ
		db	13, 10
		db	"Current vocabulary is: ", 0
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		add	eax, voc_name				; Pointer to voc name
		sub	esi, @WordSize
		mov	[esi], eax				; CONTEXT @
		call	_TYPE					; TYPE
		
		; Init print loop		
		mov	edx, [VOC_START]				; Pointer on 1st vocab
		clr	ecx					; Vocab counter
		
		
		; -------------------
		; For each vocabulary
		; -------------------
		
@@ForEachVoc:	

		; This record is used
		inc	ecx

		; Print vocab name
		pushad
		call	CR					; CR
		mov	eax, edx
		add	eax, voc_name				; Voc name
		IPUSH	WORDMAXLEN
		DISPOSE	1
		mov	PARAM0, eax
		call	DTYPE					; -TYPE
		popad

		; Print the nb of words
		pushad
		call	PDOTQ
		db	"  Words:", 0
		mov	eax, [edx.voc_wordsnb]
		IPUSH	eax
		IPUSH	5
		call	DOTR					; .R
		popad
		
		; Print the vocab father
		pushad
		cmp	ecx, 1
		jne	@@NoRoot
		call	PDOTQ
		db	"          root", 0
		jmp	@@FatherDone
@@NoRoot:	call	PDOTQ
		db	"  Father: ", 0
		mov	eax, [edx.voc_fatherptr]
		cmp	eax, -1
		je	@@NoFather
		add	eax, [VOC_START]
		add	eax, voc_name				; Voc name
		jmp	@@PrFather
@@NoFather:	mov	eax, [QuoteStrBuff]
		add	eax, size strentry
		mov	byte ptr [eax], 0
@@PrFather:	sub	esi, @WordSize
		mov	[esi], eax
		call	_TYPE					; TYPE
@@FatherDone:	popad


		; ------------------		
		; Next vocab, if any
		; ------------------
		
@@NextVoc:	add	edx, size vocentry		
		cmp	ecx, [VOC_NB]
		jne	@@ForEachVoc
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_VOCS:
VOCS		endp
endif


; ===============================================================
; = VocFind							=
; =                                                 		=
; = Usage:							=
; =	Find (search) a vocabulary name.			=
; =                     	                            	=
; = Froth stack: 	( adrw --- adrv tf ) if found		=
; =		( adrw --- ff )      if not found		=
; =                     	                            	=
; = Regs modified: none        	                           	=
; ===============================================================

		align	4
VocFind		proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	ax ebx ecx edx edi ebp


		; ---------------------------------
		; Search the name into vocabularies
		; ---------------------------------

		; Get name to test
		mov	edx, [esi]

		; Pointer to the CONTEXT vocabulary
		mov	ebp, [VOC_CONTEXT]

		; Examine all the sons of this vocabulary
@@SearchSons:	mov	edi, [VOC_START]
		clr	ecx					; Voc counter
@@NextSon:	cmp	[edi.voc_fatherptr], ebp
		jne	@@NotEqu

		; Is-it this vocab ?
		clr	ebx					; Counter
@@CmpNames:	mov	al, [edx + ebx]
		mov	ah, [edi.voc_name + ebx]
		TOUPPER	al
		TOUPPER	ah
		cmp	al, ah
		jne	@@NotEqu
		or	al, ah
		jz	@@Found
		inc	ebx
		jmp	@@CmpNames

		; Search next son, if any
@@NotEqu:	add	edi, size vocentry
		inc	ecx
		cmp	ecx, [VOC_NB]
		jne	@@NextSon

		
		; Previous vocab, if any
@@PrevVoc:	add	ebp, [VOC_START]
		mov	ebp, [ebp.voc_fatherptr]
		cmp	ebp, -1
		jne	@@SearchSons				; Examine each son for this vocabulary


		; Is-it the vocabulary FROTH ?
		mov	edi, [VOC_START]			; 1st vocabulary is FROTH
		clr	ebx					; Counter
@@CmpNames2:	mov	al, [edx + ebx]
		mov	ah, [edi.voc_name + ebx]
		TOUPPER	al
		TOUPPER	ah
		cmp	al, ah
		jne	@@NotFound
		or	al, ah
		jz	@@Found
		inc	ebx
		jmp	@@CmpNames2


		; -----------------
		; The name is found
		; -----------------
		
@@Found:		; Push vocab address
		sub	edi, [VOC_START]
		mov	[esi], edi
		
		; Push tf (true flag)
		sub	esi, @WordSize
		mov	dword ptr [esi], TRUE
		
		; Exit
		jmp	@@Exit
		
		
		; ---------------------
		; The name is NOT found
		; ---------------------
		
@@NotFound:	; Push ff (false flag)
		mov	dword ptr [esi], FALSE
		
		
		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	ebp edi edx ecx ebx ax
		ret

VocFind		endp


; ===============================================================
; = VOCABULARY							=
; =                                                 		=
; = Code for the forth word:	VOCABULARY			=
; =                     	                            	=
; = Input stream: vocabulary name				=
; ===============================================================

		align	4
VOCABULARY	proc

		; ----------------------------------------
		; Verify if the system is is execute state
		; ----------------------------------------

		call	QEXECUTE
		
		
		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; Can't create a null word
		mov	ah, ERR_NULLWORD
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		je	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		jmp	@@LL4

		; Found the end of the word
@@LL5:		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx		; Current pointer into the buffer


		; -------------------------------------------
		; Is this word already into the dictionnary ?
		; -------------------------------------------
		
		sub	esi, @WordSize
		mov	[esi], edx
		call	PFind
		mov	eax, [esi]			; Get flag (word found ?)
		add	esi, @WordSize
		cmpz	eax, FALSE
		je	@@WordNotDef
		
		; Error: word already defined
		mov	ah, ERR_VOCALRDEFW		; "Already defined as word, cannot be redefined as vocabulary."
		clr	ecx				; No line #
		call	doError
		
@@WordNotDef:


		; --------------------------------------------
		; Is this word already defined as vocabulary ?
		; --------------------------------------------
		
		sub	esi, @WordSize
		mov	[esi], edx
		call	VocFind
		mov	eax, [esi]			; Get flag (word found ?)
		add	esi, @WordSize
		cmpz	eax, FALSE
		je	@@VocNotDef
		
		; Error: word already defined
		mov	ah, ERR_VOCALRDEFV		; "Already defined as vocabulary."
		clr	ecx				; No line #
		call	doError
		
@@VocNotDef:


		; -------------------------
		; Create now the vocabulary
		; -------------------------

		; Vocab table full ?
		mov	eax, [VOC_NB]
		cmp	eax, VOC_MAXSIZE / size vocentry
		jne	@@NotFull
		mov	ah, ERR_VOCTBLFULL		; Vocabulary full
		clr	ecx				; No line #
		call	doError
@@NotFull:	

		; Found a free record
		; -------------------
		
		; At the end
		mov	ecx, [VOC_NB]
		imul	ecx, size vocentry
		add	ecx, [VOC_START]


		; Create now the vocabulary record
		; --------------------------------
		
@@CreateVoc:	

		; Vocabulary name
		clr	ebx
@@CopyName:	mov	al, [edx + ebx]
		mov	[ecx.voc_name + ebx], al
		inc	ebx
		cmpz	al, 0
		jne	@@CopyName

		; Father pointer
		mov	eax, [VOC_CURRENT]			; Current vocabulary
		mov	[ecx.voc_fatherptr], eax
		
		; Pointer to last word
		mov	[ecx.voc_lastword], -1
		
		; Nb of words defined
		mov	[ecx.voc_wordsnb], 0
		
		; Update the nb of vocabulary
		inc	[VOC_NB]
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_VOCABULARY:
VOCABULARY	endp


; ===============================================================
; = CURRENT							=
; =                                                 		=
; = Code for the forth word:	CURRENT				=
; =                     	                            	=
; = Stack: ( --- )						=
; =                     	                            	=
; = Usage:							=
; =	Make context vocabulary current.			=
; ===============================================================

		align	4
CURRENT		proc

		; -------------------------------
		; Make Context vocabulary Current
		; -------------------------------
		
		mov	eax, [VOC_CONTEXT]
		mov	[VOC_CURRENT], eax
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_CURRENT:
CURRENT		endp


code32		ends
		end
