; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	DICTMNG2.ASM						!
; 									!
; Description:	Definitions for playing with the dictionnary.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	FORGET			FORGET					!
;	TICK		TICK						!
;	DTICK		DTICK						!
;	EXECUTE		EXECUTE						!
;	RECURSE		RECURSE						!
;	MYSELF		MYSELF						!
; 	DEFERRED	DEFERRED					!
; 	DEFER		DEFER						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


if TURNKEYED eq 0
; ===============================================================
; = UpdtWordsCalled						=
; =                                                 		=
; = Usage:							=
; =	Update each word called by a forgotten word.		=
; =                     	                            	=
; = I) EDX : pointer on NIA address				=
; =                     	                            	=
; = O) none							=
; =                     	                            	=
; = M) none							=
; =                     	                            	=
; ===============================================================

		align	4
UpdtWordsCalled	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	ax ebx ecx edx
		

		; ------------------------------------
		; Which words the word to forget calls
		; ------------------------------------

		mov	ecx, edx
		sub	ecx, [NIA_START]			; The caller word
		clr	edx					; The 1st word called
		call	CRefSrch_RD
		
		
		; -----------------------------------------
		; Update the nb of uses of each word called
		; -----------------------------------------

@@Loop:		cmp	ebx, [CrossRefNb]
		je	@@NoMore
		mov	edx, ebx
		shl	edx, 2
		add	edx, [CRefNdxRefRD]
		mov	edx, [edx]
		imul	edx, size wrefentry
		add	edx, [CrossRef]
		cmp	[edx.wref_caller], ecx
		jne	@@NoMore
		
		; Now update this called word
		mov	ax, [edx.wref_calls]
		mov	edx, [edx.wref_called]
		add	edx, [NIA_START]
		sub	[edx.word_NbUses], ax

		; Update words reference (delete record, then update index)
		sub	edx, [NIA_START]
		call	CRef_del
		
		; Next word called, if any
;		inc	ebx
		jmp	@@Loop
		
@@NoMore:
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	edx ecx ebx ax
		ret

UpdtWordsCalled	endp
endif


; ===============================================================
; = doForgetWord						=
; =                                                 		=
; = Usage							=
; =	Forget a given word. Called by FORGET.			=
; =                     	                            	=
; = I) EDX : Word NIA address					=
; = O) none							=
; = M) none							=
; ===============================================================

		align	4
doForgetWord	proc

		; ---------
		; Regs used
		; ---------
		
		push	eax ebx ecx edx

		
		; -------------------
		; Is this word used ?
		; -------------------

		; Word in use ?
		cmp	[edx.word_NbUses], 0
		je	@@NotUsed
		
		; Error: word is used
		add	esp, 4 * 4				; Clean stack on error
		mov	ah, ERR_WORDISUSED			; Word is used
		add	edx, word_name				; word name
		clr	ecx					; No line #
		jmp	doError
		
@@NotUsed:
		

		; ---------------------------------------------------
		; Update the vocabulary which owns the word forgetten
		; ---------------------------------------------------
		
		; Update the nb of words
		movzx	ecx, [edx.word_vocptr]
		add	ecx, [VOC_START]
		dec	[ecx.voc_wordsnb]
		
		; Update the pointer to last word ?
		mov	ebx, edx
		sub	ebx, [NIA_START]
		cmp	[ecx.voc_lastword], ebx
		jne	@@NpUpdtLastW
		mov	ebx, [edx.word_prevword]
		mov	[ecx.voc_lastword], ebx
@@NpUpdtLastW:	


		; -----------------------------------------------------------------
		; Update the link (prev/next words) between words of the vocabulary
		; -----------------------------------------------------------------
		
		; Save EDI
		push	esi edi

		; Get pointer on prev/next word
		mov	esi, [edx.word_prevword]		; Prev word
		mov	edi, [edx.word_nextword]		; Next word
		mov	ebx, edx
		sub	ebx, [NIA_START]

		; 1st and only word for this vocabulary ?
		cmp	esi, -1
		je	@@UpdtLinkDone
		
		; Word(s) after this one ?
		cmp	edi, -1
		je	@@LastWord
		
		; There are words BEFORE and AFTER the word to forget
		mov	eax, esi
		add	eax, [NIA_START]
		mov	[eax.word_nextword], edi
		mov	eax, edi
		add	eax, [NIA_START]
		mov	[eax.word_prevword], esi
		jmp	@@UpdtLinkDone
		
		; Word to forget is the last, but there is (are) word(s) before it
@@LastWord:	mov	eax, esi
		add	eax, [NIA_START]
		mov	[eax.word_nextword], edi
		
		; Update link done
@@UpdtLinkDone:
		
		; Restore EDI
		pop	edi esi


		; ---------------------------------------------------
		; Free data memory allocated for this word (by ALLOT)
		; ---------------------------------------------------
		
		call	FreeData


if TURNKEYED eq 0
		; ---------------------------------------------
		; Update each word called by the word forgotten
		; ---------------------------------------------

		call	UpdtWordsCalled
endif


		; ---------------------
		; Now release this word
		; ---------------------

		; Update the memory used
		movzx	eax, [edx.word_CodeSize]
		sub	[CDA_SIZE], eax
		
		; Update the size of the table for NIA
		sub	[NIA_SIZE], size wordentry
		
		; Always ends the list with -3
		mov	dword ptr [edx + size wordentry], -3
		

		; ----------------
		; Return to caller
		; ----------------

		pop	edx ecx ebx eax
		ret

doForgetWord	endp


; ===============================================================
; = doKillVoc							=
; =                                                 		=
; = Usage							=
; =	Kill a given empty vocabulary. Called by doForgetWord	=
; =                     	                            	=
; = I) EDX : Voc address					=
; = O) none							=
; = M) none							=
; ===============================================================

		align	4
doKillVoc	proc

		; --------------
		; Save used regs
		; --------------
		
		push	esi edi eax ebx ecx edx
		

		; --------------------
		; Move vocabulary area
		; --------------------

		; Destination address
		mov	edi, edx
		add	edi, [VOC_START]
		
		; Source address
		mov	esi, edi
		add	esi, size vocentry
		
		; Nb of bytes to move
		mov	ecx, [VOC_NB]
		dec	ecx
		imul	ecx, size vocentry
		sub	ecx, edx

		; Move data
		cld       	             			; Inc SI and DI
		rep	movsb  			            	; Move the block


		; ----------------------------------------------------
		; Update each word which owns to any vocabulary killed
		; ----------------------------------------------------

		; Initialization: start with last word defined
		mov	edi, edx
		mov	edx, [NIA_START]
		add	edx, [NIA_SIZE]
		sub	edx, size wordentry
		
		; Forget word, if word is not deleted (by DUST)
@@Loop:		test	[edx.word_flags], FSYSTEM
		jnz	@@Stop					; We reached a system word

		; Does this word belong to a voc which has moved
		cmp	[edx.word_vocptr], di
		jb	@@Loop2
		sub	[edx.word_vocptr], size vocentry
@@Loop2:
		
		; Continue until we reach a system word
		sub	edx, size wordentry
		jmp	@@Loop
@@Stop:


		; ----------------
		; Return to caller
		; ----------------
		
		pop	edx ecx ebx eax edi esi
		ret

doKillVoc	endp


; ===============================================================
; = doForgetVoc							=
; =                                                 		=
; = Usage							=
; =	Forget a given vocabulary. Called by FORGET.		=
; =                     	                            	=
; = I) EDX : Voc address					=
; =                     	                            	=
; = O) none							=
; =                     	                            	=
; = M) none							=
; ===============================================================

		align	4
doForgetVoc	proc

		; ---------------------------
		; Save EDI which is used here
		; ---------------------------
		
		push	esi edi


		; ----------------------------------------------------------
		; Scan each vocabulary, delete each which have EDX as parent
		; ----------------------------------------------------------

		; Initializations
		lea	ebx, @@VocsList
		mov	[ebx], edx
		add	ebx, @WordSize

		; Voc nb counter
		clr	ecx
		
@@ScanVoc:	
		; Voc pointer
		imul	esi, ecx, size vocentry
		
		; Is this vocabulary parent with EDX ?
		mov	edi, esi
		add	edi, [VOC_START]
@@TestVoc:	cmp	[edi.voc_fatherptr], edx
		je	@@OK
		mov	edi, [edi.voc_fatherptr]		; Get parent of this voc
		cmp	edi, -1					; Root reached ?
		je	@@NextVoc
		add	edi, [VOC_START]
		jmp	@@TestVoc
		
@@OK:		; Update the list of vocabularies to forget
		mov	[ebx], esi				; Record voc #
		add	ebx, @WordSize

		; Next vocabulary to test, if any
@@NextVoc:	inc	ecx
		cmp	ecx, [VOC_NB]
		jne	@@ScanVoc
		
		; End list of vocs to forget with -1
		mov	dword ptr [ebx], -1


		; -----------------------------------------------------------------------
		; Now, scan each word, and forget it if it owns to any of a voc to forget
		; -----------------------------------------------------------------------

		; Initialization: start with last word defined
		mov	edx, [NIA_START]
		add	edx, [NIA_SIZE]
		sub	edx, size wordentry
		
		; Forget word, if word is not deleted (by DUST)
@@Loop:		test	[edx.word_flags], FSYSTEM
		jnz	@@Stop					; We reached a system word

		; Does this word belong to a voc to forget ?
		lea	ecx, @@VocsList
		movzx	eax, [edx.word_vocptr]
@@WordBelongVoc:	cmp	dword ptr [ecx], -1
		je	@@Loop2
		cmp	eax, [ecx]
		je	@@ForgetWord
		add	ecx, @WordSize
		jmp	@@WordBelongVoc

@@ForgetWord:	call	doForgetWord

@@Loop2:
		
		; Continue until we reach a system word
		sub	edx, size wordentry
		jmp	@@Loop
@@Stop:


		; ------------------------
		; Kill now each vocabulary 
		; ------------------------

		lea	ecx, @@VocsList
		clr	ebx
@@KillVoc:	mov	edx, [ecx]
		cmp	edx, -1
		je	@@KillVoc2
		call	doKillVoc
		inc	ebx
		add	ecx, @WordSize
		jmp	@@KillVoc
@@KillVoc2:


		; -----------------------------
		; Update the nb of vocabularies
		; -----------------------------
		
		sub	[VOC_NB], ebx

		
		; ----------------------------------------------------
		; Call automatically DUST (Garbage collector for data)
		; ----------------------------------------------------

		pop	edi esi		
		call	DUST
		

		; ----------------
		; Return to caller
		; ----------------

		ret

; ----------------
; Local variables:
; ----------------

		align	4
@@VocsList	dd	MAXNB_VOCS+1 dup (?)		; List of voc ptr (to forget)

doForgetVoc	endp


; ===============================================================
; = FORGET							=
; =                                                 		=
; = Code for the forth word:	FORGET				=
; =                     	                            	=
; = Input stream: forget <name>					=
; =                     	                            	=
; = Usage:							=
; =	Forget all froth words defined after (or equal) the	=
; =	given word. Does a garbage collection (only if the	=
; =	word given exists).					=
; ===============================================================

		align	4
FORGET		proc

		; ----------------------------------------
		; Verify if the system is is execute state
		; ----------------------------------------

		call	QEXECUTE
		
		
		; ---------------------------
		; Garbage collector
		; Discard all local variables
		; ---------------------------

		call	DUST
		call	DiscardLocalVars
		call	DUST


		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; Can't create a null word
		mov	ah, ERR_NULLWORD
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		je	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		jmp	@@LL4

		; Found the end of the word
@@LL5:		
		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx			; Current pointer into the buffer


		; -------------------------------------------
		; Is this word defined into the dictionnary ?
		; -------------------------------------------
		
		sub	esi, @WordSize
		mov	[esi], edx
		call	PFind
		mov	eax, [esi]				; Get flag (word found ?)
		add	esi, @WordSize
		cmpz	eax, FALSE
		jne	@@WordDef
		

		; --------------------------------------
		; Is this word defined as a vocabulary ?
		; --------------------------------------

		sub	esi, @WordSize
		mov	[esi], edx
		call	VocFind
		mov	eax, [esi]				; Get flag (voc found ?)
		add	esi, @WordSize				; drop
		cmpz	eax, FALSE
		je	@@VocNotDef
		
		; It's a vocabulary: try to forget it
		mov	eax, [VOC_CONTEXT]
		mov	edx, [esi]				; Pop voc addr
		add	edx, [VOC_START]
		cmp	[edx.voc_fatherptr], eax
		je	@@VocOk

		mov	ah, ERR_VOCNFORG			; "This vocabulary can't be forgotten (not son of context)"
		add	edx, voc_name				; Additionnal msg is voc name
		clr	ecx					; No line #
		call	doError

@@VocOk:		mov	edx, [esi]
		add	esi, @WordSize				; drop
		call	doForgetVoc
		jmp	@@Exit
		
@@VocNotDef:	

		; Word not defined: do nothing
		; ----------------------------
		
		jmp	@@Exit
		
@@WordDef:

		; -----------------------
		; Is this word 'system' ?
		; -----------------------

		; Get word NIA address (in EDX)
		add	esi, 4
		mov	edx, ebx
		add	edx, [NIA_START]
		
		; Word 'system' ?
		test	byte ptr [edx.word_flags], FSYSTEM
		jz	@@NotSystem
		
		; Error: word is system
		mov	ah, ERR_WORDSYSTEM			; Word is system
		add	edx, word_name
		clr	ecx					; No line #
		call	doError
		
@@NotSystem:


		; --------------------------------------------------------
		; Forget all words defined after (or equal) the given word
		; --------------------------------------------------------

		; Save given word NIA address
		mov	ecx, edx

		; Initialization: start with last word defined
		mov	edx, [NIA_START]
		add	edx, [NIA_SIZE]
		sub	edx, size wordentry
		
		; Forget word, if word is not deleted (by DUST)
@@Loop:		call	doForgetWord
		
		; Continue with previous word ?
		cmp	edx, ecx
		je	@@Stop
		sub	edx, size wordentry
		jmp	@@Loop
@@Stop:


		; ----------------------------------------------------
		; Call automatically DUST (Garbage collector for data)
		; ----------------------------------------------------

		call	DUST
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_FORGET:
FORGET		endp


; ===============================================================
; = TICK							=
; =                                                 		=
; = Code for the forth word:	TICK				=
; =                     	                            	=
; = 	   Stack:	( --- addr )				=
; = Input stream:	' word					=
; =                     	                            	=
; = Usage:	Push the code address of the given word		=
; =		(from input stream).				=
; ===============================================================

		align	4
TICK		proc


		; -----------------------------------------------
		; Parse next word. Must be defined, else an error
		; -----------------------------------------------

		call	IsolateWordEx				; NIA word in EDX

		; Get word code address
		mov	ebx, edx
		mov	eax, [edx.word_realaddr]
		IPUSH	eax

		cmp	[CompileMode], TRUE
		je	@@CompileState
		

		; -------------
		; Execute state
		; -------------

		jmp	@@done


		; -------------
		; Compile state
		; -------------

@@CompileState:	

		; Saved NIA address of called word
		push	ebx		

		; Put 'SUB ESI, 4'
		SUB_ESI_i8	@WordSize
		
		; Put 'MOV DWORD PTR [ESI], IMM32'
		MOV_xESIi8_i32
		mov		eax, PARAM0
		DISCARD		1
		call		CodeComma               	; Code,


if TURNKEYED eq 0
		; Update word cross ref
		; ---------------------

		; Point to the last word defined in the current vocabulary: word currently being defined
		mov	ecx, [VOC_CURRENT]
		add	ecx, [VOC_START]
		mov	ecx, [ecx.voc_lastword]			; NIA pointer to caller word

		; Update word cross ref, only if it's not a system word
		pop	edx
		test	[edx.word_flags], FSYSTEM
		jnz	@@System2
		sub	edx, [NIA_START]
		call	CRef_add
@@System2:
endif
		

		; ----------------
		; Return to caller
		; ----------------
		
@@done:		ret

E_TICK:
TICK		endp


; ===============================================================
; = DTICK							=
; =                                                 		=
; = Code for the forth word:	DTICK				=
; =                     	                            	=
; = 	   Stack:	( --- addr )				=
; = Input stream:	' word					=
; =                     	                            	=
; = Usage:	Push the code address of the given word		=
; =		(from input stream).				=
; ===============================================================

		align	4
DTICK		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; ----------------
		; Return to caller
		; ----------------
		
@@done:		ret

E_DTICK:
DTICK		endp


; ===============================================================
; = EXECUTE							=
; =                                                 		=
; = Code for the forth word:	EXECUTE				=
; =                     	                            	=
; = Stack:	( addr --- )					=
; =                     	                            	=
; = Usage:	Execute the word whcih the address (code)	=
; =		is given.					=
; ===============================================================

		align	4
EXECUTE		proc


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------------
		; Execute code at given address
		; -----------------------------
		
		mov	eax, PARAM0			; get <addr>
		DISCARD	1				; drop
		jmp	eax

E_EXECUTE:
EXECUTE		endp


; ===============================================================
; = RECURSE							=
; =                                                 		=
; = Code for the forth word:	RECURSE				=
; =                     	                            	=
; = Stack:	( --- )						=
; =                     	                            	=
; = Usage:	Call current definition				=
; ===============================================================

		align	4
RECURSE		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; Return to caller
		; ----------------
		
		ret

E_RECURSE:
RECURSE		endp


; ===============================================================
; = MYSELF							=
; =                                                 		=
; = Code for the forth word:	MYSELF				=
; =                     	                            	=
; = Stack:	( --- )						=
; =                     	                            	=
; = Usage:	Jump current definition				=
; ===============================================================

		align	4
MYSELF		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; Return to caller
		; ----------------
		
		ret

E_MYSELF:
MYSELF		endp


; ===============================================================
; = DEFERRED							=
; =                                                 		=
; = Code for the forth word:	DEFERRED			=
; =                     	                            	=
; = 	   Stack:	compile-time:	( --- )			=
; =			    run-time:	( --- )			=
; = Input stream:	deferred word				=
; =                     	                            	=
; = Usage:	To defer to executer definitions.		=
; ===============================================================

		align	4
DEFERRED	proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------

		; 1 ALLOT DROP
		IPUSH	1
		call	RT_ALLOT
		DISCARD	1

		; DOES>		(this ends the compilation sequence)
		call	PDOES
		dd	R_DEFERRED

		
		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_DEFERRED:	

		; Return to caller
		ret

E_DEFERRED:
DEFERRED	endp


; ===============================================================
; = DEFER							=
; =                                                 		=
; = Code for the forth word:	DEFER				=
; =                     	                            	=
; = 	   Stack:	( --- )					=
; = Input stream:	defer word1 word2			=
; =                     	                            	=
; = Usage:	Set deferred <word2> to point to <word1>.	=
; ===============================================================

		align	4
DEFER		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; Return to caller
		; ----------------
		
		ret

E_DEFER:
DEFER		endp


code32		ends
		end
