; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	DIS.ASM							!
; 									!
; Description:	Definitions for playing with the 386 disassembler.	!
; 									!
; List of FROTH definitions defined in this module:			!
;	DISASM		.DIS						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc                   		; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc
if TURNKEYED eq 0


; ---> Local variables, *static* to this module ---> -----------------------------------------------------------------------------

		align	4
OpSize_over	db	?				; Operande-size override flag (066h)
	
		align	4
AddrSize_over	db	?				; Address-size override flag (067h)
	
; <--- end of local variables <--- -----------------------------------------------------------------------------------------------


; ===============================================================
; = DISASM							=
; =                                                 		=
; = Code for the forth word:	.DIS				=
; =                     	                            	=
; = Input stream:	.DIS word				=
; ===============================================================

		align	4
DISASM		proc

		call	EdPDotQ
		db	13, 10
		db	"Warning: .DIS is not yet finished..."
		db	13, 10, 0

		; --------------
		; Temp variables
		; --------------

t_EndAddr	equ	dword ptr [ebp - 4]
		enter	1 * @WordSize, 0


		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; NIA word in EDX
		call	IsolateWordEx


		; ----------------------------------
		; Dis-asm the code of the given word
		; ----------------------------------

		; Start/End address for this definition
		mov	eax, [edx.word_realaddr]
		movzx	ebx, [edx.word_CodeSize]
		add	ebx, eax
		dec	ebx
		mov	[t_EndAddr], ebx			; Ending address
		mov	edx, eax				; Starting address

		; Dis-assemble the code
@@dis:		call	CR
		mov	eax, edx
		IPUSH	eax
		call	Z8HDOT					; 08H.
		IPUSH	':'
		call	EMIT
		call	SPACE		
		call	do_dis
		cmp	edx, [t_EndAddr]
		jbe	@@dis


		; ----------------
		; Return to caller
		; ----------------

		leave		
		ret

E_DISASM:
DISASM		endp


; ---------------------------------------------------------
; Table of instructions of 1 byte w/o prefix, w/o parameter
; ---------------------------------------------------------

NEW	MACRO	codop, name
	db	0&codop&h
	db	"&name", 0
	ENDM


InsTbl_1b_npp:	NEW	37, AAA
		NEW	3F, AAS
		NEW	98, CBW
		NEW	99, CDQ
		NEW	F8, CLC
		NEW	FC, CLD
		NEW	FA, CLI
		NEW	F5, CMC
		NEW	27, DAA
		NEW	2F, DAS
		NEW	F4, HLT
		NEW	CC, INT3
		NEW	CE, INTO
		NEW	CF, IRET
		NEW	9F, LAHF
		NEW	C9, LEAVE
		NEW	F0, LOCK
		NEW	AD, LODSD
		NEW	90, NOP
		NEW	C3, RET
		NEW	9E, SAHF
		NEW	F9, STC
		NEW	FD, STD
		NEW	FB, STI
		NEW	9B, WAIT
		NEW	0E, PUSH_CS
		NEW	1E, PUSH_DS
		NEW	06, PUSH_ES
		NEW	16, PUSH_SS
		NEW	1F, POP_DS
		NEW	07, POP_ES
		NEW	17, POP_SS
		NEW	60, PUSHAD
		NEW	61, POPAD
		NEW	9C, PUSHFD
		NEW	9D, POPFD
		db	0

is_1b_npp	proc

		push	esi

		mov	cl, [edx]				; Byte to search
		mov	esi, offset InsTbl_1b_npp

@@Search:	lodsb
		cmpz	al, 0
		je	@@NotFound				; End of table reached
		cmp	al, cl
		je	@@InstrFound

@@SkipInstr:	lodsb
		cmpz	al, 0
		jne	@@SkipInstr
		jmp	@@Search

@@NotFound:	pop	esi
		clc
		ret

@@InstrFound:	clr	ecx
@@Show:		lodsb
		cmpz	al, 0
		je	@@ShowEnd
		call	COUT					; Print AL
		inc	ecx					; Char counter
		jmp	@@Show
@@ShowEnd:	
		pop	esi
		stc
		inc	edx
		ret

is_1b_npp	endp

purge	NEW


; ------------------------------------------
; Table of instructions of relative branches
; ------------------------------------------

NEW1	MACRO	codop, deplsize, name
	db	1, 0&codop&h
	db	deplsize
	db	"&name", 0
	ENDM

NEW2	MACRO	codop1, codop2, deplsize, name
	db	2, 0&codop1&h, 0&codop2&h
	db	deplsize
	db	"&name", 0
	ENDM


InsTbl_relbr:	NEW1	EB, 1, JMP
		NEW1	E9, 4, JMP
		NEW1	E8, 4, CALL
		NEW1	7A, 1, JA
		NEW1	73, 1, JAE
		NEW1	72, 1, JB
		NEW1	76, 1, JBE
		NEW1	74, 1, JZ
		NEW2	0F, 84, 4, JZ
		db	0

is_relbranch	proc

		push	esi

		mov	cx, [edx]				; Byte/Word to search
		mov	esi, offset InsTbl_relbr

@@Search:	lodsb
		cmpz	al, 0
		je	@@NotFound				; End of table reached
		mov	bl, al					; Save codop length (1 or 2 bytes)
		cmp	al, 2
		je	@@TwoBytes
@@OneByte:	lodsb
		cmp	al, cl
		je	@@InstrFound
		jmp	@@SkipInstr
@@TwoBytes:	lodsw
		cmp	ax, cx
		je	@@InstrFound

@@SkipInstr:	lodsb						; Skip rel. branch size
@@SkipInstr2:	lodsb
		cmpz	al, 0
		jne	@@SkipInstr2
		jmp	@@Search

@@NotFound:	pop	esi
		clc
		ret

@@InstrFound:	inc	edx					; Skip codop
		cmp	bl, 1
		je	@@skipcodop
		inc	edx
@@skipcodop:
		lodsb						; Get rel. branch size
		mov	bl, al
		clr	ecx
@@Show:		lodsb
		cmpz	al, 0
		je	@@ShowEnd
		call	COUT					; Print AL
		inc	ecx					; Char counter
		jmp	@@Show
@@ShowEnd:	pop	esi

@@tab:		inc	ecx
		call	SPACE
		cmp	cl, 8
		jne	@@tab

		cmp	bl, 1
		jne	@@rel16or32

@@rel8:		clr	eax
		mov	al, [edx]				; Get rel8
		inc	edx					; Skip rel8
@@ShowAddr:	add	eax, edx
		IPUSH	eax
		call	Z8HDOT					; 08H.
		jmp	@@exit

@@rel16or32:	cmp	[OpSize_over], 1
		je	@@rel16

@@rel32:	mov	eax, [edx]				; get rel32
		add	edx, 4					; Skip rel32
		pushad
		DISPOSE	1
		add	edx, eax
		call	WHICH2
		cmp	PARAM0, -1
		jne	@@ShowWordName
		popad
		jmp	@@ShowAddr
@@ShowWordName:	call	_TYPE
		popad
		jmp	@@exit

@@rel16:	clr	eax
		mov	ax, [edx]				; get rel16
		add	edx, 2					; Skip rel16
		jmp	@@ShowAddr

@@exit:		stc
		ret

is_relbranch	endp

purge	NEW1
purge	NEW2


; ----------------------------------------
; Table of instructions of 1 byte with + r
; ----------------------------------------

print_reg	proc

		clr	eax
		mov	al, [edx]
		and	al, 00000111b				; Get 3-bits for register

		cmp	bl, 8
		je	@@reg8
	
		cmp	[OpSize_over], 1
		je	@@reg16

@@reg32:	lea	eax, [offset @@r32 + eax*4 ]
		jmp	@@reg_done
@@reg16:	lea	eax, [offset @@r16 + eax*4 ]
		jmp	@@reg_done
@@reg8:		lea	eax, [offset @@r8 + eax*4 ]
@@reg_done:	push	esi
		mov	esi, eax
@@print:	lodsb
		cmpz	al, 0
		je	@@exit
		call	COUT
		jmp	@@print
@@exit:		pop	esi 

		ret

@@r32:		db	"EAX", 0, "ECX", 0, "EDX", 0, "EBX", 0
		db	"ESP", 0, "EBP", 0, "ESI", 0, "EDI", 7

@@r16:		db	"AX", 0, 0, "CX", 0, 0, "DX", 0, 0, "BX", 0, 0
		db	"SP", 0, 0, "BP", 0, 0, "SI", 0, 0, "DI", 0, 0

@@r8:		db	"AL", 0, 0, "CL", 0, 0, "DL", 0, 0, "BL", 0, 0
		db	"AH", 0, 0, "CH", 0, 0, "DH", 0, 0, "BH", 0, 0

print_reg	endp


NEW	MACRO	codop, freg8, name
	db	0&codop&h
	db	freg8
	db	name, 0
	ENDM


InsTbl_1b_pr:	NEW	40, 0, "INC_r"
		NEW	48, 0, "DEC_r"
		NEW	50, 0, "PUSH_r"
		NEW	58, 0, "POP_r"
		NEW	90, 0, "XCHG_EAX, r"
		NEW	B0, 8, "MOV_r, 1"
		NEW	B8, 0, "MOV_r, 4"
		db	0

is_1b_pr	proc

		enter	4, 0
t_esi		equ	[ebp - 4]
		mov	[t_esi], esi

		mov	cl, [edx]				; Byte to search
		and	cl, 11111000b				; Mask the 3 last bits (registers)
		mov	esi, offset InsTbl_1b_pr

@@Search:	lodsb
		cmpz	al, 0
		je	@@NotFound				; End of table reached
		cmp	al, cl
		je	@@InstrFound

@@SkipInstr:	lodsb						; Skip flag is reg8
@@SkipInstr2:	lodsb
		cmpz	al, 0
		jne	@@SkipInstr2
		jmp	@@Search

@@NotFound:	mov	esi, [t_esi]
		leave
		clc
		ret

@@InstrFound:	clr	ecx
		lodsb						; Skip flag is reg8
		mov	bl, al
@@Show:		lodsb
		cmpz	al, 0
		je	@@ShowEnd
		cmp	al, '_'
		jne	@@NoTab
@@Tab:		call	SPACE
		inc	ecx
		cmp	cl, 8
		jne	@@Tab
		jmp	@@Show
@@NoTab:	cmp	al, 'r'
		jne	@@no_r
		call	print_reg
		jmp	@@Show
@@no_r:		cmp	al, '1'
		jne	@@no_1
		clr	eax
		inc	edx
		mov	al, [edx]
@@pr_nb:	xchg	esi, [t_esi]
		IPUSH	eax
		call	DOT
		xchg	esi, [t_esi]
		jmp	@@Show
@@no_1:		cmp	al, '4'
		jne	@@print
		cmp	[OpSize_over], 1
		je	@@reg16
@@reg32:	mov	eax, [edx+1]
		add	edx, 4
		jmp	@@pr_nb
@@reg16:	clr	eax
		mov	ax, [edx+1]
		add	edx, 2
		jmp	@@pr_nb
@@print:	call	COUT					; Print AL
		inc	ecx					; Char counter
		jmp	@@Show
@@ShowEnd:	
		mov	esi, [t_esi]
		leave
		stc
		inc	edx
		ret

is_1b_pr	endp

purge	NEW


; ===============================================================
; = do_dis							=
; =								=
; = Usage:							=
; =	Disassmble next instruction. Update EDX (pointer	=
; =	to next instruction).					=
; =								=
; = I) EDX : pointer to instruction to disassemble.		=
; =								=
; = O) EDX : pointer to instruction to disassemble.		=
; =								=
; = M) all registers.						=
; ===============================================================

		align	4
do_dis		proc

		; Clear status
		mov	[OpSize_over], 0
		mov	[AddrSize_over], 0

		; Segment overrride prefixe ?
		mov	al, [edx]
		cmp	al, 066h
		jne	@@no_66
		inc	[OpSize_over]
		inc	edx
		jmp	@@overr_done
@@no_66:	cmp	al, 067h
		jne	@@no_67
		inc	[AddrSize_over]
		inc	edx
		jmp	@@overr_done
@@no_67:
@@overr_done:

		; Is-it a single instruction 1-byte, w/o prefix, w/o parameter ?
		call	is_1b_npp
		jc	@@exit

		; Is-it a relative branch ?
		call	is_relbranch
		jc	@@exit

		; Is-it a single instruction with +r ?
		call	is_1b_pr
		jc	@@exit

		; Dont' know, just print the byte in hex
		mov	al, '['
		call	COUT
		clr	eax
		mov	al, [edx]
		IPUSH	eax
		call	Z2HDOT				; 02H.
		mov	al, ']'
		call	COUT
		inc	edx

		; Exit procedure
@@exit:		ret

do_dis		endp

endif

code32		ends
		end

