
; ===============================================
; ===					      ===
; === Constants and Structurs used by FROSTED ===
; ===					      ===
; ===============================================

	; ----------------------
	; Diverses compile flags
	; ----------------------
	
__NOHSCROLL__	equ	1


	; ---------------------
	; Max number of buffers
	; ---------------------
	
EDMAXNBBUFF	equ	40					; F1-F10  Ctrl-Shift-Alt (10 x 4)


	; ------------------------
	; Default size for buffers
	; ------------------------

DEFSZ_EDBUFF	equ	(256*1024)				;  256k
MINSZ_EDBUFF	equ	(64*1024)				;   64k mini
MAXSZ_EDBUFF	equ	(1024*1024)				; 1024k maxi

DEFSZ_SCRAPBUFF	equ	(16*1024)
MINSZ_SCRAPBUFF	equ	(8*1024)
MAXSZ_SCRAPBUFF	equ	(256*1024)

DEFSZ_EDMACROS	equ	(8*1024)				; 8K



	; -------------------------
	; Max size of a line edited
	; -------------------------

EDMAXSZLINE	equ	132					; Max length of a line
MAXHEIGHTSCREEN	equ	60					; Max height of the screen


	; ------------------------------------
	; Initial and Step size of each buffer
	; ------------------------------------
	
INITSZBUFFED	equ	16
STEPSZBUFFED	equ	1024


	; ------------------
	; Max number of tags
	; ------------------
	
MAXNBTAGS	equ	4					; F11  Shift-F11  Ctrl-F11  Alt-F11


	; --------------------
	; Max number of macros
	; --------------------
	
MAXNBMACROS	equ	26					; From Alt-A to Alt-Z


	; --------------------
	; Buffers informations
	; --------------------

buffedentry	struc	
	binf_FileName	db	MAXSZFILENAME+1 dup (0)		; FileName (complete, with path)
	binf_PrimFName	db	MAXSZFNAME+1+MAXSZEXT+1 dup (0)	; Primary file (without path)
	binf_BuffAddr	dd	?				; Address of buffer allocated dynamically for the edited text
	binf_CurrSize	dd	?				; Buffer current size, BEFORE calling 'EdPutCurLinBuff'
	binf_BuffAllSz	dd	?				; Allocated size
	binf_LinesNb	dd	?				; Actual nb of lines
	binf_1stLinAddr	dd	?				; Address of top screen line (1st line drawed on top)
	binf_CurLinAddr	dd	?				; Current line address (in the edited text)
	binf_trow	dd	?				; Current row # in text
	binf_tcol	dd	?				; Current col # in text
	binf_frow	dd	?				; First row # drawed (1st line of the screen)
	binf_fcol	dd	?				; Position of the first column shown (horizontal scrolling)
	binf_NotSaved	db	?				; Modified and not saved ?
	binf_BlockBegin	dd	?				; Selected block begin address, or -1 if no selected block
	binf_BlockEnd	dd	?				; Selected block end address
	binf_hdl	dw	?				; File handle (temporary)
	binf_seek	dd	?				; Temp variable
	binf_Tags_Row	dd	MAXNBTAGS dup (0)		; Tags (line)     (used with F11 / F12)
	binf_Tags_Col	dd	MAXNBTAGS dup (0)		; Tags (col)      (used with F11 / F12)
	binf_InMemory	db	?				; Text currently in memory ? (-> not to be reloaded)
	binf_CurrSize_S	dd	?				; Saved, when InMemory is FALSE
	binf_BuffAllSz_S dd	?				; Saved, when InMemory is FALSE
	binf_garb	db	3 dup (0)			; To have a record size of B0h bytes (176)
buffedentry	ends


	; -------------------------
	; Max size of prompt answer
	; -------------------------
	
MAXSZPROMPT	equ	80


	; ----------------------------------
	; Errors codes, specifics to FROSTED
	; ----------------------------------
	
EDERRLINTOOLONG	equ	10001					; Line too long.
EDERRNOCRLF	equ	10002					; Line separator must be CR-LF.	
EDERRCTRLCHAR	equ	10003					; Control characters non allowed
EDERRBADNB	equ	10004					; Invalid number
EDERRNOMEM	equ	10005					; Buffer memory full
EDERRSELTOOBIG	equ	10006					; Selection too big to fit inside scrap
EDERRFILETWICE	equ	10007					; A file can't be open twice.


	; -----------------------------------------------
	; Step moving right/left for horizontal scrolling
	; -----------------------------------------------
	
STEPMOVHSCROLL	equ	16


	; ---------------------------
	; Column position for comment (This is a default value, stored at init in 'ED_CommentField'
	; ---------------------------

XCOMMENT_DFLT	equ	63
XCOMMENT_MIN	equ	40
XCOMMENT_MAX	equ	72


	; ----------------------------------------
	; Line position on screen when goto a line (this is a default value, stored in 'ED_LinePosAdjust' at init)
	; ----------------------------------------

LINEPOSADJUST		equ	10
LINEPOSADJUST_MIN	equ	5
LINEPOSADJUST_MAX	equ	15
	

	; -----------
	; Help screen
	; -----------
	
HELPSCR_XSIZE	equ	80					; 80 columns
HELPSCR_YSIZE	equ	23					; 23 lines


	; --------------
	; Find / Replace
	; --------------
	
MAXSZFINDTXT	equ	60					; Max size of text to find / replace


	; -------------------------------------------------
	; Size of internal buffer used by 'EdWritFileToBuf'
	; -------------------------------------------------

EDWRITEBUFFSZ	equ	16384
	


; =================================
; ===			     ===
; === Variables used by FROSTED ===
; ===			     ===
; =================================

IFDEF		__ED__

	; -------------------------
	; Previous video mode saved	(see GetVideoMode and SetVideoMode)
	; -------------------------
	
		align	4
EdPrevVideoMode	db	?


	; ----------------------------
	; Size of the buffer allocated
	; ----------------------------

		align	4
EdTotalBuffSize		dd	?			; Total size allocated for all buffers
EdTotalScrapSize	dd	?			; Total size allocated for scrap
EdCurrBuffSize		dd	?			; Current size of buffers (in use)


	; ----------------------------------------------------
	; Pointer to internal buffer used by 'EdWritFileToBuf'
	; ----------------------------------------------------

		align	4
EdWriteBuffAddr	dd	?

	
	; ---------------------------------
	; Screen size (in rows and columns)
	; ---------------------------------

		align	4
EdColsNb	dw	?
		align	4
EdRowsNb	dw	?
		align	4
EdFirstRowNb	dw	?				; If 25*80 : 1
		align	4
EdLastRowNb	dw	?				; If 25*80 : 23
		align	4
EdMsgRowNb	dw	?				; Line # for user prompt


	; -----------------------
	; Current buffer # edited
	; -----------------------
	
		align	4
EdCurrBufferNb	dd	?


	; ------------------------------------------------------
	; Buffer to store the current line (where the cursor is)
	; ------------------------------------------------------
	
		align	4
EdCurrLineBuff	db	EDMAXSZLINE*2+2 dup (' ')		; Buffer to hold the current line (expanded, as shown on screen)
		align	4
EdEmptyLine	db	EDMAXSZLINE*2+2 dup (' ')


	; --------------------------
	; Mode insert or overwrite ?
	; --------------------------
	
		align	4
EdInsertMode	db	?


	; -----------------------------------------
	; Pointers to all application's definitions
	; -----------------------------------------

		align	4
EdWordsTable	dd	MAXNB_WORDS dup (?)		; Table of pointer of all words
EdWordsNb	dd	?				; Nb of user defined words


	; -----------------------------------------------------
	; Do we use fast screen functions ? (direct video mode)
	; -----------------------------------------------------

		  align	4
EdFastScreen	  db	?
		  align	4
EdScreenLinesAddr dd	MAXHEIGHTSCREEN dup (?)		; Address of the start of each line


	; -----------------------------------------------------------------------
	; Video attributes/colors for editor elements (copy from FROTH variables)
	; -----------------------------------------------------------------------

		align	4
edc_sll		db	?
		align	4
edc_sli		db	?
		align	4
edc_txn		db	?
		align	4
edc_txs		db	?
		align	4
edc_hln		db	?
		align	4
edc_hla		db	?
		align	4
edc_hlk		db	?


	; ---------
	; The scrap
	; ---------

		align	4
EdScrap		dd	?				; Pointer to scrap
EdScrapSize	dd	?				; Size of the current scrap (in bytes)
EdScrapNbLines	dd	?				; Nb of lines into the scrap


	; ------
	; Macros
	; ------
	
		align	4
EdMacros	dd	?				; Pointer to macros region
EdMacrosSize	dd	?				; Current size for macros
EdMacRecording	db	?				; Recording macro is active ?


	; --------------------------------
	; Status line buffer (to print it)
	; --------------------------------

		align	4
EdStatusLine	db	" B..  Line# .....  Col# ...  Lines ....  Size ...... Mode ...  File ............"
		db	"                       FROSTED v"
		db	FROTH_VERSION + '0', '.', (FROTH_RELEASE/10) + '0', (FROTH_RELEASE MOD 10) + '0', '.', FROTH_UNDERREL + 'a', " "
		db	(FROTH_DATE_MONTH / 10) + '0', (FROTH_DATE_MONTH mod 10) + '0', '-'
		db	(FROTH_DATE_DAY   / 10) + '0', (FROTH_DATE_DAY   mod 10) + '0', '-'
		db	(FROTH_DATE_YEAR  / 10) + '0', (FROTH_DATE_YEAR  mod 10) + '0'
		db	" [OS]"
		align	4
EdFrostedTittle	db	"FROSTED v"
		db	FROTH_VERSION + '0', '.', (FROTH_RELEASE/10) + '0', (FROTH_RELEASE MOD 10) + '0', '.', FROTH_UNDERREL + 'a', " "
		db	(FROTH_DATE_MONTH / 10) + '0', (FROTH_DATE_MONTH mod 10) + '0', '-'
		db	(FROTH_DATE_DAY   / 10) + '0', (FROTH_DATE_DAY   mod 10) + '0', '-'
		db	(FROTH_DATE_YEAR  / 10) + '0', (FROTH_DATE_YEAR  mod 10) + '0'
		db	"   [Oliver SINGLA]                                       ", 0
		db	"                                                    "				; !!!
		align	4
EdPromptBuffer	db	MAXSZPROMPT+1 dup (?)
		align	4
EdPromptFind	db	MAXSZFINDTXT+1 dup (?)
		align	4
EdMsgIns	db	"INS"
		align	4
EdMsgOvr	db	"OVR"
		align	4
EdMsgConfLdFile	db	"Current buffer is not saved. Confirm nevertheless load another file (Y/N) ? ", 0
		align	4
EdMsgLoadFile	db	"Load which file ? ", 0
		align	4
EdMsgSaveFile	db	"Save in which file ? ", 0
		align	4
EdMsgFileAlExist	db	"File already exists. Confirm nevertheless replace file (Y/N) ? ", 0
		align	4
EdMsgErrCont	db	" Press ESC to continue ", 0
		align	4
EdMsgErrUnknow	db	"Error # ", 0								; Err code unknow
		align	4
EdMsgGotoLine	db	"Goto line ? ", 0
		align	4
EdMsgRecordMac	db	"Learn macro, type a letter (A-Z for Alt-.):  (you'll press ^C to stop recording)", 0
		align	4
EdMsgGotoWordNo	db	"No word reference Word ref are only available after compilation (LOAD). Hit ESC", 0
		align	4
EdMsgGotoWord	db	"<    PgUp Pgn> Move selection   <CR> Edit the selected word   <ESC> abort", 0
		align	4
EdMsgGotoWordFT	db	"This word was not defined within a source file. Hit <ESC>", 0
		align	4
EdMsgConfAbortBf db	"Current buffer is not saved. Confirm nevertheless abort buffer (Y/N) ? ", 0


	; --------------
	; Find / Replace
	; --------------

		align	4
EdFOpt_Word	db	?					; Word On/Off ?
		align	4
EdFOpt_Case	db	?					; Case sensitive On/Off ?
		align	4
EdFOpt_AllBuff	db	?					; On all buffers ?

		align	4
EdMsgFind	db	"Search ? ", 0
		align	4
EdMsgFindOpt	db	"^W: Word search:off   ^C: case sensitive:off   ^A: on all buffers:off", 0
		align	4
EdMsgOn		db	"ON ", 0
		align	4
EdMsgOff	db	"OFF", 0


	; ---------
	; Error msg
	; ---------

MSG		MACRO	Code, Msg
		dw	Code
		db	Msg, 0
		endm

EdMsgErr:	MSG	2,		 "FileName is invalid or not found."
		MSG	3,		 "PathName is invalid or not found."
		MSG	4,		 "No handles are available table is full)."
		MSG	5,		 "Bad access mode."
		MSG	EDERRLINTOOLONG,  "Line too long."
		MSG	EDERRNOCRLF,	 "Line separator must be CR-LF."
		MSG	EDERRCTRLCHAR,	 "Control characters non allowed."
		MSG	EDERRBADNB,	 "Invalid number."
		MSG	EDERRNOMEM,	 "Buffer memory full."
		MSG	EDERRSELTOOBIG,	 "Selection too big to fit inside scrap."
		MSG	EDERRFILETWICE,	 "A file can't be open twice."
		dw	0


	; -----------------------------------------------------
	; Working buffer to print decimal number (see PrNumber)
	; -----------------------------------------------------
	
		align	4
		db	20 dup (?)
EdDecNbBuff	db	20 dup (?)
		align	4
EdDecNbBase	dd	10

	
	; -------------------
	; Buffer informations
	; -------------------

		align	4
EdBuffInfos	buffedentry	EDMAXNBBUFF  dup (<>)


	; --------------
	; Diverses flags
	; --------------
	
		align	4
EdExitEd	db	?			; Exit editor ?		0 = NO,  1 = Yes, no load,  2 = Yes, load
		align	4
EdWasACmd	db	?			; Last char was a command ?
		align	4
EdWordsTableOk	db	?			; Words table needs to be computed ?


	; --------------------------
	; Table of Keys -> Functions
	; --------------------------
	
KF	MACRO	key1, key2, funct, param
	dw	key1
	dw	key2
	dd	funct
	ifb <param>
	dw	?
	else
	dw	param
	endif
	endm

		align	4
EdFunctTable:	

		KF	K_INS,		1,	EdToggleInsOvr		; Toggle INS/OVR mode

		KF	K_CR,		0,	EdCarrReturn		; Carriage return
		KF	K_TAB,		0,	EdTab			; Tab = every 4 characters

		KF	K_HOME,		1,	EdMoveHome		; Start of line
		KF	K_END,		1,	EdMoveEnd		; End of line
		KF	K_LEFT,		1,	EdMoveLeft		; Move left
		KF	K_RIGHT,		1,	EdMoveRight		; Move right
		KF	K_CTRL_LEFT	1,	EdMoveWordLeft		; Move to left word
		KF	K_CTRL_RIGHT	1,	EdMoveWordRight		; Move to right word
		KF	K_UP,		1,	EdMoveUp			; Move up
		KF	K_SHIFT_UP,	1,	EdMoveUp			; Move up   (with SHIFT)
		KF	K_DOWN,		1,	EdMoveDown		; Move down
		KF	K_SHIFT_DOWN,	1,	EdMoveDown		; Move down (with SHIFT)
		KF	K_PGUP,		1,	EdPageUp			; Page up
		KF	K_PGDN,		1,	EdPageDown		; Page down
		KF	K_CTRL_UP,	1,	EdScrollToUp		; Scroll to up
		KF	K_CTRL_DN,	1,	EdScrollToDown		; Scroll to down
		KF	K_CTRL_HOME,	1,	EdMoveBeginTxt		; Move to begin of text
		KF	K_CTRL_END,	1,	EdMoveEndTxt		; Move to end   of text
		KF	K_CTRL_PGUP,	1,	EdMoveTopScr		; Move to top of screen
		KF	K_CTRL_PGDN,	1,	EdMoveBottomScr		; Move to bottom of screen
		KF	K_ALT_CR,	1,	EdGotoLine		; Goto line #
		KF	K_CTRL_CR,	1,	EdGotoWord		; List then goto a word
		KF	K_SHIFT_CR,	1,	EdGtoCommentField		; Goto comment field

		KF	K_DEL,		1,	EdDelChar		; Delete character under cursor
		KF	K_BS,		0,	EdBSChar			; Delete character left
		KF	K_CTRL_INS,	1,	EdDelToEOL		; Delete to EOL
		KF	K_CTRL_DEL,	1,	EdDelRightWord		; Delete right word
		KF	K_ALT_DEL,	1,	EdDelLine		; Delete line
		KF	K_ALT_UP,	1,	EdJoinLine		; Join line
		KF	K_ALT_DN,	1,	EdBreakLine		; Break line
		KF	K_ALT_INS,	1,	EdInsLine		; Insert line

		KF	K_CTRL_L,	0,	EdLoadFile		; Load a asked file
		KF	K_CTRL_S,	0,	EdSaveFile		; Save file
		KF	K_CTRL_U,	0,	EdSaveAllBuff		; Save all buffers
		KF	K_CTRL_V,	0,	EdSaveFileAs		; Save file as
		KF	K_CTRL_B,	0,	EdSaveBlock		; Save block in file
		KF	K_CTRL_N,	0,	EdInsertFile		; Insert file
		KF	K_CTRL_T,	0,	EdAbortBuffer		; Abort buffer

		KF	K_CTRL_X,	0,	EdExitEditor		; Exit Editor: Save, Quit, Load
		KF	K_CTRL_Q,	0,	EdQuitEditor		; Quit Editor: Save, Quit

		KF	K_CTRL_E,	0,	EdBuffSelection		; Buffer selection
		KF	K_F1,		0,	EdSelectBuffer,	 0	; Select buffer #1
		KF	K_F2,		0,	EdSelectBuffer,	 1	; Select buffer #2
		KF	K_F3,		0,	EdSelectBuffer,	 2	; Select buffer #3
		KF	K_F4,		0,	EdSelectBuffer,	 3	; Select buffer #4
		KF	K_F5,		0,	EdSelectBuffer,	 4	; Select buffer #5
		KF	K_F6,		0,	EdSelectBuffer,	 5	; Select buffer #6
		KF	K_F7,		0,	EdSelectBuffer,	 6	; Select buffer #7
		KF	K_F8,		0,	EdSelectBuffer,	 7	; Select buffer #8
		KF	K_F9,		0,	EdSelectBuffer,	 8	; Select buffer #9
		KF	K_F10,		0,	EdSelectBuffer,	 9	; Select buffer #10
		KF	K_F11,		0,	EdMarkLine,	 0	; Mark current line
		KF	K_F12,		0,	EdGtoMarkedLine,   0	; Goto marked line

		KF	K_SHIFT_F1,	0,	EdSelectBuffer,	10	; Select buffer #11
		KF	K_SHIFT_F2,	0,	EdSelectBuffer,	11	; Select buffer #12
		KF	K_SHIFT_F3,	0,	EdSelectBuffer,	12	; Select buffer #13
		KF	K_SHIFT_F4,	0,	EdSelectBuffer,	13	; Select buffer #14
		KF	K_SHIFT_F5,	0,	EdSelectBuffer,	14	; Select buffer #15
		KF	K_SHIFT_F6,	0,	EdSelectBuffer,	15	; Select buffer #16
		KF	K_SHIFT_F7,	0,	EdSelectBuffer,	16	; Select buffer #17
		KF	K_SHIFT_F8,	0,	EdSelectBuffer,	17	; Select buffer #18
		KF	K_SHIFT_F9,	0,	EdSelectBuffer,	18	; Select buffer #19
		KF	K_SHIFT_F10,	0,	EdSelectBuffer,	19	; Select buffer #20
		KF	K_SHIFT_F11,	0,	EdMarkLine,	 1	; Mark current line
		KF	K_SHIFT_F12,	0,	EdGtoMarkedLine,   1	; Goto marked line

		KF	K_CTRL_F1,	0,	EdSelectBuffer,	20	; Select buffer #21
		KF	K_CTRL_F2,	0,	EdSelectBuffer,	21	; Select buffer #22
		KF	K_CTRL_F3,	0,	EdSelectBuffer,	22	; Select buffer #23
		KF	K_CTRL_F4,	0,	EdSelectBuffer,	23	; Select buffer #24
		KF	K_CTRL_F5,	0,	EdSelectBuffer,	24	; Select buffer #25
		KF	K_CTRL_F6,	0,	EdSelectBuffer,	25	; Select buffer #26
		KF	K_CTRL_F7,	0,	EdSelectBuffer,	26	; Select buffer #27
		KF	K_CTRL_F8,	0,	EdSelectBuffer,	27	; Select buffer #28
		KF	K_CTRL_F9,	0,	EdSelectBuffer,	28	; Select buffer #29
		KF	K_CTRL_F10,	0,	EdSelectBuffer,	29	; Select buffer #30
		KF	K_CTRL_F11,	0,	EdMarkLine,	 2	; Mark current line
		KF	K_CTRL_F12,	0,	EdGtoMarkedLine,   2	; Goto marked line

		KF	K_ALT_F1,	0,	EdSelectBuffer,	30	; Select buffer #31
		KF	K_ALT_F2,	0,	EdSelectBuffer,	31	; Select buffer #32
		KF	K_ALT_F3,	0,	EdSelectBuffer,	32	; Select buffer #33
		KF	K_ALT_F4,	0,	EdSelectBuffer,	33	; Select buffer #34
		KF	K_ALT_F5,	0,	EdSelectBuffer,	34	; Select buffer #35
		KF	K_ALT_F6,	0,	EdSelectBuffer,	35	; Select buffer #36
		KF	K_ALT_F7,	0,	EdSelectBuffer,	36	; Select buffer #37
		KF	K_ALT_F8,	0,	EdSelectBuffer,	37	; Select buffer #38
		KF	K_ALT_F9,	0,	EdSelectBuffer,	38	; Select buffer #39
		KF	K_ALT_F10,	0,	EdSelectBuffer,	39	; Select buffer #40
		KF	K_ALT_F11,	0,	EdMarkLine,	 3	; Mark current line
		KF	K_ALT_F12,	0,	EdGtoMarkedLine, 	 3	; Goto marked line
		
		KF	K_5,		0,	EdHelp			; Ask help
		
		KF	K_NK_MINUS,	0,	EdCopySelToBin		; Copy selection to scrap
		KF	K_NK_PLUS,	0,	EdPasteBin		; Paste scrap to text
		KF	K_NK_STAR,	0,	EdCopyLineToBin		; Copy line to scrap
		KF	K_NK_SLASH,	0,	EdCutLineToBin		; Cut line to scrap
		
		KF	K_CTRL_F,	0,	EdFind			; Find occurence
		KF	K_CTRL_R,	0,	EdReplace		; Find and replace occurence
		KF	K_CTRL_A,	0,	EdLastFindRepl		; Redo last Find or Replace
		
		KF	K_CTRL_C,	0,	EdRecordMacro		; Start record a macro
		
		KF	K_CTRL_Y,	0,	EdInsFileHeader		; Insert file header
		KF	K_CTRL_Z,	0,	EdInsWordHeader		; Insert word header

		dw	0


	; -----------
	; Help screen
	; -----------
	
		align	4
EdHelpScreen:	db	"ͻ", 0
		db	" [FROSTED] (Froth integrated source editor)    [v"
		db	FROTH_VERSION + '0', '.', (FROTH_RELEASE/10) + '0', (FROTH_RELEASE MOD 10) + '0', '.', FROTH_UNDERREL + 'a', " "
		db	(FROTH_DATE_MONTH / 10) + '0', (FROTH_DATE_MONTH mod 10) + '0', '-'
		db	(FROTH_DATE_DAY   / 10) + '0', (FROTH_DATE_DAY   mod 10) + '0', '-'
		db	(FROTH_DATE_YEAR  / 10) + '0', (FROTH_DATE_YEAR  mod 10) + '0'
		db	"]   Oliver SINGLA ", 0
		db	"Ķ", 0
		db	"{}    {}    Prev/Next char      [^]{}    [^]{}    Prev/Next word   [^]{Y}  Ins File header", 0
		db	"{}    {}    Prev/Next line      [^]{}    [^]{}    Scroll Up/Down   [^]{Z}  Ins Word header", 0
		db	"{PgUp} {PgDn} Page Up/Down        [^]{PgUp} [^]{PgDn} Begin/End screen [s]{CR} Gto commt field", 0
		db	"{Home} {End}  Begin/End line      [^]{Home} [^]{End}  Begin/End text                      ", 0
		db	"{CR}        Skip to next line   [@]{CR}         Goto line #      [^]{CR}  List/Goto word", 0
		db	"{DEL}       Del char under curs {}           Del char left                       ", 0
		db	"[^]{INS}      Del to end of line  [^]{DEL}        Del right word   [@]{DEL} Delete line   ", 0
		db	"[@]{}        Join line           [@]{}          Break line       [@]{INS} Insert line   ", 0
		db	"[^]{L}        Load a new file     [^]{S}          Save file        [^]{V}   Save file as  ", 0
		db	"[^]{B}        Save block in file  [^]{N}          Insert a file                       ", 0
		db	"{F1}...{F10}  Select buffer #     [^]{E}          Buffer selection [^]{T}   Abort buffer  ", 0
		db	"[^]{Q}        Save, Quit          [^]{X}          Save, Quit, Load [^]{U}   Save all buff.", 0
		db	"{INS}       Toggle ins/ovr      [^]{C}          Macro learn                         ", 0
		db	"[^]{F}        Find                [^]{R}          Replace          [^]{A}   Last find/repl", 0
		db	"{F11}       Mark current line   {F12}         Goto marked line                    ", 0
		db	" {-}        Copy sel. to scrap  {+}           Paste scrap                         ", 0
		db	" {*}        Copy line to scrap  {/}           Cut line to scrap                   ", 0
		db	"                            [^] means Ctrl   [@] mean Alt                         ", 0
		db	"  To make selection, press [SHIFT] then move cursor througth line with {} and {}  ", 0
		db	" Press ESC ͼ", 0
		db	0
	

ENDIF		; __ED__


global	EdPrevVideoMode:		byte
global	EdTotalBuffSize:		dword
global	EdTotalScrapSize:		dword
global	EdCurrBuffSize:		dword
global	EdWriteBuffAddr:		dword
global	EdColsNb:		word
global	EdRowsNb:		word
global	EdFirstRowNb:		word
global	EdLastRowNb:		word
global	EdMsgRowNb:		word
global	EdCurrBufferNb:		dword
global	EdCurrLineBuff:		byte:		EDMAXSZLINE*2+2
global	EdEmptyLine:		byte:		EDMAXSZLINE*2+2
global	EdInsertMode:		byte
global	EdWordsTable:		dword:		MAXNB_WORDS
global	EdWordsNb:		dword
global	EdFastScreen:		byte
global	EdScreenLinesAddr:	dword:		MAXHEIGHTSCREEN
global	edc_sll:			byte
global	edc_sli:			byte
global	edc_txn:			byte
global	edc_txs:			byte
global	edc_hln:			byte
global	edc_hla:			byte
global	edc_hlk:			byte
global	EdScrap:			dword
global	EdScrapSize:		dword
global	EdScrapNbLines:		dword
global	EdMacros:		dword
global	EdMacrosSize:		dword
global	EdMacRecording:		byte
global	EdStatusLine:		byte
global	EdFrostedTittle:		byte
global	EdPromptBuffer:		byte:		MAXSZPROMPT+1
global	EdPromptFind:		byte:		MAXSZFINDTXT+1
global	EdMsgIns:		byte
global	EdMsgOvr:		byte
global	EdMsgConfLdFile:		byte
global	EdMsgFileAlExist:		byte
global	EdMsgLoadFile:		byte
global	EdMsgSaveFile:		byte
global	EdMsgErrCont:		byte
global	EdMsgErr:		near
global	EdMsgErrUnknow:		byte
global	EdMsgGotoLine:		byte
global	EdMsgRecordMac:		byte
global	EdMsgGotoWordNo:		byte
global	EdMsgGotoWord:		byte
global	EdMsgGotoWordFT:		byte
global	EdMsgFind:		byte
global	EdMsgFindOpt:		byte
global	EdFOpt_Word:		byte
global	EdFOpt_Case:		byte
global	EdFOpt_AllBuff:		byte
global	EdMsgOff:		byte
global	EdMsgOn:			byte
global	EdDecNbBuff:		byte:		20
global	EdDecNbBase:		dword
global	EdBuffInfos:		buffedentry:	EDMAXNBBUFF
global	EdSelBlockBegin:		dword
global	EdSelBlockEnd:		dword
global	EdExitEd:		byte
global	EdWasACmd:		byte
global	EdWordsTableOk:		byte
global	EdFunctTable:		near


; ==========================================
; ===				    	 ===
; === Command keys to use with FROSTED   ===
; ===				   	 ===
; ==========================================

K_CTRL_A	equ	001h		; Ctrl-A		Again (Find or Replace)
K_CTRL_B	equ	002h		; Ctrl-B		Save block into a file
K_CTRL_C	equ	003h		; Ctrl-C		Learn macro
K_CTRL_E	equ	005h		; Ctrl-E		Buffer selection
K_CTRL_F	equ	006h		; Ctrl-F		Find
K_BS		equ	008h		; BS		Backspace
K_TAB		equ	009h		; TAB		Tabulation
K_CTRL_L	equ	00Ch		; Ctrl-L		Load file
K_CR		equ	00Dh		; CR		Carriage return
K_CTRL_N	equ	00Eh		; Ctrl-N		Insert a file
K_CTRL_Q	equ	011h		; Ctrl-Q		Quit without saving
K_CTRL_R	equ	012h		; Ctrl-R		Replace
K_CTRL_S	equ	013h		; Ctrl-S		Save file
K_CTRL_T	equ	014h		; Ctrl-T		Abort buffer
K_CTRL_U	equ	015h		; Ctrl-U		Save all buffers
K_CTRL_V	equ	016h		; Ctrl-V		Save file as
K_CTRL_W	equ	017h		; Ctrl-W		Find/Repl: Word search options
K_CTRL_X	equ	018h		; Ctrl-X		Quit with saving
K_CTRL_Y	equ	019h		; Ctrl-Y		Insert file header
K_CTRL_Z	equ	01Ah		; Ctrl-Z		Insert word header
K_EOF		equ	01Ah		; Ctrl-Z		Mark end of file
K_ESC		equ	01Bh		; ESC		Abort current operation

K_NUM_CR	equ	19Ch		; CR numeric kbd	Goto comment field
K_5		equ	14Ch		; 5		Ask help
K_INS		equ	152h		; INS		Toggle insert / overwrite mode
K_DEL		equ	153h		; DEL		Delete character under cursor
K_LEFT		equ	14Bh		; Left		Move to left
K_RIGHT		equ	14Dh		; Right		Move to right
K_UP		equ	148h		; Up		Move to up
K_DOWN		equ	150h		; Down		Move to down
K_SHIFT_UP	equ	248h		; Up		Move to up	(with SHIFT down)
K_SHIFT_DOWN	equ	250h		; Down		Move to down	(with SHIFT down)
K_PGUP		equ	149h		; PageUp		Page up
K_PGDN		equ	151h		; PageDn		Page down
K_HOME		equ	147h		; Home		Start of line
K_END		equ	14Fh		; End		End of line
K_SHIFT_UP	equ	248h		; Shift-Up	Line selection, up
K_SHIFT_DN	equ	250h		; Shift-Down	Line selection, down
K_SHIFT_CR	equ	21Ch		; Shift-CR	Goto comment field
K_CTRL_UP	equ	448h		; Ctrl-Up	Scroll to up
K_CTRL_DN	equ	450h		; Ctrl-Down	Scroll to down
K_CTRL_LEFT	equ	44Bh		; Ctrl-Left	Move to left word
K_CTRL_RIGHT	equ	44Dh		; Ctrl-Right	Move to right word
K_CTRL_HOME	equ	447h		; Ctrl-Home	Move to begin of text
K_CTRL_END	equ	44Fh		; Ctrl-End	Move to end   of text
K_CTRL_PGUP	equ	449h		; Ctrl-PgUp	Move to top    of screen
K_CTRL_PGDN	equ	451h		; Ctrl-PgDn	Move to bottom of screen
K_CTRL_INS	equ	452h		; Ctrl-INS	Delete to EOL
K_CTRL_DEL	equ	453h		; Ctrl-DEL	Delete right word
K_CTRL_CR	equ	41Ch		; Ctrl CR	Goto a word
K_ALT_UP	equ	848h		; Alt-Up		Join line
K_ALT_DN	equ	850h		; Alt-Dn		Break line
K_ALT_CR	equ	81Ch		; Alt CR		Goto line #
K_ALT_DEL	equ	853h		; Alt-DEL	Delete line
K_ALT_INS	equ	852h		; Alt-INS	Insert line

K_NK_MINUS	equ	14Ah		; -		Copy to scrap
K_NK_PLUS	equ	14Eh		; +		Paste scrap
K_NK_STAR	equ	137h		; *		Copy current line to scrap
K_NK_SLASH	equ	1B5h		; /		Cut current line to scrap
K_CTRL_STAR	equ	437h		; Ctrl-*		Next used buffer
K_CTRL_SLASH	equ	4B5h		; Ctrl-/		Prev used buffer
K_ALT_STAR	equ	837h		; Alt-*		Next buffer
K_ALT_SLASH	equ	8B5h		; Alt-/		Prev buffer

K_F1		equ	13Bh		; F1		Select buffer #1
K_F2		equ	13Ch		; F2		Select buffer #2
K_F3		equ	13Dh		; F3		Select buffer #3
K_F4		equ	13Eh		; F4		Select buffer #4
K_F5		equ	13Fh		; F5		Select buffer #5
K_F6		equ	140h		; F6		Select buffer #6
K_F7		equ	141h		; F7		Select buffer #7
K_F8		equ	142h		; F8		Select buffer #8
K_F9		equ	143h		; F9		Select buffer #9
K_F10		equ	144h		; F10		Select buffer #10
K_F11		equ	145h		; F11		Mark line
K_F12		equ	146h		; F12		Goto marked line

K_SHIFT_F1	equ	254h		; Shift F1	Select buffer #21
K_SHIFT_F2	equ	255h		; Shift F2	Select buffer #22
K_SHIFT_F3	equ	256h		; Shift F3	Select buffer #23
K_SHIFT_F4	equ	257h		; Shift F4	Select buffer #24
K_SHIFT_F5	equ	258h		; Shift F5	Select buffer #25
K_SHIFT_F6	equ	259h		; Shift F6	Select buffer #26
K_SHIFT_F7	equ	25Ah		; Shift F7	Select buffer #27
K_SHIFT_F8	equ	25Bh		; Shift F8	Select buffer #28
K_SHIFT_F9	equ	25Ch		; Shift F9	Select buffer #29
K_SHIFT_F10	equ	25Dh		; Shift F10	Select buffer #30
K_SHIFT_F11	equ	245h		; Shift-F11	Mark line
K_SHIFT_F12	equ	246h		; Shift-F12	Goto marked line

K_CTRL_F1	equ	45Eh		; Ctrl F1	Select buffer #11
K_CTRL_F2	equ	45Fh		; Ctrl F2	Select buffer #12
K_CTRL_F3	equ	460h		; Ctrl F3	Select buffer #13
K_CTRL_F4	equ	461h		; Ctrl F4	Select buffer #14
K_CTRL_F5	equ	462h		; Ctrl F5	Select buffer #15
K_CTRL_F6	equ	463h		; Ctrl F6	Select buffer #16
K_CTRL_F7	equ	464h		; Ctrl F7	Select buffer #17
K_CTRL_F8	equ	465h		; Ctrl F8	Select buffer #18
K_CTRL_F9	equ	466h		; Ctrl F9	Select buffer #19
K_CTRL_F10	equ	467h		; Ctrl F10	Select buffer #20
K_CTRL_F11	equ	445h		; Ctrl-F11	Mark line
K_CTRL_F12	equ	446h		; Ctrl-F12	Goto marked line

K_ALT_F1	equ	868h		; Alt F1		Select buffer #31
K_ALT_F2	equ	869h		; Alt F2		Select buffer #32
K_ALT_F3	equ	86Ah		; Alt F3		Select buffer #33
K_ALT_F4	equ	86Bh		; Alt F4		Select buffer #34
K_ALT_F5	equ	86Ch		; Alt F5		Select buffer #35
K_ALT_F6	equ	86Dh		; Alt F6		Select buffer #36
K_ALT_F7	equ	86Eh		; Alt F7		Select buffer #37
K_ALT_F8	equ	86Fh		; Alt F8		Select buffer #38
K_ALT_F9	equ	870h		; Alt F9		Select buffer #39
K_ALT_F10	equ	871h		; Alt F10	Select buffer #40
K_ALT_F11	equ	845h		; Alt-F11	Mark line
K_ALT_F12	equ	846h		; Alt-F12	Goto marked line


; ==============================================
; ===					===
; Public functions defined in: Ed.asm	===
; ===					===
; ==============================================
	
global		EdInit:			proc
global		Editor:			proc
global		EdNotAvailable:		proc
global		EdShowNotSavedF:		proc
global		EdShowLigNb:		proc
global		EdShowColNb:		proc
global		EdShowNbLines:		proc
global		EdShowCurrSize:		proc
global		EdShowInsOvrMod:		proc
global		EdShowFileName:		proc
global		EdPrStatusLine:		proc
global		EdInstallFrame:		proc
global		EdSelectBuffer:		proc
global		EdTakeChar:		proc
global		EdCurrBuffIsMod:		proc
global		EdAbortBuffer:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_Util.asm	===
; ===					===
; ==============================================

global		EdPrNumber:		proc
global		EdError:			proc
global		EdCls:			proc
global		EdClrStatusLine:		proc
global		EdClrMsgLine:		proc
global		EdPrMsgStatus:		proc
global		EdPrMsgMsg:		proc
global		EdPrompt:		proc
global		EdConfirm:		proc
global		EdCompCurLinSz:		proc
global		EdCompTextSize:		proc
global		EdFindPrevLine:		proc
global		EdFindNextLine:		proc	
global		EdPutCurLinBuff:		proc
global		EdGetCurLinBuff:		proc
global		EdPosCur:		proc
global		EdDrawCurrLine:		proc
global		EdRedrawPage:		proc
global		EdInactiveBlock:		proc
global		EdUnloadBuffers:		proc
global		EdRequestMem:		proc
global		EdGiveBackMem:		proc
global		EdDiskWriteBuffer:	proc
global		EdDiskReadBuffer:		proc
global		EdUpdatePosTags:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_bios.asm	===
; ===					===
; ==============================================
	
global		EdGetVideoMode:		proc
global		EdSetVideoMode:		proc
global		EdGetRowsNb:		proc
global		EdGetScreenAddr:		proc
global		Ed_C:			proc
global		Ed_AC:			proc
global		Ed_XYAC:			proc
global		Ed_XYNAC:		proc
global		Ed_XYAS:			proc
global		EdGotoXY:		proc
global		EdWhereXY:		proc
global		EdScrollUp:		proc
global		EdScrollDown:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_cmd.asm	===
; ===					===
; ==============================================

global		EdExitEditor:		proc
global		EdQuitEditor:		proc
global		EdToggleInsOvr:		proc
global		EdMoveLeft:		proc
global		EdMoveRight:		proc
global		EdMoveWordLeft:		proc
global		EdMoveWordRight:		proc
global		EdMoveUp:		proc
global		EdMoveDown:		proc
global		EdPageDown:		proc
global		EdPageUp:		proc
global		EdMoveHome:		proc
global		EdMoveTopScr:		proc
global		EdMoveBottomScr:		proc
global		EdMoveEnd:		proc
global		EdDoMoveEnd:		proc
global		EdMoveBeginTxt:		proc
global		EdDoMoveBeginTxt:		proc
global		EdMoveEndTxt:		proc
global		EdDoMoveEndTxt:		proc
global		EdScrollToUp:		proc
global		EdScrollToDown:		proc
global		EdCarrReturn:		proc
global		EdTab:			proc
global		EdGotoLine:		proc
global		EdDoGotoLine:		near
global		EdGtoCommentField:	proc


; ==============================================
; ===					===
; Public functions defined in: Ed_cmd2.asm	===
; ===					===
; ==============================================

global		EdDelChar:		proc
global		EdBSChar:		proc
global		EdDelToEOL:		proc
global		EdDelRightWord:		proc
global		EdDelLine:		proc
global		EdJoinLine:		proc
global		EdBreakLine:		proc
global		EdInsLine:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_cmd3.asm	===
; ===					===
; ==============================================

global		EdCopySelToBin:		proc
global		EdPasteBin:		proc
global		EdCopyLineToBin:		proc
global		EdCutLineToBin:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_cmd4.asm	===
; ===					===
; ==============================================

global		EdFind:			proc
global		EdReplace:		proc
global		EdLastFindRepl:		proc
global		EdMarkLine:		proc
global		EdGtoMarkedLine:		proc
global		EdBuffSelection:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_cmd5.asm	===
; ===					===
; ==============================================

global		EdRecordMacro:		proc
global		EdInsFileHeader:		proc
global		EdInsWordHeader:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_GotoW.asm	===
; ===					===
; ==============================================

global		EdLineNbNotAvail:		proc
global		EdGotoWord:		proc


; ==============================================
; ===					===
; Public functions defined in: Ed_file.asm	===
; ===					===
; ==============================================

global		EdLoadFile:		proc
global		EdSaveFile:		proc
global		EdSaveFileAs:		proc
global		EdSaveAllBuff:		proc
global		EdInsertFile:		proc
global		EdSaveBlock:		proc



if UNDERDEV

SHOWBUFF	macro	num
		pushad
		mov	bl, num
		call	DbgShowBuff
		popad
		endm


SHOWTEXT	macro	num
		pushad
		mov	bl, num
		call	DbgShowText
		popad
		endm

endif

