; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED_CMD.ASM						!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

if TURNKEYED eq 0

include		Ed.inc					; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdQuitEditor						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	We have been asked to exit from editor. 		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdQuitEditor	proc

		; ----------------
		; Save all buffers
		; ----------------

		call	EdSaveAllBuff
		jc	@@Exit					; Abort ?

		
		; ----------------
		; We leave FROSTED
		; ----------------
		
		mov	[EdExitEd], TRUE


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdQuitEditor	endp


; ===============================================================
; = EdExitEditor						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	We have been asked to exit from editor. 		=
; =								=
; = I)	none							=
; =								=
; = O)	FLAGS							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdExitEditor	proc

		; ----------------
		; Save all buffers
		; ----------------

		call	EdSaveAllBuff
		jc	@@Exit					; Abort ?

		
		; ----------------
		; We leave FROSTED
		; ----------------
		
		mov	[EdExitEd], 2


		; Return to caller
		; ----------------
		
@@Exit:		ret

EdExitEditor	endp


; ===============================================================
; = EdToggleInsOvr						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Toggle Insert/Overwrite mode.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdToggleInsOvr	proc

		; -------------------------------------
		; Toggle the Insertion / Overwrite mode
		; -------------------------------------

		cmp	[EdInsertMode], FALSE
		je	@@SetToIns
		
		; Set to OVR
		mov	[EdInsertMode], FALSE
		jmp	@@ShowMode

		; Set to INS
@@SetToIns:	mov	[EdInsertMode], TRUE

		; Show mode
@@ShowMode:	call	EdShowInsOvrMod


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdToggleInsOvr	endp


; ===============================================================
; = EdMoveLeft							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move cursor to left, by a character.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveLeft	proc

		; -------------------------------------------------------------------------
		; If (#Col != #HCol) then #Col--	(no horizontal scroll in this case)
		; -------------------------------------------------------------------------
		
		mov	eax, [ebp.binf_tcol]			; Current col
		cmp	eax, [ebp.binf_fcol]			; Compare with the 1st column printed
		je	@@ColEqHCol
		
		; Just backward the cursor
		dec	eax					; Prev col
		mov	[ebp.binf_tcol], eax			; Current col
		
		; Print the current column #
		call	EdShowColNb

		; Update the cursor position on screen
		call	EdPosCur
		
		; Exit
		jmp	@@Exit


		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		

		; ----------------------------------------------
		; Current col position is the 1st column printed
		; ----------------------------------------------
		
@@ColEqHCol:	; !!!


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveLeft	endp


; ===============================================================
; = EdMoveWordLeft						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move cursor to left, by a word.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveWordLeft	proc

		; -------------------
		; The actual position
		; -------------------
		
		mov	ebx, [ebp.binf_tcol]
		
		
		; -----------------
		; Move to next word
		; -----------------
		
		; Begin of line ?
		cmpz	bx, 0
		jne	@@NotBegin
		
		; Do nothing if on the beginning of the text
		mov	eax, [ebp.binf_CurLinAddr]
		cmp	eax, [ebp.binf_BuffAddr]
		je	@@Exit

		call	EdMoveUp				; Goto previous line
		jmp	EdDoMoveEnd				; End of this line
		
@@NotBegin:
		
		; Move pointer
		lea	esi, EdCurrLineBuff
		dec	bx
		cmpz	bx, 0
		je	@@CL3
		cmp	byte ptr [esi + ebx], ' '
		jne	@@CL2
@@CL4:		dec	bx
		cmpz	bx, 0
		je	@@CL3
		cmp	byte ptr [esi + ebx], ' '
		je	@@CL4
@@CL2:		dec	bx
		cmpz	bx, 0
		je	@@CL3
		cmp	byte ptr [esi + ebx], ' '
		jne	@@CL2
		inc	bx
@@CL3:

		; Update variables
		mov	[ebp.binf_tcol], ebx
		
		; Print the current column #
		call	EdShowColNb

		; Update the cursor position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveWordLeft	endp


; ===============================================================
; = EdMoveRight							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move cursor to right, by a character.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveRight	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		

		; --------------------
		; Last col available ?
		; --------------------
		
		mov	eax, [ebp.binf_tcol]
		cmp	ax, [EdColsNb]
		je	@@Exit
		
		
		; -----------
		; Scrolling ?
		; -----------
		
		sub	eax, [ebp.binf_fcol]
		inc	eax
		cmp	ax, [EdColsNb]
		jne	@@NoHScroll

		ifdef __NOHSCROLL__
		jmp	@@Exit
		endif

		; Update the position of the first column shown (horizontal scrolling)
		mov	ebx, [ebp.binf_fcol]
		add	ebx, STEPMOVHSCROLL
		mov	[ebp.binf_fcol], ebx
		
		; Do not exceed max line width
		add	ebx, [ebp.binf_tcol]
		cmp	ebx, (EDMAXSZLINE	- 2)			; Max length of a line
		jle	@@LenOk
		sub	ebx, (EDMAXSZLINE - 2)
		sub	[ebp.binf_tcol], ebx
		jmp	@@Redraw
@@LenOk:
		
		; Update cursor position
		inc	[ebp.binf_tcol]					; Next col
		
		; Redraw the complete page
@@Redraw:	call	EdRedrawPage

		; Print the current column #
@@Next:		call	EdShowColNb

		; Update the cursor position on screen
		call	EdPosCur

		; Exit
		jmp	@@Exit
	

		; ----------------
		; No hor scrolling
		; ----------------
		
@@NoHScroll:	; Update cursor position
		inc	[ebp.binf_tcol]				; Next col
		jmp	@@Next


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveRight	endp


; ===============================================================
; = EdMoveWordRight						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move cursor to right, by a word.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveWordRight	proc

		; --------------------------------------
		; Compute the length of the current line
		; --------------------------------------

		call	EdCompCurLinSz				; New length in EAX


		; -------------------
		; The actual position
		; -------------------
		
		mov	ebx, [ebp.binf_tcol]
		
		
		; -----------------
		; Move to next word
		; -----------------
		
		; End of line ?
		cmp	ax, bx
		je	EdCarrReturn
		
		; Move pointer
		lea	esi, EdCurrLineBuff
		cmp	byte ptr [esi + ebx], ' '
		jne	@@CR2
@@CR4:		inc	bx
		cmp	bx, ax
		je	@@CR3
		cmp	byte ptr [esi + ebx], ' '
		je	@@CR4
		jmp	@@CR3
@@CR2:		inc	bx
		cmp	bx, ax
		je	@@CR3
		cmp	byte ptr [esi + ebx], ' '
		jne	@@CR2
		jmp	@@CR4
@@CR3:		

@@CR5:		
		ifdef __NOHSCROLL__
		cmp	bx, [EdColsNb]
		jl	@@LenOk
		movzx	ebx, [EdColsNb]
		dec	ebx
@@LenOk:		
		endif

		; Update variables
		mov	[ebp.binf_tcol], ebx
		
		; Print the current column #
		call	EdShowColNb

		; Update the cursor position on screen
		call	EdPosCur



		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveWordRight	endp


; ===============================================================
; = EdMoveUp							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move cursor to up, by a line.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveUp		proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		

		; ----------------------------------
		; Top of text and screen: do nothing
		; ----------------------------------
		
		cmp	[ebp.binf_trow], 0
		je	@@Exit
		
		
		; ------------------------------------------
		; Is SHIFT is pressed, update selected block
		; ------------------------------------------

		call	isShiftKey		
		jnc	@@NoShift
		
@@NoShift:

		
		; ----------------------
		; On top of the screen ?
		; ----------------------
		
		mov	eax, [ebp.binf_trow]
		cmp	eax, [ebp.binf_frow]
		jne	@@NotOnTop
		
		; Just scroll
		call	EdScrollToDown
		jmp	@@Exit


@@NotOnTop:

		; -----------------------------------
		; We are not on the top of the screen
		; -----------------------------------
		
		; Find line address of previous line
		mov	esi, [ebp.binf_CurLinAddr]
		call	EdFindPrevLine
		mov	[ebp.binf_CurLinAddr], esi
		
		; Goto on previous line
		dec	[ebp.binf_trow]
		
		; Bring the new line into editing buffer
		mov	ecx, esi
		call	EdGetCurLinBuff
		
		; Show new row # on status line
		call	EdShowLigNb
		
		; Update cursor position on screen
		call	EdPosCur

		
		; ----------------
		; Return to caller
		; ----------------

@@Exit:		ret

EdMoveUp	endp


; ===============================================================
; = EdMoveDown							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move cursor to down, by a line.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveDown	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		
		
		; ---------------------------------
		; Do nothing if we are on last line
		; ---------------------------------
		
		mov	eax, [ebp.binf_trow]
		inc	eax
		cmp	eax, [ebp.binf_LinesNb]
		je	@@Exit
		
		
		; ------------------------------------------
		; Is SHIFT is pressed, update selected block
		; ------------------------------------------

		call	isShiftKey		
		jnc	@@NoShift
		
		; Update selected block start address ?
		mov	eax, [ebp.binf_CurLinAddr]
		cmp	[ebp.binf_BlockBegin], -1
		jne	@@BlockExist
		
		; Block is new
@@BlockNew:	mov	[ebp.binf_BlockBegin], eax
		jmp	@@UpdtBlockEnd

		; Cancel the previous block selection ?
@@BlockExist:	cmp	eax, [ebp.binf_BlockEnd]
		je	@@UpdtBlockEnd
		call	EdInactiveBlock
		jmp	@@BlockNew
		
		; Update selected block end address
@@UpdtBlockEnd:	mov	esi, [ebp.binf_CurLinAddr]
		call	EdFindNextLine
		mov	[ebp.binf_BlockEnd], esi
		
		; Redraw the current line (marked as selected)
		call	EdDrawCurrLine
		
@@NoShift:


		; ------------------------
		; On last line of screen ?
		; ------------------------
		
		mov	eax, [ebp.binf_trow]
		sub	eax, [ebp.binf_frow]
		inc	ax
		mov	bx, [EdLastRowNb]
		sub	bx, [EdFirstRowNb]		
		inc	bx
		cmp	ax, bx
		jne	@@NotOnBottom
		
		; Just scroll
		call	EdScrollToUp
		jmp	@@Exit

		
@@NotOnBottom:

		; --------------------------------------
		; We are not on the bottom of the screen
		; --------------------------------------
		
		; Find line address of next line
		mov	esi, [ebp.binf_CurLinAddr]
@@FLAPL1:	cmp	byte ptr [esi], 0Dh
		je	@@FLAPL2
		inc	esi
		jmp	@@FLAPL1
@@FLAPL2:	inc	esi
		inc	esi					; Be after 0Dh 0Ah
		mov	[ebp.binf_CurLinAddr], esi
		
		; Goto on next line
		inc	[ebp.binf_trow]
		
		; Bring the new line into editing buffer
		mov	ecx, esi
		call	EdGetCurLinBuff
		
		; Show new row # on status line
		call	EdShowLigNb
		
		; Update cursor position on screen
		call	EdPosCur
		

		; ----------------
		; Return to caller
		; ----------------

@@Exit:		ret

EdMoveDown	endp


; ===============================================================
; = EdPageUp							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move page up.						=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdPageUp	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


		; --------------------------------
		; Do nothing if we are on 1st line
		; --------------------------------
		
		cmp	[ebp.binf_trow], 0
		je	@@Exit
		

		; ---------------------------
		; Find start of previous page
		; ---------------------------

		; Address of top screen line (1st line drawed on top)
		mov	esi, [ebp.binf_1stLinAddr]
		
		; Line counter
		clr	ecx
		movzx	edx, [EdRowsNb]
		dec	edx
		dec	edx					; Size of a screen

		; Find previous line address
@@PrevLine:	call	EdFindPrevLine
		
		; Top of text reached ? If yes, goto top of text
		jc	EdDoMoveBeginTxt

		; Update position in text		
		dec	[ebp.binf_trow]
		dec	[ebp.binf_frow]
		
		; End of page reached
		inc	ecx
		cmp	ecx, edx
		jne	@@PrevLine

		; Save 1st line address
		mov	[ebp.binf_1stLinAddr], esi


		; ----------------------------
		; Find address of current line
		; ----------------------------

		; Address of top screen line (1st line drawed on top)
		mov	esi, [ebp.binf_1stLinAddr]
		
		; Line counter
		clr	cx

		; Current screen line position
		mov	edx, [ebp.binf_trow]
		sub	edx, [ebp.binf_frow]

		; Current line reached ? If yes, it's ok
@@NextLine2:	cmp	cx, dx
		je	@@NextLine3
		
		; Find next line address
		call	EdFindNextLine
		
		; End of text reached ? If yes, goto end of text
		jc	EdDoMoveEndTxt

		; Scan next line
		inc	cx
		jmp	@@NextLine2

		; Move current line to buffer
@@NextLine3:	mov	[ebp.binf_CurLinAddr], esi
		mov	ecx, esi
		call	EdGetCurLinBuff


		; ----------------------
		; Draw the previous page
		; ----------------------

		; Draw page
		call	EdRedrawPage
		
		; Print the current line #
		call	EdShowLigNb
				
		; Position the cursor on the current position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdPageUp	endp


; ===============================================================
; = EdPageDown							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move page down.						=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdPageDown	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


		; ----------------------------------
		; Do nothing if on last line of text
		; ----------------------------------
		
		mov	eax, [ebp.binf_trow]
		inc	eax
		cmp	eax, [ebp.binf_LinesNb]
		je	@@Exit


		; -----------------------
		; Find start of next page
		; -----------------------

		; Address of top screen line (1st line drawed on top)
		mov	esi, [ebp.binf_1stLinAddr]
		
		; Line counter
		clr	cx
		mov	dx, [EdRowsNb]
		dec	dx
		dec	dx					; Size of a screen

		; Find next line address
@@NextLine:	call	EdFindNextLine
		
		; End of text reached ? If yes, goto end of text
		jc	EdDoMoveEndTxt

		; Update position in text		
		inc	[ebp.binf_trow]
		inc	[ebp.binf_frow]
		
		; End of page reached
		inc	cx
		cmp	cx, dx
		jne	@@NextLine

		; Save 1st line address
		mov	[ebp.binf_1stLinAddr], esi


		; ----------------------------
		; Find address of current line
		; ----------------------------

		; Address of top screen line (1st line drawed on top)
		mov	esi, [ebp.binf_1stLinAddr]
		
		; Line counter
		clr	cx

		; Current screen line position
		mov	edx, [ebp.binf_trow]
		sub	edx, [ebp.binf_frow]

		; Current line reached ? If yes, it's ok
@@NextLine2:	cmp	cx, dx
		je	@@NextLine3
		
		; Find next line address
		call	EdFindNextLine
		
		; End of text reached ? If yes, goto end of text
		jc	EdDoMoveEndTxt

		; Scan next line
		inc	cx
		jmp	@@NextLine2

		; Move current line to buffer
@@NextLine3:	mov	[ebp.binf_CurLinAddr], esi
		mov	ecx, esi
		call	EdGetCurLinBuff


		; ------------------
		; Draw the next page
		; ------------------
		
		; Draw page
		call	EdRedrawPage
		
		; Print the current line #
		call	EdShowLigNb
				
		; Position the cursor on the current position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdPageDown	endp


; ===============================================================
; = EdMoveHome							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the start of the line.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveHome	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


		; ---------------------------------------------
		; Redraw page ? (because text was hor scrolled)
		; ---------------------------------------------
		
		cmp	[ebp.binf_fcol], 0
		je	@@GotoCol0
		
		; 1st col to draw is #0
		mov	[ebp.binf_fcol], 0
		
		; Redraw page
		call	EdRedrawPage
		

		; --------------
		; Goto on col #0
		; --------------

@@GotoCol0:	mov	[ebp.binf_tcol], 0


		; ---------------
		; Print the col #
		; ---------------

		mov	[ebp.binf_fcol], 0
		call	EdShowColNb


		; ------------------------------------
		; Update the cursor position on screen
		; ------------------------------------
		
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveHome	endp


; ===============================================================
; = EdMoveEnd							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the end of the line.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveEnd	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


EdDoMoveEnd:

		; --------------------------------------
		; Compute the length of the current line
		; --------------------------------------

		call	EdCompCurLinSz				; New length in EAX


		; ----------------------------
		; No Hor scrolling is needed ?
		; ----------------------------
		
		cmp	ax, [EdColsNb]
		jge	@@HorScroll
		

		; Redraw page ? (because text was hor scrolled)
		; ---------------------------------------------
		
		cmp	[ebp.binf_fcol], 0
		je	@@NotRedraw
		
		; 1st col to draw is #0
		mov	[ebp.binf_fcol], 0
		
		; Redraw page
		call	EdRedrawPage
		
@@NotRedraw:	
		; Update the col #
		mov	[ebp.binf_tcol], eax

		; Print the col #
		mov	[ebp.binf_fcol], 0
		call	EdShowColNb
		
		; Update the cursor position on screen
		call	EdPosCur

		; Ok
		jmp	@@Exit
		


		; --------------
		; Need to scroll
		; --------------

@@HorScroll:

		ifdef __NOHSCROLL__
		movzx	eax, [EdColsNb]
		dec	eax
		jmp	@@NotRedraw
		endif

		; Initialize 1st col to draw
		mov	[ebp.binf_fcol], 0

		; Next hor scroll		
@@Loop:		mov	ebx, [ebp.binf_fcol]
		add	ebx, STEPMOVHSCROLL
		mov	[ebp.binf_fcol], ebx
		
		; Cursor position is now visible ?
		mov	ebx, eax
		sub	ebx, [ebp.binf_fcol]
		cmp	bx, [EdColsNb]
		jge	@@Loop

		; Redraw page
		call	EdRedrawPage
		
		; Update the col #
		mov	[ebp.binf_tcol], eax

		; Print the col #
		call	EdShowColNb
		
		; Update the cursor position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveEnd	endp


; ===============================================================
; = EdMoveTopScr						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the top of the screen.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveTopScr	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


		; ------------------------------
		; Pointer to current line edited
		; ------------------------------

		; Update current line buffer
		mov	ecx, [ebp.binf_1stLinAddr]		; Address of top screen line (1st line drawed on top)
		mov	[ebp.binf_CurLinAddr], ecx		; Current line address (in the edited text)
		
		; Bring this line into current buffer		
		call	EdGetCurLinBuff

		; Redraw the entire status line
		mov	eax, [ebp.binf_frow]
		mov	[ebp.binf_trow], eax
		call	EdPrStatusLine
		
		; Update the cursor position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveTopScr	endp


; ===============================================================
; = EdMoveBottomScr						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the bottom of the screen.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveBottomScr	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


		; ------------------------
		; Find end of current page
		; ------------------------

		; Address of top screen line (1st line drawed on top)
		mov	esi, [ebp.binf_1stLinAddr]
		
		; Line counter
		clr	cx
		mov	dx, [EdRowsNb]
		dec	dx
		dec	dx					; Size of a screen
		dec	dx

		; Update current line #
		mov	eax, [ebp.binf_frow]
		mov	[ebp.binf_trow], eax

		; Find next line address
@@NextLine:	call	EdFindNextLine
		
		; End of text reached ? If yes, goto end of text
		jc	EdDoMoveEndTxt

		; Update position in text		
		inc	[ebp.binf_trow]
		
		; End of page reached
		inc	cx
		cmp	cx, dx
		jne	@@NextLine

		; Bring this line into current buffer		
		mov	[ebp.binf_CurLinAddr], esi		; Current line address (in the edited text)
		mov	ecx, esi
		call	EdGetCurLinBuff

		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Update the cursor position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveBottomScr	endp


; ===============================================================
; = EdMoveBeginTxt						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the start of the text.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdMoveBeginTxt	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


EdDoMoveBeginTxt:

		; Save used registers
		pushad

		; 1st line is #0 
		mov	eax, [ebp.binf_BuffAddr]
		mov	[ebp.binf_1stLinAddr], eax
		mov	[ebp.binf_CurLinAddr], eax
		mov	[ebp.binf_trow], 0
		mov	[ebp.binf_frow], 0
		
		; 1st col to draw is #0
		mov	[ebp.binf_tcol], 0
		mov	[ebp.binf_fcol], 0

		; In case of a call outside editor session, do not update the screen
		cmp	[EditorActive], TRUE
		jne	@@Done

		; Redraw page
		call	EdRedrawPage
		
		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Update the cursor position on screen
		call	EdPosCur
		
@@Done:		


		; ----------------
		; Return to caller
		; ----------------
		
		popad
		ret

EdMoveBeginTxt	endp


; ===============================================================
; = EdMoveEndTxt						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the start of the text.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdMoveEndTxt	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


EdDoMoveEndTxt:

		; Find the end of the text
		call	EdCompTextSize
		mov	esi, [ebp.binf_BuffAddr]
		dec	eax					; 1: EOF
		add	esi, eax
		mov	[ebp.binf_CurLinAddr], esi		; Save temp value
		
		; Init row #
		mov	eax, [ebp.binf_LinesNb]
		dec	eax
		mov	[ebp.binf_frow], eax
		mov	[ebp.binf_trow], eax
		
		; Top of the text ?
		cmp	esi, [ebp.binf_BuffAddr]
		je	@@DrawPage
		
		; Line counter
		clr	cx
		mov	dx, [EdRowsNb]
		dec	dx
		dec	dx					; Size of a screen

		; Find previous line address
@@PrevLine:	call	EdFindPrevLine

		; Top of text reached ? If yes, stop
		jc	@@Top

		; Update the 1st line drawn		
		dec	[ebp.binf_frow]

		; End of page reached
		inc	cx
		cmp	cx, dx
		jne	@@PrevLine

		; Update the 1st line drawn		
		inc	[ebp.binf_frow]

		; Save 1st line address
@@Top:		mov	[ebp.binf_1stLinAddr], esi


		; -----------------------------------
		; Address of current line (last line)
		; -----------------------------------
		
		mov	esi, [ebp.binf_CurLinAddr]
		call	EdFindPrevLine
		mov	[ebp.binf_CurLinAddr], esi


		; -----------------
		; Draw now the page
		; -----------------

@@DrawPage:
		
		; 1st col to draw is #0
		mov	[ebp.binf_tcol], 0
		mov	[ebp.binf_fcol], 0
		
		; Redraw page
		call	EdRedrawPage
		
		; Redraw the entire status line
		call	EdPrStatusLine

		; Move to end of line and update the cursor position on screen
		call	EdMoveEnd


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdMoveEndTxt	endp


; ===============================================================
; = EdScrollToUp						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Scroll one line, to up.					=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdScrollToUp	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		
		
		; ---------------------------------
		; Do nothing if we are on last line
		; ---------------------------------
		
		mov	eax, [ebp.binf_trow]
		inc	eax
		cmp	eax, [ebp.binf_LinesNb]
		je	@@Exit


		; -----------------------------
		; Scroll the area of the screen
		; -----------------------------

		mov	ah, [edc_txn]				; New line Attr/Color		
		mov	al, 1					; Scroll 1 line
		mov	bx, [EdFirstRowNb]
		mov	bh, bl					; Upper row
		clr	bl					; Upper left col
		mov	cx, [EdLastRowNb]
		mov	ch, cl					; Lower row
		mov	dx, [EdColsNb]
		mov	cl, dl
		dec	cl					; Lower right col
		call	EdScrollUp


		; -----------------------------------------
		; Update the address of the 1st line drawed
		; -----------------------------------------
		
		mov	esi, [ebp.binf_1stLinAddr]
		call	EdFindNextLine
		mov	[ebp.binf_1stLinAddr], esi
		

		; -------------------------------
		; Update the position in the text
		; -------------------------------
		
		inc	[ebp.binf_trow]
		inc	[ebp.binf_frow]
		call	EdShowLigNb
		
		
		; -----------------------------
		; Find last screen line address
		; -----------------------------
		
		; Address of top screen line (1st line drawed on top)
		mov	esi, [ebp.binf_1stLinAddr]
		
		; Line counter
		clr	cx
		mov	dx, [EdRowsNb]
		dec	dx
		dec	dx					; Size of a screen
		dec	dx

		; Find next line address
@@NextLine:	call	EdFindNextLine
		
		; End of text reached ? If yes, goto end of text
		jc	@@EmptyLine

		; End of page reached
		inc	cx
		cmp	cx, dx
		jne	@@NextLine

		; Bring this line into current buffer		
		mov	ecx, esi
		call	EdGetCurLinBuff
		jmp	@@DrawLine

		; An empty line
@@EmptyLine:	lea	ecx, EdEmptyLine
		call	EdGetCurLinBuff

		; Draw this line
@@DrawLine:	mov	ax, [EdColsNb]				; Nb of chars to print (width of the screen in AL)
		mov	ah, [edc_txn]				; Attr/Color		
		mov	bx, [EdLastRowNb]
		mov	bh, bl					; Row # in BH
		clr	bl					; Col #
		lea	edx, EdCurrLineBuff			; Current line buffer
		add	edx, [ebp.binf_fcol]			; Hor scrolling: 1st col to draw
		call	Ed_XYAS
		
		; Update screen lines length
		movzx	ecx, bh
		
@@FLSLA2:
		
		; ----------------------------
		; Get current line into buffer
		; ----------------------------

		; Update current line address		
		mov	esi, [ebp.binf_CurLinAddr]
		call	EdFindNextLine
		mov	[ebp.binf_CurLinAddr], esi

		; Get current line address
		mov	ecx, esi
		call	EdGetCurLinBuff
		
		; Update the cursor position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdScrollToUp	endp


; ===============================================================
; = EdScrollToDown						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Scroll one line, to down.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdScrollToDown	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff
		
		
		; ----------------------------------------------------
		; Do nothing if the 1st line drawed is the top of text
		; ----------------------------------------------------
		
		mov	eax, [ebp.binf_1stLinAddr]
		cmp	eax, [ebp.binf_BuffAddr]
		je	@@Exit


		; -----------------------------
		; Scroll the area of the screen
		; -----------------------------
		
		mov	ah, [edc_txn]				; Attr/Color		
		mov	al, 1					; Scroll 1 line
		mov	bx, [EdFirstRowNb]
		mov	bh, bl					; Upper row
		clr	bl					; Upper left col
		mov	cx, [EdLastRowNb]
		mov	ch, cl					; Lower row
		mov	dx, [EdColsNb]
		mov	cl, dl
		dec	cl					; Lower right col
		call	EdScrollDown
		
		
		; -----------------------------------------
		; Update the address of the 1st line drawed
		; -----------------------------------------
		
		mov	esi, [ebp.binf_1stLinAddr]
		call	EdFindPrevLine
		mov	[ebp.binf_1stLinAddr], esi
		

		; -------------------------------
		; Update the position in the text
		; -------------------------------
		
		dec	[ebp.binf_trow]
		dec	[ebp.binf_frow]
		call	EdShowLigNb
		
		
		; -----------------------------
		; Find last screen line address
		; -----------------------------
		
		; Bring this line into current buffer		
		mov	ecx, esi					; 1st line drawed on the screen
		call	EdGetCurLinBuff

		; Draw this line
		mov	ax, [EdColsNb]				; Nb of chars to print (width of the screen in AL)
		mov	ah, [edc_txn]				; Attr/Color		
		mov	bx, [EdFirstRowNb]
		mov	bh, bl					; Row # in BH
		clr	bl					; Col #
		lea	edx, EdCurrLineBuff			; Current line buffer
		add	edx, [ebp.binf_fcol]			; Hor scrolling: 1st col to draw
		call	Ed_XYAS
		
		; Update screen lines length
		movzx	ecx, bh
		
		
		; ----------------------------
		; Get current line into buffer
		; ----------------------------

		; Update current line address		
		mov	esi, [ebp.binf_CurLinAddr]
		call	EdFindPrevLine
		mov	[ebp.binf_CurLinAddr], esi
		
		; Get current line address
		mov	ecx, esi
		call	EdGetCurLinBuff
		
		; Update the cursor position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdScrollToDown	endp


; ===============================================================
; = EdCarrReturn						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move cursor to start of next line. Create this line	=
; =	if not exists.						=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdCarrReturn	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff					; EDX : addr of next line


		; ---------------------------------------
		; End of text ? If yes, insert a new line
		; ---------------------------------------
		
		cmp	byte ptr [edx], K_EOF
		jne	@@GotoNextLine

		; Room for 2 more bytes ?
		mov	eax, 2
		call	EdRequestMem
		jc	@@Exit

		; Add a new line
		mov	byte ptr [edx],     0Dh
		mov	byte ptr [edx + 1], 0AH
		mov	byte ptr [edx + 2], K_EOF
		
		; Update the nb of lines of the text
		inc	[ebp.binf_LinesNb]		
		call	EdShowNbLines

		; Current buffer is modified
		call	EdCurrBuffIsMod
		
		; Cancel the block selection, if any
		call	EdInactiveBlock
		
		
		; --------------------------------
		; Just goto the start of next line
		; --------------------------------
		
@@GotoNextLine:	
		
		; We scroll the screen ?
		mov	eax, [ebp.binf_trow]
		sub	eax, [ebp.binf_frow]
		inc	eax
		mov	bx, [EdLastRowNb]
		sub	bx, [EdFirstRowNb]
		inc	bx
		cmp	ax, bx
		jne	@@NoScroll

		; Scroll
		call	EdScrollToUp
		jmp	@@Suite
		
		; Bring this line into editing buffer
@@NoScroll:	mov	[ebp.binf_CurLinAddr], edx
		mov	ecx, edx
		call	EdGetCurLinBuff

		; Goto on next line: row
		inc	[ebp.binf_trow]
		call	EdShowLigNb
		
		; Goto on col #0
@@Suite:	mov	[ebp.binf_tcol], 0
		mov	eax, [ebp.binf_fcol]
		push	eax
		mov	[ebp.binf_fcol], 0
		call	EdShowColNb

		; Redraw the page ? (because text was hor scrolled)
		pop	eax
		cmpz	eax, 0
		je	@@NotToRedraw
		call	EdRedrawPage
@@NotToRedraw:		
		
		; Update cursor position on screen
		call	EdPosCur
		
		
		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdCarrReturn	endp


; ===============================================================
; = EdTab							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	TAB is replaced by four spaces.				=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	all registers.						=
; ===============================================================

		align	4
EdTab		proc

		; -------------------------------
		; Go to next column multiple of 4
		; -------------------------------

		mov	cl, 4
@@AddSpace:	push	cx
		mov	al, ' '
		call	EdTakeChar
		pop	cx
		dec	cl
		jz	@@AddSpace2
		mov	eax, [ebp.binf_tcol]		; Current col # in text
		and	al, 3
		cmp	al, 3
		jne	@@AddSpace
@@AddSpace2:	


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdTab		endp


; ===============================================================
; = EdGotoLine							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the beginning of a line #.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdGotoLine	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff				; EDX : addr of next line


		; -----------------------
		; Ask for the line number
		; -----------------------

		lea	edx, EdMsgGotoLine
		mov	bl, 5
		call	EdPrompt
		jc 	@@Exit					; Abort
		
		; Ascii->binary conversion
		call	EdNumber
		jnc	@@Ok
		mov	ax, EDERRBADNB				; Invalid number
		call	EdError
		jmp	@@Exit
@@Ok:		dec	eax					; Number is now in EAX

		; Column #0
		clr	ebx


; ----------------------------------------------------
; Enter point : EAX is the line #, EBX is the column #
; ----------------------------------------------------

EdDoGotoLine:

		; ----------------
		; Search this line
		; ----------------
		
		; Initialize line #
		clr	ecx
		
		; Starting address
		mov	esi, [ebp.binf_BuffAddr]

		; Line asked reached ?
@@Loop:		cmp	ecx, eax
		je	@@LineFound

		; Find next line address
		call	EdFindNextLine
		
		; End of text reached ? If yes, goto end of text
		jc	EdDoMoveEndTxt
		
		; Continue the search
		inc	ecx
		jmp	@@Loop
		
		
		; ---------------
		; Line # is found
		; ---------------

		; Initialize line #
@@LineFound:	mov	[ebp.binf_trow], ecx
		mov	[ebp.binf_frow], ecx

		; Bring the new line into editing buffer
		mov	[ebp.binf_CurLinAddr], esi
		mov	ecx, esi
		call	EdGetCurLinBuff

		; Asked col bigger than line length ?
		call	EdCompCurLinSz				; Length in EAX
		cmp	ebx, eax
		jl	@@LenOk
		mov	ebx, eax
		dec	ebx
		cmpz	ebx, 0
		jge	@@LenOk
		clr	ebx
@@LenOk:

		; Save asked column
		mov	[ebp.binf_tcol], ebx

; !!!		
		; Saved 1st drawn column
		mov	[ebp.binf_tcol], ebx
	
		; Store in EDX the line position adjusted
		call	ED_LINEPOSADJUST
		mov	edx, [eax]
		cmp	edx, LINEPOSADJUST_MIN
		jl	@@NOK
		cmp	edx, LINEPOSADJUST_MAX
		jle	@@OK
@@NOK:		mov	edx, LINEPOSADJUST
@@OK:

		; Try to go back for LINEPOSADJUST lines
		clr	ecx
@@Adjust:	cmp	esi, [ebp.binf_BuffAddr]
		je	@@AdjustDone
		call	EdFindPrevLine
		dec	[ebp.binf_frow]
		inc	ecx
		cmp	cx, dx
		jne	@@Adjust
@@AdjustDone:	mov	[ebp.binf_1stLinAddr], esi
		
		; Draw page
		call	EdRedrawPage
		
		; Print the current line #
		call	EdShowLigNb
				
		; Position the cursor on the current position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------

@@Exit:
		
		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Update the cursor position on screen
		call	EdPosCur

		; Return
		ret

EdGotoLine	endp


; ===============================================================
; = EdGtoCommentField						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Move to the beginning of the comment field.		=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	all registers.						=
; ===============================================================

		align	4
EdGtoCommentField	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff				; EDX : addr of next line


		; ------------------
		; Goto comment field
		; ------------------
		
		; Update column
		call	ED_COMMENTFIELD
		mov	eax, [eax]
		cmp	eax, XCOMMENT_MIN
		jl	@@NOK
		cmp	eax, XCOMMENT_MAX
		jle	@@OK
@@NOK:		mov	eax, XCOMMENT_DFLT			; Value default if incorrect
@@OK:		mov	[ebp.binf_tcol], eax
		
		; Update 1st drawn column
		mov	eax, [ebp.binf_fcol]
		mov	[ebp.binf_fcol], 0
		cmpz	eax, 0
		je	@@Ok
		call	EdRedrawPage				; Redraw all page
@@Ok:
		
		; Print the current column #
		call	EdShowColNb

		; Update the cursor position on screen
		call	EdPosCur

		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
EdGtoCommentField	endp

endif

code32		ends
		end
