; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED_CMD2.ASM						!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32 		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

if TURNKEYED eq 0

include		Ed.inc					; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdDelChar							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Delete the character under the cursor.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdDelChar	proc

		; --------------------------
		; Current buffer is modified
		; --------------------------

		call	EdCurrBuffIsMod


		; ----------------------------------
		; Cancel the block selection, if any
		; ----------------------------------
		
		call	EdInactiveBlock


		; -------------------------------------
		; Delete the character under the cursor
		; -------------------------------------

		; Move line
		lea	edi, EdCurrLineBuff				
		add	edi, [ebp.binf_tcol]				; Destination address
		mov	esi, edi
		inc	esi						; Source address
		mov	ecx, EDMAXSZLINE
		sub	ecx, [ebp.binf_tcol]				; Length of the block
		dec	ecx
		cld
		rep	movsb  			                	; Move/Copy the block
		
		; Add a blank character at line's end
		mov	byte ptr [EdCurrLineBuff + EDMAXSZLINE - 1], ' '

		; Redraw the line
		call	EdDrawCurrLine
		
		; Put cursor on right position on screen
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdDelChar	endp


; ===============================================================
; = EdBSChar							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Delete the character left the cursor.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdBSChar		proc

		; -------------------		
		; Do nothing if col 0
		; -------------------
		
		cmp	[ebp.binf_tcol], 0
		je	@@Exit
		
		
		; -----------------------
		; Move one character left
		; -----------------------
		
		call	EdMoveLeft


		; -------------------------------------
		; Delete the character under the cursor
		; -------------------------------------

		call	EdDelChar


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdBSChar		endp


; ===============================================================
; = EdDelToEOL							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Delete from cursor to End Of Line			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdDelToEOL	proc

		; ----------------------------------
		; Cancel the block selection, if any
		; ----------------------------------
		
		call	EdInactiveBlock


		; ---------------------------------
		; Delete from cursor to End Of Line
		; ---------------------------------
		
		; Compute the length of the current line
		call	EdCompCurLinSz				; New length in EAX
		
		; Current position into line
		mov	ebx, [ebp.binf_tcol]
		
		; Clear to EOL
		lea	edx, EdCurrLineBuff
@@Loop:		cmp	bx, ax
		je	@@Done
		mov	byte ptr [edx + ebx], ' '
		inc	bx
		jmp	@@Loop
@@Done:

		; Redraw the current line
		call	EdDrawCurrLine

		; Current buffer is modified
		call	EdCurrBuffIsMod
		
		; Put screen cursor on right position
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdDelToEOL	endp


; ===============================================================
; = EdDelRightWord						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Delete right word.					=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdDelRightWord	proc

		; --------------------------------------
		; Compute the length of the current line
		; --------------------------------------

		call	EdCompCurLinSz				; New length in EAX
		mov	ecx, eax


		; -------------------
		; The actual position
		; -------------------
		
		mov	ebx, [ebp.binf_tcol]
		mov	edx, ebx
		
		
		; ----------------------------
		; Do nothing if at end of line
		; ----------------------------
		
		cmp	bx, ax
		je	@@Exit
		push	eax					; Save length
		
		
		; ----------------------------------
		; Cancel the block selection, if any
		; ----------------------------------

		call	EdInactiveBlock


		; -----------------
		; Move to next word
		; -----------------
		
		; Move pointer to new position
		lea	esi, EdCurrLineBuff
		cmp	byte ptr [esi + ebx], ' '
		jne	@@CR2
@@CR4:		inc	bx
		cmp	bx, ax
		je	@@CR3
		cmp	byte ptr [esi + ebx], ' '
		je	@@CR4
		jmp	@@CR3
@@CR2:		inc	bx
		cmp	bx, ax
		je	@@CR3
		cmp	byte ptr [esi + ebx], ' '
		jne	@@CR2
		jmp	@@CR4
@@CR3:		

		; Move portion to be deleted
		push	edx					; Save actual position
		add	esi, ebx				; Source address in ESI
		lea	edi, EdCurrLineBuff + edx		; Destination address
		mov	edx, esi
		sub	edx, offset EdCurrLineBuff
		sub	ecx, edx				; Length of the block
		push	ecx					; Save length of the block
		cld						; Inc ESI and EDI
		rep	movsb  			        	; Move/Copy the block
		
		; Blank the end of the line
		pop	ecx edx eax				; Pop actual position (edx), block length (ecx), line length
		lea	edi, EdCurrLineBuff
		add	edx, ecx
		add	edi, edx				; EDI points to starting address
		lea	ecx, [eax - edx]			; Size of the block to write
		mov	al, ' '					; Value to write
		cld						; Inc EDI and ESI
		rep	stosb					; Write the block


		; ----
		; Done
		; ----

		; Redraw the current line
		call	EdDrawCurrLine

		; Current buffer is modified
		call	EdCurrBuffIsMod
		
		; Put screen cursor on right position
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdDelRightWord	endp


; ===============================================================
; = EdDelLine							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Delete current line.					=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdDelLine	proc

		; ---------------------------------
		; Do nothing if we are on last line
		; ---------------------------------
		
		mov	eax, [ebp.binf_trow]
		inc	eax
		cmp	eax, [ebp.binf_LinesNb]
		je	@@Exit
		

		; ----------------------------------
		; Cancel the block selection, if any
		; ----------------------------------
		
		call	EdInactiveBlock


		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff


		; --------------------------
		; Current buffer is modified
		; --------------------------
		
		call	EdCurrBuffIsMod


		; -------------------
		; Delete current line
		; -------------------

		; Destination address
		mov	edi, [ebp.binf_CurLinAddr]		; Current line addr
		
		; Source address
		mov	ecx, edi
		call	EdGetCurLinBuff				; Bring the current line into editing buffer
		mov	esi, ecx
		
		; Nb of characters to move
		mov	ecx, [ebp.binf_BuffAddr]		; Buffer addr
		call	EdCompTextSize				; Text size in EAX		
		add	ecx, eax
		sub	ecx, esi
		
		; Move text
		cld						; Inc ESI and EDI
		rep	movsb  			                ; Move/Copy the block
		
		
		; ------------------------
		; Get current line address
		; ------------------------
		
		mov	ecx, [ebp.binf_CurLinAddr]
		call	EdGetCurLinBuff
		
		
		; ----------------------------------
		; Update the nb of lines of the text
		; ----------------------------------
		
		dec	[ebp.binf_LinesNb]		
		call	EdShowNbLines
		

		; -------------------------------------------------
		; Update the position of tagged marks / definitions
		; -------------------------------------------------

		mov	edi, -1					; Offset in nb of lines		
		call	EdUpdatePosTags


		; ---------------
		; Redraw the page
		; ---------------

		; Goto 1st column
		mov	[ebp.binf_tcol], 0
		mov	[ebp.binf_fcol], 0
		
		; Redraw the text page
		call	EdRedrawPage
		
		; Put screen cursor on right position
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdDelLine	endp


; ===============================================================
; = EdJoinLine							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Join next line to current line.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdJoinLine	proc

		call	EdNotAvailable
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdJoinLine	endp


; ===============================================================
; = EdBreakLine							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Join break line to current line.			=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdBreakLine	proc

		call	EdNotAvailable
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdBreakLine	endp


; ===============================================================
; = EdInsLine							=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Insert a line.						=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdInsLine	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff


		; ----------------------------------
		; Cancel the block selection, if any
		; ----------------------------------
		
		call	EdInactiveBlock


		; -----------------------
		; Room for 2 more bytes ?
		; -----------------------
		
		mov	eax, 2
		call	EdRequestMem
		jc	@@Exit


		; --------------------------
		; Current buffer is modified
		; --------------------------
		
		call	EdCurrBuffIsMod


		; -----------------------------
		; Goto to start of current line
		; -----------------------------
		
		call	EdCarrReturn
		call	EdMoveUp
		

		; -----------------
		; Insert a new line
		; -----------------

		; Source address
		mov	esi, [ebp.binf_BuffAddr]		; Buffer addr
		call	EdCompTextSize				; Text size in EAX		
		add	esi, eax
		dec	esi					; ESI : last char of the text
		
		; Destination address
		mov	edi, esi
		add	edi, 2
		
		; Nb of characters to move
		mov	ecx, esi
		sub	ecx, [ebp.binf_CurLinAddr]
		inc	ecx
		
		; Move text
		std						; Dec ESI and EDI
		rep	movsb  			                ; Move/Copy the block
		cld
		
		
		; ----------------
		; Put the new line
		; ----------------

		mov	esi, [ebp.binf_CurLinAddr]
		mov	byte ptr [esi],     0Dh
		mov	byte ptr [esi + 1], 0Ah		
		
		
		; ----------------------------------
		; Update the nb of lines of the text
		; ----------------------------------
		
		inc	[ebp.binf_LinesNb]		
		call	EdShowNbLines


		; -------------------------------------------------
		; Update the position of tagged marks / definitions
		; -------------------------------------------------

		mov	edi, 1				; Offset in nb of lines		
		call	EdUpdatePosTags


		; ---------------
		; Redraw the page
		; ---------------
		
		; Redraw the text page
		call	EdRedrawPage
		
		; Put screen cursor on right position
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdInsLine	endp

endif

code32		ends
		end
