; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED_CMD3.ASM						!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc					; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

if TURNKEYED eq 0

include		Ed.inc						; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdCopySelToBin						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Copy selection to scratch.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdCopySelToBin	proc

		; ---------------------------
		; If no selection, do nothing
		; ---------------------------
		
		mov	ecx, [ebp.binf_BlockBegin]
		cmp	ecx, -1
		je	@@Exit


		; -------------------------------------
		; Selection too big to fit into scrap ?
		; -------------------------------------
		
		; Compute size of the selection (in bytes)
		sub	ecx, [ebp.binf_BlockEnd]
		neg	ecx
		
		; Not too big ?
		mov	eax, [EdTotalScrapSize]
		cmp	ecx, eax
		jbe	@@SizeOk
		mov	ax, EDERRSELTOOBIG			; Selection too big to fit inside scrap
		call	EdError
		call	EdPrStatusLine				; Redraw the entire status line
		call	EdPosCur				; Put screen cursor on right position
		jmp	@@Exit
@@SizeOk:
		

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		push	ecx		
		call	EdPutCurLinBuff
		pop	ecx


		; ------------------------------------
		; Save the size of the text into scrap
		; ------------------------------------

		; in bytes		
		mov	[EdScrapSize], ecx
		
		; Compute the number of lines
		clr	ecx					; Line counter
		mov	esi, [ebp.binf_BlockBegin]
		mov	edi, [ebp.binf_BlockEnd]
@@CompNbLines:	cmp	byte ptr [esi], 0Dh
		jne	@@NotCR
		inc	ecx
@@NotCR:	inc	esi
		cmp	esi, edi
		jne	@@CompNbLines
		mov	[EdScrapNbLines], ecx


		; ----------------------------
		; Copy selected block to scrap
		; ----------------------------
		
		; Source address
		mov	esi, [ebp.binf_BlockBegin]
		
		; Destination address
		mov	edi, [EdScrap]
		
		; The nb of bytes
		mov	ecx, [EdScrapSize]
		
		; Move block
		cld
		rep	movsb
		

		; ----------------------------------------
		; Inactive the current selection selection
		; ----------------------------------------
			
		call	EdInactiveBlock


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdCopySelToBin	endp


; ===============================================================
; = EdPasteBin							=
; =                                     	            	=
; = Called by 'IsKeyACommand' : 				=
; = 	Copy scratch to text.					=
; =                       		                       	=
; = I)	none							=
; =                                 	                	=
; = O)	none							=
; =                                         	        	=
; = M)	all registers.						=
; ===============================================================

		align	4
EdPasteBin	proc

		; ----------------------------
		; Do nothing if scrap is empty
		; ----------------------------
		
		mov	eax, [EdScrapSize]
		cmpz	eax, 0
		je	@@Exit


		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


		; -----------------
		; Ask for new space
		; -----------------
		
		mov	eax, [EdScrapSize]
		call	EdRequestMem
		jc	@@Exit

	
		; -------------------------------------------------
		; Move the text, make a 'hole' where scrap will fit
		; -------------------------------------------------
		
		; Source address
		call	EdCompTextSize					; Size of the text into EAX
		mov	esi, eax
		add	esi, [ebp.binf_BuffAddr]
		dec	esi						; ESI : last char of the text
		
		; Destination address
		mov	edi, esi
		add	edi, [EdScrapSize]
		
		; The nb of bytes to move
		mov	ecx, esi
		sub	ecx, [ebp.binf_CurLinAddr]
		inc	ecx

		; Move block
		std							; Dec ESI and EDI
		rep	movsb  			                		; Move/Copy the block
		cld
		
		
		; ------------------------------------
		; Now insert the scrap into the 'hole'
		; ------------------------------------

		; Source address
		mov	esi, [EdScrap]
		
		; Destination address
		mov	edi, [ebp.binf_CurLinAddr]
		
		; The nb of bytes to move
		mov	ecx, [EdScrapSize]

		; Move block
		cld
		rep	movsb


		; --------------------------
		; Current buffer is modified
		; --------------------------
		
		call	EdCurrBuffIsMod


		; ----------------------------------
		; Update the nb of lines in the text
		; ----------------------------------

		; Update
		mov	eax, [EdScrapNbLines]		
		add	[ebp.binf_LinesNb], eax

		; Print the nb of lines
		call	EdShowNbLines
		

		; -------------------------------------------------
		; Update the position of tagged marks / definitions
		; -------------------------------------------------

		mov	edi, [EdScrapNbLines]			; Offset in nb of lines		
		call	EdUpdatePosTags


		; ---------------
		; Redraw the page
		; ---------------
		
		; Redraw the text page
		mov	eax, [ebp.binf_BlockBegin]
		mov	[ebp.binf_BlockBegin], -1
		call	EdRedrawPage
		mov	[ebp.binf_BlockBegin], eax
		
		; Put screen cursor on right position
		call	EdPosCur


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdPasteBin	endp


; ===============================================================
; = EdCopyLinToBin						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Copy current line to scratch.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdCopyLineToBin	proc

		; ----------------------------------------
		; Inactive the current selection selection
		; ----------------------------------------
			
		call	EdInactiveBlock


		; -----------------------------------------------------------
		; Update Begin/End address of pseudo-selection (current line)
		; -----------------------------------------------------------
		
		; Begin
		mov	esi, [ebp.binf_CurLinAddr]
		mov	[ebp.binf_BlockBegin], esi
		mov	edi, esi
		
		; End
		call	EdFindNextLine
		mov	[ebp.binf_BlockEnd], esi


		; ------------------------------------
		; Save the size of the text into scrap
		; ------------------------------------

		; In bytes
		mov	ecx, esi
		sub	ecx, edi
		mov	[EdScrapSize], ecx
		
		; The number of lines
		mov	[EdScrapNbLines], 1


		; -------------------------------------
		; Selection too big to fit into scrap ?
		; -------------------------------------

		mov	eax, [EdTotalScrapSize]
		cmp	ecx, eax
		jbe	@@SizeOk
		mov	ax, EDERRSELTOOBIG			; Selection too big to fit inside scrap
		call	EdError
		call	EdPrStatusLine				; Redraw the entire status line
		call	EdPosCur				; Put screen cursor on right position
		jmp	@@Exit
@@SizeOk:
		

		; -------------------------------
		; Put current line in edited text
		; -------------------------------
		
		call	EdPutCurLinBuff


		; ----------------------------
		; Copy selected block to scrap
		; ----------------------------
		
		; Source address
		mov	esi, [ebp.binf_BlockBegin]
		
		; Destination address
		mov	edi, [EdScrap]
		
		; The nb of bytes
		mov	ecx, [EdScrapSize]
		
		; Move block
		cld
		rep	movsb
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdCopyLineToBin	endp


; ===============================================================
; = EdCutLinToBin						=
; =								=
; = Called by 'IsKeyACommand' : 				=
; = 	Cut current line to scratch.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdCutLineToBin	proc

		; ------------------
		; Copy line to scrap
		; ------------------
		
		call	EdCopyLineToBin
		
		
		; ---------------------
		; Then delete this line
		; ---------------------
		
		call	EdDelLine


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdCutLineToBin	endp

endif

code32		ends
		end
