; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED_FILE.SM						!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

if TURNKEYED eq 0

include		Ed.inc					; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdTakeFileToBuf						=
; =								=
; = Usage:							=
; =	Move the current line read from file to text buffer.	=
; =								=
; = I)	EDX : pointer on line read from file.			=
; =								=
; = O)	CF  : 1 if error, 0 if not error			=
; =	ECX : expanded line length				=
; = 	EBX : pointer in expanded line				=
; =								=
; = M)	None, else FLAGS.					=
; =								=
; = Notes:							=
; =	- The file has not to be closed.			=
; ===============================================================

		align	4
EdTakeLineToBuf	proc

		; ------------------
		; Save all registers
		; ------------------
		
		push	eax edx edi
		lea	edi, @@Line
		
		
		; -------------------
		; Move each character
		; -------------------
		
		clr	ebx						; Counter
@@NextChar:	mov	al, [edx]					; Read character
		cmpz	al, 0
		je	@@EndLine
		cmp	al, K_TAB
		je	@@Tab
		cmp	al, 020h
		jae	@@CharOk
		
		; Error: control char not allowed
call	PrAX
call	CharIn
		mov	ax, EDERRCTRLCHAR				; Control characters not allowed
		call	EdError
		jmp	@@Error

		; Change TAB chars into space chars
@@Tab:		mov	al, ' '
@@Tab2:		mov	ah, bl
		inc	ah
		and	ah, 07h
		je	@@CharOk
		mov	[edi], al
		inc	edi
		inc	ebx
		cmp	ebx, EDMAXSZLINE
		jl	@@Tab2

		; Error: control char not allowed
		mov	ax, EDERRLINTOOLONG				; "Line too long"
		call	EdError
		jmp	@@Error

		; Character is OK
@@CharOk:	mov	[edi], al
		inc	edi
		inc	edx
		inc	ebx
		jmp	@@NextChar


		; -------------------
		; End of line reached
		; -------------------

@@EndLine:	
		; End-of-line mark
		mov	byte ptr [edi], 0Dh
		mov	byte ptr [edi + 1], 0Ah
;		add	edi, 2
		
		; Update the number of lines
		inc	[ebp.binf_LinesNb]
		call	EdShowNbLines
		

		; --------
		; No error
		; --------
		
@@NoError:	clc					; CF = 0
		jmp	@@Exit
		

		; --------
		; No error
		; --------
		
@@Error:	stc					; CF = 1


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		mov	ecx, ebx
		add	ecx, 2				; Expanded line length
		pop	edi edx eax
		lea	ebx, @@Line			; Private buffer -> line expanded
		ret

		align	4
@@Line		db	EDMAXSZLINE+3 dup (?)		; Private buffer to decompact line

EdTakeLineToBuf	endp


; ===============================================================
; = EdLoadFileToBuf						=
; =								=
; = Usage:							=
; =	Load the open file to current buffer.			=
; =								=
; = I)	BX  : file handle					=
; =								=
; = O)	AX  : if error, error code (with CF=1).			=
;	CF  : 1 if error, 0 if not error			=
; =								=
; = M)	None, else AX and FLAGS.				=
; =								=
; = Notes:							=
; =	- The file has not to be closed.			=
; ===============================================================

		align	4
EdLoadFileToBuf	proc

		; ------------------
		; Save all registers
		; ------------------
		
		pushad
		
		
		; ----------------------------
		; Some initial initializations
		; ----------------------------
		
		; Size of the text in buffer
		mov	[ebp.binf_LinesNb], 0
		call	EdShowNbLines
		
		; File pointer current position
		mov	[ebp.binf_seek], 0
		
		; Current position of text (next line)
		mov	edi, [ebp.binf_BuffAddr]
		
		
		; ----------------------------
		; Read next line from the file
		; ----------------------------
		
@@ReadNextLine:	mov	bx, [ebp.binf_hdl]				; File handle
		mov	ecx, EDMAXSZLINE+3				; Max bytes to read
		lea	edx, EdCurrLineBuff				; Buffer address
		call	doReadFile					; in EBX: bytes read from the file
		jnc	@@ReadFileOk
		call	EdError
		jmp	@@Error
@@ReadFileOk:

		; End-of-file reached ?
		cmpz	ebx, 0
		je	@@EndOfFile

		; Store a '\0' at the end of the line
		mov	byte ptr [edx + ebx], 0
		
		
		; ----------------------------------------------------
		; Find either: - end of line (0D + 0A), or end of file
		; ----------------------------------------------------
		
		clr	ebx						; Char counter
@@FindEOL:	mov	al, [edx + ebx]					; Get next char in the line
		cmpz	al, 0
		je	@@EndLineFound
		cmp	al, 0Dh
		je	@@CrFound
		inc	ebx
		cmp	ebx, EDMAXSZLINE
		jle	@@FindEOL
		mov	ax, EDERRLINTOOLONG				; "Line too long"
		call	EdError
		jmp	@@Error
		
@@CrFound:	mov	byte ptr [edx + ebx], 0				; End of the line
		inc	ebx
		cmp	byte ptr [edx + ebx], 0Ah
		jne	@@ErrCrLf
		inc	ebx
		jmp	@@EndLineFound

@@ErrCrLf:	mov	ax, EDERRNOCRLF					; "Line separator must be CR-LF."
		call	EdError
		jmp 	@@Error
		

@@EndLineFound:	

		; Seek file pointer after line read
		push	ax ebx ecx
		mov	al, SEEK_SET					; Methof of file pointer moving
		add	[ebp.binf_seek], ebx
		mov	ecx, [ebp.binf_seek]				; New offset position	
		mov	bx, [ebp.binf_hdl]				; File handle
		call	doSeekFile
		jnc	@@SeekFileOk
		call	EdError
		jmp	@@Error
@@SeekFileOk:	pop	ecx ebx ax

		; Move line to buffer
		call	EdTakeLineToBuf	; I) EDX:line read from file (compacted)   O) ECX:real line length  EBX:buffer
		jc	@@Error

		; Ask for memory
		mov	edx, [ebp.binf_BuffAddr]			; Save buffer address (can be moved)
		mov	eax, ecx
		call	EdRequestMem
		jc	@@Error

		; Update destination address (perhaps EdRequestMem has moved buffer !)
		mov	eax, [ebp.binf_BuffAddr]			; New addr
		sub	eax, edx					; - Old Addr -> offset
		add	edi, eax

		; Move local buffer
		push	esi edi ecx
		mov	esi, ebx					; ESI = Source address
									; EDI = Destination addres
									; ECX = Nb bytes to move
		clc							; Inc ESI/EDI
		rep	movsb						; Move data
		pop	ecx edi esi

		; Update new position
		add	edi, ecx


		; ---------------------
		; Read next line (loop)
		; ---------------------
		
		jmp	@@ReadNextLine
		

		; -------------------
		; End-of-file reached
		; -------------------
		
@@EndOfFile:	

		; Store EOF mark at the of the text
		mov	eax, 1
		call	EdRequestMem				; Ask for memory
		mov	byte ptr [edi], K_EOF			; Ctrl-Z at end of text

		
		; --------
		; No error
		; --------
		
@@NoError:	popad
		clc						; CF = 0
		ret						; jmp	@@Exit


		; --------
		; No error
		; --------
		
@@Error:		popad
		stc						; CF = 1


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdLoadFileToBuf	endp


; ===============================================================
; = VrfFileTwice						=
; =								=
; = Usage:							=
; =	Verify that a file is not open twice.			=
; =								=
; = I)	EDX : pointer to filename				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
VrfFileTwice	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx esi


		; --------------------------------------
		; Find 1st character to primary filename
		; --------------------------------------

		; Find the end of the filename
		clr	ecx
@@FindEndFN:	cmp	byte ptr [edx + ecx], 0
		je	@@FindEndFN2
		inc	ecx
		jmp	@@FindEndFN
@@FindEndFN2:

		; Find the beginning of the primary name ('\' or begin)
@@FindEndFN3:	cmpz	ecx, 0
		je	@@FindEndFN5
		cmp	byte ptr [edx + ecx], '\'
		je	@@FindEndFN4
		cmp	byte ptr [edx + ecx], '/'
		je	@@FindEndFN4
		dec	ecx
		jmp	@@FindEndFN3
@@FindEndFN4:	inc	ecx
@@FindEndFN5:	add	edx, ecx


		; ------------------------------------
		; Search in all buffers, less this one
		; ------------------------------------
		
		clr	ecx					; Buffers counter
@@SrchBuff:	cmp	ecx, [EdCurrBufferNb]
		je	@@SrchBuff2
		imul	esi, ecx, size buffedentry
		add	esi, offset EdBuffInfos
		clr	ebx
@@SrchBuff3:	mov	al, [edx + ebx]
		TOUPPER	al
		mov	ah, [esi.binf_PrimFName + ebx]
		TOUPPER	ah
		cmp	al, ah
		jne	@@SrchBuff2
		or	al, ah
		jz	@@Found
		inc	ebx
		jmp	@@SrchBuff3
@@SrchBuff2:	inc	ecx
		cmp	cl, EDMAXNBBUFF
		jne	@@SrchBuff		

		jmp	@@NotFound


		; ----------------
		; Return to caller
		; ----------------

@@NotFound:	clc
		jmp	@@Exit

@@Found:		mov	ax, EDERRFILETWICE			; "A file can't be open twice"
		call	EdError
		stc
		
@@Exit:		pop	esi edx ecx ebx eax
		ret

VrfFileTwice	endp


; ===============================================================
; = EdLoadFile							=
; =								=
; = Usage:							=
; =	Load a asked file into current buffer. Is called 	=
; =	also by 'EdGotoWord'.					=
; =								=
; = I)	EAX = ptr to filename to load, or zero.			=
; =								=
; = O)	none							=
; =								=
; = M)	NONE.							=
; ===============================================================

		align	4
EdLoadFile	proc

		; ------------------
		; Save all registers
		; ------------------
		
		pushad
		

		; ------------------------------------------------------
		; Is a filename given (case if called by 'EdGotoWord') ?
		; ------------------------------------------------------

		mov	edx, eax		
		mov	[@@SavedEAX], eax			; Save initial value EAX given on entry
		cmpz	eax, 0
		jne	@@OpenNow
		

		; -----------------------------------
		; If buffer not saved, ask to confirm 
		; -----------------------------------
		
		cmp	[ebp.binf_NotSaved], FALSE
		je	@@Ok
		
		; Ask for confirmation
		lea	edx, EdMsgConfLdFile
		call	EdConfirm
		jnc	@@Exit					; Abort or Not confirmation
		
@@Ok:

		; ----------------------------
		; Ask for the filename to load
		; ----------------------------
		
		lea	edx, EdMsgLoadFile
		mov	bl, 40
		call	EdPrompt
		jc 	@@Exit					; Abort


		; -------------------------------------------------------------
		; If extension is not given, add .FTH extension to the filename
		; -------------------------------------------------------------

@@OpenNow:	call	EdAddExtFTH


		; -----------------------------------------------------------
		; Verify that this file is not already open in another buffer
		; -----------------------------------------------------------
		
		call	VrfFileTwice
		jc	@@Exit


		; --------------------
		; Try to open the file
		; --------------------

		; Redraw the entire status line
		call	EdPrStatusLine
		
		mov	ebx, edx				; Pointer to filename
		mov	ax, 010h				; Access mode: inherited + Deny both + Read
		clr	cl					; Set CL=0 (not create new file)
		call	doWideOpenFile
		jnc	@@OpenOk
		call	EdError
		jmp	@@Exit
@@OpenOk:
		; Save the handle
		mov	[ebp.binf_hdl], bx

		mov	eax, ebp
		add	eax, binf_FileName			; Pointer on filename (start)
		mov	ebx, eax
		add	ebx, MAXSZFILENAME - 1			; Pointer on max end
@@Loop:		cmp	eax, ebx
		je	@@Stop
		cmp	byte ptr [ebx], '\'
		je	@@Found
		cmp	byte ptr [ebx], ':'
		je	@@Found
		dec	ebx
		jmp	@@Loop
		
@@Found:	inc	ebx

@@Stop:		


		; ----------------------------------------
		; Give back the memory used by this buffer
		; ----------------------------------------

		; Give back memory
		mov	eax, [ebp.binf_BuffAllSz]
		sub	eax, INITSZBUFFED
		jz	@@GiveBackDone
		call	EdGiveBackMem
@@GiveBackDone:


		; -------------------------------------
		; Read the file into the current buffer
		; -------------------------------------
		
		; Clear buffer size
		mov	[ebp.binf_CurrSize], 0

		; Load file
		call	EdLoadFileToBuf
		
		
		; --------------
		; Close the file
		; --------------
		
		mov	bx, [ebp.binf_hdl]
		call	doCloseFile
		jnc	@@CloseOk
		call	EdError
		jmp	@@Exit
@@CloseOk:

		; -------------------------------------
		; Update the filename of current buffer
		; -------------------------------------

		; Zero the filename
		clr	eax					; Cunter
		mov	ecx, ebp
		add	ecx, binf_FileName			; Destination
@@ZerFileName:	mov	byte ptr [ecx + eax], 0
		inc	eax
		cmp	eax, MAXSZFILENAME
		jl	@@ZerFileName
		
		; Store filename
		cmp	[@@SavedEAX], 0				; Is a filename given (case if called by 'EdGotoWord') ?
		jne	@@StoreFN2	
		lea	ebx, EdPromptBuffer			; Source
		jmp	@@StoreFN
@@StoreFN2:	mov	ebx, [@@SavedEAX]	
@@StoreFN:	clr	eax					; Char counter
@@UpdtFN:	mov	dl, [ebx + eax]
		TOUPPER	dl
		mov	[ecx + eax], dl
		cmpz	dl, 0
		je	@@UpdtFN2
		inc	eax
		jmp	@@UpdtFN
@@UpdtFN2:


		; --------------------------------------
		; Find 1st character to primary filename
		; --------------------------------------

		; Find the end of the filename
		clr	ecx
@@FindEndFN:	cmp	byte ptr [ebp.binf_FileName + ecx], 0
		je	@@FindEndFN2
		inc	ecx
		jmp	@@FindEndFN
@@FindEndFN2:

		; Find the beginning of the primary name ('/' or begin)
@@FindEndFN3:	cmpz	ecx, 0
		je	@@FindEndFN5
		cmp	byte ptr [ebp.binf_FileName + ecx], '/'
		je	@@FindEndFN4
		cmp	byte ptr [ebp.binf_FileName + ecx], '\'
		je	@@FindEndFN4
		dec	ecx
		jmp	@@FindEndFN3
@@FindEndFN4:	inc	ecx
@@FindEndFN5:	

		; Copy the primary filename
		mov	edx, ecx
		clr	ecx
		dec	ecx
		dec	edx
@@CopyFN:	inc	ecx
		inc	edx
		mov	al, [ebp.binf_FileName + edx]
		mov	[ebp.binf_PrimFName + ecx], al
		cmpz	al, 0
		jne	@@CopyFN
		
		; Draw filename on screen
		call	EdShowFileName


		; --------------------------------------------------------------------------------
		; Line # informations is no longer available for the words recorded into this file
		; --------------------------------------------------------------------------------
		
		cmp	[@@SavedEAX], 0			; Is a filename given (case if called by 'EdGotoWord') ?
		jne	@@StillAvail
		call	EdLineNbNotAvail
@@StillAvail:


		; ----------------------------------
		; Move cursor to the top of the text
		; ----------------------------------
		
		; #row and #col
		mov	[ebp.binf_frow], 0			; First row # drawed (1st line of the screen)
		mov	[ebp.binf_fcol], 0			; Col # (real col, in expanded line, as show on screen)
		mov	[ebp.binf_trow], 0			; Row #
		mov	[ebp.binf_tcol], 0			; Col #
		
		; Modified and not saved ?
		mov	[ebp.binf_NotSaved], 0

		; Current line address (in the edited text)
		mov	eax, [ebp.binf_BuffAddr]		; Address of buffer allocated dynamically for the edited text
		mov	[ebp.binf_1stLinAddr], eax		; Address of top screen line (1st line drawed on top)
		mov	[ebp.binf_CurLinAddr], eax		; Current line address (in the edited text)
		

		; ----------------------------
		; The text is loaded in memory
		; ----------------------------
		
		mov	[ebp.binf_InMemory], TRUE

		
		; ----------------------------
		; Now redraw the complete page
		; ----------------------------
		
		call	EdRedrawPage


		; ----------------------------------------
		; Bring 1st line of text to editing buffer
		; ----------------------------------------

		mov	ecx, [ebp.binf_CurLinAddr]
		call	EdGetCurLinBuff
		
		; Done
		jmp	@@ExitOk


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		

		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Update the cursor position on screen
@@ExitOk:	call	EdPosCur

		; Return
		popad
		ret

		align	4
@@SavedEAX	dd	?				; Save initial value EAX given on entry

EdLoadFile	endp


; ===============================================================
; = EdWritFileToBuf						=
; =								=
; = Usage:							=
; =	Save the current buffer to the open file.		=
; =								=
; = I)	None.							=
; =								=
; = O)	AX  : if error, error code (width CF=1).		=
;	CF  : 1 if error, 0 if not error			=
; =								=
; = M)	None, else AX and FLAGS.				=
; =								=
; = Notes:							=
; =	- The file has not to be closed.			=
; ===============================================================

		align	4
EdWritFileToBuf	proc

		; ------------------
		; Save all registers
		; ------------------
		
		pushad


		; ---------------------------------
		; Scan each line, try to compact it
		; ---------------------------------
		
		mov	esi, [ebp.binf_BuffAddr]
		mov	edi, [EdWriteBuffAddr]			; Temp buffer for write
		mov	[@@Used], 0				; Temp buffer size used
		
		; Use tab (compact spaces to tab) ?
		call	ED_QUSETAB
		mov	eax, [eax]
		mov	[@@UseTab], al
		
		
@@NextLine:	

		; End of text ?
		; -------------
		
		cmp	byte ptr [esi], K_EOF
		je	@@SaveData
		
		
		; Compact the line
		; ----------------
		
		clr	ebx				; Counter into source
		clr	ecx				; Counter into destination
		mov	[@@InString], FALSE		; Flag: string ? If true, do not compact with TAB

@@GetChar:	mov	al, [esi + ebx]
		cmp	al, 0Dh
		je	@@EndChar
		cmp	al, '"'
		jne	@@NotQuote
		mov	[@@InString], TRUE
		jmp	@@StoreChar
@@NotQuote:	cmp	al, ' '
		jne	@@StoreChar
		cmp	[@@UseTab], FALSE
		je	@@StoreChar
		cmp	[@@InString], TRUE
		je	@@StoreChar
		mov	edx, ebx

@@LL1:		cmp	byte ptr [esi + edx + 1], ' '
		jne	@@LL3
		mov	ah, dl
		inc	ah
		test	ah, 07h
		jnz	@@LL2
		mov	ebx, edx
		mov	al, K_TAB
		jmp	@@StoreChar
@@LL2:		inc	edx
		jmp	@@LL1

@@LL3:		mov	byte ptr [edi + ecx], ' '
		inc	ebx
		inc	ecx
		cmp	ebx, edx
		jg	@@GetChar
		jmp	@@LL3

@@StoreChar:	mov	[edi + ecx], al
		inc	ebx
		inc	ecx
		jmp	@@GetChar

@@EndChar:	mov	byte ptr [edi + ecx], 0Dh
		mov	byte ptr [edi + ecx + 1], 0Ah
		add	ecx, 2
		add	esi, ebx
		add	esi, 2
		add	edi, ecx


		; Enought place into temp write buffer ?
		; --------------------------------------

		add	[@@Used], ecx				; Update size used		
		cmp	[@@Used], (EDWRITEBUFFSZ - EDMAXSZLINE)	; Compare with max size - max line length
		jb	@@NextLine


		; Save data to file
		; -----------------

@@SaveData:	
		cmpz	ecx, 0
		je	@@NoError

		mov	ecx, [@@Used]				; ECX : Nb of bytes to write
		mov	bx, [ebp.binf_hdl]			; File handle
		mov	edx, [EdWriteBuffAddr]			; Buffer address
		call	doWriteFile				; in EBX: bytes writen the file
		jnc	@@WriteFileOk
		call	EdError
		jmp	@@Error
@@WriteFileOk:	cmp	ebx, ecx
		jne	@@Error

		mov	edi, [EdWriteBuffAddr]			; Temp buffer for write
		mov	[@@Used], 0				; Temp buffer size used

		cmp	byte ptr [esi], K_EOF
		je	@@NoError


		; Find next line start address
		; ----------------------------
		
		jmp	@@NextLine


		; --------
		; No error
		; --------
		
@@NoError:	popad
		clc					; CF = 0
		jmp	@@Exit


		; --------
		; No error
		; --------
		
@@Error:	popad
		stc					; CF = 1


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

		align	4
@@Used		dd	?				; Buffer used
@@InString	db	?				; Flag: string ? If true, do not compact with TAB
		align	4
@@UseTab	db	?				; Use tab ?

EdWritFileToBuf	endp

		
; ===============================================================
; = EdSaveFile							=
; =								=
; = Usage:							=
; =	Save current buffer to file.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	All registers						=
; ===============================================================

		align	4
EdSaveFile	proc

		; ---------------------------------------------------
		; Do 'Save as' command if filename is not still given
		; ---------------------------------------------------
		
		cmp	byte ptr [ebp.binf_FileName], 0
		jne	@@NameOk
		call	EdSaveFileAs
		jmp	@@Exit
@@NameOk:


		; ----------------------
		; Try to create the file
		; ----------------------

		mov	ebx, ebp
		add	ebx, binf_FileName			; Pointer to filename
		mov	ax, 00h					; File attribute : normal
		call	doCreateFile
		cmpz	ax, 0
		je	@@OpenOk
		Call	EdError
		jmp	@@ExitNOk
@@OpenOk:
		; Save the handle
		mov	[ebp.binf_hdl], bx


		; ------------------------------------
		; Write the file to the current buffer
		; ------------------------------------

		call	EdPutCurLinBuff
		call	EdWritFileToBuf
		
		
		; --------------
		; Close the file
		; --------------
		
		mov	bx, [ebp.binf_hdl]
		call	doCloseFile
		cmpz	ax, 0
		je	@@CloseOk
		call	EdError
		jmp	@@ExitNOk
@@CloseOk:


		; -----------------------
		; Update the 'touch' flag
		; -----------------------
		
		mov	[ebp.binf_NotSaved], 0

		; Redraw the touch flag
		call	EdShowNotSavedF
		
		; Exit, no error
		jmp	@@ExitOk
		


		; ----------------
		; Return to caller
		; ----------------
		
@@ExitNOk:		

		; Redraw the entire status line
		call	EdPrStatusLine


@@ExitOk:
		
		; Update the cursor position on screen
		call	EdPosCur

		; Return
@@Exit:		ret

EdSaveFile	endp


; ===============================================================
; = EdSaveFileAs						=
; =								=
; = Usage:							=
; =	Save current buffer to a given file.			=
; =								=
; = I)	none							=
; =								=
; = O)	CF = 0, ok    CF = 1: abort operation			=
; =								=
; = M)	All registers						=
; ===============================================================

		align	4
EdSaveFileAs	proc

		; -----------------------------
		; Ask for the filename to write
		; -----------------------------

		lea	edx, EdMsgSaveFile
		mov	bl, MAXSZFILENAME
		call	EdPrompt
		jc 	@@Exit						; Abort


		; -------------------------------------------------------------
		; If extension is not given, add .FTH extension to the filename
		; -------------------------------------------------------------

		call	EdAddExtFTH


		; -----------------------------------------------------------
		; Verify that this file is not already open in another buffer
		; -----------------------------------------------------------
		
		call	VrfFileTwice
		jc	@@Exit

		
		; ------------------------------
		; Does this file already exist ?
		; ------------------------------

		; Does this file exist ?
		mov	ebx, edx					; Pointer to filename
		call	doQFileExist
		cmpz	bl, FALSE
		je	@@NewFile

		; Ask for confirmation
		lea	edx, EdMsgFileAlExist
		call	EdConfirm
		jnc	@@Exit						; Abort or Not confirmation

		
		; -------------------------------------
		; Update the filename of current buffer
		; -------------------------------------
@@NewFile:

		; Zero the filename
		clr	eax						; Counter
		mov	ecx, ebp
		add	ecx, binf_FileName				; Destination
@@ZerFileName:	mov	byte ptr [ecx + eax], 0
		inc	eax
		cmp	eax, MAXSZFILENAME
		jl	@@ZerFileName
		
		; Store filename
		clr	eax						; Char counter
		lea	ebx, EdPromptBuffer				; Source
@@UpdtFN:	mov	dl, [ebx + eax]
		mov	[ecx + eax], dl
		cmpz	dl, 0
		je	@@UpdtFN2
		inc	eax
		jmp	@@UpdtFN
@@UpdtFN2:


		; --------------------------------------
		; Find 1st character to primary filename
		; --------------------------------------

		; Find the end of the filename
		clr	ecx
@@FindEndFN:	cmp	byte ptr [ebp.binf_FileName + ecx], 0
		je	@@FindEndFN2
		inc	ecx
		jmp	@@FindEndFN
@@FindEndFN2:

		; Find the beginning of the primary name ('\' or begin)
@@FindEndFN3:	cmpz	ecx, 0
		je	@@FindEndFN5
		cmp	byte ptr [ebp.binf_FileName + ecx], '\'
		je	@@FindEndFN4
		cmp	byte ptr [ebp.binf_FileName + ecx], '/'
		je	@@FindEndFN4
		dec	ecx
		jmp	@@FindEndFN3
@@FindEndFN4:	inc	ecx
@@FindEndFN5:	

		; Copy the primary filename
		mov	edx, ecx
		clr	ecx
		dec	ecx
		dec	edx
@@CopyFN:	inc	ecx
		inc	edx
		mov	al, [ebp.binf_FileName + edx]
		mov	[ebp.binf_PrimFName + ecx], al
		cmpz	al, 0
		jne	@@CopyFN
		
		; Draw filename on screen
		call	EdShowFileName


		; ---------------------
		; Save buffer into file
		; ---------------------
		
		call	EdSaveFile
		

		; --------		
		; No abort
		; --------
		
		clc						; CF = 0


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		

		; Save CF
		pushf

		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Update the cursor position on screen
		call	EdPosCur
		
		; Restore CF
		popf

		; Return
		ret

EdSaveFileAs	endp


; ===============================================================
; = EdSaveAllBuff						=
; =								=
; = Usage:							=
; =	Save all modified buffers.				=
; =								=
; = I)	none							=
; =								=
; = O)	CF=0 : ok    CF=1 : abort				=
; =								=
; = M)	All registers						=
; ===============================================================

		align	4
EdSaveAllBuff	proc

		; -------------------------------
		; Put current line in edited text
		; -------------------------------

		call	EdPutCurLinBuff
		

		; -------------------
		; Save current buffer
		; -------------------
		
		mov	ebx, [EdCurrBufferNb]
		mov	[@@SavedBuffNb], ebx
		
		
		; -----------------------------------------------------
		; Select each buffer, save-it if modified and not saved
		; -----------------------------------------------------
		
		; Initializations
		lea	ebp, EdBuffInfos
		clr	ah
		
		; Save-it ?
@@Save1:		cmp	[ebp.binf_NotSaved], 0
		je	@@Save2
		
		; Filename is given ?
		cmp	byte ptr [ebp.binf_FileName], 0
		je	@@Save3
		pushad
		call	EdSaveFile
		popad
		jmp	@@Save2
		
		; Filename is not given
@@Save3:		push	eax ebp
		call	EdRedrawPage
		call	EdSaveFileAs
		pop	ebp eax
		jc	@@Exit
		
		; Next buffer, if any
@@Save2:		add	ebp, size buffedentry
		inc	ah
		cmp	ah, EDMAXNBBUFF
		jne	@@Save1
		
		
		; ----------------------
		; Restore current buffer
		; ----------------------
		
		mov	ebx, [@@SavedBuffNb]
		mov	[EdCurrBufferNb], ebx
		
		
		; --------
		; No abort
		; --------
		
		clc						; CF = 0


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

		align	4
@@SavedBuffNb	dd	?

EdSaveAllBuff	endp

		
; ===============================================================
; = EdSaveBlock							=
; =								=
; = Usage:							=
; =	Save current selection into a specified file.		=
; =								=
; = I)	none							=
; =								=
; = O)	none.							=
; =								=
; = M)	All registers						=
; ===============================================================

		align	4
EdSaveBlock	proc

		call	EdNotAvailable		



		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdSaveBlock	endp

		
; ===============================================================
; = EdInsertFile						=
; =								=
; = Usage:							=
; =	Insert a file at current position.			=
; =								=
; = I)	none							=
; =								=
; = O)	none.							=
; =								=
; = M)	All registers						=
; ===============================================================

		align	4
EdInsertFile	proc

		call	EdNotAvailable		


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdInsertFile	endp

endif
		
code32		ends
		end
