; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	ED_UTIL.ASM						!
; 									!
; Description:	Definitions for playing with FROSTED, built-in editor.	!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

if TURNKEYED eq 0

include		Ed.inc					; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = EdPrNumber							=
; =								=
; = Print a decimal number, left justified.			=
; =								=
; = I)	EAX : the number to print				=
; =	BH  : row (from 0 to ..)				=
; =	BL  : column (from 0 to ..)				=
; =	CL  : the attribute					=
; =	CH  : width of the field				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdPrNumber	proc

		; ------------------
		; Save all registers
		; ------------------
		
		pushad

		; ESI points to current digit location
		lea	esi, EdDecNbBuff
		clr	edi					; Digit counter

		; Divide current value by 10 (base)
@@Loop:		clr	edx
		div	[EdDecNbBase]

		; Store digit
		add	dl, '0'
		mov	[esi], dl
		inc	di
		dec	esi
		cmpz	eax, 0
		jne	@@Loop

		; Complete with ' ' (left justified)
@@Stop:		inc	esi
		lea	ebp, EdDecNbBuff + 1
@@Compl:		mov	dx, di
		cmp	dl, ch
		jge	@@StopCompl
		mov	byte ptr [ebp], ' '
		inc	di
		inc	ebp
		jmp	@@Compl
@@StopCompl:


		; ----------------
		; Print the number
		; ----------------

		mov	ax, di					; Nb of chars to print (line width)
		mov	ah, cl 					; Attribute / Color
		mov	edx, esi 				; The string to print
		call	Ed_XYAS


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		popad
		ret

EdPrNumber	endp


; ===============================================================
; = EdError							=
; =								=
; = Usage:							=
; =	Print the I/O error msg (in AX)				=
; =								=
; = I)	AX : error code (MS/DOS and specific)			=
; =								=
; = O)	none							=
; =								=
; = M)	All registers						=
; ===============================================================

		align	4
EdError		proc

		; -------------------------------------------
		; Reload EBP: pointer to current buffer infos
		; -------------------------------------------

		mov	ebp, [EdCurrBufferNb]
		imul	ebp, size buffedentry
		add	ebp, offset EdBuffInfos


		; ---------------------
		; Erase the status line
		; ---------------------
		
		call	EdClrStatusLine


		; ---------------------------------
		; Search the msg, depends of code #
		; ---------------------------------

		lea	edx, EdMsgErr
		mov	si, -1					; If -1, dont't print the error # (found in table)
@@Compare:	mov	bx, [edx]
		cmp	bx, ax
		je	@@PrMsg
		cmpz	bx, 0
		je	@@Unknow
		inc	edx
@@Next:		inc	edx
		cmp	byte ptr [edx], 0
		jne	@@Next
		inc	edx
		jmp	@@Compare
		
@@Unknow:	lea	edx, EdMsgErrUnknow - 2
		mov	si, ax					; Save the error #


		; -----------------
		; Print the message	(pointed by EDX)
		; -----------------

@@PrMsg:		add	edx, 2					; 2 = size word
		mov	ah, [edc_sll]				; Attr / Color
		clr	al					; Asking to compute msg length
		clr	bh					; Row #
		clr	bl					; Col #
		call	Ed_XYAS
		
		
		; -------------------
		; Print the error # ?
		; -------------------

		cmp	si, -1
		je	@@PrintErrNbDone
		push	ax
		clr	bh					; Row #
		mov	bl, al					; Col #
		movzx	eax, si					; Nb to print (error #)
		mov	cl, [edc_sli]				; Attr / Color
		mov	ch, 5					; Width field
		call	EdPrNumber
		pop	ax
		add	al, 5
		
@@PrintErrNbDone:

		
		; -------------------------------------
		; Add the msg " Press ESC to continue "
		; -------------------------------------

		mov	cl, al

		mov	ah, [edc_sll]				; Attr / Color
		lea	edx, EdMsgErrCont				; Pointer on msg
		clr	bh					; Row #
		mov	bl, al					; Col # (it's the length of previous msg)
		clr	al					; Asking to compute msg length
		call	Ed_XYAS


		; --------------------------------
		; Put cursor at the end of the msg
		; --------------------------------

		clr	ah					; Row #
		add	al, cl					; Col # 
		call	EdGotoXY


		; ----------------
		; Wait for ESC key
		; ----------------
		
@@WaitEsc:	call	CharIn
		cmp	ax, K_ESC
		jne	@@WaitEsc
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

EdError		endp


; ===============================================================
; = EdCls							=
; =								=
; = Usage:							=
; = 	Erase all the page					=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdCls		proc

		; Initializations
		mov	bx, [EdRowsNb]				; Nb of lines
		mov	bh, bl
		dec	bh

		; Erase line
@@Erase:		mov	ah, [edc_txn]				; Attr / Color
		mov	al, ' '					; Erase line
								; Row # (in BH)
		clr	bl					; Col # (in BL)
		mov	cx, [EdColsNb]				; Nb of chars to print (in CL)
		call	Ed_XYNAC

		; Next line to erase
		dec	bh
		cmpz	bh, 0
		jge	@@Erase
		
		; Return to caller
		ret

EdCls		endp


; ===============================================================
; = EdClrStatusLine						=
; =								=
; = Usage:							=
; =	Erase the status line					=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	None.							=
; ===============================================================

		align	4
EdClrStatusLine	proc

		; -------------------
		; Save used registers
		; -------------------
		
		pushad


		; ----------------------------
		; Erase the entire status line
		; ----------------------------

		mov	ah, [edc_sll]				; Color / Attr
		mov	al, ' '					; Character to print
		clr	bh					; Row #
		clr	bl					; We start at column #0
		mov	cx, [EdColsNb]				; The width of the screen (in CL)
		call	Ed_XYNAC


		; ----------------
		; Return to caller
		; ----------------

		popad		
		ret
		
EdClrStatusLine	endp

		
; ===============================================================
; = EdClrMsgLine						=
; =								=
; = Usage:							=
; = 	Clear the msg line (last line on screen).		=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	None.							=
; ===============================================================

		align	4
EdClrMsgLine	proc

		; Save all registers
		pushad

		; Erase line		
		mov	ah, [edc_sll]				; Attr / Color
		mov	al, ' '					; Erase line
		mov	bx, [EdMsgRowNb]
		mov	bh, bl					; Row # (status line)
		clr	bl					; Column #
		mov	cx, [EdColsNb]				; Width of the line
		call	Ed_XYNAC

		; Exit
		popad
		ret

EdClrMsgLine	endp


; ===============================================================
; = EdPrMsgStatus						=
; =								=
; = Usage:							=
; =	Print a msg on status line.				=
; =								=
; = I)	EDX : points on msg to print.				=
; =								=
; = O)	none							=
; =								=
; = M)	all registers						=
; =								=
; = Note:							=
; =	The string pointed by EDX is endded with '\0'.		=
; ===============================================================

		align	4
EdPrMsgStatus	proc

		; -----------------
		; Erase status line
		; -----------------
		
		call	EdClrStatusLine


		; ----------------------------
		; Compute length of the string
		; ----------------------------
		
		clr	ecx
@@StrLen:	cmp	byte ptr [edx + ecx], 0
		je	@@EndString
		inc	ecx
		jmp	@@StrLen
@@EndString:
		

		; ------------------------
		; Print msg on status line
		; ------------------------

		mov	al, cl					; Length of msg
		mov	ah, [edc_sll]				; Attr / Color
		clr	bh					; Row #
		clr	bl					; Col #
		call	Ed_XYAS


		; --------------------------------
		; Put cursor at the end of the msg
		; --------------------------------

		mov	ah, ch					; Row #
		mov	al, cl					; Col #
		call	EdGotoXY


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
EdPrMsgStatus	endp


; ===============================================================
; = EdPrMsgMsg							=
; =								=
; = Usage:							=
; =	Print a msg on msg line.				=
; =								=
; = I)	EDX : points on msg to print.				=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; =								=
; = Note:							=
; =	The string pointed by EDX is endded with '\0'.		=
; ===============================================================

		align	4
EdPrMsgMsg	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax bx


		; --------------
		; Erase msg line
		; --------------
		
		call	EdClrMsgLine


		; ----------------------------
		; Compute length of the string
		; ----------------------------
		
		clr	eax
@@StrLen:	cmp	byte ptr [edx + eax], 0
		je	@@EndString
		inc	eax
		jmp	@@StrLen
@@EndString:
		

		; ------------------------
		; Print msg on status line
		; ------------------------

		mov	ah, [edc_sll]				; Attr / Color
		mov	bx, [EdMsgRowNb]
		mov	bh, bl					; Row # (status line)
		clr	bl					; Col #
		call	Ed_XYAS


		; ----------------
		; Return to caller
		; ----------------

		pop	bx eax		
		ret
		
EdPrMsgMsg	endp


; ===============================================================
; = EdPrompt							=
; =								=
; = Usage:							=
; =	Print a prompt msg, then ask for the parameter.		=
; =								=
; = I)	EDX : points on msg to print.				=
;	BL  : the max length for the text to input		=
; =								=
; = O)	CF = 1 if operation is aborted				=
; =	CF = 0 if operation is not aborted (CR)			=
; =	EDX  points on the answer.				=
; =	BH   is the length of the answer.			=
; =								=
; = M)	FLAGS							=
; =	All register.						=
; =								=
; = Note:							=
; =	The string pointed by EDX is endded with '\0'.		=
; ===============================================================

		align	4
EdPrompt		proc

		; ------------------------
		; Print msg on status line
		; ------------------------
		
		call	EdPrMsgStatus


		; ---------------
		; Read the answer
		; ---------------

		lea	edx, EdPromptBuffer			; Buffer to store the answer
		clr	bh					; Nb of char read		

@@ReadChar:	call	CharIn
		cmp	ax, K_ESC
		je	@@Abort
		cmp	ax, K_CR
		jne	@@Rd2
		
		; CR, ok ?
		cmpz	bh, 0
		je	@@ReadChar				; Not accept CR if answer is empty...
		jmp	@@AnswerOk
		
@@Rd2:		; Convert lowercase to uppercase
		cmp	ax, 'a'
		jl	@@Rd3
		cmp	ax, 'z'
		jg	@@Rd3
		sub	al, 020h
		jmp	@@RdTake
		
@@Rd3:		cmp	ax, K_BS
		jne	@@Rd4

		; BS
		; --
		cmpz	bh, 0
		je	@@ReadChar				; No char !
		
		; Cursor left
		push	ax
		call	EdWhereXY
		dec	al
		call	EdGotoXY
		
		; Print a space
		mov	al, ' '
		call	Ed_C
		pop	ax
		
		; Update nb of char in the answer
		dec	bh

		; Store the character into the buffer
		dec	edx

		; Continue
		jmp	@@ReadChar
		
@@Rd4:		cmp	ax, '.'
		je	@@RdTake
		cmp	ax, '$'
		je	@@RdTake
		cmp	ax, '\'
		je	@@RdTake
		cmp	ax, '/'
		je	@@RdTake
		cmp	ax, '-'
		je	@@RdTake
		cmp	ax, '_'
		je	@@RdTake
		cmp	ax, '='
		je	@@RdTake
		cmp	ax, '0'
		jl	@@Error
		cmp	ax, '9'
		jle	@@RdTake
		cmp	ax, '@'
		jl	@@Error
		cmp	ax, 'Z'
		jle	@@RdTake
		
@@Error:		; Error, char not ok
		jmp	@@ReadChar


@@RdTake:	; Char is OK
;		mov	ah, [edc_sll]				; Attr / Color
		call	Ed_C					; Print char

		; Update cursor position
		push	ax
		call	EdWhereXY
		inc	al
		call	EdGotoXY
		pop	ax
		
		; Update nb of char in the answer
		inc	bh

		; Store the character into the buffer
		mov	[edx], al
		inc	edx

		; Do again
		jmp	@@ReadChar
		

		; ---------------
		; Abort operation
		; ---------------
		
@@Abort:

		; Return with CF = 1
		stc
		jmp	@@Exit


		; ------------
		; Answer is OK
		; ------------
		
@@AnswerOk:

		; Return answer address
		mov	byte ptr [edx], 0		; Put '\0' at the end of the string
		lea	edx, EdPromptBuffer		; Buffer to store the answer

		; Return with CF = 0
		clc


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pushad
		pushfd
		
		; -----------------
		; Erase status line
		; -----------------
		
		call	EdClrStatusLine


		popfd
		popad
		ret

EdPrompt		endp


; ===============================================================
; = EdConfirm							=
; =								=
; = Usage:							=
; =	Print a msg of confirmation, then wait for Y or N.	=
; =								=
; = I)	EDX : points on msg to print.				=
; =								=
; = O)	CF = 1 if answer is Y					=
; =	CF = 0 if answer is N					=
; =								=
; = M)	FLAGS							=
; =								=
; = Notes:							=
; =	- The string pointed by EDX is endded with '\0'.	=
; =	- The status line is NOT redrawn after this function.	=
; ===============================================================

		align	4
EdConfirm	proc

		; ------------------
		; Save register used
		; ------------------
		
		push	eax ebx ecx edx
		

		; ------------------------
		; Print msg on status line
		; ------------------------
		
		call	EdPrMsgStatus


		; ----------------------------
		; Wait for user answer: Y or N
		; ----------------------------
		
@@WaitAnswer:	call	CharIn
		cmp	ax, 'Y'
		je	@@Yes
		cmp	ax, 'y'
		je	@@Yes
		cmp	ax, 'N'
		je	@@No
		cmp	ax, 'n'
		je	@@No
		cmp	ax, K_ESC
		je	@@No
		jmp	@@WaitAnswer
		

		; -------------
		; Answer is YES
		; -------------

@@Yes:		

		; Return with CF = 1
		stc
		jmp	@@Exit
		

		; ------------
		; Answer is NO
		; ------------

@@No:

		; Return with CF = 0
		clc


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		pop	edx ecx ebx eax
		ret

EdConfirm	endp


; ===============================================================
; = EdCompCurLinSz						=
; =								=
; = Usage:							=
; =	Compute the size of the current line edited.		=
; =								=
; = I)	none							=
; =								=
; = O)	EAX : length of the line				=
; =								=
; = M)	None.							=
; ===============================================================

		align	4
EdCompCurLinSz	proc

		; -------------
		; Register used
		; -------------
		
		push	ebx


		; -------------------
		; Compute line length
		; -------------------
		
		lea	ebx, EdCurrLineBuff
		mov	eax, EDMAXSZLINE - 1
@@Len1:		cmpz	eax, 0
		je	@@End
		cmp	byte ptr [ebx + eax - 1], ' '
		jne	@@End
		dec	eax
		jmp	@@Len1

@@End:		


		; ----------------
		; Return to caller
		; ----------------

		pop	ebx		
		ret

EdCompCurLinSz	endp


; ===============================================================
; = EdCompTextSize						=
; =								=
; = Usage:							=
; =	Compute the size of the current line edited.		=
; =								=
; = I)	none							=
; =								=
; = O)	EAX : length of the text.				=
; =								=
; = M)	None.							=
; ===============================================================

		align	4
EdCompTextSize	proc

		; -------------
		; Register used
		; -------------
		
		push	ebx


		; -------------------
		; Compute text length
		; -------------------
		
		clr	eax
		mov	ebx, [ebp.binf_BuffAddr]
@@Len1:		cmp	byte ptr [ebx + eax], K_EOF
		je	@@End
		inc	eax
		jmp	@@Len1

@@End:		inc	eax


		; ----------------
		; Return to caller
		; ----------------

		pop	ebx		
		ret

EdCompTextSize	endp


; ===============================================================
; = EdFindPrevLine						=
; =								=
; = Usage							=
; = 	Find the address of the previous line.			=
; =								=
; = I)	ESI : point on current line.				=
; =								=
; = O)	ESI : point on prev line.				=
; =	CF  = 1 if top is reached, else CF = 0			=
; =								=
; = M)	ESI and CF						=
; =								=
; = Note							=
; =	Never called for the 1st line of text !!!		=
; ===============================================================

		align	4
EdFindPrevLine	proc

		; Are we on the text ?
		cmp	esi, [ebp.binf_BuffAddr]
		je	@@TopText

		; Find previous CR
@@AA1:		dec	esi
		cmp	byte ptr [esi], 0Dh
		jne	@@AA1

		; Find previous CR
@@BB1:		cmp	esi, [ebp.binf_BuffAddr]
		je	@@TopText
		dec	esi
		cmp	byte ptr [esi], 0Dh
		jne	@@BB1
		add	esi, 2


		; Not the top of the text
		; -----------------------
		
@@NotTopText:	clc					; CF = 0
		ret					; jmp	@@Exit
		
		
		; The top of the text
		; -------------------
		
@@TopText:	stc					; CF = 1
		
		
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdFindPrevLine	endp


; ===============================================================
; = EdFindNextLine						=
; =								=
; = Usage							=
; = 	Find the address of the next line.			=
; =								=
; = I)	ESI : point on current line.				=
; =								=
; = O)	ESI : point on next line.				=
; =	CF  = 1 if end is reached, else CF = 0			=
; =								=
; = M)	ESI and CF						=
; =								=
; = Note							=
; =	Never called for the last line of text !!!		=
; ===============================================================

		align	4
EdFindNextLine	proc

		; -------------------------
		; The current line is empty
		; -------------------------
		
		cmp	byte ptr [esi], 0Dh
		jne	@@LL1

		add	esi, 2
		
		cmp	byte ptr [esi + 2], 1Ah
		je	@@EndText
		jmp	@@NotEndText


		; -----------------------------
		; The current line is not empty
		; -----------------------------
		
@@LL1:		

@@LL2:		inc	esi
		cmp	byte ptr [esi], 0Dh
		jne	@@LL2

		add	esi, 2		

		cmp	byte ptr [esi + 2], 1Ah
		je	@@EndText
	

		; Not the end of the text
		; -----------------------
		
@@NotEndText:	clc					; CF = 0
		ret					; jmp	@@Exit
		
		
		; The end of the text
		; -------------------
		
@@EndText:	stc					; CF = 1
		
		
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdFindNextLine	endp


; ===============================================================
; = EdPutCurLinBuff						=
; =								=
; = Usage:							=
; =	Put current line in edited text.			=
; =								=
; = I)	none							=
; =								=
; = O)	ECX : start of current line into text.			=
; =	EDX : start of next    line into text.			=
; =								=
; = M)	ECX EDX							=
; ===============================================================

		align	4
EdPutCurLinBuff	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx esi edi
		

		; ------------------------------------------------
		; Compute the length of the line (into the buffer)
		; ------------------------------------------------

		call	EdCompCurLinSz					; New length in EAX


		; ----------------------------------
		; Compute length of line in the text
		; ----------------------------------
		
		mov	edx, [ebp.binf_CurLinAddr]
		clr	ebx						; Length
@@Len1:		cmp	byte ptr [edx], 0Dh
		je	@@Len2
		inc	ebx
		inc	edx
		jmp	@@Len1
@@Len2:		
		; EDX : pointer to next line
		inc	edx
		inc	edx
		add	edx, eax
		sub	edx, ebx					; EDX points now on next line
		

		; --------------------------
		; Enought room into buffer ?
		; --------------------------

		cmp	eax, ebx
		jle	@@Smaller
		push	eax
		sub	eax, ebx					; EAX is the nb of chars added
		call	EdRequestMem
		pop	eax
		jc	@@Exit
@@Smaller:


		; --------------------------
		; Same length ? (EAX == EBX)
		; --------------------------
		
		cmp	eax, ebx
		jne	@@NotEqu
		
		; Simply move the line to text
@@PutLine:	lea	esi, EdCurrLineBuff				; Source address
		mov	edi, [ebp.binf_CurLinAddr]			; Destination address
		mov	ecx, eax					; Length of the block
		cld
		rep	movsb  			                	; Move/Copy the block
		
		jmp	@@Exit
		

		; ---------------------------------		
		; New line is SHORTER than old line
		; ---------------------------------		
		
@@NotEqu:	jg	@@Longer

		; Move the text (to top)
		mov	esi, [ebp.binf_CurLinAddr]
		add	esi, ebx					
		sub	esi, eax					; Source address
		mov	edi, [ebp.binf_CurLinAddr]			; Destination address
		push	eax
		call	EdCompTextSize
		mov	ecx, eax
		pop	eax
		add	ecx, [ebp.binf_BuffAddr]
		sub	ecx, edi
		add	ecx, eax
		sub	ecx, ebx					; Length of the block
		cld
		rep	movsb  			                	; Move/Copy the block
		

		; Now move buffer to text
		jmp	@@PutLine


		; --------------------------------		
		; New line is LONGER than old line
		; --------------------------------		
		
@@Longer:	
		; Move the text (to bottom)
		mov	esi, [ebp.binf_BuffAddr]
		push	eax
		call	EdCompTextSize
		mov	ecx, eax
		pop	eax
		add	esi, ecx
		dec	esi						; Source address
		mov	edi, esi
		add	edi, eax
		sub	edi, ebx					; Destination address
		mov	ecx, esi
		sub	ecx, [ebp.binf_CurLinAddr]
		inc	ecx						; Size of the block to move
		std
		rep	movsb  			                	; Move/Copy the block
		cld

		; Now move buffer to text
		jmp	@@PutLine



		; ----------------
		; Return to caller
		; ----------------

@@Exit:		
		pop	edi esi ebx eax
		mov	ecx, [ebp.binf_CurLinAddr]			; ECX : start of current line in text
		ret

EdPutCurLinBuff	endp


; ===============================================================
; = EdGetCurLinBuff						=
; =								=
; = Usage:							=
; =								=
; = I)	ECX : Address of the start of current line.		=
; =								=
; = O)	ECX : Address of the start of next line.		=
; =								=
; = M)	ECX							=
; ===============================================================

		align	4
EdGetCurLinBuff	proc

		; --------------
		; Save registers
		; --------------
		
		push	ax ebx edx
		

		; ---------------
		; Initializations
		; ---------------

		lea	edx, EdCurrLineBuff			; Destination address
		clr	ebx					; Counter (source)
		
		
		; ---------
		; Make line
		; ---------
		
		cmp	byte ptr [ecx], K_EOF
		je	@@EndLine
		
@@NextChar:	mov	al, [ecx + ebx]				; Read char from text
		cmp	al, 0Dh					; End of line ?
		je	@@EndLine
		mov	[edx + ebx], al				; Store char in buffer
		inc	ebx
		jmp	@@NextChar


		; ---------------------------------
		; End of line: end line with blanks
		; ---------------------------------
		
@@EndLine:	add	ecx, ebx
		add	ecx, 2
		mov	al, ' '
@@Compl:		cmp	bx, EDMAXSZLINE
		je	@@Ok
		mov	[edx + ebx], al
		inc	ebx
		jmp	@@Compl

@@Ok:


		; ----------------
		; Return to caller
		; ----------------
		
		pop	edx ebx ax
		ret

EdGetCurLinBuff	endp


; ===============================================================
; = EdPosCur							=
; =								=
; = Usage:							=
; =	Position the cursor on the current position on screen.	=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdPosCur		proc

		; Save used registers
		push	eax ebx

		; Compute #row
		mov	eax, [ebp.binf_trow]
		sub	eax, [ebp.binf_frow]
		inc	eax
		mov	ah, al					; AH is the #row
		
		; Compute #col
		mov	ebx, [ebp.binf_tcol]
		sub	ebx, [ebp.binf_fcol]
		mov	al, bl					; AL is the #col

		; Move cursor
		call	EdGotoXY
		
		; Return to caller
		pop	ebx eax
		ret

EdPosCur		endp


; ===============================================================
; = EdDrawCurrLine						=
; =								=
; = Usage:							=
; =	Redraw the entire current line. After draw, the cursor	=
; =	is not put at its right place.				=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	all registers.						=
; ===============================================================

		align	4
EdDrawCurrLine	proc

		; ----------
		; The line #
		; ----------

		mov	edi, [ebp.binf_trow]
		sub	edi, [ebp.binf_frow]


		; -------------------
		; Compute line length
		; -------------------
		
		call	EdCompCurLinSz				; Line length in AX


		; ------------------------------
		; Is the current line selected ?
		; ------------------------------

		mov	bh, [edc_txn]				; Attr / Color
		cmp	[ebp.binf_BlockBegin], -1
		je	@@Done		
		mov	esi, [ebp.binf_CurLinAddr]
		cmp	esi, [ebp.binf_BlockBegin]
		jb	@@Done
		cmp	esi, [ebp.binf_BlockEnd]
		jae	@@Done
		mov	bh, [edc_txs]				; Attr / Color
@@Done:


		; -------------
		; Draw the line 
		; -------------
		
		mov	ax, [EdColsNb]				; Nb of chars to print (width of the screen in AL)
		mov	ah, bh					; Color / Attribute
		mov	bx, [EdColsNb]
		mov	ebx, [ebp.binf_trow]
		sub	ebx, [ebp.binf_frow]			
		inc	bl
		mov	bh, bl					; Row # in BH
		clr	bl					; Col #
		lea	edx, EdCurrLineBuff			; Current line buffer
		add	edx, [ebp.binf_fcol]			; Hor scrolling: 1st col to draw
		call	Ed_XYAS


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdDrawCurrLine	endp


; ===============================================================
; = EdRedrawPage						=
; =								=
; = Usage:							=
; = 	Redraw the entire page, from current position.		=
; =								=
; = I)	none							=
; = O)	none							=
; = M)	none							=
; ===============================================================

		align	4
EdRedrawPage	proc

		; -------------------
		; Save used registers
		; -------------------
		
		pushad


		; -----------------------------------
		; Save pointer on current line edited
		; -----------------------------------
		
		mov	eax, [ebp.binf_CurLinAddr]
		push	eax


		; ---------------
		; Initializations
		; ---------------
		
		mov	esi, 1				; Line counter
		mov	edi, [ebp.binf_frow]		; Real line drawed
		mov	ecx, [ebp.binf_1stLinAddr]		; Address of top screen line (1st line drawed on top)
		
		
		; ---------
		; Draw line
		; ---------

@@DrawLine:

		; End of text reached ?
		cmp	byte ptr [ecx], K_EOF
		je	@@EndText

		; Get line from text to editing buffer
		mov	edx, ecx
		call	EdGetCurLinBuff

		; Is the current line selected ?
		mov	bh, [edc_txn]				; Attr / Color
		cmp	[ebp.binf_BlockBegin], -1
		je	@@Done		
		cmp	edx, [ebp.binf_BlockBegin]
		jb	@@Done
		cmp	edx, [ebp.binf_BlockEnd]
		jae	@@Done
		mov	bh, [edc_txs]				; Attr / Color
@@Done:		

		; Draw line
		mov	ax, [EdColsNb]				; Nb of chars to print (in AL)
		mov	ah, bh					; Color / Attribute
		mov	bx, si
		mov	bh, bl					; Row #
		clr	bl					; Col #
		lea	edx, EdCurrLineBuff			; Current line buffer
		add	edx, [ebp.binf_fcol]			; Hor scrolling: 1st col to draw
		call	Ed_XYAS

		
		; -----------------
		; Next line, if any
		; -----------------
		
		inc	esi
		inc	edi
		cmp	si, [EdLastRowNb]
		jle	@@DrawLine
		jmp	@@Exit
		
		
		; -----------------------
		; End of text was reached
		; -----------------------
		
@@EndText:	cmp	si, [EdLastRowNb]
		jg	@@Exit

		; Erase line
		mov	ah, [edc_txn]				; Attr / Color
		mov	al, ' '					; Erase line
		mov	bx, si
		mov	bh, bl					; Row #
		clr	bl					; Col #
		mov	cx, [EdColsNb]				; Nb of chars to print (in CL)
		call	Ed_XYNAC

		; Next line to erase
		inc	esi
		jmp	@@EndText


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		

		; Restore pointer on current line edited
		; --------------------------------------

		; Bring the previous line (before the call) into editing buffer
		pop	ecx		
		mov	[ebp.binf_CurLinAddr], ecx
		call	EdGetCurLinBuff

		popad
		ret

EdRedrawPage	endp


; ===============================================================
; = EdInactiveBlock						=
; =								=
; = Usage:							=
; = 	Mask a block, if any					=
; =								=
; = I)	none							=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdInactiveBlock	proc

		; No block present ?
		cmp	[ebp.binf_BlockBegin], -1
		je	@@Exit
		
		; Put current line in edited text
		push	ecx edx
		call	EdPutCurLinBuff
		pop	edx ecx

		; No more block
		mov	[ebp.binf_BlockBegin], -1

		; Redraw the text page
		call	EdRedrawPage
		
		; Put screen cursor on right position
		call	EdPosCur

		; Return to caller
@@Exit:		ret

EdInactiveBlock	endp


; ===============================================================
; = EdMoveBuffers_Add						=
; =								=
; = Usage:							=
; = 	Move buffers to add space for the current buffer.	=
; =								=
; = I)	EAX : memory asked					=
; =	EBX : current buffer #					=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdMoveBuffers_Add	proc

		; ------------------
		; Save all registers
		; ------------------

		pushad


		; ---------------
		; Move the blocks
		; ---------------

		; Only if not last buffer
		mov	ecx, ebx
		inc	ecx
		cmp	ecx, EDMAXNBBUFF
		je	@@NotMove

		; Compute source address
		mov	edx, EDMAXNBBUFF - 1
		imul	edx, size buffedentry
		add	edx, offset EdBuffInfos
		mov	esi, [edx.binf_BuffAddr]
		add	esi, [edx.binf_BuffAllSz]
		dec	esi
		
		; Compute destination address
		mov	edi, esi
		add	edi, eax
		
		; Number of bytes to move
		mov	edx, ebx
		inc	edx					; # of next buffer
		imul	edx, size buffedentry
		add	edx, offset EdBuffInfos
		mov	ecx, [edx.binf_BuffAddr]
		sub	ecx, esi
		neg	ecx
		inc	ecx

		; Move blocks
		std						; dec ESI and EDI
		rep	movsb					; Move/Copy the block
		cld
@@NotMove:


		; -----------------------------------
		; Update current block allocated size
		; -----------------------------------

		imul	edx, ebx, size buffedentry
		add	edx, offset EdBuffInfos
		add	[edx.binf_BuffAllSz], eax
		

		; -------------------------
		; Update each block address
		; -------------------------
		
		; Start on next block
@@NextBlk:	inc	ebx
		cmp	ebx, EDMAXNBBUFF
		je	@@StopUpdt

		; Update address
		imul	edx, ebx, size buffedentry
		add	edx, offset EdBuffInfos
		add	[edx.binf_BuffAddr], eax		; Address of buffer allocated dynamically for the edited text
		add	[edx.binf_1stLinAddr], eax		; Address of top screen line (1st line drawed on top)
		add	[edx.binf_CurLinAddr], eax		; Current line address (in the edited text)
		
		; Next block
		jmp	@@NextBlk
@@StopUpdt:		


		; ---------------------------------------
		; Update current buffer size used (total)
		; ---------------------------------------
		
		add	[EdCurrBuffSize], eax


		; ----------------
		; Return to caller
		; ----------------

		popad		
		ret

EdMoveBuffers_Add	endp


; ===============================================================
; = EdUnloadBuffers						=
; =								=
; = Usage:							=
; = 	Try to unload buffers to save space.			=
; =								=
; = I)	EAX : memory asked					=
; = O)	CF = 0 if ok, CF = 1 if out of memory.			=
; = M)	none							=
; ===============================================================

		align	4
EdUnloadBuffers	proc

		; --------------
		; Save used regs
		; --------------
		
		pushad
		

		; -------------------------------------------
		; Scan each buffer, except the current buffer
		; -------------------------------------------
		
		clr	ecx					; Buffer counter
@@Scan:		cmp	ecx, [EdCurrBufferNb]
		je	@@Next

		; Pointer to infos		
		imul	edx, ecx, size buffedentry
		add	edx, offset EdBuffInfos
		
		; Text into buffer ?
		cmp	[edx.binf_BuffAllSz], INITSZBUFFED
		je	@@Next
		
		; Buffer already un-loaded ?
		cmp	[edx.binf_InMemory], FALSE
		je	@@Next

		; Unload the buffer to disk
		mov	ebx, ecx
		call	EdDiskWriteBuffer
		jc	@@NOK

		; Is enought space now ?
		mov	ebx, [EdCurrBuffSize]			; Current memory used
		add	ebx, STEPSZBUFFED
		cmp	ebx, [EdTotalBuffSize]			; Compare with memory allocated
		jb	@@OK					; Enought memory

@@Next:		inc	ecx
		cmp	cx, EDMAXNBBUFF
		jne	@@Scan


		; ------
		; Not OK
		; ------

@@NOK:		popad
		stc
		ret


		; --
		; OK
		; --

@@OK:		popad
		clc		
		ret

EdUnloadBuffers	endp


; ===============================================================
; = EdRequestMem						=
; =								=
; = Usage:							=
; = 	Ask for memory for the current buffer.			=
; =								=
; = I)	EAX : memory asked					=
; =								=
; = O)	CF = 0 if ok, CF = 1 if out of memory.			=
; =								=
; = M)	CF							=
; ===============================================================

		align	4
EdRequestMem	proc

		; --------------
		; Registers used
		; --------------

		push	ebx ecx


		; -----------------------------------------------------
		; Buffer current size, BEFORE calling 'EdPutCurLinBuff'
		; -----------------------------------------------------

		mov	ebx, [ebp.binf_CurrSize]
		add	ebx, eax
		cmp	ebx, [ebp.binf_BuffAllSz]
		jb	@@OK
		
		
		; -------------------------
		; Is there enought memory ?
		; -------------------------

		mov	ecx, [EdCurrBuffSize]			; Current memory used
		add	ecx, eax
		add	ecx, STEPSZBUFFED
		cmp	ecx, [EdTotalBuffSize]			; Compare with memory allocated
		jb	@@EnoughtMem				; Enought memory


		; ----------------------------------------
		; Try to unload some buffers to save space
		; ----------------------------------------

		call	EdUnloadBuffers
		jc	@@NOK					; If no space

@@EnoughtMem:
		
		
		; --------------
		; Move buffers !
		; --------------

		push	eax
		mov	ecx, [ebp.binf_BuffAllSz]
		neg	ecx					; Found a multiple of STEPSZBUFFED
@@FoundMulti:	add	ecx, STEPSZBUFFED
		cmp	ecx, eax
		jl	@@FoundMulti
		mov	eax, ecx
		mov	ebx, [EdCurrBufferNb]
		call	EdMoveBuffers_Add
		pop	eax
		jmp	@@OK


		; -------------------
		; Out of memory error
		; -------------------

@@NOK:		pushad
		mov	ax, EDERRNOMEM				; Buffer memory full
		call	EdError

		; Redraw current line
		call	EdDrawCurrLine

		; Redraw the entire status line
		call	EdPrStatusLine
		
		; Restore cursor to right position
		call	EdPosCur
		popad

		stc						; CF = 1
		jmp	@@Exit


		; --------
		; No error
		; --------

@@OK:		
		; Update current buffer size
		add	[ebp.binf_CurrSize], eax
		
		; Some updates, only into editor session
		cmp	[EditorActive], TRUE
		jne	@@EdNotActive
		call	EdShowCurrSize
		call	EdPosCur				; Restore cursor to right position
@@EdNotActive:

		clc						; CF = 0

		
		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	ecx ebx
		ret

EdRequestMem	endp


; ===============================================================
; = EdGiveBackMem						=
; =								=
; = Usage:							=
; = 	Give back the memory used the current buffer.		=
; =								=
; = I)	EAX : memory given back.				=
; =								=
; = O)	none.							=
; =								=
; = M)	none.							=
; ===============================================================

		align	4
EdGiveBackMem	proc

		; ------------------
		; Save all registers
		; ------------------

		pushad


		; ---------------
		; Move the blocks
		; ---------------

		; Only if not last buffer
		mov	ebx, [EdCurrBufferNb]
		mov	ecx, ebx
		inc	ecx
		cmp	ecx, EDMAXNBBUFF
		je	@@MoveDone

		; Compute source address
		mov	edx, ebx
		inc	edx					; Buffer # after the one shrunk
		imul	edx, size buffedentry
		add	edx, offset EdBuffInfos
		mov	esi, [edx.binf_BuffAddr]
		
		; Compute destination address
		mov	edi, esi
		sub	edi, eax
		
		; Number of bytes to move
		mov	edx, EDMAXNBBUFF - 1			; # of next buffer
		imul	edx, size buffedentry
		add	edx, offset EdBuffInfos
		mov	ecx, [edx.binf_BuffAddr]
		add	ecx, [edx.binf_BuffAllSz]
		sub	ecx, esi

		; Move blocks
		cld						; Inc ESI and EDI
		rep	movsb					; Move/Copy the block

@@MoveDone:


		; -------------------------
		; Update each block address
		; -------------------------
		
		; Start on next block
		mov	ebx, [EdCurrBufferNb]
@@NextBlk:	inc	ebx
		cmp	ebx, EDMAXNBBUFF
		je	@@StopUpdt

		; Update address
		imul	edx, ebx, size buffedentry
		add	edx, offset EdBuffInfos
		sub	[edx.binf_BuffAddr], eax		; Address of buffer allocated dynamically for the edited text
		sub	[edx.binf_1stLinAddr], eax		; Address of top screen line (1st line drawed on top)
		sub	[edx.binf_CurLinAddr], eax		; Current line address (in the edited text)
		
		; Next block
		jmp	@@NextBlk

@@StopUpdt:		


		; -------------------------------
		; Update current buffer size used
		; -------------------------------
		
		; Update infos
		mov	[ebp.binf_CurrSize], 3			; 3 for: 0D, 0A, 1A
		mov	[ebp.binf_BuffAllSz], INITSZBUFFED

		; Total
		sub	[EdCurrBuffSize], eax


		; ----------------
		; Return to caller
		; ----------------

		popad		
		ret

EdGiveBackMem	endp


; ===============================================================
; = EdDiskWriteBuffer						=
; =								=
; = Usage:							=
; = 	Un-load current buffer on disk, to save memory.		=
; =								=
; = I)	EBX : buffer nb (0-39)					=
; =								=
; = O)	CF = 0 if ok, CF = 1 if disk error.			=
; =								=
; = M)	CF							=
; ===============================================================

		align	4
EdDiskWriteBuffer	proc

		; --------------
		; Registers used
		; --------------

		push	ebp eax ebx ecx edx


		; -----------------
		; Save the buffer ?
		; -----------------

		; Pointer to buffer infos
		mov	[@@BuffNb], ebx
		imul	ebp, ebx, size buffedentry
		add	ebp, offset EdBuffInfos
		
		; Save file, if needed
		cmp	[ebp.binf_NotSaved], FALSE
		je	@@SavedOk
		push	ebx
		call	EdSaveFile
		pop	ebx
		jc	@@Exit
@@SavedOk:


		; -------------
		; Open the file
		; -------------

		; Build the filename
		mov	eax, ebx
		mov	bl, 10
		idiv	bl		
		lea	edx, @@FileName
		add	al, '0'
		mov	[edx], al
		add	ah, '0'
		mov	[edx + 1], ah

		; Delete any previous file
		call	doDeleteFile

		; Open the file
		lea	ebx, @@FileName0
		mov	ax, 6092h				; R/W
		mov	cl, 1					; Authorize file creation
		call	doWideOpenFile
		jc	@@Exit
		mov	[@@FileHdl], bx


		; -------------------------------
		; Write the entire buffer to file
		; -------------------------------

		call	EdCompTextSize		
		mov	ecx, eax				; Bytes to write
		mov	edx, [ebp.binf_BuffAddr]		; Pointer to buffer
		mov	bx, [@@FileHdl]				; File handle
		call	doWriteFile
		jnc	@@NoError
		
		; Error, Close file
		mov	bx, [@@FileHdl]
		call	doCloseFile
		stc
		jmp	@@Exit

		; Close the file		
@@NoError:	mov	bx, [@@FileHdl]
		call	doCloseFile


		; ----------------------------------------
		; Give back the memory used by this buffer
		; ----------------------------------------

		; Save previous size
		mov	eax, [ebp.binf_BuffAllSz]
		mov	[ebp.binf_BuffAllSz_S], eax
		mov	eax, [ebp.binf_CurrSize]
		mov	[ebp.binf_CurrSize_S], eax

		; Give back memory
		mov	ebx, [@@BuffNb]
		mov	ecx, [EdCurrBufferNb]
		mov	[EdCurrBufferNb], ebx
		mov	eax, [ebp.binf_BuffAllSz]
		sub	eax, INITSZBUFFED
		call	EdGiveBackMem
		mov	[EdCurrBufferNb], ecx
		
		; Mark this buffer not in memory
		mov	[ebp.binf_InMemory], FALSE


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	edx ecx ebx eax ebp
		ret

		align	4
@@FileName0	db	"\FROTH\TEMP\BUFF_"
@@FileName	db	2 dup (?)
@@FileName2	db	".TMP", 0
		align	4
@@FileHdl	dw	?					; File handle
		align	4
@@BuffNb	dd	?

EdDiskWriteBuffer	endp


; ===============================================================
; = EdDiskReadBuffer						=
; =								=
; = Usage:							=
; = 	Re-load current buffer from disk, to save memory.	=
; =								=
; = I)	EAX : buffer nb (0-39)					=
; =								=
; = O)	CF = 0 if ok, CF = 1 if disk error.			=
; =								=
; = M)	CF							=
; ===============================================================

		align	4
EdDiskReadBuffer	proc

		; --------------
		; Registers used
		; --------------

		push	ebp eax ebx ecx edx


		; -------------
		; Open the file
		; -------------

		; Pointer to buffer infos
		imul	ebp, eax, size buffedentry
		add	ebp, offset EdBuffInfos

		; Build the filename
		mov	bl, 10
		idiv	bl		
		lea	edx, @@FileName
		add	al, '0'
		mov	[edx], al
		add	ah, '0'
		mov	[edx + 1], ah

		; Delete any previous file
		call	doDeleteFile

		; Open the file
		lea	ebx, @@FileName0
		mov	ax, 2090h				; R/O
		call	doWideOpenFile
		mov	[@@FileHdl], bx
		jc	@@Exit


		; --------------------------------
		; Read the entire buffer from file
		; --------------------------------

		mov	ecx, 0FFFFFFh				; Bytes to read (maxi)
		mov	edx, [ebp.binf_BuffAddr]		; Pointer to buffer
		call	doReadFile
		jnc	@@NoError
		
		; Error, Close file
		mov	bx, [@@FileHdl]
		call	doCloseFile
		stc
		jmp	@@Exit

		; Close the file		
@@NoError:	mov	bx, [@@FileHdl]
		call	doCloseFile


		; --------------------
		; Copy saved variables 
		; --------------------
		
		mov	eax, [ebp.binf_BuffAllSz_S]		; Saved, when InMemory is FALSE
		mov	[ebp.binf_BuffAllSz], eax
		mov	eax, [ebp.binf_CurrSize_S]		; Saved, when InMemory is FALSE
		mov	[ebp.binf_CurrSize], eax
		
		; Buffer now into memory
		mov	[ebp.binf_InMemory], TRUE

		
		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	edx ecx ebx eax ebp
		ret

		align	4
@@FileName0	db	"\FROTH\TEMP\BUFF_"
@@FileName	db	2 dup (?)
@@FileName2	db	".TMP", 0
		align	4
@@FileHdl	dw	?					; File handle

EdDiskReadBuffer	endp


; ===============================================================
; = EdUpdatePosTags						=
; =								=
; = Usage:							=
; =	Update the position of tagged marks / definitions	=
; =								=
; = I)	EDI : offset in nb of lines (may be negative).		=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdUpdatePosTags	proc

		; --------------------------------------------------------------------
		; Scan each word cross-ref-ed, and update its line-position, if needed
		; --------------------------------------------------------------------

		; Current line #
		mov	ebx, [ebp.binf_trow]

		; No needed ?
		cmpz	edi, 0
		je	@@Exit

		; Words counter
		clr	ecx

@@ScanWords:	
		; End of list ?
		cmp	ecx, [EdWordsNb]
		je	@@ScanWordsDone

		; Pointer on word NIA address		
		mov	edx, [EdWordsTable + ecx * 4]
		add	edx, [NIA_START]

		; Is this line into the intervall ?
		cmp	[edx.word_LineNb], ebx
		jl	@@NextWord
		
		; Update this line #
		add	[edx.word_LineNb], edi
		
		; Scan next word, if any
@@NextWord:	inc	ecx
		jmp	@@ScanWords
		
@@ScanWordsDone:


		; ---------------------------------------------------------------------------
		; Scan each tag set for this buffer, then update its line position, of needed
		; ---------------------------------------------------------------------------
		
		; Tag counter
		clr	ecx
		
@@ScanTags:	

		; End of tags ?
		cmp	cx, MAXNBTAGS
		je	@@ScanTagsDone
		
		; Is this line into the intervall ?
		cmp	[ebp.binf_Tags_Row + ecx * 4], ebx
		jl	@@NextTag
		
		; Update this line #
		add	[ebp.binf_Tags_Row + ecx * 4], edi
		
		; Scan next tag, if any
@@NextTag:	inc	ecx
		jmp	@@ScanTags
		
@@ScanTagsDone:


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

EdUpdatePosTags	endp

endif

code32		ends
		end
