; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	EQU.ASM							!
; 									!
; Description:	Constants, Structurs, Macros, Externs.			!
; 									!
; ----------------------------------------------------------------------+

	PAGE	66, 250
	SMART

	; ------------------------
	; Configuration parameters
	; ------------------------

	; if 1, developpement version (under debug)
UNDERDEV	equ	1

	; If 1, turnkey edition (no built-in editor, no trace manager)
TURNKEYED	equ	0

	; If 1, test stack overflow 
TESTSTACK	equ	1

	; If 1, DOS is running in REAL mode (beurk !), so DOS calls needs buffers BELOW 1Mb (see QUOTE)
DOSREAL		equ	1

	; FROTH version
FROTH_VERSION		equ	0				; 0.
FROTH_RELEASE		equ	49				; .49
FROTH_UNDERREL		equ	0				; 'a'
FROTH_DATE_DAY		equ	18				; 11
FROTH_DATE_MONTH	equ	04				; apr
FROTH_DATE_YEAR		equ	1996				; 1996


	; --------
	; Keyboard
	; --------

kbentry		struc
	kb_symbname	db	8 dup (?)			; Symbolic name
	kb_ScanCode	dw	?				; Scan code (E0xx or xx)
	kb_AsciiCode	dw	?				; Ascii code: normal
	kb_ShiftCode	dw	?				; Ascii code: shift held
	kb_CtrlCode	dw	?				; Ascii code: ctrl held
	kb_AltCode	dw	?				; Ascii code: alt held
kbentry		ends

KB_MAXNBRECORDS	equ	110					; Max nb of keys on the keyboard
KB_FIFOSZ	equ	32					; Size of the FIFO stack

EXTKEYS		equ	001h					; Bit set on extended keys (no shift, no ctrl, no alt)
SHIFTKEYS	equ	002h					; Bit set on shift-ed keys
CTRLKEYS	equ	004h					; Bit set on ctrl-ed keys
ALTKEYS		equ	008h					; Bit set on alt-ed keys


	; ---------
	; Serial IT
	; ---------

MAXNB_ITS		equ	8			; Max nb of serial IT which can be handled
MAXNB_CHARS_ITPCK	equ	4			; 4 chars max in a serial it packet
MAXNB_DEF_ITPCK		equ	10			; 10 packet definitions max in serial IT


	; ---------------------
	; Max number of THREADS
	; ---------------------
	
MAXNB_THREADS	equ	32


	; ------------
	; System stack		SEE PMODE.ASM !!!
	; ------------
	
SSTACKSIZE	equ	1000h


	; ----------------
	; Parameters stack
	; ----------------

PSTACKSIZE	equ	40h					; Froth stack is 64 bytes size, can contains 16 int (32-bits)


	; -----------------------
	; Second parameters stack
	; -----------------------

RSTACKSIZE	equ	20h				; Froth 2nd stack is 20 bytes size, can contains 8 int (32-bits)


	; ----------------------
	; Max size of a filename
	; ----------------------
	
MAXSZFILENAME	equ	63
MAXSZFNAME	equ	8
MAXSZEXT	equ	3


	; ----------------------
	; Some special filenames
	; ----------------------
	
DEFKBD_FILENAME		equ	<"\FROTH\KB_US.FKB", 0>
if TURNKEYED eq 0
HELP_FILENAME		equ	<"\FROTH\FROTH.HLP">
STARTUP_FILENAME	equ	<"\FROTH\STARTUP.FTH", 0>	; File loaded when FROTH is started
else
STARTUP_FILENAME	equ	<"\FROTH\STARTUPT.FTH", 0>
endif
SHUTDOWN_FILENAME	equ	<"\FROTH\SHUTDOWN.FTH", 0>	; File loaded when FROTH is endded
ifdef UNDERDEV
WORDSLIST_FILENAME	equ	<"\FROTH\WORDSLST.TXT", 0>
endif


	; ----------
	; Vocabulary
	; ----------

WORDMAXLEN	equ	32				; Max length of a word
MAXNB_VOCS	equ	50				; Max nb of vocabularies

vocentry	struc					; Record size = 36 bytes
	voc_name	db	WORDMAXLEN+1 dup (?)	; Vocabulary name
	voc_fatherptr	dd	?			; Father pointer
	voc_lastword	dd	?			; Pointer to the last word
	voc_wordsnb	dd	?			; Number of words
vocentry	ends

VOC_MAXSIZE	equ	(MAXNB_VOCS * size vocentry)	; Max size of VOC_AREA


	; --------------------
	; NIA : Name Info Area
	; --------------------

FSYSTEM		equ	01h
FIMMEDIATE	equ	02h
FMACRO		equ	04h
FRUNTIME	equ	08h
FBUILDDOES	equ	10h				; Is it a defining-word ?
FNOCOMPILE	equ	20h				; The word can't be compiled (ex: INFOS)
FSMUDGE		equ	40h				; This bit is set while the word is not ended
FCOLONDEF	equ	80h				; Does this definition build with : ?
	
BSYSTEM		equ	 0
BIMMEDIATE	equ	 1
BMACRO		equ	 2
BRUNTIME	equ	 3
BBUILDDOES	equ	 4
BNOCOMPILE	equ	 5
BSMUDGE		equ	 6
BCOLONDEF	equ	 7
	
wordentry	struc						; Size record is 0xD4 (212) bytes
	word_name	db	WORDMAXLEN+1 dup (0)		; Word name. If word_name[0] == 0, the record is free
	word_flags	db	?				; Immediate ? System ? Runtime ? Macro ?
	word_vocptr	dw	?				; Vocab pointer
	word_prevword	dd	?				; Previous word (in same vocabulary), or -1
	word_nextword	dd	?				; Next word (in same vocabulary), or -1
	word_realaddr	dd	?				; Real address of the code
	word_runtaddr	dd	?				; Run-time portion address, or 0 if not a defined-word
	word_EndCode	dd	?				; Last addr+1 of code (AFTER ret)
	word_CodeSize	dw	?				; Size of the code
	word_DefiningWord dd	?				; If created by a DW, this is a ptr to the defining-word
	word_NbUses	dw	?				; Nb of uses for this word
	word_FileName	db	MAXSZFNAME+1+MAXSZEXT+1 dup (0)	; FileName where this word is defined
	word_LineNb	dd	?				; ED_GOTOW: line # in the file where the word is defined
	word_garb	db	3 dup (?)			; 81 + 3 = 84 bytes
wordentry	ends

MAXNB_WORDS	equ	2000
NIA_WSIZE	equ	(MAXNB_WORDS * size wordentry)		; Size allocated for NIA (20000h)

	; About methods and instances
ZDOES		equ	"does>"
L_ZDOES		equ	5+1



if TURNKEYED eq 0
	; ----------------------
	; Cross words references
	; ----------------------

MAXNBWREF	equ	5000
wrefentry	struc
	wref_caller	dd	?			; NIA pointer to the word caller
	wref_called	dd	?			; NIA pointer to the word called
	wref_calls	dw	?			; Nb of uses
wrefentry	ends
endif


	; --------------------------
	; Data Directory area (DDIR)
	; --------------------------

; Flags for ddir_flags
QDATA_AUTO	equ	01h
QDATA_DYNA	equ	02h
QDATA_MARKDEL	equ	04h
QDATA_DELDONE	equ	08h
	
ddirentry	struc					; Record size is: 0x10 (16) bytes
	ddir_NextFreed	dd	?			; If not -1, this field points to next record freed
	ddir_wordnb	dw	?			; Reference to the word (word nb) to multiply with size wordentry
	ddir_blksize	dd	?			; Block size
	ddir_blkaddr	dd	?			; Block address (relative to the start of Data Dict)
	ddir_threadnb	db	?			; Thread nb (which thread owns this data ?)
	ddir_flags	db	?			; Flags: QDATA_AUTO QDATA_DYNA QDATA_MARKDEL
ddirentry	ends

DDIR_WSIZE	equ	6000h				; Window size of DDIR: room for 1536s allocations


	; ---------------------
	; Local Dict Area (LDA)
	; ---------------------

LDA_MAXNB	equ	24					; Max nb of words into the Local Dictionnary Area (LDA)	

lwordentry	struc
	lword_name	db	WORDMAXLEN+1 dup (0)		; Word name. If word_name[0] == 0, the record is free
	lword_blocknbad	dd	?				; Block nb address (address after the 'CALL [ESI + N_xxx]'
	lword_rtaddr	dd	?				; Run time nb addr
	lword_garb	db	3 dup (?)			; To have a record of 40 bytes size
lwordentry	ends


	; --------------------------
	; STRING (Fixed and Dynamic)
	; --------------------------

strentry		struc
	str_currlen	dd	?		; Current length
	str_currsize	dd	?		; Current size allocated
	str_memblknb	dd	?		; Memory block nb (ref to DDIR) (0, 1, 2, ... )
	str_incsize	dw	?		; Increment size (if dynamic string), else 0 (if fixed string)
	str_wordref	dw	?		; Reference to the definition (NIA)
strentry		ends
sizestrentry	=	size strentry
ESCCHAR		equ	'\'			; Character used as escape char into literal string
STRMAXLEN	equ	4194304


	; ------------
	; Input buffer
	; ------------

IBUFFMAXS	equ	255					; Max line length
WBUFFMAXS	equ	255					; Max word length
MAXNBLOADFILES	equ	5					; Max depth for loading files
LOADFILEBUFFSZ	equ	8192					; Size of the buffer to read load files

loadfileentry	struc
	loadfile_hdl		dw	?				; File handle (if 0, file is not opened)
	loadfile_buff		dd	?				; Buffer addr, where the file is read
	loadfile_inFile		dd	?				; Current position in file (seek)
	loadfile_inBuff		dd	?				; Current position in buffer
	loadfile_name		db	MAXSZFILENAME+1 dup (?)		; FileName
	loadfile_linenb		dd	?				; Current line # into the file
	loadfile_linsaddr	dd	?				; Current line starting address
	loadfile_colnb		dw	?				; Current col  # into the file
;	loadfile_garb		db	0 dup (?)			; To have a record of 88 bytes
loadfileentry	ends

	; Move file pointer methods
SEEK_SET	equ	0						; Seeks from beginning of file
SEEK_CUR	equ	1						; Seeks from current location
SEEK_END	equ	2						; Seeks from end of file


	; ---------------------------
	; Numeric output buffer (NOB)
	; ---------------------------

NOBMAXSZ		equ	128					; Max size for <# # #>	


	; ------------------------------------------------------------
	; Max size of a string which can be drawn by !STRING functions
	; ------------------------------------------------------------
	
MAXSZ_SCREENSTR	equ	132


if TURNKEYED eq 0
	; ---------
	; HELP file
	; ---------

hlpentry	struc
	hlp_fileptr	dd	?				; Pointer into the file
	hlp_wordname	db	WORDMAXLEN+1 dup (?)		; Word
	ends
MAXNB_SYSTEMWORDS	equ	1000				; Max nb of system words
MAXSZ_HELP_BUFFER	equ	4096				; Max size of the buffer for the help of a given word
endif


if TURNKEYED eq 0
	; -------------
	; TRACE manager
	; -------------

DEFSZ_TRACEBUFFER	equ	2000h				; Default size for trace manager buffer
endif


	; ------------------
	; Exit resolve stack
	; ------------------

EXITSTACKSZ	equ	16					; Max nb of EXIT into a definition


	; ------------------------
	; Control structures stack
	; ------------------------

CONTROLSTACKSZ	equ	24					; Max depth of control stack


	; ----------
	; Loop Stack
	; ----------
	
loopentry	struc
	loop_start	dd	?				; Start value
	loop_end	dd	?				; End value
	loop_index	dd	?				; Loop index
	loop_step	dd	?				; Step value
loopentry	ends

LOOPSTACKMAXNB	equ	5					; Max number of loop at the same time


	; ----------
	; Case stack
	; ----------
	
CASESTACKMAXNB	equ	5					; Max number of case..endcase levels
	

	; ---------------------------------------------------------------
	; READ/WRITE data greater than 64k or above 1Mb into MS/DOS files
	; ---------------------------------------------------------------

FILEATOMSIZE	equ	32768


	; ------------------------
	; Max value of file handle
	; ------------------------

MAXNB_FILEHDL	equ	254


	; ---------------------------
	; Messages (constant strings)
	; ---------------------------

CMSG_COMPILING	equ	<"Compiling ", 0>
CMSG_ERR_RETED	equ	<13, 10, "Press (CR) to edit file or (ESC) to return to interpret ", 0>


	; --------------------------------
	; Push an immediate value on stack
	; --------------------------------
	
IPUSH	MACRO	value
		sub	esi, @WordSize
		mov	dword ptr [esi], value
	ENDM


	; ---------------------
	; INTEGER run-time call
	; ---------------------

DOINT	macro	no
		mov	eax, [edi.ddir_blkaddr + no]
		sub	esi, @WordSize
		mov	[esi], eax
	endm


DOVAR		MACRO	refvar
		local	next
		call	next
next:		add	dword ptr [esp], 4+6+3+2+1
		mov	eax, [edi.ddir_blkaddr + refvar]
		if refvar lt 128
		nop
		nop
		nop
		endif
		sub	esi, @WordSize
		mov	PARAM0, eax
		ret
		ENDM


	; ------------------
	; Diverses constants
	; ------------------

FALSE		equ	0				; FALSE value
TRUE		equ	-1				; TRUE value


	; -------------------------------------
	; Colors which can be used in Text mode
	; -------------------------------------
	
BLACK		equ	00h
BLUE		equ	01h
GREEN		equ	02h
CYAN		equ	03h
RED		equ	04h
MAGENTA		equ	05h
BROWN		equ	06h
LIGHTGRAY	equ	07h
DARKGRAY	equ	08h
LIGHTBLUE	equ	09h
LIGHTGREEN	equ	0Ah
LIGHTCYAN	equ	0Bh
LIGHTRED	equ	0Ch
LIGHTMAGENTA	equ	0Dh
YELLOW		equ	0Eh
WHITE		equ	0Fh


	; --------------------------------
	; Keys to use for editing (prompt)
	; --------------------------------
	
EKEY_CR		equ	00Dh					; Carriage return
EKEY_BS		equ	008h					; BackSpace (delete left char)
EKEY_HOME	equ	147h					; Home key
EKEY_END	equ	14Fh					; End key
EKEY_LEFT	equ	14Bh					; Left key
EKEY_RIGHT	equ	14Dh					; Right key
EKEY_DOWN	equ	150h					; Down key
EKEY_UP		equ	148h					; Up key
EKEY_PGUP	equ	149h					; PgUp
EKEY_PGDN	equ	151h					; PgDn
EKEY_CTRL_LEFT	equ	44Bh					; Ctrl-Left key
EKEY_CTRL_RIGHT	equ	44Dh					; Ctrl-Right key
EKEY_CTRL_HOME	equ	447h					; Ctrl-Home	Move to begin of text
EKEY_CTRL_END	equ	44Fh					; Ctrl-End	
EKEY_INS	equ	152h					; Insert a space
EKEY_DEL	equ	153h					; Delete char under cursor
EKEY_ESC	equ	01Bh					; ESC key (used into TRACE)


	; --------------------------------
	; To define a FROTH word as global
	; --------------------------------

function	MACRO	name, rt_name
		global	name:		proc
		global	E_&name:	proc
		ifnb	<rt_name>
			global	RT_&name:	proc
		endif
		ENDM


	; --------------------------------------
	; Public functions defined in: FROTH.ASM
	; --------------------------------------

global		CtrlBreakRecov:		proc
global		Abort:			proc
function	Quit
global		Excp1:			proc
function	doError


	; --------------------------------------
	; Public functions defined in: Debug.asm
	; --------------------------------------

function	VStack1
function	VStack2
function	VStack3
function	VStack4
function	VStack5
function	VStack6
function	VStack7
function	VStack8
function	OStack
global		VRStack:	proc
global		PrStackDepth:	proc
function	DOTS
function	DUMP
if UNDERDEV
function	WDUMP
endif
if TURNKEYED eq 0
function	WHICH
global		WHICH2:		proc
function	showVOC
function	showNIA
function	showCDA
function	showDDIR
function	showCREF
endif
if UNDERDEV
global		PrEAX:		proc
global		PrAX:		proc
global		PrAL:		proc
global		PrintEAX:	proc
global		PrintAX:	proc
global		PrintAL:	proc
global		PrintMsg:	proc
endif


	; ------------------------------------
	; Public functions defined in: KBD.ASM
	; ------------------------------------

function	KBDFILE
global		KBDFILE0:	proc
global		InstallKbIt:	proc
global		DesinstallKbIt:	proc
if UNDERDEV
global		debug_charin:	proc
endif
function	CharIn
function	KEY
function	QKEY
function	SCANCODE
function	QALTKEY
function	QCTRLKEY
function	QALTKEY
function	QSHIFTKEY
global		isShiftKey:	proc
function	KEYINDICE
function	QKEYSTATE


	; -------------------------------------
	; Public functions defined in: KEYS.ASM
	; -------------------------------------

global		UpdateIndexKeys:	proc
function	SHIFT_ED
function	CTRL_ED
function	ALT_ED
function	KEYPAD_ED
function	KEY_INS
function	KEY_DEL
function	KEY_HOME
function	KEY_END
function	KEY_PGUP
function	KEY_PGDN
function	KEY_UP
function	KEY_DOWN
function	KEY_LEFT
function	KEY_RIGHT
function	KEY_ESC
function	KEY_PLUS
function	KEY_MINUS
function	KEY_ALT_A
function	KEY_ALT_Z
function	KEY_ALT_E
function	KEY_ALT_R
function	KEY_ALT_T
function	KEY_ALT_Y
function	KEY_ALT_U
function	KEY_ALT_I
function	KEY_ALT_O
function	KEY_ALT_P
function	KEY_ALT_Q
function	KEY_ALT_S
function	KEY_ALT_D
function	KEY_ALT_F
function	KEY_ALT_G
function	KEY_ALT_H
function	KEY_ALT_J
function	KEY_ALT_K
function	KEY_ALT_L
function	KEY_ALT_M
function	KEY_ALT_W
function	KEY_ALT_X
function	KEY_ALT_C
function	KEY_ALT_V
function	KEY_ALT_B
function	KEY_ALT_N
function	KEY_F1
function	KEY_F2
function	KEY_F3
function	KEY_F4
function	KEY_F5
function	KEY_F6
function	KEY_F7
function	KEY_F8
function	KEY_F9
function	KEY_F10
function	KEY_F11
function	KEY_F12


if TURNKEYED eq 0
	; --------------------------------------
	; Public functions defined in: TRACE.ASM
	; --------------------------------------

global		doTraceInit:	proc
function	TRACE_PAGE
function	TRACE_CLEAR
function	TRACE_BEGIN
function	TRACE_END
function	TRACE_DOT
function	TRACE_DOTSTACK
function	TRACE_TYPE
endif


	; ---------------------------------------
	; Public functions defined in: THREAD.ASM
	; ---------------------------------------

function	MAXNBTHREADS
function	ResetThreads
function	NBTHREADS
function	NEWTHREAD
function	doNewThread
function	ENDTHREAD
function	KILLTHREAD
function	YIELD
function	ACTIVETHREAD
	

	; ---------------------------------------
	; Public functions defined in: VARIED.ASM
	; ---------------------------------------

global		GetArgc:	proc
global		GetProgName:	proc
global		prGreatingMsg:	proc
function	WORDSIZE
function 	QOK
function	QSAFE
function	QSPEED
function	QOPTIMIZE
function 	QIARITHOVERFLOW
function 	QCHAT
function 	QREDEF
function	BYE
function	PBYE
function	QUOTELOAD
global		QUOTELOAD0:	proc
global		QuoteLoadErr:	proc
if TURNKEYED eq 0
function	WINFOS
function	LOAD
function	ED
endif
function	RND
function	INFOS
if TURNKEYED eq 0
function	MEM
function	LIMITS
endif
function	PMODE
function	CTRLBREAK_OFF
function	CTRLBREAK_REST
function	GETTIME
function	GETDATE
function	TIMEANDDATE
function	GETTICKS
function	DELAY
function	FROTHVERSION
function	FROTHDATE
function	LANGUAGE
function	ENGLISH
function	FRENCH
global		EdNumber:		proc
global		EdPDotQ:		proc
global		EdAddExtFTH:		proc
	

if TURNKEYED eq 0
	; ------------------------------------
	; Public functions defined in: USE.ASM
	; ------------------------------------

function	USE
ifdef UNDERDEV
function	USINGQ
endif
endif


	; -----------------------------------------
	; Public functions defined in: INTRPRET.ASM
	; -----------------------------------------

function	SHARPIF
function	SHARPELSE
function	SHARPENDIF
function	QLOAD
global		AutoFind:		proc
function	INTERPRET
global		Execute:		proc
function	QERROR
function	WORDNAME


	; -----------------------------------------
	; Public functions defined in: DictMang.asm
	; -----------------------------------------

global		isOptimize:		proc
global		isQIArithOverfl:	proc
global		QPrev_WasNb:		proc
global		QPrev_WasCall:		proc
function	QCOMPILE
function	QEXECUTE
global		PFind:			proc
function	BUILD
function	DynVarNotAll
function	isDynVarAlAll
global		IsolateWordEx:		proc
function	doBuild
function	doLocalBuild
function	COLON
global		VrfCtrlStrPending:	proc
global		PurgeAutoVars:		proc
function	SEMICOLON
function	IMMEDIATE
function	LITERAL
function	LBRACKET
function	RBRACKET
function	COMPILE
function	LPARENT
function	RPARENT
function	SLASHSLASH
function	WORDS
function	XWORDS
if TURNKEYED eq 0
function	VOCS
endif
global		VocFind:		proc
function	VOCABULARY
function	CURRENT
global		doForgetWord:		proc
function	FORGET

	; defined in DictMng2.asm
function	TICK
function	DTICK
function	EXECUTE
function	RECURSE
function	MYSELF
function	DEFERRED,		RT_DEFERRED
function	DEFER
	

	; ---------------------------------------
	; Public functions defined in: Struct.asm
	; ---------------------------------------

function	_STRUCT
function	_ENDSTRUCT


if TURNKEYED eq 0

	; -----------------------------------------
	; Public functions defined in: CrossRef.asm
	; -----------------------------------------

global		CRef_add:		proc
global		CRefSrch_RD:		proc
global		CRefSrch_DR:		proc
global		CRef_del:		proc
function	UPDATE
endif


	; ----------------------------------------
	; Public functions defined in: SaveExe.asm
	; ----------------------------------------
	
function	SAVE_EXE
function	QTURNKEY
function	QFROTHTURNKEY
function	NBCMDARGS
function	CMDARGS


	; ------------------------------------
	; Public functions defined in: BIO.ASM
	; ------------------------------------

global		COUT:		proc
function	ReadLine
global		POUT:		proc
function	EMIT
function	PEMIT
function	PTYPE
function	PCR
function	PSPACE
function	RdLoadFileLine
global		doOpenFile:	proc
global		doWideOpenFile:	proc
global		doCreateFile:	proc
global		doReadFile:	proc
global		doWriteFile:	proc
global		doSeekFile:	proc
global		doFilePos:	proc
global		doFileSize:	proc
global		doCloseFile:	proc
global		doDeleteFile:	proc
global		doQFileExist:	proc


	; --------------------------------------
	; Public functions defined in: NumIO.asm
	; --------------------------------------

function	_BASE
function	DECIMAL
function	HEXADECIMAL
function	BSHARP
function	SHARP
function	SHARPQ
function	USHARP
function	HOLD
function	HOLDSTR
function	SIGN
function	SHARPS
function	USHARPS
function	ESHARP
function	BSHARPSE
function	NUMBER
function	DOT
function	DOTR
function	DOTL
function	UDOT
function	HDOT
function	Z2HDOT
function	Z4HDOT
function	Z8HDOT

	; --------------------------------------
	; Public functions defined in: Stack.asm
	; --------------------------------------

global		isSpeed:	proc
global		isSafe:		proc
function	RT_QENOUGHT
function	QENOUGHT
function	DEPTH
function	DROP
function	DROPS
function	DDROP
function	_DUP
function	PICK
function	DDUP
function	SWAP
function	DSWAP
function	OVER
function	ABOVE
function	UNDER
function	BELOW
function	TUCK
function	ROT
function	MROT
function	ROLL
function	MROLL
function	TOR
function	FROMR
function	R
function	RSWAP
function	RDUP
function	ROVER
function	RABOVE
function	RDROP
function	RUNDER
function	RBELOW
function	RROT
function	RMROT


	; ----------------------------------------
	; Public functions defined in: Memory.asm
	; ----------------------------------------
	
function	MEMFREE

function	RAT
function	PAT
function	AT

function	RCAT
function	PCAT
function	CAT

function	RWAT
function	PWAT
function	WAT

function	PSTORE
function	STORE

function	PCSTORE
function	CSTORE

function	PWSTORE
function	WSTORE

function	PLUSSTORE
function	MINUSSTORE
function	STARSTORE
function	SLASHSTORE
function	QUEST
function	CQUEST
function	WQUEST
function	DYNAMIC
function	ALLOCATE
function	FREE
function	CodeComma
function	CodeWComma
function	CodeCComma
function	QALLOTFAIL
function	RT_ALLOT
function	ALLOT
function	RT_BALLOT
function	BALLOT
function	RT_WALLOT
function	WALLOT
function	RT_ALLOTMORE
function	ALLOTMORE
function	RT_BALLOTMORE
function	BALLOTMORE
function	RT_WALLOTMORE
function	WALLOTMORE
function	DOES
function	PDOES
function	LocalVar
function	DATASALVAGE
function	DUSTTHRESHOLD
function	DUST
function	UnAllot
global		ReAlloc:		proc
function	BOOLEAN,		RT_BOOLEAN
function	IBOOLEAN,		RT_IBOOLEAN
function	BOOL_FALSESTORE
function	BOOL_TRUESTORE
function	BOOL_QFALSE
function	BOOL_QTRUE
function	IINTEGER,		RT_IINTEGER
function	INTEGER,		RT_INTEGER
function	CONSTANT,		RT_CONSTANT


	; -----------------------------------------
	; Public functions defined in: Memory2.asm
	; -----------------------------------------
	
function	BMOVE
function	PBMOVE
function	FILL
function	FreeLocalVar
global		FreeData:		proc
global		DiscardLocalVars:	proc
global		DiscardThreadLVar:	proc


	; --------------------------------------
	; Public functions defined in: Arith.asm
	; --------------------------------------
	
function	NEGATE
function	MIN
function	MAX
function	UMIN
function	UMAX
function	PLUS
function	MINUS
function	STAR
function	SLASH
function	SLASHMOD
function	IMOD
function 	STARSLASH
function 	STARSLASHMOD
function	IABS

	
	; --------------------------------------
	; Public functions defined in: Logic.asm
	; --------------------------------------
	
function	VFALSE
function	VTRUE
function	EQUAL
function	NEQUAL
function	LESS
function	LESSEQ
function	GREAT
function	GREATEQ
function	ULESS
function	ULESSEQ
function	UGREAT
function	UGREATEQ
function	IAND
function	IOR
function	IXOR
function	INVERT
function	ISHL
function	ISHR
function	SETBIT
function	RESETBIT
function	COMPLBIT
function	GETBIT

	
	; ----------------------------------------
	; Public functions defined in: Control.asm
	; ----------------------------------------

function	EXIT	
function	BEGIN
function	AGAIN
function	UNTIL
function	FIF
function	FELSE
function	FENDIF
function	I
function	J
function	K
function	L
function	M
function	doTimes
function	TIMES
function	doEndTimes
function	ENDTIMES
function	doDo
function	DO
function	doWhen
function	WHEN
function	doNotWhen
function	NOTWHEN
function	doEndDo
function	ENDDO
function	doPEndDo
function	PENDDO
function	doPUEndDo
function	PUENDDO
function	doMake
function	MAKE
function	doPlusMake
function	PLUSMAKE
function	doEndMake
function	ENDMAKE
function	_WHILE
function	_REPEAT
function	CASE
function	OF
function	ENDOF
function 	TIPOF
function	ENDCASE
function	doCase
function	doOf
function	doEndCase

	
	; ---------------------------------------
	; Public functions defined in: LocVar.asm
	; ---------------------------------------
	
function	ZEROLOC
function	ONELOC
function	TWOLOC
function	THREELOC
function	AT0
function	AT1
function	AT2
function	AT3
function	STORE0
function	STORE1
function	STORE2
function	STORE3
function	PLUSSTORE0
function	PLUSSTORE1
function	PLUSSTORE2
function	PLUSSTORE3
function	LOOP0
function	LOOP1
function	LOOP2
function	LOOP3


	; ---------------------------------------
	; Public functions defined in: Screen.asm
	; ---------------------------------------
	
function	NBPAGES
function	PAGENB
function	NBROWS
function	NBCOLS
function	SCREENBCKG
function	QUPDATECURSOR
global		doGetScreenInf:		proc
function	SETVIDEOMODE
function	GETVIDEOMODE
function	SELECTPAGE
function	CHANGEPAGE
function	CURSOR
function	GOTOXY
function	GOTOX
function	GOTOY
function	WHEREXY
function	WHEREX
function	WHEREY
function	BACKFOREGROUND
function	ATXYCHAR
function	ATXYACHAR
global		doGetVideoPage:		proc
function	STOREXYCHAR
function	STOREXYACHAR
function	STOREXYNCHAR
function	STOREXYNACHAR
function	ATTRCLS
function	CLS
function	STORESTRING
global		STORESTRING0:		proc
function	STOREASTRING
function	STOREXYSTRING
function	STOREXYASTRING
function	STOREXYLSTRING
function	TOXYLSTRING_CHAR
function	TOXYLSTRING_ATTR
function	ACLRLINE
function	CLRLINE
function	ACLREOL
function	CLREOL
function	LOADCHARSET_8x8
function	LOADCHARSET_8x14
function	LOADCHARSET_8x16
function	VBLACK
function	VBLUE
function	VGREEN
function	VCYAN
function	VRED
function	VMAGENTA
function	VBROWN
function	VLIGHTGRAY
function	VDARKGRAY
function	VLIGHTBLUE
function	VLIGHTGREEN
function	VLIGHTCYAN
function	VLIGHTRED
function	VLIGHTMAGENTA
function	VYELLOW
function	VWHITE


	; ---------------------------------------
	; Public functions defined in: String.asm
	; ---------------------------------------
	
function	_TYPE
function	DTYPE
function	TYPED
function	CR
function	SPACE
function	SPACES
function	PDOTQ
function	PQUOTE
function	DOTQUOTE
function	DOTQUOTE
function	QUOTE
function	CHARUPCASE
function	CHARLOCASE
function	FSTRING,		RT_FSTRING
function	DSTRING,		RT_DSTRING
function	CSTRING,		RT_CSTRING
function	QSTRING
function	QFIXEDSTRING
function	QDYNAMICSTRING
function	NULLSTR
function	STR_LEN
function	STR_MLEN
function	STR_UPCASE
function	STR_LOCASE
function	STR_EQU
function	STR_LESS
function	STR_LESSEQ
function	STR_GREAT
function	STR_GREATEQ
function	STR_STORE
function	STR_ADD
function	ADD_STR
function	STR_REPL
function	STR_FIND
function	STR_COPY
function	STR_COPY_TOC
function	STR_COPY_CHARS
function	STR_CONVNB
function	STR_SHRINK
function	STR_SRCHCHR
function	STR_SRCHSTR
function	STR_MONTH
function	INPUT

	
	; -------------------------------------
	; Public functions defined in: File.asm
	; -------------------------------------

function	FILEIOSTATUS
function	FLAG_RO
function	FLAG_RW
function	FLAG_WO
function	OPENFILE
function	CLOSEFILE
function	EMITFILE
function	WRITEFILE
function	READFILE
function	LINESEP
function	WRITELINE
function	READLINE
function	QEOF
function	SEEKFILE
function	FILEPOS
function	SEEKSET
function	SEEKCUR
function	SEEKEND
function	FILESIZE
function	DELETEFILE
function	QFILEEXIST

	
	; ---------------------------------------
	; Public functions defined in: Serial.asm
	; ---------------------------------------

function	SERIALOPEN
function	SERIALCLOSE
function	NONEPARITY
function	ODDPARITY
function	EVENPARITY
function	SERIALDEFPCK
function	SERIALPCKQ
function	SERIALPCKREAD
function	SERIALPARAM
function	SERIALIN
function	SERIALINQ
function	SERIALREAD
function	SERIALFLUSHIN
function	SERIALOUT
function	SERIALWRITE
function	SERIALFLUSHOUT
function	SERIALABORT

	
	; ---------------------------------------
	; Public functions defined in: PortIO.asm
	; ---------------------------------------

function	INPORT
function	OUTPORT
function	WINPORT
function	WOUTPORT
function	CINPORT
function	COUTPORT


	; ------------------------------------
	; Public functions defined in: HGR.ASM
	; ------------------------------------

global		InitGraphVars:	proc
function 	M320x200x16
function 	M640x200x16
function 	M640x350x4
function 	M640x350x16
function 	M640x480x2
function 	M640x480x16
function 	M320x200x256
function 	M800x600x16
function	GNBROWS
function 	GNBCOLS
function 	GNBCOLORS
function 	GNBPAGES
function 	GNOTFILLED
function 	GFILLED
function	CLEAR
function	SELECTGPAGE
function 	CLIP
function	GMODE_COPY
function	GMODE_XOR
function	GMODE_AND
function	GMODE_OR
function	SETPIXEL
function	GETPIXEL
function 	LINE
function 	HLINE
function 	VLINE
function 	BOX
function 	VIDEOREGION,		RT_VIDEOREGION
function 	GETVIDEO
function 	PUTVIDEO
function 	SPRITE,			RT_SPRITE
function 	DEFCOLOR_SPRITE
function 	DEF_SPRITE
function 	INIT_SPRITE
function 	DRAW_SPRITE
	

if TURNKEYED eq 0
	; ---------------------------------------
	; Public functions defined in: Editor.asm
	; ---------------------------------------
	
function	EDC_SLL
function	EDC_SLI
function	EDC_TXN
function	EDC_TXS
function	EDC_HLN
function	EDC_HLA
function	EDC_HLK
function	ED_VIDEOMODE
function	ED_QDIRECTVIDEO
function 	ED_QUSETAB
function 	ED_COMMENTFIELD
function	ED_LINEPOSADJUST
function	ED_KEY
endif
function	ED_DIR

	
	; ------------------------------------
	; Public functions defined in: Asm.asm
	; ------------------------------------
	
function	OPENASMBRACKET
function	CLOSEASMBRACKET


if TURNKEYED eq 0
	; ---------------------------------------
	; Public functions defined in: Disasm.asm
	; ---------------------------------------
	
function	DISASM
endif


	; -------------------------------------
	; Public functions defined in: Last.asm
	; -------------------------------------

global		_StartAddr:	near
global		_EndAddr:	near


	
	; ==============================
	; ===                        ===
	; === General purpose MACROS ===
	; ===                        ===
	; ==============================

	; ----------------------------------
	; Print a string (from a FROTH word)
	; ----------------------------------

PRSTR		MACRO	str
		call	PDOTQ					; rt_."
		db	str, 0
		ENDM
		

	; ----------------------------------------------------
	; Convert an address for DOS/BIOS real mode (seg, ofs)
	; ----------------------------------------------------

HILO		MACRO	seg, ofs, reg

		push	eax
		mov	eax, reg
		add	eax, [_code32a]
		push	ax
		shr	eax, 4
		mov	[seg], ax
		pop	ax
		and	ax, 0Fh
		mov	[ofs], ax
		pop	eax
		
		ENDM


		
	; ------------------------------------------------------------
	; Convert a 8-bit register value to upper char (if lower char)
	; ------------------------------------------------------------

TOUPPER		macro	reg
		local	NoLower
		cmp	reg, 'a'
		jb	NoLower
		cmp	reg, 'z'
		ja	NoLower
		sub	reg, 'a' - 'A'
NoLower:
		endm


	; ----------------------------------------------------------
	; Convert a 8-bit register char value to a digit (base 2-26)
	; ----------------------------------------------------------

TODIGIT		macro	reg
		local	NotDig, Done, IsHexDig

		cmp	reg, '0'
		jb	NotDig
		cmp	reg, '9'
		ja	IsHexDig
		sub	reg, '0'
		jmp	Done

IsHexDig:	cmp	reg, 'Z'
		ja	NotDig
		cmp	reg, 'A'
		jb	NotDig
		sub	reg, 55
		jmp	Done

NotDig:		mov	reg, -1
Done:		
		endm


	; -------------------------------------
	; Colors which can be used in Text mode
	; -------------------------------------
	
COLOR	MACRO	id, background, foreground
id	db	(background shl 4) + foreground
	ENDM


	; ----------------
	; Clear a register
	; ----------------

clr	macro	reg
	xor	reg, reg
	endm


	; -------------------------
	; Test a register with zero
	; -------------------------

cmpz	macro	reg, zero
	or	reg, reg
	endm


	; -------------------------------------------------
	; Set a FROTH variable with a given immediate value
	; Get a FROTH variable value and store into a register
	; -------------------------------------------------

SETFVAR	macro	var, value
	mov	eax, [edi + ddir_blkaddr + var]
	mov	dword ptr [eax], value
	endm

GETFVAR	macro	reg, var
	mov	reg, [edi + ddir_blkaddr + var]
	mov	reg, [reg]
	endm

DO_AT		macro
		push	eax
		mov	eax, PARAM0
		mov	eax, [eax]
		mov	PARAM0, eax
		pop	eax
		endm


DO_STORE	macro
		push	eax ebx
		lodsd						; Pop <addr>
		mov	ebx, eax
		lodsd						; Pop <n>
		mov	[ebx], eax
		pop	ebx eax
		endm


	; ----------------------------------------
	; Access to data on FROTH parameters stack
	; ----------------------------------------

PARAM0		equ	<dword ptr [esi + 0 * @WordSize]>
PARAM1		equ	<dword ptr [esi + 1 * @WordSize]>
PARAM2		equ	<dword ptr [esi + 2 * @WordSize]>
PARAM3		equ	<dword ptr [esi + 3 * @WordSize]>
PARAM4		equ	<dword ptr [esi + 4 * @WordSize]>
PARAM5		equ	<dword ptr [esi + 5 * @WordSize]>
PARAM6		equ	<dword ptr [esi + 6 * @WordSize]>
PARAM7		equ	<dword ptr [esi + 7 * @WordSize]>
DISPOSE		macro	nbargs
		sub	esi, nbargs * @WordSize
		endm
DISCARD		macro	nbargs
		add	esi, nbargs * @WordSize
		endm


; ------------------
; Verify stack frame
; ------------------

	; Test primary stack underflow (on definition entry)
	; --------------------------------------------------
	
VSTACK		macro	NbArgs
		ifdef TESTSTACK
			if NbArgs eq 1
				call	VStack1
			elseif NbArgs eq 2
				call	VStack2
			elseif NbArgs eq 3
				call	VStack3
			elseif NbArgs eq 4
				call	VStack4
			elseif NbArgs eq 5
				call	VStack5
			elseif NbArgs eq 6
				call	VStack6
			elseif NbArgs eq 7
				call	VStack7
			elseif NbArgs eq 8
				call	VStack8
			else
				ERR	<VSTACK: bad argument>
			endif
		endif
		endm


	; Test primary stack overflow (on definition exit)
	; ------------------------------------------------

OVSTACK		macro
		ifdef TESTSTACK
			call	OStack
		endif
		endm

	
	
; -------------------------
; Verify second stack frame
; -------------------------

VRSTACK		macro	nbargs
		ifdef TESTSTACK
			call	VRStack
			db	nbargs
		endif
		endm
	
	
if UNDERDEV

global	EdGotoXY:	proc
global	EdWhereXY:	proc

ifdef __MAIN__
SaveEAX		dd	?
endif
global		SaveEAX:		dword

B_SHOW		macro	num
		pushad

		mov	[SaveEAX], eax
		call	EdWhereXY
		push	ax
;		GETFVAR	eax, D_NBROWS
mov	eax, 24
		dec	eax
		mov	ah, al				; row #
		clr	al				; col #
		call	EdGotoXY
		mov	al, num
		pushad
		call	COUT
		mov	al, ' '
		call	COUT
		popad
		mov	eax, [SaveEAX]

		endm

SHOW		macro	reg
		push	ax
		mov	ax, reg
		call	PrAX
		pop	ax
		endm

ESHOW		macro	reg
		push	eax
		mov	eax, reg
		call	PrEAX
		pop	eax
		endm

E_SHOW		macro
		call	debug_charin
		cmp	ax, 03h
		je	BYE
		pop	ax
		call	EdGotoXY
		popad
		endm
		
STOP		macro	num
		B_SHOW	num
		E_SHOW
		endm

		
SEE		macro	num
		B_SHOW	num
		pop	ax
		call	EdGotoXY
		popad
		endm

		
TRACE		macro	num, reg
		B_SHOW	num
		ESHOW	reg
		E_SHOW
		endm
endif		

