; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	FILE.ASM						!
; 									!
; Description:	Definitions for playing with file I/O.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	FILEIOSTATUS	FILEIOSTATUS					!
;	FLAG_RO		R/O						!
;	FLAG_WO		W/O						!
;	FLAG_RW		R/W						!
;	OPENFILE	Open-File					!
;	CLOSEFILE	Close-File					!
;	EMITFILE	Emit-File					!
;	WRITEFILE	Write-File					!
;	READFILE	Read-File					!
;	WRITELINE	Write-Line					!
;	READLINE	Read-Line					!
;	SEEKFILE	Reposition-File					!
;	SEEKSET		SeekSet						!
;	SEEKCUR		SeekCur						!
;	SEEKEND		SeekEnd						!
;	FILEPOS		File-Position					!
;	FILESIZE	File-Size					!
;	DELETEFILE	Delete-File					!
;	QFILEEXIST	?FileExist					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; ===============================================================
; = FILEIOSTATUS                             			=
; =								=
; = Code for the froth word:    FileIOStatus           		=
; =								=
; = Stack:  ( --- n )                                         	=
; ===============================================================

		align	4
FILEIOSTATUS   	proc

		; Push file IO status
		movzx	eax, [FileIOStatus]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E_FILEIOSTATUS:
FILEIOSTATUS	endp


; ===============================================================
; = FLAG_RO                             			=
; =								=
; = Code for the froth word:    R/O           			=
; =								=
; = Stack:  ( --- n )						=
; ===============================================================

		align	4
FLAG_RO   	proc

		; Push value on stack
		IPUSH	00h

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_FLAG_RO:
FLAG_RO		endp


; ===============================================================
; = FLAG_WO                             			=
; =								=
; = Code for the froth word:    W/O           			=
; =								=
; = Stack:  ( --- n )						=
; ===============================================================

		align	4
FLAG_WO   	proc

		; Push value on stack
		IPUSH	01h

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_FLAG_WO:
FLAG_WO		endp


; ===============================================================
; = FLAG_RW                             			=
; =								=
; = Code for the froth word:    R/W           			=
; =								=
; = Stack:  ( --- n )						=
; ===============================================================

		align	4
FLAG_RW   	proc

		; Push value on stack
		IPUSH	02h

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_FLAG_RW:
FLAG_RW		endp


; ===============================================================
; = OPENFILE                             			=
; =								=
; = Code for the froth word:    Open-File           		=
; =								=
; = Stack:  ( addrz n1 --- n2 )					=
; ===============================================================

		align	4
OPENFILE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2
		call	SWAP					; SWAP
		call	QSTRING					; ?String
		call	SWAP					; SWAP


		; --------------------
		; Try to open the file
		; --------------------
		
		; Pointer to filename
		mov	ebx, PARAM1
		
		; Access mode
		mov	eax, PARAM0
		or	ax, 2090h			; b13: no int24h   b7: don't pass handle to child   b4: no sharing

		; Authorize new file creation ? (update ZF flag)
		mov	ecx, PARAM0
		and	cl, 3				; CL == 0 if R/O, else CL != 0
		
		; Call BIOS
		call	doWideOpenFile
		jc	@@ErrOpen
		
		
		; --------------------------------
		; No error, return the file handle
		; --------------------------------
		
		DISCARD	1				; drop
		movzx	eax, bx
		mov	PARAM0, eax			; File handle
		mov	[FileIOStatus], cx		; DOS action (0:file opened, 1:new file)
		mov	[FileEof + eax], FALSE		; fh ?EOF false!
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@ErrOpen:	DISCARD	1				; drop
		mov	PARAM0, -1
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_OPENFILE:
OPENFILE	endp


; ===============================================================
; = CLOSEFILE                             			=
; =								=
; = Code for the froth word:    Close-File           		=
; =								=
; = Stack:  ( n --- )						=
; ===============================================================

		align	4
CLOSEFILE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------
		; Try to close the file
		; ---------------------
		
		; File handle
		mov	ebx, PARAM0
		DISCARD	1					; drop
		
		; Call BIOS
		call	doCloseFile
		jc	@@Error
		
		
		; --------------------------------
		; No error, return the file handle
		; --------------------------------
		
		mov	[FileIOStatus], 0			; No error
		mov	[FileEof + ebx], FALSE			; fh ?EOF false!
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		ret

E_CLOSEFILE:
CLOSEFILE	endp


; ===============================================================
; = LINESEP							=
; =                                                 		=
; = Code for the forth word:	LineSep$			=
; =                     	                            	=
; = Stack:	( --- addrz )					=
; ===============================================================

		align	4
LINESEP		proc

		call	@@next
@@next:		add	dword ptr [esp], 4+6+3+2+1
		mov	eax, [edi.ddir_blkaddr + D_LINESEP]
		IPUSH	eax
		ret
		
		; 4+4+2+2
		add	PARAM0, size strentry
		
		; Return to caller
@@Exit:		ret

E_LINESEP:
LINESEP		endp


; ===============================================================
; = EMITFILE                             			=
; =								=
; = Code for the froth word:    Emit-File           		=
; =								=
; = Stack:  ( n1 c --- n2 )					=
; ===============================================================

		align	4
EMITFILE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; ---------------------------
		; Write the character to file
		; ---------------------------
		
		; File handle
		mov	ebx, PARAM1

		; Bytes to write
		mov	ecx, 1
		
		; Pointer to buffer
		mov	edx, [QuoteStrBuff]
		add	edx, size strentry
		mov	eax, PARAM0				; Get <c>
		mov	[edx], al
		
		; Call BIOS to perform asked operation
		DISCARD	1
		call	doWriteFile
		jc	@@Error
		
		
		; --------
		; No error
		; --------

		mov	PARAM0, ebx				; Bytes writen		
		mov	[FileIOStatus], 0			; No error
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	mov	PARAM0, -1				; No byte writen
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_EMITFILE:
EMITFILE	endp


; ===============================================================
; = WRITEFILE                             			=
; =								=
; = Code for the froth word:    Write-File           		=
; =								=
; = Stack:  ( n1 addr n2 --- n3 )				=
; ===============================================================

		align	4
WRITEFILE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; ----------------------
		; Write the data to file
		; ----------------------
		
		; File handle
		mov	ebx, PARAM2

		; Bytes to write (ECX) and pointer to buffer (EDX)
		mov	edx, PARAM1					; Buffer address
		mov	ecx, PARAM0					; Length
		
		; Call BIOS to perform asked operation
		DISCARD	2						; drop drop
		call	doWriteFile
		jc	@@Error
		
		
		; --------
		; No error
		; --------
		
		mov	PARAM0, ebx					; Bytes writen		
		mov	[FileIOStatus], 0				; No error
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:		mov	dword ptr PARAM0, -1			; Bytes writen
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_WRITEFILE:
WRITEFILE	endp


; ===============================================================
; = READFILE                             			=
; =								=
; = Code for the froth word:    Read-File           		=
; =								=
; = Stack:  ( fh addr n2 --- n3 )				=
; ===============================================================

		align	4
READFILE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; -----------------------
		; Read the data from file
		; -----------------------
		
		; File handle
		mov	ebx, PARAM2

		; Bytes to read (ECX) and pointer to buffer (EDX)
		mov	edx, PARAM1				; Buffer address
		mov	ecx, PARAM0				; Length

		; Memory protection
		; !!!
		
		; Call BIOS to perform asked operation
		DISCARD	2					; drop drop
		call	doReadFile
		jc	@@Error
		
		
		; --------
		; No error
		; --------

		mov	PARAM0, ebx				; Bytes read		
		mov	[FileIOStatus], 0			; No error
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	mov	dword ptr PARAM0, -1			; Bytes read = none
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_READFILE:
READFILE	endp


; ===============================================================
; = WRITELINE                             			=
; =								=
; = Code for the froth word:    Write-Line           		=
; =								=
; = Stack:  ( fh addrz --- n )					=
; ===============================================================

		align	4
WRITELINE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2

		; Verify if a valid string
		call	QSTRING					; ?String


		; ------------------------
		; Write the string to file
		; ------------------------
		
		; File handle
		mov	ebx, PARAM1

		; Bytes to write (ECX) and pointer to buffer (EDX)
		mov	edx, PARAM0				; String address
		clr	ecx
@@StrLen:	cmp	byte ptr [edx + ecx], 0
		je	@@StrLen2
		inc	ecx
		jmp	@@StrLen
@@StrLen2:
		
		; Call BIOS to perform asked operation
		call	doWriteFile
		jc	@@Error
		

		; --------------------------------
		; Write the line separator to file
		; --------------------------------

		; File handle
		mov	ecx, PARAM1
		mov	PARAM1, ebx				; Store length written
		mov	ebx, ecx

		; Bytes to write (ECX) and pointer to buffer (EDX)
		mov	edx, [edi + ddir_blkaddr + D_LINESEP]	; String address (Line Seperator)
		add	edx, size strentry
		clr	ecx
@@StrLenX:	cmp	byte ptr [edx + ecx], 0
		je	@@StrLenX2
		inc	ecx
		jmp	@@StrLenX
@@StrLenX2:

		; Call BIOS to perform asked operation
		call	doWriteFile
		jc	@@Error

		
		; --------
		; No error
		; --------

		DISCARD	1					; drop		
		add	PARAM0, ebx				; Bytes writen		
		mov	[FileIOStatus], 0			; No error
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	DISCARD	1					; drop
		mov	PARAM0, -1				; No byte writen
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_WRITELINE:
WRITELINE	endp


; ===============================================================
; = READLINE                             			=
; =								=
; = Code for the froth word:    Read-Line           		=
; =								=
; = Stack:  ( fh addrz n1 --- n2 )				=
; ===============================================================

		align	4
READLINE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3
		call	SWAP					; SWAP
		call	QSTRING					; ?String
		call	SWAP					; SWAP


		; ------------------------
		; Is-it a dynamic string ?
		; ------------------------

		call	OVER					; OVER
		call	QFIXEDSTRING				; ?FixedString
		call	DROP					; drop



		; -----------------------
		; Read max line from file
		; -----------------------
		
		; File handle
		mov	ebx, PARAM2				; File handle

		; Bytes to read (ECX) and pointer to buffer (EDX)
		mov	edx, PARAM1				; Buffer address
		mov	ecx, PARAM0				; Length

		; Call BIOS to perform asked operation
		call	doReadFile
		jc	@@Error
		push	ebx
		

		; ---------------------------------------------
		; Search the <EOL> seperator into the line read
		; ---------------------------------------------

		; Store a <NULL> at the end of the line read into buffer
		mov	eax, PARAM1				; buffer address
		mov	byte ptr [eax + ebx], 0

		; Scan string buffer
		push	edi
		mov	edi, [edi + ddir_blkaddr + D_LINESEP]	; String address (Line Seperator)
		add	edi, size strentry
		mov	edx, PARAM1				; buffer address
		clr	ecx
@@scan0:	clr	ebx
@@scan:		mov	ah, [edi + ebx]
		cmpz	ah, 0
		je	@@scan3
		mov	al, [edx + ebx]
		cmpz	al, 0
		je	@@scan3
		cmp	al, ah
		je	@@scan2
		inc	edx
		inc	ecx
		jmp	@@scan0
@@scan2:	inc	ebx
		jmp	@@scan
@@scan3:	pop	edi


		; -----------------------
		; Reposition file pointer
		; -----------------------

		; Offset
		mov	eax, PARAM2				; get <fh>
		mov	edx, ecx				; saving
		pop	ecx					; Byte read
		cmp	ecx, 0
		jne	@@NotEof
@@Eof:		mov	[FileEof + eax], TRUE
		jmp	@@EofDone
@@NotEof:	mov	[FileEof + eax], FALSE
@@EofDone:
		sub	ecx, edx				; Length
		neg	ecx
		add	ecx, ebx				; Skip line separator

		; File handle
		mov	ebx, PARAM2

		; Method of moving
		mov	eax, SEEK_CUR				; Seeks from current location

		; Call BIOS to perform asked operation
		call	doSeekFile
		jc	@@Error
		

		; -----------
		; Line length
		; -----------

		mov	ecx, PARAM1				; Buffer address
		DISCARD	2					; drop drop
		mov	PARAM0, edx				; Line length
		mov	[ecx.str_currlen - size strentry], edx	; Update length in the string header
		mov	byte ptr [ecx + edx], 0			; String zero terminal
		jmp	@@Exit


		; ----------------------------
		; There is an error on reading
		; ----------------------------

@@Error:	DISCARD	2					; drop
		mov	PARAM0, -1				; No byte read
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_READLINE:
READLINE	endp


; ===============================================================
; = QEOF							=
; =								=
; = Code for the forth word:	?EOF				=
; =								=
; = Stack:	( fh --- ff ) if not end-of-file after Read-Line=
; =		( fh --- tf ) if end-of-file after Read-Line	=
; ===============================================================

		align	4
QEOF		proc

		; Verify stack frame
		VSTACK	1

		; Push value on stack
		mov	eax, PARAM0
		movzx	eax, [FileEof + eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_QEOF:
QEOF		endp


; ===============================================================
; = SEEKFILE                             			=
; =								=
; = Code for the froth word:    Reposition-File           	=
; =								=
; = Stack:  ( fh n ofs1 --- ofs2 )				=
; ===============================================================

		align	4
SEEKFILE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; -----------------------
		; Read the data from file
		; -----------------------
		
		; File handle
		mov	ebx, PARAM2

		; Method of moving
		mov	eax, PARAM1

		; Offset
		mov	ecx, PARAM0
		DISCARD	2					; drop drop

		; Call BIOS to perform asked operation
		call	doSeekFile
		jc	@@Error
		
		
		; --------
		; No error
		; --------

		mov	PARAM0, ebx				; File position
		mov	[FileIOStatus], 0			; No error
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	mov	dword ptr PARAM0, -1			; File position error
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_SEEKFILE:
SEEKFILE	endp


; ===============================================================
; = SEEKSET                             			=
; =								=
; = Code for the froth word:    SeekSet           		=
; =								=
; = Stack:  ( --- n )						=
; ===============================================================

		align	4
SEEKSET   	proc

		; Push value on stack
		IPUSH	SEEK_SET
		
		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_SEEKSET:
SEEKSET		endp


; ===============================================================
; = SEEKCUR                             			=
; =								=
; = Code for the froth word:    SeekCur           		=
; =								=
; = Stack:  ( --- n )						=
; ===============================================================

		align	4
SEEKCUR   	proc

		; Push value on stack
		IPUSH	SEEK_CUR
		
		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_SEEKCUR:
SEEKCUR		endp


; ===============================================================
; = SEEKEND                             			=
; =								=
; = Code for the froth word:    SeekEnd           		=
; =								=
; = Stack:  ( --- n )						=
; ===============================================================

		align	4
SEEKEND   	proc

		; Push value on stack
		IPUSH	SEEK_END
		
		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_SEEKEND:
SEEKEND		endp


; ===============================================================
; = FILEPOS                             			=
; =								=
; = Code for the froth word:    File-Position          		=
; =								=
; = Stack:  ( n1 --- n2 )					=
; ===============================================================

		align	4
FILEPOS   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------
		; Read the data from file
		; -----------------------
		
		; File handle
		mov	ebx, PARAM0

		; Call BIOS to perform asked operation
		call	doFilePos
		jc	@@Error
		
		
		; --------
		; No error
		; --------

		mov	PARAM0, ebx				; File position
		mov	[FileIOStatus], 0			; No error
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	mov	dword ptr PARAM0, -1			; File position error
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_FILEPOS:
FILEPOS		endp


; ===============================================================
; = FILESIZE                             			=
; =								=
; = Code for the froth word:    File-Size           		=
; =								=
; = Stack:  ( fh --- n )					=
; ===============================================================

		align	4
FILESIZE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; -----------------------
		; Read the data from file
		; -----------------------
		
		; File handle
		mov	ebx, PARAM0

		; Call BIOS to perform asked operation
		call	doFileSize
		jc	@@Error
		
		
		; --------
		; No error
		; --------

		mov	PARAM0, ebx				; File size
		mov	[FileIOStatus], 0			; No error
		jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	mov	dword ptr PARAM0, -1			; File size error
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_FILESIZE:
FILESIZE	endp


; ===============================================================
; = DELETEFILE                             			=
; =								=
; = Code for the froth word:    Delete-File           		=
; =								=
; = Stack:  ( addrz --- )					=
; ===============================================================

		align	4
DELETEFILE   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1

		; Verify if a valid string
		call	QSTRING					; ?String


		; --------------------
		; Try to open the file
		; --------------------
		
		; Pointer to filename
		mov	ebx, PARAM0			; Get filename
		DISCARD	1				; drop
		
		; Call DOS
		call	doDeleteFile
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_DELETEFILE:
DELETEFILE	endp


; ===============================================================
; = QFILEEXIST                             			=
; =								=
; = Code for the froth word:    ?FileExist           		=
; =								=
; = Stack:  ( addrz --- f )					=
; ===============================================================

		align	4
QFILEEXIST   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1

		; Verify if a valid string
		call	QSTRING					; ?String


		; -------
		; Ask DOS
		; -------
		
		; Get filename
		mov	ebx, PARAM0

		; Call BIOS to perform asked operation
		call	doQFileExist
		jc	@@Error
		
		
		; --------
		; No error
		; --------

		mov	dword ptr PARAM0, ebx		; File found / not found
		mov	[FileIOStatus], ax		; No error
		ret					; jmp	@@Exit
		
		
		; ----------------------------
		; There is an error on opening
		; ----------------------------

@@Error:	mov	PARAM0, -1			; Error
		mov	[FileIOStatus], ax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_QFILEEXIST:
QFILEEXIST	endp


code32		ends
		end
