; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	H_LINE_10.ASM						!
; 									!
; Description:	Definitions for playing with graphics.			!
; 		(drawing line).						!
; 									!
; ----------------------------------------------------------------------+

Line_10		proc

		enter	8, 0

		push	esi edi

		movzx	esi, VSP.vsp_cbRow

		; check for horizontal line
		mov	bx, y2
		sub	bx, y1				; BX = y2 - y1
		jnz	L000

		call	_HorzLine_10
		jmp	L009

		; check for vertical line
L000:		mov	cx, x2
		sub	cx, x1				; CX = x2 - x1
		jnz	L001

		call	_VertLine_10
		jmp	L009

		; come here if not horizontal or vertical
L001:		jns	L002				; jump if x2 > x1

		; at this point x2 < x1, so reverse the endpoints
		neg	cx				; CX = dX
		neg	bx				; BX = dY

		mov	ax, x2				; exchange x1 and x2
		xchg	ax, x1
		mov	x2, ax

		mov	ax, y2				; exchange y1 and y2
		xchg	ax, y1
		mov	y2, ax

		; at this point, dX (in register CX) is positive
L002:		or	bx, bx				; test sign of dY
		jns	L004				; jump if slope is positive

		neg	bx				; make dY positive
		neg	esi				; negate the y-increment

		; select the appropriate routine for the slope of line
L004:		mov	LineFn, offset _LoSlopeLine_10
		cmp	bx, cx
		jl	L005				; jump if dY < dX (slope < 1)

		mov	LineFn, offset _HiSlopeLine_10
		xchg	bx, cx				; exchange dY and dX

		; calculate the initial decision variable and increments
L005:		shl	bx, 1				; BX = Incr1 = 2*dY
		mov	dx, bx
		sub	dx, cx				; DX = d = 2*dY-dX
		mov	Incr2, dx
		sub	Incr2, cx			; Incr2 = 2*(dY-dX)

		; calculate the first pixel address
		push	cx dx				; preserve dX and d

		push	word ptr x1
		push	word ptr y1
		call	PixelAddr_10			; DX:AX -> buffer

		; set up the Graphics Controller
		call	_SetGC_10 			; (leaves DX = 3CEh)
		mov	ax, 0B05h			; AH = 0Bh (write mode 3, read mode 1)
		out	dx, ax

		mov	ax, 7				; AH = 0 (Color Don't Care value)
		out	dx, ax

		; create a bit mask for the first pixel
		mov	ax, 8008h			; AH = 80h (bit mask)
							; AL = 8 (Bit Mask reg #)
		mov	cl, x1	
		and	cl, 7				; CL = x1 MOD 8
		shr	ah, cl				; AH = initial bit mask

		pop	dx				; DX = d
		pop	cx				; CX = dX
		inc	cx				; CX = # of pixels to draw
	
							; at this point:
							;  EDI -> first pixel
							;   AH = bit mask for 1st pixel
							;   BX = Incr1
							;   CX = # pixels to draw
							;   DX = decision variable
							;   SI = y-increment

		call	LineFn				; draw the line

		; restore the default Graphics Controller state
L009:		mov	dx, 3CEh
		clr	ax				; AH = 0, AL = 0
		out	dx, ax				; Set/Reset

		inc	ax				; AH = 0, AL = 1
		out	dx, ax				; Enable Set/Reset

		mov	al, 3				; AH = 0, AL = 3
		out	dx, ax				; Data Rotate/Func Select

		mov	al, 5				; AH = 0, AL = 5
		out	dx, ax				; Mode

		mov	ax, 0FF08h			; AH = 11111111b, AL = 8
		out	dx, ax				; Bit Mask

		; restore registers and return
		pop	edi esi
		leave
		ret	5*2

Line_10		endp


_VertLine_10	proc

		mov	ax, y1				; AX = y1
		mov	cx, y2
		sub	cx, ax				; CX = y2-y1
		jge	L011				; jump if dY >= 0
	
		add	ax, cx				; AX = y2
		neg	cx				; force dY >= 0

L011:		inc	cx				; CX = # of pixels to draw
		push	cx				; preserve CX

		; point to the bottommost pixel
		push	word ptr x1
		push	ax
		call	PixelAddr_10
	
		call	_SetGC_10 			; set up Graphics Controller
	
		mov	ax, 8008h			; AH = 80h (bit mask)
							; AL = 8 (Bit Mask register number)
		mov	cl, x1
		and	cl, 7				; CL = x1 MOD 8
		shr	ah, cl
		out	dx, ax
	
		pop	cx				; CX = # of pixels

		; draw the line
L012:		and	fs:[edi], al			; set pixel
		add	edi, esi				; increment to next line
		loopw	L012
	
		ret

_VertLine_10	endp



_HorzLine_10	proc

		; ensure that x1 < x2 (move left to right)
		movzx	edi, word ptr x1			; DI = x1
		movzx	esi, word ptr x2			; SI = x2
		cmp	si, di
		jae	L020

		xchg	esi, edi				; DI = leftmost x-coordinate

		; compute the bit masks for the leftmost and rightmost pixel(s)
L020:		mov	bx, 0FF80h			; BH = FFh (left mask)
							; BL = 80h (right mask)
		mov	cx, di
		and	cl, 7				; CL = x1 MOD 8
		shr	bh, cl				; BH = left mask
		mov	cx, si
		and	cl, 7				; CL = x2 MOD 8
		sar	bl, cl				; BL = right mask

		; compute the offset of the first byte in the scan line
		mov	ax, y1
		mul	VSP.vsp_cbRow			; AX = y * (bytes per row)

		; compute byte offsets for x1 and x2
		shr	di, 3				; DI = byte offset of x1
		shr	si, 3				; SI = byte offset of x2

		mov	cx, si
		sub	cx, di				; CX = difference of byte offsets

		; point to the first byte in the line
		movzx	eax, ax				; movzx eax, ax
		add	edi, eax
		add	edi, VSP.vsp_VBuf			; EDI -> 1st byte
	
		call	_SetGC_10 			; set up Graphics Controller

		jcxz	L023				; jump if only 1 byte in the line

		; set pixels in the leftmost byte of the line
		mov	ah, bh				; AH = bit mask for 1st byte
		mov	al, 8				; AL = 8 (Bit Mask register number)
		out	dx, ax				; update Graphics Controller
	
		and	fs:[edi], ah			; update the pixels

		dec	cx
		jz	L022				; jump if only 1 byte remains

		; draw the middle of the line
		mov	ah, 0FFh				; AH = bit mask
		out	dx, ax				; update Bit Mask register
		
L021:		inc	edi				; point to the next byte
		and	fs:[edi], ah			; update pixels
		loopw	L021

		; set pixels in the rightmost byte of the line
L022:		mov	ah, bl				; AH = bit mask for last byte
		out	dx, ax				; update Graphics Controller
		and	fs:[edi+1], ah			; update bit planes

		ret

		; come here if only one byte in the line
L023:		mov	ah, bl
		and	ah, bh				; AH = bit mask for the byte
		mov	al, 8				; AL = 8 (Bit Mask register number)
		out	dx, ax

		and	fs:[edi], ah			; set the pixels
		ret

_HorzLine_10	endp


_LoSlopeLine_10	proc


		mov	al, ah				; AL = initial bit mask

L031:		or	ah, al				; mask current pixel position
		ror	al, 1				; rotate pixel value
		jc	L034				; jump if bit mask rotated to
							;  leftmost pixel position

		; bit mask not rotated around
		or	dx, dx				; test sign of d
		jns	L032				; jump if d >= 0
		
		add	dx, bx				; d = d + Incr1
		loopw	L031

		; at this point, all but the last byte of pixels have been set
		and	fs:[edi], ah			; set remaining pixel(s)
		ret

		; come here when d >= 0
L032:		add	dx, Incr2			; d = d + Incr2
		and	fs:[edi], ah			; update bit planes
		mov	ah, al				; AH = bit mask for next pixel
		add	edi, esi				; increment y
		loopw	L031
		ret

		; bit mask rotated around
L034:		and	fs:[edi], ah			; update bit planes
		mov	ah, al				; AH = bit mask for next pixel
		inc	edi				; increment x

		or	dx, dx				; test sign of d
		jns	L035				; jump if non-negative

		add	dx, bx				; d = d + Incr1
		loopw	L031
		ret

L035:		add	dx, Incr2			; d = d + Incr2
		add	edi, esi				; increment y
		loopw	L031
		ret

_LoSlopeLine_10	endp


_HiSlopeLine_10	proc

L041:		and	fs:[edi], ah			; update bit planes
		add	edi, esi				; increment y

L042:		or	dx, dx				; test sign of d
		jns	L043				; jump if d >= 0

		add	dx, bx				; d = d + Incr1
		loopw	L041				; loop until all pixels are drawn
		ret

L043:		add	dx, Incr2			; d = d + Incr2
		ror	ah, 1				; rotate bit mask
		adc	edi, 0				; increment DI if mask has rotated to
							;  leftmost pixel position
		loopw	L041
		ret

_HiSlopeLine_10	endp


_SetGC_10	proc

		push	ax

		mov	dx, 3CEh			; Graphics Controller I/O addr

		mov	ah, ucPixel			; AH = pixel value
		clr	al				; AL = Set/Reset reg #
		out	dx, ax

		mov	ax, 0F01h			; AH = 1111b (bit plane mask for
							;  Enable Set/Reset)
		out	dx, ax				; AL = Enable Set/Reset reg #

		mov	ah, VSP.vsp_cRMW		; AH = Read-Modify-Write bits
		mov	al, 3				; AL = Data Rotate/Func Select reg #
		out	dx, ax

		pop	ax
		ret

_SetGC_10	endp

