; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	H_LINE_10.ASM						!
; 									!
; Description:	Definitions for playing with graphics.			!
; 		(drawing line).						!
; 									!
; ----------------------------------------------------------------------+

Line_13		proc

		enter	8, 0

		push	esi edi

		movzx	esi, VSP.vsp_cbRow

		; check for horizontal line
		mov	bx, y2
		sub	bx, y1				; BX = y2 - y1
		jnz	L100

		call	_HorzLine_13
		jmp	L109

		; check for vertical line
L100:		mov	cx, x2
		sub	cx, x1				; CX = x2 - x1
		jnz	L101

		call	_VertLine_13
		jmp	L109

		; come here if not horizontal or vertical
L101:		jns	L102				; jump if x2 > x1

		; at this point x2 < x1, so reverse the endpoints
		neg	cx				; CX = dX
		neg	bx				; BX = dY

		mov	ax, x2				; exchange x1 and x2
		xchg	ax, x1
		mov	x2, ax

		mov	ax, y2				; exchange y1 and y2
		xchg	ax, y1
		mov	y2, ax

		; at this point, dX (in register CX) is positive
L102:		or	bx, bx				; test sign of dY
		jns	L104				; jump if slope is positive

		neg	bx				; make dY positive
		neg	esi				; negate the y-increment

		; select the appropriate routine for the slope of line
L104:		mov	LineFn, offset _LoSlopeLine_13
		cmp	bx, cx
		jl	L105				; jump if dY < dX (slope < 1)

		mov	LineFn, offset _HiSlopeLine_13
		xchg	bx, cx				; exchange dY and dX

		; calculate the initial decision variable and increments
L105:		shl	bx, 1				; BX = Incr1 = 2*dY
		mov	dx, bx
		sub	dx, cx				; DX = d = 2*dY-dX
		mov	Incr2, dx
		sub	Incr2, cx			; Incr2 = 2*(dY-dX)

		; calculate the first pixel address
		push	cx dx				; preserve dX and d

		push	word ptr x1
		push	word ptr y1
		call	PixelAddr_13			; DX:AX -> buffer

		; set up the Graphics Controller
		call	_SetGC_13 			; (leaves DX = 3CEh, AH = pixel value)

		pop	dx				; DX = d
		pop	cx				; CX = dX
		inc	cx				; CX = # of pixels to draw

							; at this point:
							;  EDI -> first pixel
							;   AH = pixel value
							;   BX = Incr1
							;   CX = # pixels to draw
							;   DX = decision variable
							;   SI = y-increment

		call	LineFn				; draw the line

		; restore the default Graphics Controller state
L109:		mov	dx, 3CEh
		mov	ax, 3				; AH = 0, AL = 3
		out	dx, ax				; Data Rotate/Func Select

		; restore registers and return
		pop	edi esi
		leave
		ret	5*2

Line_13		endp


_VertLine_13	proc

		mov	ax, y1				; AX = y1
		mov	cx, y2
		sub	cx, ax				; CX = y2-y1
		jge	L111				; jump if dY >= 0

		add	ax, cx				; AX = y2
		neg	cx				; force dY >= 0

L111:		inc	cx				; CX = # of pixels to draw

		; point to the bottommost pixel
		push	word ptr x1
		push	ax
		call	PixelAddr_13

		call	_SetGC_13

		; draw the line
L112:		mov	al, ah				; AL = pixel value
		xchg	fs:[edi], al
		add	edi, esi			; increment y
		loopw	L112

		ret

_VertLine_13	endp


_HorzLine_13	proc

		; ensure that x1 < x2 (move left to right)
		mov	di, word ptr x1			; DI = x1
		mov	si, word ptr x2			; SI = x2
		cmp	si, di
		jae	L120

		xchg	si, di				; DI = leftmost x-coordinate

L120:		mov	cx, si
		sub	cx, di				; CX = dX
		inc	cx				; CX = # of pixels

		; point to the leftmost pixel
		push	di
		push	word ptr y1
		call	PixelAddr_13

		call	_SetGC_13
		mov	al, ah				; AL = pixel value

		; draw the line
L121:		mov	ah, fs:[edi]			; latch the current pixel value
		mov	fs:[edi], al			; update the pixel
		inc	edi
		loopw	L121

		ret
		
_HorzLine_13	endp


_LoSlopeLine_13	proc

L131:		mov	al, ah
		xchg	al, fs:[edi]			; update the pixel
		inc	edi				; point to next pixel

		or	dx, dx				; test sign of d
		jns	L132				; jump if d >= 0

		add	dx, bx				; d = d + Incr1
		loopw	L131
		ret

		; come here when d >= 0
L132:		add	dx, Incr2			; d = d + Incr2
		add	edi, esi			; increment y
		loopw	L131
		ret

_LoSlopeLine_13	endp


_HiSlopeLine_13	proc

L141:		mov	al, ah
		xchg	al, fs:[edi]			; update the pixel
		add	edi, esi				; increment y

		or	dx, dx				; test sign of d
		jns	L142				; jump if d >= 0

		add	dx, bx				; d = d + Incr1
		loopw	L141				; loop until all pixels are drawn
		ret

L142:		add	dx, Incr2			; d = d + Incr2
		inc	edi				; point to next pixel
		loopw	L141
		ret

_HiSlopeLine_13	endp


_SetGC_13	proc

		mov	dx, 3CEh			; Graphics Controller I/O addr
	
		mov	ah, VSP.vsp_cRMW		; AH = Read-Modify-Write bits
		mov	al, 3				; AL = Data Rotate/Func Select reg #
		out	dx, ax

		mov	ah, ucPixel			; AH = pixel value
		ret

_SetGC_13	endp

