; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	MEMORY.ASM						!
; 									!
; Description:	Definitions for playing with the extern interpret.	!
; 									!
; List of FROTH definitions defined in this module:			!
;	SHARPIF		#IF						!
;	SHARPELSE	#ELSE						!
;	SHARPENDIF	#ENDIF						!
;	QLOAD		?LOAD						!
;	INTERPRET	INTERPRET					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc                   		; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = SHARPIF							=
; =                                                 		=
; = Code for the forth word:	#IF				=
; =                     	                            	=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
SHARPIF		proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SHARPIF:
SHARPIF		endp


; ===============================================================
; = SHARPELSE							=
; =                                                 		=
; = Code for the forth word:	#ELSE				=
; =                     	                            	=
; = Stack:	( --- )						=
; ===============================================================

		align	4
SHARPELSE	proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SHARPELSE:
SHARPELSE	endp


; ===============================================================
; = SHARPENDIF							=
; =                                                 		=
; = Code for the forth word:	#ENDIF				=
; =                     	                            	=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
SHARPENDIF	proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SHARPENDIF:
SHARPENDIF	endp


; ===============================================================
; = QLOAD							=
; =                                                 		=
; = Code for the forth word:	?LOAD				=
; =                     	                            	=
; = Stack:	( --- )						=
; =                     	                            	=
; = Usage:	Issue an error if interpreted from a 		=
; =		file loaded.					=
; ===============================================================

		align	4
QLOAD		proc

		; -------------------------
		; Error is a file is loaded
		; -------------------------
		
		cmp	[LoadFilesNb], 0
		je	@@Ok

		; "Cannot be compiled from a loaded file."
		mov	ah, ERR_NOTLOADED			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
@@Ok:		ret

E_QLOAD:
QLOAD		endp


; ===============================================================
; = isDynVarAlAll						=
; =                                                         	=
; = Function: is a dynamic variable already allocated ?		=
; =                                                     	=
; = I) none                         				=
; = O) none (never return)                  			=
; = M) never return                     			=
; ===============================================================

isDynVarAlAll	proc

		; Dynamic variable already allocated ?
		cmp	[AllocFlag], TRUE
		je	@@AlAlloc
		
		; Free this variable ?
		cmp	[FreeFlag], TRUE
		je	@@FreeDynVar

		; Done
		ret
		
@@AlAlloc:

		; "DYNAMIC word already allocated"
		mov	ah, ERR_DYNVARALALL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError
		
		
		; -------------------------
		; Free the dynamic variable
		; -------------------------

@@FreeDynVar:

		; Drop address on stack
		DISCARD	1

		; Get dynamic variable begin address
		mov	edx, [esp]
		sub	edx, 26-9
		mov	[DynVarAddr], edx
		mov	ecx, edx

		; Release the memory
		mov	edx, [edx + 2]
		sub	edx, ddir_blkaddr			; Memory block nb
		call	UnAllot

		; Put: 'CALL DynVarNotAll' (5 bytes)
		CALL_rel32					; CALL rel32
		lea	eax, DynVarNotAll
		add	eax, [OfsDictReloc]
		sub	eax, [DynVarAddr]
		sub	eax, 4					; 4 is for REL32
		call	CodeComma               		; Code,

		; Put compile-time procedure RELATIVE address
		mov	eax, [ecx + 22]				; EAX is compile-time RELATIVE address
		call	CodeComma				; Code,

		; Reset Free flag
		mov	[FreeFlag], FALSE

		; Done
		mov	[DynVarAddr], 0
		
		; Do NOT execute next instruction: 'JMP run_time procedure'
		add	esp, @WordSize
		ret
		
E_isDynVarAlAll:
isDynVarAlAll	endp


; ===============================================================
; = _CreateDynamicVar						=
; =                                                         	=
; = Local functions.						=
; =                                                     	=
; = I) EAX : NIA word ptr					=
; = O) none (never return)                  			=
; = M) never return                     			=
; ===============================================================

_CreateDynamicVar proc

		; --------------
		; Save used regs
		; --------------
		
		pushad


		; ---------------------------------
		; This is a proc callable with CALL
		; ---------------------------------

		; Compile: CALL $
		push	eax
		CALL_rel32
		clr	eax
		call	CodeComma               			; Code,
		pop	eax

		; Compile: ADD DWORD PTR [ESP], 4+6+3+2+1
		ADD_xESPi8_i8	0, 4+6+3+2+1


		; ----------------------------------
		; Put: 'CALL DynVarNotAll' (5 bytes)
		; ----------------------------------
		
		compCALL_rel32	DynVarNotAll
		

		; --------------------------------------
		; Put compile-time procedure REL address
		; --------------------------------------

		push	eax
		mov	eax, [eax.word_realaddr]		; EAX is compile-time address
		sub	eax, [CDA_START]
		call	CodeComma				; Code,
		pop	eax


		; ----------------------------------------------------------
		; Store 12 NOPs : a call to a variable is normally 11 bytes:
		;	CALL	$+0				5
		;	ADD	DWORD PTR [ESP], 4+6+3+2+1	4
		;	MOV	EAX, [EDI + n_var]		6
		;	SUB	ESI, 4				3
		;	MOV	[ESI], EAX			2
		;	RET					1
		; ----------------------------------------------------------
		
		_NOP_	((6+3+2+1)-(5+4))


		; ----------------------------------------------------------------------------------
		; Put 'CALL rel32' to a function which verify that dyna var is not already allocated
		; ----------------------------------------------------------------------------------

		compCALL_rel32	isDynVarAlAll


		; -------------------------------------
		; Put 'JMP rel32' to run-time procedure
		; -------------------------------------

		; Store codop
		JMP_rel32
		push	eax
		mov	eax, [eax.word_runtaddr]		; EAX is run-time address
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 for size REL32
		call	CodeComma				; Code,
		pop	eax


		; ----------------------------------------------------
		; Store the RELATIVE address to compile-time procedure
		; ----------------------------------------------------
		
		mov	eax, [eax.word_realaddr]		; EAX is compile-time address
		sub	eax, [CDA_START]
		call	CodeComma				; Code,


		; -------
		; align 4
		; -------
		
@@Align4:	mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		and	al, 3
		jz	@@Align4Done
		_NOP_
		jmp	@@Align4
@@Align4Done:


		; -----------------------------------------------------
		; Update the end code address of the word being defined
		; -----------------------------------------------------

		; Pointer on the word NIA address		
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		add	eax, [NIA_START]

		; Compute the size of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		sub	ebx, [eax.word_realaddr]
		
		; Store the code size
		mov	[eax.word_CodeSize], bx
		
		; Update the end of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		dec	ebx
		mov	[eax.word_EndCode], ebx


		; ----------
		; Reset flag
		; ----------
		
		mov	[DynamicFlag], FALSE

		
		; ----------------
		; Return to caller
		; ----------------

		popad		
		ret

_CreateDynamicVar endp


; ===============================================================
; = _PutMacro							=
; =                                                         	=
; = Usage:							=
; =    Called by INTERPRET, store a given word macro.		=
; =                                                     	=
; = I) EDX : ptr to NIA word.					=
; = O) none							=
; = M) all regs							=
; ===============================================================

		align	4
_PutMacro	proc

		; --------------
		; Temp variables
		; --------------

t_OStackAddr	equ	byte ptr [ebp - 4]
t_into		equ	byte ptr [ebp - 8]
		enter	2 * @WordSize, 0


		; ----------------------------------------
		; Pointer to Begin/End on the code to copy
		; ----------------------------------------
		
		mov	ecx, [edx.word_realaddr]			; Begin
		mov	edx, [edx.word_EndCode]				; End


		; -----------------------------------------------------------------
		; If ?speed if FALSE and macro begins with VSTACK n, skip this call
		; -----------------------------------------------------------------
		
		; 1st instruction is CALL ?
		cmp	byte ptr [ecx], 0E8h
		jne	@@speed30
		mov	eax, [ecx + 1]
		lea	eax, [eax + ecx + 5]

		; ?safe		
		call	QSAFE						; ?safe
		call	BOOL_QTRUE					; true?		
		lodsd
		cmpz	ebx, FALSE
		je	@@speed10

		CALL_rel32						; CALL rel32
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4						; 4 for REL32
		call	CodeComma               			; Code,
@@speed10:	add	ecx, 5

@@speed30:	

		; ------------------------------------
		; Integer arithemetic overflow error ?
		; ------------------------------------
		
		mov	[t_into], FALSE
		lea	ebx, [edx - 1]
		cmp	ebx, ecx
		jbe	@@LL51
		cmp	byte ptr [ebx], INTO_codeop
		jne	@@LL51
		
		; Yes, INTO
		call	isQIArithOverfl					; ?iarithoverflow
		jc	@@LL50
		mov	[t_into], TRUE
@@LL50:		dec	edx
@@LL51:


		; ------------------
		; Skip CALL OSTACK ?
		; ------------------

		mov	[t_OStackAddr], FALSE
		lea	ebx, [edx - 5]
		cmp	ebx, ecx
		jbe	@@LL30
		
		; Instruction is CALL ?
		cmp	byte ptr [ebx], 0E8h				; CALL ostack ?
		jne	@@LL30
		
		; ?safe		
		call	QSAFE						; ?safe
		call	BOOL_QTRUE					; true?		
		lodsd
		cmpz	ebx, FALSE
		je	@@LL10

		mov	[t_OStackAddr], TRUE
@@LL10:		sub	edx, 5
		
@@LL30:		


		; -----------------------
		; Now copy the macro code
		; -----------------------

@@CopyMacro:	mov	al, [ecx]
		call	CodeCComma					; CodeC,
		inc	ecx
		cmp	ecx, edx
		jb	@@CopyMacro


		; ---------------------
		; Compile CALL OSTACK ?
		; ---------------------

		cmp	[t_OStackAddr], FALSE
		je	@@LL90
		compCALL_rel32	OStack					; CALL rel32
@@LL90:		


		; --------------
		; Compile INTO ?
		; --------------

		cmp	[t_into], FALSE
		je	@@LL91
		_INTO_							; into
@@LL91:


		; ----------------
		; Return to caller
		; ----------------
		
		leave
		ret

_PutMacro	endp


; ===============================================================
; = AutoFind							=
; =                                                         	=
; = Usage:							=
; =	Search a given word name into automatic symbols 	=
; =	table.							=
; =                                                     	=
; = I) EDX : pointer to word name.				=
; =                                                         	=
; = O) CF  : 1 if not found, 0 is found.			=
; =    ECX : pointer to LDA (lword_entry).			=
; =                                                         	=
; = M) ECX and CF						=
; =                                                         	=
; ===============================================================

		align	4
AutoFind	proc

		; Save used regs
		push	eax ebx

		; Search the word into local words table
		clr	ecx
@@SrchLW:	cmp	ecx, [LDA_SIZE]
		je	@@WordNotFound
		clr	ebx
		add	ecx, [LDA_START]
@@SrchLW2:	mov	al, [ecx.lword_name + ebx]
		mov	ah, [edx + ebx]
		TOUPPER	al
		TOUPPER	ah
		cmp	al, ah
		jne	@@SrchLW3
		or	al, ah
		jz	@@LWordFound
		inc	ebx
		jmp	@@SrchLW2
@@SrchLW3:	sub	ecx, [LDA_START]
		add	ecx, size lwordentry
		jmp	@@SrchLW
		
		; Word is found into local words table
@@LWordFound:	clc
		jmp	@@Exit

		; Word not found
@@WordNotFound:	stc

		; Return to caller
@@Exit:		pop	ebx eax
		ret

AutoFind	endp


; ===============================================================
; = _Compile_inb                           			=
; =                                                         	=
; = Usage:	Compile integer literal number.			=
; =                                                     	=
; = I) ( n --- )						=
; =                                                         	=
; = O) 								=
; =                                                         	=
; = M) 								=
; ===============================================================

		align	4
_Compile_inb	proc

		; Put 'SUB ESI, 4'
		SUB_ESI_i8	@WordSize

		; Stack overflow run-time check ?		
		call		isSafe
		jnc		@@NoOvTest
		compCALL_rel32	OStack					; CALL   OStack
@@NoOvTest:

		; Put 'MOV DWORD PTR [ESI], IMM32'
		MOV_xESIi8_i32
		lodsd
		call	CodeComma               			; Code,

		ret

_Compile_inb	endp


; ===============================================================
; = INTERPRET                           			=
; =                                                         	=
; = Forth extern interpret.                 			=
; =                                                     	=
; = I) none                         				=
; =                                                         	=
; = O) none (never return)                  			=
; =                                                         	=
; = M) never return                     			=
; ===============================================================

global		InterpretAddr:	proc				; See 'ALLOT' and 'CALLOT'

		align	4
INTERPRET	proc

@@GetLine:

		; -----------------------------------------
		; Exit to host os if running a turnkey .exe
		; -----------------------------------------
		
		cmp	[ExeFile], TRUE
		je	BYE


		; -------------------------
		; Only one thread is active
		; -------------------------

		call	ResetThreads
		
		
		; -------------------------------------
		; Reset the flag comment to end of line
		; -------------------------------------
		
		mov	[CDA_SLASHSLASH], FALSE


		; --------------------------
		; Are we loading from file ?
		; --------------------------

		cmp	[LoadFilesNb], 0
		jne	@@LoadFromFile

	
		; ---------------------
		; Reading from terminal
		; ---------------------


		; Does greating message printed ?
		; -------------------------------
		
		cmp	[GreatingMsgPr], FALSE
		jne	@@GreatingMsg
		cmp	[ExeFile], TRUE			; Are-we executing a saved file.exe ?
		je	@@GreatingMsg

		; Print greating msg
		call	prGreatingMsg

if TURNKEYED eq 0
		; Initialize TRACE manager
		call	doTraceInit
endif

		; Greating msg done
		mov	[GreatingMsgPr], TRUE
		
@@GreatingMsg:
		
		

		; Skip a line
		; -----------
		
		call	CR					; CR


		; Print the nb of words on the stack, if mode is execute (not compile state)
		; ----------------------------------------------------------------------------
		
		; Print the nb of words on the stack, if mode is execute (not compile state)
		cmp	[CompileMode], TRUE
		je	@@LL1
		call	PrStackDepth
@@LL1:


		; Read a line of text from terminal
		; ---------------------------------
		
		lea	edx, INPUTBUFF
		mov	ecx, IBUFFMAXS
		call	ReadLine
		

		; SPACE
		; -----

		call	SPACE					; Space
		
		jmp	@@AnalLine
		
		
		; ---------------------------
		; Loading from a file ("load)
		; ---------------------------

@@LoadFromFile: 

		; Read a line from the current file loaded (see "load)
		; ----------------------------------------------------

		call	RdLoadFileLine

	
		; Initialization: EDX points to the start of the input buffer
		; -----------------------------------------------------------
		
@@AnalLine: 


		; ===============================================   =============================================
		; Strip space chars: goto on first not blank char   (It's the top of the LOOP to treat each word)
		; ===============================================   =============================================

@@GtoFirstNB:	
		; If reading from a file, record the current column position
		mov	eax, [LoadFilesNb]
		cmpz	eax, 0
		je	@@GtoFirstNB0
		dec	eax
		imul	eax, size loadfileentry
		add	eax, [LoadFilesInfos]
		mov	ebx, edx
		sub	ebx, [eax.loadfile_linsaddr]
		inc	ebx
		mov	[eax.loadfile_colnb], bx			; Current position in the buffer		

@@GtoFirstNB0:	mov 	al, [edx]
		cmpz 	al, 0
		je  	@@GetLine
		cmp	al, 09h						; TAB key
		je	@@GtoFirstNB3
		cmp 	al, ' '
		jne 	@@GtoFirstNB2
@@GtoFirstNB3:	inc 	edx
		jmp 	@@GtoFirstNB0
@@GtoFirstNB2:


		; First, save current ptr: the start of the word
		; ----------------------------------------------
		
		mov 	[IBUFFPTR], edx			; Current pointer into the buffer
		mov 	ecx, edx


		; Goto now on the first blank, or end of buffer
		; ---------------------------------------------

@@GtoFirstBl:  	mov 	al, [ecx]
		cmpz 	al, 0
		jz  	@@EndWord
		cmp 	al, 09h				; TAB key
		je  	@@EndWord
		cmp 	al, ' '
		je  	@@EndWord
		inc 	ecx
		jmp 	@@GtoFirstBl


		; Found the end of the word
		; -------------------------

@@EndWord:	mov 	byte ptr [ecx], 0           	; Now: EDX=word start, ECX=word end
		mov 	[IBUFFPTR2], ecx            	; Save pointer to end word


		; Is comment to end of line active ?
		; ----------------------------------
		
		cmp	[CDA_SLASHSLASH], TRUE
		je	@@NextWord


		; Is-it ')' (end of comment block)
		; --------------------------------
		
		cmp 	[CDA_COMMENT], 0      		; Inside block comment ?
		je  	@@IsBracketOn
		cmp 	word ptr [edx], '(' 	        ; Is this ')\0' ?
		je 	@@IsBracketOn             	; If not, ignore word; because comment is ON
		cmp 	word ptr [edx], ')' 	        ; Is this ')\0' ?
		jne 	@@NextWord              	; If not, ignore word; because comment is ON


		; Are we into [ ... ] mode ?
		; --------------------------
		
@@IsBracketOn:	cmp	[CDA_BRACKET_MODE], FALSE
		je	@@IsWordInDict
		
		; Is-it ]] ?
		cmp	word ptr [edx], ']]'			; ']:'
		jne	@@NotEndBrack
		call	CLOSEASMBRACKET
		jmp	@@NextWord
@@NotEndBrack:	

		; Is-it a valid number ?
		call	_BASE					; BASE
		DO_AT						; @
		call	TOR					; >R
		call	HEXADECIMAL				; HEXADECIMAL
		sub	esi, @WordSize
		mov	[esi], edx
		call	NUMBER
		call	FROMR					; R>
		call	_BASE					; BASE
		push	ebx
		DO_STORE					; !
		pop	ebx
		mov 	eax, [esi]
		add 	esi, @WordSize
		cmpz 	eax, 0
		jne	@@BrackNbOk
		
		; "Not a valid hexadecimal number in [[ ]] mode."
		mov	ah, ERR_NOTNBBRACKMODE			; Error #
		clr	ecx					; No line #
		call	doError

@@BrackNbOk:	lodsd
		cmp	eax, 100h
		jae	@@Brack2
		call	CodeCComma				; C,
		jmp	@@NextWord
@@Brack2:	cmp	eax, 10000h
		jae	@@Brack3
		call	CodeWComma				; W,
		jmp	@@NextWord
@@Brack3:	call	CodeComma				; ,
		jmp	@@NextWord
		


		; Is the word into the dictionary ?
		; ---------------------------------
		
@@IsWordInDict:	sub	esi, @WordSize
		mov	PARAM0, edx
		call	PFind


		; If found, execute or compile it
		; -------------------------------
		
		lodsd			             		; Get flag (word found ?)
		cmpz	eax, FALSE
		je	@@IsItLocalWrd


		; Word is found: is-it a 'Build> ... Does> ...' construction ?
		; ------------------------------------------------------------
		
		mov 	eax, ebx
		add 	eax, [NIA_START]
		test	byte ptr [eax.word_flags], FBUILDDOES
		jz	@@NotBuild

		mov	[CDA_CURR_WORD], ebx			; Save NIA address of current word (see 'CreateLocal')
		DISCARD	1					; DROP NIA address
		cmp	[CompileMode], TRUE
		je	@@BuildComp

		; Build the name
		call	doBuild					; rt_build>
		

		; Update word cross ref
		; ---------------------

		; Point to the last word defined in the current vocabulary: word currently being defined
		mov	ecx, [VOC_CURRENT]
		add	ecx, [VOC_START]
		mov	ecx, [ecx.voc_lastword]			; NIA pointer to caller word

		; Update the pointer to the defining-word
		push	ecx
		add	ecx, [NIA_START]
		mov	[ecx.word_DefiningWord], ebx
		pop	ecx
		
if TURNKEYED eq 0
		; Update word cross ref, only if it's not a system word
		mov	edx, ebx
		add	edx, [NIA_START]
		test	[edx.word_flags], FSYSTEM
		jnz	@@System2
		sub	edx, [NIA_START]
		call	CRef_add
@@System2:
endif
		
		; Create a DYNAMIC variable ?
		cmp	[DynamicFlag], FALSE
		je	@@NoDynamic
		call	_CreateDynamicVar
		jmp	InterpretAddr				; Done
@@NoDynamic:	call	[eax.word_realaddr]			; Execute the compile-time sequence

InterpretAddr:	jmp	@@NextWord
@@BuildComp:	call	doLocalBuild				; rtLocalBuild>
		jmp	@@NextWord
		
@@NotBuild:
		

		; Word is found: execute or compile ?
		; -----------------------------------
		
		cmp 	[CompileMode], FALSE
		je  	@@ExecWord       	       		; If mode execute (not compilation mode)
		test 	[eax.word_flags], FIMMEDIATE  		; Is word immediate ?
		jnz  	@@ExecWord				; Execute this immediate word

		
		; Try to compile a run-time word ?
		; --------------------------------
		
		test 	byte ptr [eax.word_flags], FRUNTIME
		jz	@@NotRT
		
		; "Run time words can't be compiled inside user definitions."
		mov	ah, ERR_RTWCANTBECOMPILED		; Error #
		lea	edx, [eax.word_name]			; Word name
		clr	ecx					; No line #
		call	doError
		
@@NotRT:

		
		; Try to compile a word not allowed to be compiled ?
		; --------------------------------------------------
		
		test 	byte ptr [eax.word_flags], FNOCOMPILE
		jz	@@NotNoToComp
		
		; "This system word can't be compiled inside user definitions."
		mov	ah, ERR_CANTBECOMPILED			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@NotNoToComp:

		; ------------------
		; Compile word found
		; ------------------

		; Update word cross ref
		; ---------------------

		; Save used registers
		push	ecx edx

		; Point to the last word defined in the current vocabulary: word currently being defined
		mov	ecx, [VOC_CURRENT]
		add	ecx, [VOC_START]
		mov	ecx, [ecx.voc_lastword]			; NIA pointer to caller word
		
		; The word found NIA address (relative to NIA_START) is on top stack
		DISCARD	1
		mov	edx, ebx				; NIA pointer to called word
		
if TURNKEYED eq 0
		; Update word cross ref, only if it's not a system word
		mov	eax, edx
		add	eax, [NIA_START]
		test	[eax.word_flags], FSYSTEM
		jnz	@@System
		call	CRef_add
@@System:
endif
		
		; Restore used registers
		pop	edx ecx


		; ---------------
		; Is-it a macro ?
		; ---------------

		call	QSPEED					; ?speed
		call	BOOL_QTRUE				; true?		
		lodsd
		cmpz	edx, FALSE
		je	@@PutCallWord
ifdef THIS_IS_BUGGED
		test 	byte ptr [eax.word_flags], FMACRO
		jz	@@PutCallWord
		mov	edx, eax
		call	_PutMacro
		jmp	@@NextWord
endif


		; Put 'CALL rel_32' code operation
		; --------------------------------

@@PutCallWord:	CALL_rel32					; CALL rel32

		add 	ebx, [NIA_START]
		mov	eax, [ebx.word_realaddr]
		mov	[LastWrdCompiled], eax			; Address of the last waord compiled
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 for REL32
		call	CodeComma               		; Code,

		; Reset ALLOC and FREE found flag (while compiling)
		mov	[AllocFound], FALSE
		mov	[FreeFound], FALSE
		
		; Now parse next word		
		jmp	@@NextWord


		; ------------------
		; Execute word found
		; ------------------

@@ExecWord:	
		; Try to execute a run-time not immediate word ?
		test	[eax.word_flags], FRUNTIME		; Is word run-time ?
		jz	@@ExecWordOk
		test 	[eax.word_flags], FIMMEDIATE  		; Is word immediate ?
		jnz	@@ExecWordOk
		
		; "Run time words can't be executed in execute state."
		lea	edx, [eax.word_name]			; Additionnal msg
		mov	ah, ERR_RTWCANTBEEXECUTED		; Error #
		clr	ecx					; No line #
		call	doError
		
		
@@ExecWordOk:	mov	eax, PARAM0
		DISCARD	1
		call	eax
Execute:							; INPORTANT!!! See 'doError': stop searching into stack !
		jmp	@@NextWordTst


		; ------------------------------------
		; Word not found. Is-it a local word ?
		; ------------------------------------
		
@@IsItLocalWrd:	

		; But before, it's an error if ALLOC or FREE were called before this word
		; -----------------------------------------------------------------------
		
		cmp	[AllocFound], TRUE
		je  	@@WnfNan
		cmp	[FreeFound], TRUE
		je  	@@WnfNan


		; Search the word into local words table
		call	AutoFind
		jc	@@IsItANb				; If word is not found

		; Word is found into local words table
		; ------------------------------------

		; Put 'CALL_rel32 LocalVar'
		compCALL_rel32	LocalVar
		
		; Store block nb addr cell relative addr
		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		sub	eax, [ecx.lword_blocknbad]		; Block nb address (address after the 'CALL LocalVar')
		neg	eax
		call	CodeComma               		; Code,

		; Store run time number addr into next cell
		CALL_rel32
		mov	eax, [ecx.lword_rtaddr]
		add	eax, [NIA_START]
		mov	eax, [eax]
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 is for REL32
		call	CodeComma               		; Code,
		
		; Continue compilation and analyse next word
		jmp	@@NextWord


		; Word not found. Is-it a valid number ?
		; -------------------------------------
		
@@IsItANb:	sub	esi, @WordSize
		mov	[esi], edx
		call	NUMBER
		mov 	eax, [esi]
		add 	esi, @WordSize
		cmpz 	eax, 0
		jne	@@ItIsANb
		
	
		; Not a number (in the current base). Is-it an hexadecimal number (0x...) ?
		; -------------------------------------------------------------------------
		
		cmp	word ptr [edx], 'x0'
		jne	@@NotHexNb
		sub	esi, @WordSize
		call	_BASE					; BASE
		DO_AT						; @
		call	TOR					; >R
		call	HEXADECIMAL				; HEXADECIMAL
		mov	[esi], edx
		add	dword ptr [esi], 2			; Skip '0x'
		call	NUMBER					; Try to convert from ASCII to BINARY
		call	FROMR					; R>
		call	_BASE					; BASE
		DO_STORE					; !
		lodsd
		cmpz 	eax, 0
		jne	@@ItIsANb
@@NotHexNb:

		
		; Not a word, not a nb: is-it a vocabulary ?
		; ------------------------------------------
		
		sub	esi, 4
		mov	[esi], edx
		call	VocFind
		mov 	eax, [esi]
		add 	esi, 4
		cmpz 	eax, 0
		jne	@@IsVoc

		
		; Is-it a constant character ?
		; ----------------------------

		cmp	byte ptr [edx], 027h			; Quote ?	'
		jne	@@NoConstChar
		mov 	eax, [IBUFFPTR2]           		; Save pointer to end word
		sub	eax, edx				; AX is the length of the word
		cmp	ax, 5
		jg	@@NoConstChar
		
		; One char ?
		cmp	ax, 2
		jne	@@No1Char
		movzx	eax, byte ptr [edx + 1]
@@PushConstCh:	sub	esi, 4
		mov	[esi], eax
		jmp	@@ItIsANb
		
		; Two chars ?
@@No1Char:	cmp	ax, 3
		jne	@@No2Chars
		clr	eax
		mov	ah, [edx + 1]
		mov	al, [edx + 2]
		jmp	@@PushConstCh
		
		; Three chars ?
@@No2Chars:	cmp	ax, 4
		jne	@@No3Chars
		clr	eax
		mov	ah, [edx + 1]
		shl	eax, 8
		mov	ah, [edx + 2]
		mov	al, [edx + 3]
		jmp	@@PushConstCh
		
		; Four chars ?
@@No3Chars:	cmp	ax, 5
		jne	@@NoConstChar
		clr	eax
		mov	ah, [edx + 1]
		mov	al, [edx + 2]
		shl	eax, 16
		mov	ah, [edx + 3]
		mov	al, [edx + 4]
		jmp	@@PushConstCh
		
@@NoConstChar:


		; ------------------------
		; Nothing: it's an error !
		; ------------------------

		jmp  	@@WnfNan
		

		; It's a vocabulary : change context vocabulary
		; ---------------------------------------------
		
@@IsVoc:	lodsd
		mov	[VOC_CONTEXT], eax
		jmp	@@NextWord
		

		; It's a number: in execute state, push it on stack, in compile mode, compile number
		; ----------------------------------------------------------------------------------
		
@@ItIsANb:	cmp	[CompileMode], TRUE
		je	@@CompilNb


		; ---------------------------------------------------------------------------
		; Execute state: push the number onto the stack (it's already done by NUMBER)
		; ---------------------------------------------------------------------------
		
		jmp	@@NextWordTst


		; ----------------------------------------------
		; Compile state: 'compile' the number  ( n --- )
		; ----------------------------------------------
		
@@CompilNb:	call	_Compile_inb
		jmp	@@NextWord


		; --------------------------------
		; Word not found and not a number
		; --------------------------------

@@WnfNan:	mov	ah, ERR_WNFNAN				; Code # error
		clr	ecx					; No line #
		call	doError


		; ============================================
		; Found next word into the input stream buffer
		; ============================================

@@NextWordTst:
		; Test primary stack overflow
		ifdef TESTSTACK
		call	OStack
		endif
		
@@NextWord:	mov	edx, [IBUFFPTR2]
		inc	edx
		jmp	@@GtoFirstNB

		
		align	4
@@wordname	db	WORDMAXLEN+1 dup (?)
		align	4
@@wordnametoS	db	WORDMAXLEN+1 dup (?)
		align	4
@@CharMethPos	dd	?
@@LdaAddr	dd	?
@@edx		dd	?

E_INTERPRET:
INTERPRET	endp


; ===============================================================
; = QERROR							=
; =                                                 		=
; = Code for the forth word:	"ERROR				=
; =                                                 		=
; = Stack:	( addrz --- )					=
; ===============================================================

		align	4
QERROR		proc

		; Verify stack frame
		VSTACK	1

		; Do the given message
		mov	ah, -1					; NO error #
		mov	edx, PARAM0				; Additionnal msg is the given msg
		DISCARD	1					; DROP
		clr	ecx					; No line #
		jmp	doError

E_QERROR:
QERROR		endp


; ===============================================================
; = RT_WORDNAME							=
; =                                                 		=
; = Code for the forth word:	RT_WordName$			=
; =                                                 		=
; = Stack:	( --- addrz )					=
; ===============================================================

MRT_WORDNAME	MACRO	depl

		align	4

		if	depl eq 0
RT_WORDNAME_0	proc
		elseif	depl eq 1
RT_WORDNAME_1	proc
		endif

		; Verify stack frame overflow
		DISPOSE	1
		OVSTACK
		mov	PARAM0, -1				; In case of error...

		; Pointer into the caller's space
		mov	edx, [esp + depl * @WordSize]

		; Scan all words
		clr	ebx
		mov	ecx, [NIA_START]
@@Scan:		cmp	byte ptr [ecx.word_name], 0
		je	@@WordEmpty
		
		; Word address is this one ?
		cmp	edx, [ecx.word_realaddr]
		jb	@@NotThisWord
		cmp	edx, [ecx.word_EndCode]
		jbe	@@ThisWord
		
@@NotThisWord:
@@WordEmpty:	add	ecx, size wordentry
		add	ebx, size wordentry
		cmp	ebx, [NIA_SIZE]				; !!!
		jl	@@Scan
		jmp	@@Exit					; Error...

		; Word is found
@@ThisWord:	add	ecx, word_name

		; Copy string and compute length
		lea	edx, WordNameBuff
		clr	ebx
		dec	ebx
@@CopyWordName:	inc	ebx
		mov	al, [ecx + ebx]
		mov	[edx + ebx + size strentry], al
		cmpz	al, 0
		jne	@@CopyWordName

		mov	[edx.str_currlen], ebx	 	; Current length
		mov	[edx.str_currsize], ebx		; Current size allocated
		mov	[edx.str_memblknb], -1		; NO Memory block nb (ref to DDIR)
		mov	[edx.str_incsize], 0		; Increment size (if dynamic string), else 0 (if fixed string)
		mov	[edx.str_wordref], 0		; Reference to the definition (NIA)

		add	edx, size strentry
		mov	PARAM0, edx

@@Exit:
		
		; Return to caller
		ret
		
		if	depl eq 0
E_RT_WORDNAME_0:
		elseif	depl eq 1
E_RT_WORDNAME_1:
		endif

		endp
		endm


; ===============================================================
; = WORDNAME							=
; =                                                 		=
; = Code for the forth word:	WordName$			=
; =                                                 		=
; = Stack:	( --- addrz )					=
; ===============================================================

MRT_WORDNAME	0
MRT_WORDNAME	1

		align	4
WORDNAME	proc

		; Verify if the system is in compile state
		call	QCOMPILE

		; Compile the call		
		cmp	[CDA_DOES_CALLED], TRUE				; Is 'DOES>' called ?
		je	@@Does

		compCALL_rel32	RT_WORDNAME_0
		jmp	@@Exit
				
@@Does:		compCALL_rel32	RT_WORDNAME_1
		jmp	@@Exit
				
		; Return to caller
@@Exit:		ret
		
E_WORDNAME:
WORDNAME	endp


code32		ends
		end
