; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	KBD.ASM							!
; 									!
; Description:	Definitions for playing with the keyboard.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	KEY		KEY						!
;	QKEY		?KEY						!
;	QALTKEY		?ALTKEY						!
;	QCTRLKEY	?CTRLKEY					!
;	QSHIFTKEY	?SHIFTKEY					!
;	KEYINDICE	KEYINDICE					!
;	QKEYSTATE	?KEYSTATE					!
;	KBDFILE		KBDFILE						!
; 									!
; ----------------------------------------------------------------------+

		.486p
		jumps
		locals
		smart

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

		align	4

include		pmode.inc					; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; ===============================================================
; = KbItHandler							=
; =								=
; = Usage:							=
; =	The keyboard interrupt handle (it #1)			=
; ===============================================================

		align	4
KbItHandler	proc

		; --------------------
		; Saved used registers
		; --------------------
		
		push	eax ebx ecx edx ds


		; ----------------
		; Set data segment
		; ----------------
		
		mov	ds, cs:_seldata

	        
		; -------------------------
		; Read keyboard data/status
		; -------------------------

		; Read data from keyboard	        
		in	al, 60h
        
		; 2nd char of two sequences chars ?
		cmp	[TwoSeq], 0
		je	@@IsTwoSeq
		mov	[TwoSeq], 0
		mov	ah, 0E0h
		jmp	@@SeqDone
	        
	        	; Is-it a Two-Sequence key (prefix is $E0) ?
@@IsTwoSeq:	cmp	al, 0E0h
		jne	@@OneSeq

		; 1st char of two sequences chars ?
		mov	[TwoSeq], 1
		jmp	@@Done
		
		; One sequence only
@@OneSeq:	clr	ah					; AH <- 0

@@SeqDone:	


		; ------------------------------------------
		; Is-it a depressed key ? or a pressed key ?
		; ------------------------------------------
		
		lea	edx, KbdTable
		clr	ecx					; This will be the key indice
		btr	ax, 7
		jnc	@@NoDepressed


		; ------------------
		; A key is depressed
		; ------------------

		; Skip Ctrl-Break
		add	edx, size kbentry
		inc	ecx
		
		; Is-it left SHIFT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@XNotShiftL
		and	[ShiftState], 0FEh			; Clear bit #0
		jmp	@@Done
@@XNotShiftL:	add	edx, size kbentry
		inc	ecx

		; Is-it right SHIFT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@XNotShiftR
		and	[ShiftState], 0FDh			; Clear bit #1
		jmp	@@Done
@@XNotShiftR:	add	edx, size kbentry
		inc	ecx

		; Is-it left CTRL key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@XNotCtrlL
		and	[CtrlState], 0FEh			; Clear bit #0
		jmp	@@Done
@@XNotCtrlL:	add	edx, size kbentry
		inc	ecx

		; Is-it right CTRL key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@XNotCtrlR
		and	[CtrlState], 0FDh			; Clear bit #1
		jmp	@@Done
@@XNotCtrlR:	add	edx, size kbentry
		inc	ecx

		; Is-it left ALT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@XNotAltL
		and	[AltState], 0FEh			; Clear bit #0
		jmp	@@Done
@@XNotAltL:	add	edx, size kbentry
		inc	ecx

		; Is-it right ALT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@XNotAltR
		and	[AltState], 0FDh			; Clear bit #1
		jmp	@@Done
@@XNotAltR:	add	edx, size kbentry
		inc	ecx

		; Find scan code into the table
@@FindScan0:	cmp	ax, [edx.kb_ScanCode]
		je	@@ScanFound0
		add	edx, size kbentry
		inc	ecx
		cmp	[edx.kb_ScanCode], 0
		jne	@@FindScan0

		; Scan code not found
		jmp	@@Done

		; Update each key state (held or not held)
@@ScanFound0:	mov	byte ptr [ScanStates + ecx], 0

		; Done
		jmp	@@Done
		
		; 
		; ----------------
		; A key is pressed
		; ----------------
		
@@NoDepressed:

		; Is-it Ctrl-BREAK ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@NoCtrlBreak

		; Running a turnkey .exe ?
		cmp	[ExeFile], TRUE
		je	@@DoNotCtrlBreak

		; Depending the state of 'NoCtrlBreak'
		cmp	[NoCtrlBreak], 0
		je	@@DoCtrlBreak
@@DoNotCtrlBreak: mov	[CtrlBreakEvent], TRUE
		jmp	@@Done					; Not allowed
@@DoCtrlBreak:
			
		; Re-initialize the FIFO: discard any character waiting
		mov	[FIFO_CharsNb], 0				; Nb of chars available into the FIFO
		mov	[FIFO_head], 0				; Head pointer into FIFO
		mov	[FIFO_tail], 0				; Tail pointer into FIFO

		; Put 'return' address
		mov	al, 20h
		out	20h, al					; Validate IT
		db	01Fh					; pop ds
		pop	edx ecx ebx eax				; Clean stack
		push	ds
		mov	ds, cs:_seldata
		cmp	[_sysbyte0], 3				; DPMI 32-bits ?
		db	01Fh					; pop ds
		jne	@@NotDPMI
		iret
@@NotDPMI:	pop	eax					; Procedure stopped address
		add	esp, 8					; pop values push by interruption
		push	eax
		jmp	CtrlBreakRecov

@@NoCtrlBreak:	add	edx, size kbentry
		inc	ecx

		; Is-it left SHIFT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@NotShiftL
		or	[ShiftState], 01h
		jmp	@@Done
@@NotShiftL:	add	edx, size kbentry
		inc	ecx

		; Is-it right SHIFT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@NotShiftR
		or	[ShiftState], 02h
		jmp	@@Done
@@NotShiftR:	add	edx, size kbentry
		inc	ecx

		; Is-it left CTRL key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@NotCtrlL
		or	[CtrlState], 01h
		jmp	@@Done
@@NotCtrlL:	add	edx, size kbentry
		inc	ecx

		; Is-it right CTRL key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@NotCtrlR
		or	[CtrlState], 02h
		jmp	@@Done
@@NotCtrlR:	add	edx, size kbentry
		inc	ecx

		; Is-it left ALT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@NotAltL
		or	[AltState], 01h
		jmp	@@Done
@@NotAltL:	add	edx, size kbentry
		inc	ecx

		; Is-it right ALT key ?
		cmp	ax, [edx.kb_ScanCode]
		jne	@@NotAltR
		or	[AltState], 02h
		jmp	@@Done
@@NotAltR:	add	edx, size kbentry
		inc	ecx

		; Find scan code into the table
@@FindScan:	cmp	ax, [edx.kb_ScanCode]
		je	@@ScanFound
		add	edx, size kbentry
		inc	ecx
		cmp	[edx.kb_ScanCode], 0
		jne	@@FindScan

		; Scan code not found
		jmp	@@Done

@@ScanFound:

		; Update each key state (held or not held)
		mov	byte ptr [ScanStates + ecx], 1

		; Is-it a shift-ed key ?
		test	[ShiftState], 03h
		jz	@@NoShift
		mov	bx, [edx.kb_ShiftCode]
		cmp	bx, -1
		jne	@@StoreChar
		cmp	ah, 0E0h
		je	@@ShiftTwoSeq
		mov	bx, ax
		mov	bh, SHIFTKEYS
		jmp	@@StoreChar
@@ShiftTwoSeq:	mov	bx, ax
		mov	bh, SHIFTKEYS
		add	bl, 80h
		jmp	@@StoreChar
@@NoShift:

		; Is-it a ctrl-ed key ?
		test	[CtrlState], 03h
		jz	@@NoCtrl
		mov	bx, [edx.kb_CtrlCode]
		cmp	bx, -1
		jne	@@StoreChar
		cmp	ah, 0E0h
		je	@@CtrlTwoSeq
		mov	bx, ax
		mov	bh, CTRLKEYS
		jmp	@@StoreChar
@@CtrlTwoSeq:	mov	bx, ax
		mov	bh, CTRLKEYS
		add	bl, 80h
		jmp	@@StoreChar
@@NoCtrl:

		; Is-it an alt-ed key ?
		test	[AltState], 03h
		jz	@@NoAlt
		mov	bx, [edx.kb_AltCode]
		cmp	bx, -1
		jne	@@StoreChar
		cmp	ah, 0E0h
		je	@@AltTwoSeq
		mov	bx, ax
		mov	bh, ALTKEYS
		jmp	@@StoreChar
@@AltTwoSeq:	mov	bx, ax
		mov	bh, ALTKEYS
		add	bl, 80h
		jmp	@@StoreChar
@@NoAlt:

		; A 'normal' char: no shift, not ctrl, no alt
		mov	bx, [edx.kb_AsciiCode]
		cmp	bx, -1
		jne	@@StoreChar
		cmp	ah, 0E0h
		je	@@NormTwoSeq
		mov	bx, ax
		mov	bh, EXTKEYS
		jmp	@@StoreChar
@@NormTwoSeq:	mov	bx, ax
		mov	bh, EXTKEYS
		add	bl, 80h

		; Store the character in FIFO
@@StoreChar:	

		; FIFO full ?
		cmp	[FIFO_CharsNb], KB_FIFOSZ
		je	@@Done					; Discard char if FIFO is full
		
		; Store char into FIFO and update head pointer
		mov	eax, [FIFO_head]
		mov	[FIFO + eax * 2], bx
		inc	eax
		cmp	ax, KB_FIFOSZ
		jne	@@FifoOk
		clr	eax
@@FifoOk:	mov	[FIFO_head], eax

		; An other char is available
		inc	[FIFO_CharsNb]
		
		
		; ------------------
		; Validate interrupt
		; ------------------
		
@@Done:		mov	al, 20h
		out	20h, al


		; ----------------------
		; Restore used registers
		; ----------------------
		
		db	01Fh					; pop ds
		pop	edx ecx ebx eax


		; ---------------------
		; Return form interrupt
		; ---------------------

		sti
		iretd

KbItHandler	endp


; ===============================================================
; = InstallKbIt							=
; =								=
; = Usage:							=
; =	Install the keyboard interrupt handle (it #1)		=
; =								=
; = I) none							=
; =								=
; = O) none							=
; =								=
; = M) all							=
; ===============================================================

		align	4
InstallKbIt	proc

		; Initialize the FIFO
		mov	[FIFO_CharsNb], 0				; Nb of chars available into the FIFO
		mov	[FIFO_head], 0				; Head pointer into FIFO
		mov	[FIFO_tail], 0				; Tail pointer into FIFO

		; Some initializations
		mov	[TwoSeq], 0				; $E0 sequence ?
		mov	[ShiftState], 0				; Flag (0 | 1 | 2): is a key SHIFT held ?
		mov	[CtrlState], 0				; Flag (0 | 1 | 2): is a key CTRL held ?
		mov	[AltState], 0				; Flag (0 | 1 | 2): is a key ALT held ?

		; Initialize the state of each key (scan code is the indice of this array)
		lea	edx, ScanStates
		clr	ecx
@@ZScanStates:	mov	byte ptr [edx + ecx], 0
		inc	cl
		cmp	cl, KB_MAXNBRECORDS
		jne	@@ZScanStates

		; Save previous vector (protected mode)
		mov	bl, 1					; Irq #1 is keyboard
		lea	eax, _getirqvect
		call	[eax]
		mov	[OldProtMIrq1], edx

		; Store new vector
		lea	edx, KbItHandler
		add	edx, [OfsDictReloc]
		lea	eax, _setirqvect
		call	[eax]
	
		; Save previous vector (real mode)
		push	edi
		lea	edi, kbrmirqbuff
		lea	ecx, _rmpmirqset
		call	ecx
		mov	[OldRealMIrq1], eax
		pop	edi
		
		; Is own handler installed ?
		mov	[KbdItInstalled], TRUE

		; Return to caller
		ret

InstallKbIt	endp


; ===============================================================
; = DesinstallKbIt						=
; =								=
; = Usage:							=
; =	Desinstall the keyboard interrupt handle (it #1)	=
; =								=
; = I) none							=
; =								=
; = O) none							=
; =								=
; = M) all							=
; ===============================================================

		align	4
DesinstallKbIt	proc

		; Is own handler installed ?
		cmp	[KbdItInstalled], FALSE
		je	@@Exit

		; Real-Protected interrupt
		mov	bl, 1					; Irq #1 is keyboard
		mov	eax, [OldRealMIrq1]
		lea	ecx, _rmpmirqfree
		call	ecx
		
		; Protected mode vector
		mov	edx, [OldProtMIrq1]
		lea	eax, _setirqvect
		call	[eax]

		; Is own handler installed ?
		mov	[KbdItInstalled], FALSE

		; Return to caller
@@Exit:		ret

DesinstallKbIt	endp


; ===============================================================
; = CharIn							=
; =								=
; = Usage:							=
; =	Get a char. Wait until a char is available.		=
; =								=
; = I) none							=
; =								=
; = O) AX : ascii code						=
; =								=
; = M) EAX FLAGS						=
; ===============================================================

if UNDERDEV
debug_charin	proc
@@Wait:		cmp	[FIFO_CharsNb], 0
		je	@@Wait
		push	ebx
		cli
		mov	ebx, [FIFO_tail]
		movzx	eax, [FIFO + ebx * 2]			; Get char
		inc	ebx
		cmp	bx, KB_FIFOSZ
		jne	@@Ok
		clr	ebx
@@Ok:		mov	[FIFO_tail], ebx
		dec	[FIFO_CharsNb]
		sti
		pop	ebx
		ret
debug_charin	endp
endif		
		
		align	4
CharIn		proc

		; Wait until a char is available
@@Wait:		cmp	[FIFO_CharsNb], 0
		je	@@Wait

		; Push value on stack
		push	ebx
		cli
		mov	ebx, [FIFO_tail]
		movzx	eax, [FIFO + ebx * 2]			; Get char
		inc	ebx
		cmp	bx, KB_FIFOSZ
		jne	@@Ok
		clr	ebx
@@Ok:		mov	[FIFO_tail], ebx
		dec	[FIFO_CharsNb]
		sti
		pop	ebx

		; Return to caller
		ret

E_CharIn:
CharIn		endp


; ===============================================================
; = KEY								=
; =								=
; = Code for the forth word:	KEY				=
; =								=
; = Stack:	( --- c )					=
; ===============================================================

		align	4
KEY		proc

		; Wait until a char is available
@@Wait:		cmp	[FIFO_CharsNb], 0
		jne	@@Available
		cmp	[ActiveThreadsNb], 1
		je	@@Wait
		call	YIELD					; Yield to another active thread
		jmp	@@Wait
@@Available:

		; Push value on stack
		push	ebx
		cli
		mov	ebx, [FIFO_tail]
		movzx	eax, [FIFO + ebx * 2]			; Get char
		inc	ebx
		cmp	bx, KB_FIFOSZ
		jne	@@Ok
		clr	ebx
@@Ok:		mov	[FIFO_tail], ebx
		dec	[FIFO_CharsNb]
		sti
		pop	ebx

		; Push character on stack
		IPUSH	eax

		; Return to caller		
		ret
E_KEY:
KEY		endp


; ===============================================================
; = QKEY							=
; =								=
; = Code for the forth word:	?KEY				=
; =								=
; = Stack:	( --- ff ) if no char				=
; =		( --- c tf ) if char				=
; ===============================================================

		align	4
QKEY		proc

		; Is a char avaiable ?
		DISPOSE	1
		cmp	[FIFO_CharsNb], 0			; Nb of chars available into the FIFO
		jne	@@Char

		; No char
		mov	PARAM0, FALSE				; push ff
		jmp	@@exit
		
		; At least a char is available
@@Char:		mov	PARAM0, TRUE				; push tf

		; Verify stack frame overflow, then return to caller
@@exit:		OVSTACK
		ret

E_QKEY:
QKEY		endp


; ===============================================================
; = QALTKEY							=
; =								=
; = Code for the froth word: ?AltKey				=
; =								=
; = Stack: ( --- f )						=
; ===============================================================

		align	4
QALTKEY		proc

		; Push key state
		movzx	eax, [AltState]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK
		ret

E_QALTKEY:
QALTKEY		endp


; ===============================================================
; = QCTRLKEY							=
; =								=
; = Code for the froth word: ?CtrlKey				=
; =								=
; = Stack: ( --- f )						=
; ===============================================================

		align	4
QCTRLKEY	proc

		; Push key state
		movzx	eax, [CtrlState]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK
		ret

E_QCTRLKEY:
QCTRLKEY	endp


; ===============================================================
; = QSHIFTKEY							=
; =								=
; = Code for the froth word: ?ShiftKey				=
; =								=
; = Stack: ( --- f )						=
; ===============================================================

		align	4
QSHIFTKEY	proc

		; Push key state
		movzx	eax, [ShiftState]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK
		ret

E_QSHIFTKEY:
QSHIFTKEY	endp


; ===============================================================
; = isShiftKey							=
; =								=
; = Usage:							=
; =	Tells (to editor) if a SHIFT key is down.		=
; =								=
; = I) none							=
; =								=
; = O) CF = 1, yes a SHIFT key is down.				=
; =    CF = 0, no						=
; =								=
; = M) flags.							=
; ===============================================================

		align	4
isShiftKey	proc

		; Query key state
		cmp	[ShiftState], 0
		je	@@NotDown

@@Down:		stc
		jmp	@@Exit

@@NotDown:	clc

		; return
@@Exit:		ret

isShiftKey	endp


; ===============================================================
; = KEYINDICE							=
; =								=
; = Code for the froth word: KeyIndice				=
; =								=
; = Stack: ( c --- n )						=
; ===============================================================

		align	4
KEYINDICE	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; --------------------------------
		; Search key in the keyboard table
		; --------------------------------

		; Searching
		lea	edx, KbdTable
		mov	eax, [esi]
		clr	ecx

@@FindKey:	

		; Ascii code
		cmp	ax, 100h
		jae	@@NotAscii
		cmp	ax, [edx.kb_AsciiCode]
		je	@@Found					; Ascii 
		cmp	ax, [edx.kb_ShiftCode]		
		je	@@Found					; Shift
		cmp	ax, [edx.kb_CtrlCode]		
		je	@@Found					; Ctrl
		cmp	ax, [edx.kb_AltCode]		
		je	@@Found					; Alt
		jmp	@@Next
@@NotAscii:

		; Extended key
		cmp	ah, EXTKEYS
		jne	@@NotExtend
		cmp	al, 80h
		jae	@@Extend2
		mov	bx, ax
		clr	bh
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@Extend2:	mov	bh, 0E0h
		mov	bl, al
		sub	bl, 80h
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@NotExtend:
		
		; Shift-ed key
		cmp	ah, SHIFTKEYS
		jne	@@NotShift
		cmp	[edx.kb_ShiftCode], -1
		jne	@@Next
		cmp	al, 80h
		jae	@@Shift2
		mov	bx, ax
		clr	bh
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@Shift2:	mov	bh, 0E0h
		mov	bl, al
		sub	bl, 80h
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@NotShift:
		
		; Ctrl-ed key
		cmp	ah, CTRLKEYS
		jne	@@NotCtrl
		cmp	[edx.kb_CtrlCode], -1
		jne	@@Next
		cmp	al, 80h
		jae	@@Ctrl2
		mov	bx, ax
		clr	bh
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@Ctrl2:	mov	bh, 0E0h
		mov	bl, al
		sub	bl, 80h
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@NotCtrl:
		
		; Alt-ed key
		cmp	ah, ALTKEYS
		jne	@@NotAlt
		cmp	[edx.kb_AltCode], -1
		jne	@@Next
		cmp	al, 80h
		jae	@@Alt2
		mov	bx, ax
		clr	bh
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@Alt2:		mov	bh, 0E0h
		mov	bl, al
		sub	bl, 80h
		cmp	bx, [edx.kb_ScanCode]
		jne	@@Next
		je	@@Found
@@NotAlt:
		
		; Next
@@Next:		inc	ecx
		add	edx, size kbentry
		cmp	[edx.kb_ScanCode], 0
		jne	@@FindKey

		; Code not found
		mov	dword ptr [esi], -1
		jmp	@@Ret

		; Code found
@@Found:	mov	dword ptr [esi], ecx		


		; ------
		; Return
		; ------
		
@@Ret:		ret

E_KEYINDICE:
KEYINDICE	endp


; ===============================================================
; = QKEYSTATE							=
; =								=
; = Code for the froth word: ?KeyState				=
; =								=
; = Stack: ( n --- f )						=
; ===============================================================

		align	4
QKEYSTATE	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ----------
		; Push state
		; ----------
		
		mov	eax, PARAM0
		bound	eax, qword ptr @@limits			; Check value against bound
		mov	al, [ScanStates + eax]
		mov	PARAM0, eax
		

		; ------
		; Return
		; ------
		
@@Ret:		ret

@@limits	dd	0
		dd	KB_MAXNBRECORDS-1			; Max nb of keys on the keyboard
E_QKEYSTATE:
QKEYSTATE	endp


; ===============================================================
; = KBDFILE							=
; =								=
; = Code for the forth word:	KbdFile				=
; =								=
; = Stack:  ( addrz --- )					=
; ===============================================================

		align	4
KBDFILE		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1

		; Verify if a valid string
		call	QSTRING					; ?String


KBDFILE0:							; Enter for cold start
		; ---------------------------------
		; Try to open the given file in R/O
		; ---------------------------------

		; Call BIOS to open in R/O		
		mov	ebx, [esi]				; Pop filename addr
		add	esi, @WordSize
		mov	ax, 10h                			; 0 001 0000 -> (inherited, deny both, read)
		call	doOpenFile
		jc	QuoteLoadErr


		; No error on opening in R/O, now read the file header into buffer
		; ----------------------------------------------------------------

		mov	[@@FileHdl], bx				; Save file handle
		mov	ecx, 2         				; Bytes to read
		lea	edx, @@Header				; Buffer address
		call	doReadFile                    		; in EBX: bytes read from the file
		jc	QuoteLoadErr
		cmp	bl, 2
		je	@@ReadOk

		
		; Error on reading
		; ----------------

@@ReadNOk:	mov	ah, ERR_BADKBFRMT			; Bad keyboard file format
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

	
		; ---------------------------------
		; Read the keyboard table in memory
		; ---------------------------------

		; Read table
@@ReadOk:	mov	bx, [@@FileHdl]				; Get file handle
		movzx	ecx, [@@Header]				; Nb of records
		imul	ecx, size kbentry				; Bytes to read
		lea	edx, KbdTable				; Buffer address
		call	doReadFile                    		; in EBX: bytes read from the file
		jc	QuoteLoadErr
		cmp	ebx, ecx
		jne	@@ReadNOk
		
		; Add an empty record at the end
		movzx	ecx, [@@Header]				; Nb of records
		imul	ecx, size kbentry
		lea	edx, KbdTable
		add	edx, ecx
		mov	[edx.kb_ScanCode], 0			; 0 means end of table
		
		
		; --------------
		; Close the file
		; --------------

		mov	bx, [@@FileHdl]				; Get file handle
		call	doCloseFile
		jc	QuoteLoadErr


		; --------------------------------------
		; Update the index for all function keys
		; --------------------------------------
		
		call	UpdateIndexKeys


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
; ---> Local variables
		align	4
@@FileHdl	dw	?				; File handle
		align	4
@@Header	dw	?				; Header: Nb of records
		
E_KBDFILE:
KBDFILE		endp


code32		ends
		end

