; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	KEYS.ASM						!
; 									!
; Description:	Definitions for playing with serial I/O.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	SHIFT_ED	SHIFT-ED					!
;	CTRL_ED		CTRL-ED						!
;	ALT_ED		ALT-ED						!
;	KEYPAD-ED	KEYPAD-ED					!
;	KEY_...		KEY_INS						!
; 	....		....						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Ed.inc                      		; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = UpdateIndexKeys						=
; =								=
; = Usage:							=
; =	Update Ascii/Shift/Ctrl/Alt keys.         		=
; =								=
; = I) ECX							=
; = O) none							=
; = M) none							=
; ===============================================================

		align	4
UpdateKeyTable	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx
		mov	edx, ecx
		
		
		; -------------------------------------------
		; A 'normal' char: no shift, not ctrl, no alt
		; -------------------------------------------
		
		mov	bx, [edx.kb_AsciiCode]
		cmp	bx, -1
		jne	@@NormDone
		mov	ax, [edx.kb_ScanCode]
		cmp	ah, 0E0h
		je	@@NormTwoSeq
		mov	bx, ax
		mov	bh, EXTKEYS
		jmp	@@NormDone
@@NormTwoSeq:	mov	bx, ax
		mov	bh, EXTKEYS
		add	bl, 80h
@@NormDone:	mov	[edx.kb_AsciiCode], bx
		

		; ----------------
		; A 'shifted' char
		; ----------------
		
		mov	bx, [edx.kb_ShiftCode]
		cmp	bx, -1
		jne	@@ShiftDone
		mov	ax, [edx.kb_ScanCode]
		cmp	ah, 0E0h
		je	@@NormTwoSeq2
		mov	bx, ax
		mov	bh, SHIFTKEYS
		jmp	@@ShiftDone
@@NormTwoSeq2:	mov	bx, ax
		mov	bh, SHIFTKEYS
		add	bl, 80h
@@ShiftDone:	mov	[edx.kb_ShiftCode], bx
		

		; -----------------
		; A 'ctrl-ed' char
		; -----------------
		
		mov	bx, [edx.kb_CtrlCode]
		cmp	bx, -1
		jne	@@CtrlDone
		mov	ax, [edx.kb_ScanCode]
		cmp	ah, 0E0h
		je	@@NormTwoSeq3
		mov	bx, ax
		mov	bh, CTRLKEYS
		jmp	@@CtrlDone
@@NormTwoSeq3:	mov	bx, ax
		mov	bh, CTRLKEYS
		add	bl, 80h
@@CtrlDone:	mov	[edx.kb_CtrlCode], bx
		

		; ---------------
		; A 'Alt-ed' char
		; ---------------
		
		mov	bx, [edx.kb_AltCode]
		cmp	bx, -1
		jne	@@AltDone
		mov	ax, [edx.kb_ScanCode]
		cmp	ah, 0E0h
		je	@@NormTwoSeq4
		mov	bx, ax
		mov	bh, ALTKEYS
		jmp	@@AltDone
@@NormTwoSeq4:	mov	bx, ax
		mov	bh, ALTKEYS
		add	bl, 80h
@@AltDone:	mov	[edx.kb_AltCode], bx
		

		; ----------------
		; Return to caller
		; ----------------
		
		pop	edx ecx ebx eax
		ret

UpdateKeyTable	endp


; ===============================================================
; = UpdateIndexKeys						=
; =								=
; = Usage:							=
; =	Update the index for all function keys.			=
; =								=
; = I) none							=
; = O) none							=
; = M) All regs.						=
; ===============================================================

		align	4
UpdateIndexKeys	proc

		; ------------------------------
		; Scan all defined function keys
		; ------------------------------
		
		lea	edx, IK_INS
		
		; Search the symbolic name into kbd file
@@Scan:		add	edx, 4					; Pointer to symb name to search
		lea	ecx, KbdTable
@@Scan5:		clr	ebx					; Char counter
@@Scan2:		mov	al, [edx + ebx]
		mov	ah, [ecx.kb_symbname + ebx]
		cmp	al, ah
		jne	@@Scan3
		or	al, ah
		jz	@@Scan4
		inc	ebx
		jmp	@@Scan2
		
@@Scan3:		add	ecx, size kbentry
		cmp	[ecx.kb_ScanCode], 0			; 0 means end of table
		jne	@@Scan5
		call	EdPDotQ
		db	"Invalid key file, at ", 0
		mov	eax, edx
		call	PrEAX
		call	DesinstallKbIt				; If previously installed
		jmp	_exit
		
		
		; i-key found
@@Scan4:		call	UpdateKeyTable
		mov	eax, ecx
		add	eax, kb_AsciiCode
		mov	[edx - 4], eax

		; Next i-key
@@Scan6:		inc	edx
		cmp	byte ptr [edx], 0
		jne	@@Scan6
		inc	edx
		cmp	dword ptr [edx], -1
		jne	@@Scan


		; ----------------
		; Return to caller
		; ----------------
		
		ret

UpdateIndexKeys	endp


; ===============================================================
; = SHIFT_ED							=
; =								=
; = Code for the forth word:	Shift_Ed			=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
SHIFT_ED		proc

		; Push value
		IPUSH	0200h

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E_SHIFT_ED:
SHIFT_ED		endp


; ===============================================================
; = CTRL_ED							=
; =								=
; = Code for the forth word:	Ctrl_Ed				=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
CTRL_ED		proc

		; Push value
		IPUSH	0400h

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_CTRL_ED:
CTRL_ED		endp


; ===============================================================
; = ALT_ED							=
; =								=
; = Code for the forth word:	Alt_Ed				=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
ALT_ED		proc

		; Push value
		IPUSH	0800h

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_ALT_ED:
ALT_ED		endp


; ===============================================================
; = KEYPAD_ED							=
; =								=
; = Code for the forth word:	KeyPad_Ed			=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
KEYPAD_ED	proc

		; Push value
		IPUSH	80h

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_KEYPAD_ED:
KEYPAD_ED	endp


; ===============================================================
; = KEY_xxx							=
; =								=
; = Code for the froth words:	KEY_xxx				=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

ikey		MACRO	iname

		align	4
KEY_&iname	proc

		; Push key-code address
		sub	esi, @WordSize
		mov	eax, [IK_&iname] 
		movzx	eax, word ptr [eax]
		and	ax, 0F7Fh				; Inhib bit 7 (Extended keyboard)
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_KEY_&iname:
KEY_&iname	endp

		endm

		ikey	INS
		ikey	DEL
		ikey	HOME
		ikey	END
		ikey	PGUP
		ikey	PGDN
		ikey	UP
		ikey	DOWN
		ikey	LEFT
		ikey	RIGHT
		ikey	ESC
		ikey	PLUS
		ikey	MINUS

		ikey	F1
		ikey	F2
		ikey	F3
		ikey	F4
		ikey	F5
		ikey	F6
		ikey	F7
		ikey	F8
		ikey	F9
		ikey	F10
		ikey	F11
		ikey	F12
		
purge		ikey
		

; ===============================================================
; = KEY_ALT_x							=
; =								=
; = Code for the froth words:	KEY_ALT_x			=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

ikey		MACRO	iname

		align	4
KEY_&iname	proc

		; Push key-code address
		sub	esi, @WordSize
		mov	eax, [IK_&iname] 
		movzx	eax, word ptr [eax + 6]
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_KEY_&iname:
KEY_&iname	endp

		endm

		ikey	ALT_A
		ikey	ALT_Z
		ikey	ALT_E
		ikey	ALT_R
		ikey	ALT_T
		ikey	ALT_Y
		ikey	ALT_U
		ikey	ALT_I
		ikey	ALT_O
		ikey	ALT_P
		ikey	ALT_Q
		ikey	ALT_S
		ikey	ALT_D
		ikey	ALT_F
		ikey	ALT_G
		ikey	ALT_H
		ikey	ALT_J
		ikey	ALT_K
		ikey	ALT_L
		ikey	ALT_M
		ikey	ALT_W
		ikey	ALT_X
		ikey	ALT_C
		ikey	ALT_V
		ikey	ALT_B
		ikey	ALT_N
		
purge		ikey


code32		ends
		end

