; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	LOCVAR.ASM						!
; 									!
; Description:	Definitions for playing with local variables.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	BASE		BASE						!
;	DECIMAL		DECIMAL						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = 0LOC 1LOC 2LOC 3LOC               				=
; =                                                         	=
; = Code for the forth words:	0LOC	1LOC	2LOC	3LOC	=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

NLOC		MACRO	name, ename, sz

		align	4
name		proc

		; Verify if the system is in compile state
		call	QCOMPILE

		; Store 386 op: ENTER n, 0
		ENTER_n	sz*@WordSize

		; Update LocVar, then ; will put 386 op LEAVE
		mov	[LocVar], sz				; 0, 1 (0LOC), 2 (1LOC), 3 (2LOC), 4 (3LOC)

		; Return to caller
		ret

ename:
name		endp

		ENDM

		NLOC	ZEROLOC, E_ZEROLOC, 1
		NLOC	ONELOC, E_ONELOC, 2
		NLOC	TWOLOC, E_TWOLOC, 3
		NLOC	THREELOC, E_THREELOC, 4


; ===============================================================
; = @0 @1 @2 @3		               				=
; =                                                         	=
; = Code for the forth words:	@0	@1	@2	@3	=
; =                                                         	=
; = Stack:  ( --- n )						=
; ===============================================================

ATN		MACRO	name, ename, sz

		align	4
name		proc

		; Verify if the system is in compile state
		call	QCOMPILE

		; Store 386 op: MOV EAX, [EBP + i8]
		MOV_EAX_xEBPi8	-sz*@WordSize		; MOV	EAX, [EBP - n*4]

		; Push value on stack
		SUB_ESI_i8	@WordSize		; SUB	ESI, 4
		MOV_xESIi8_EAX				; MOV	[ESI], EAX
	
		; Return to caller
		ret

ename:
name		endp

		ENDM

		ATN	AT0, E_AT0, 1
		ATN	AT1, E_AT1, 2
		ATN	AT2, E_AT2, 3
		ATN	AT3, E_AT3, 4


; ===============================================================
; = !0 !1 !2 !3 	               				=
; =                                                         	=
; = Code for the forth words:	!0	!1	!2	!3	=
; =                                                         	=
; = Stack:  ( n --- )						=
; ===============================================================

STOREN		MACRO	name, ename, sz

		align	4
name		proc

		; Verify if the system is in compile state
		call	QCOMPILE

		; Get value from stack
		_LODSD_
	
		; Store 386 op: MOV EAX, [EBP + i8]
		MOV_xEBPi8_EAX	-sz*@WordSize		; MOV	[EBP - n*4], EAX

		; Return to caller
		ret

ename:
name		endp

		ENDM

		STOREN	STORE0, E_STORE0, 1
		STOREN	STORE1, E_STORE1, 2
		STOREN	STORE2, E_STORE2, 3
		STOREN	STORE3, E_STORE3, 4


; ===============================================================
; = +!0 +!1 +!2 +!3 	               				=
; =                                                         	=
; = Code for the forth words:	+!0	+!1	+!2	+!3	=
; =                                                         	=
; = Stack:  ( n --- )						=
; ===============================================================

PLUSSTOREN	MACRO	name, ename, sz

		align	4
name		proc

		; Verify if the system is in compile state
		call	QCOMPILE

		; Get value from stack
		_LODSD_
	
		; Store 386 op: ADD EAX, [EBP + i8]
		ADD_xEBPi8_EAX	-sz*@WordSize		; ADD	[EBP - n*4], EAX

		; Return to caller
		ret

ename:
name		endp

		ENDM

		PLUSSTOREN	PLUSSTORE0, E_PLUSSTORE0, 1
		PLUSSTOREN	PLUSSTORE1, E_PLUSSTORE1, 2
		PLUSSTOREN	PLUSSTORE2, E_PLUSSTORE2, 3
		PLUSSTOREN	PLUSSTORE3, E_PLUSSTORE3, 4


; ===============================================================
; = LOOP0 LOOP1 LOOP2 LOOP3 	               			=
; =                                                         	=
; = Code for the forth words:	LOOP0	LOOP1	LOOP2	LOOP3	=
; =                                                         	=
; = Stack:  ( --- )						=
; ===============================================================

		align	4
cLOOPN		proc

		; Verify if the system is in compile state
		call	QCOMPILE

		; Verify that 'Begin' was called before
		; -------------------------------------
		
		mov	edx, [ControlStackPtr]		
		mov	eax, [edx]
		add	edx, @WordSize
		cmp	eax, 'Begi'
		je	@@BeginCalled
		
		; No, it's an error
		mov	ah, ERR_BEGINNC				; "'Begin' not called beforehand"
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@BeginCalled:			

		; Store 386 op: DEC DWORD PTR [EBP + i8]
		DEC_xEBP					; DEC	DWORD PTR [EBP - n*4]
		pop	eax
		call	CodeCComma				; CodeC,


		; Put 'CALL ZBranch' offset
		; -------------------------
		
		; Relative address to jump
		mov	ebx, [edx]				; Address to jump
		add	edx, @WordSize
		mov	[ControlStackPtr], edx

		; Put the 32-bits offset into the Code Dictionnary
		mov	eax, ebx
		sub	eax, 1					; Size of the codop 'JE rel8'
		sub	eax, [CDA_SIZE]				; HERE
		sub	eax, 1					; 1 for REL8
		cmp	eax, -128
		jl	@@LongJe
		cmp	eax, 127
		jg	@@LongJe
		JNE_rel8					; Put 'JNZ rel8'	operation
		call	CodeCComma				; CodeC,
		jmp	@@ShortLongJe
@@LongJe:	JNE_rel32					; Put 'JNZ rel32'	operation
		mov	eax, ebx
		sub	eax, [CDA_SIZE]				; HERE
		sub	eax, 4					; 1 for REL32
		call	CodeComma				; Code,		
@@ShortLongJe:

		; Return to caller
		ret

cLOOPN		endp


LOOPN		MACRO	name, ename, sz

		align	4
name		proc

		push	large -sz*@WordSize
		jmp	cLOOPN

ename:
name		endp

		ENDM


		LOOPN		LOOP0, E_LOOP0, 1
		LOOPN		LOOP1, E_LOOP1, 2
		LOOPN		LOOP2, E_LOOP2, 3
		LOOPN		LOOP3, E_LOOP3, 4


code32		ends
		end

