; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	LOGIC.ASM						!
; 									!
; Description:	Definitions for playing with integer logic.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	VALSE		FALSE						!
;	VTRUE		TRUE						!
;	==		EQUAL						!
;	!=		NEQUAL						!
;	<		LESS						!
;	<=		LESSEQ						!
;	>		GREAT						!
;	>=		GREATEQ						!
;	U<		ULESS						!
;	U<=		ULESSEQ						!
;	U>		UGREAT						!
;	U>=		UGREATEQ					!
;	AND		IAND						!
;	OR		IOR						!
;	XOR		IXOR						!
;	INVERT		INVERT						!
;	SHR		ISHR						!
;	SHL		ISHL						!
;	SETBIT		SETBIT						!
;	RESETBIT	RESETBIT					!
;	COMPLBIT	COMPLBIT					!
;	GETBIT		GETBIT						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = VFALSE							=
; =                                                 		=
; = Froth word:		FALSE					=
; =                     	                            	=
; = Stack:		( --- ff )				=
; ===============================================================

		align	4
VFALSE		proc

		cmp	[CompileMode], TRUE
		je	@@Compile


		; ------------
@@Execute:	; Execute mode
		; ------------

		IPUSH	FALSE			; Push value on stack
		OVSTACK				; Verify stack frame overflow
		ret				; Return to caller


		; ------------
@@Compile:	; Compile mode
		; ------------

		mov	eax, FALSE
		COMP_NB
		ret				; Return to caller
		
E_VFALSE:
VFALSE		endp


; ===============================================================
; = VTRUE							=
; =                                                 		=
; = Froth word:		TRUE					=
; =                     	                            	=
; = Stack:		( --- 1 )				=
; ===============================================================

		align	4
VTRUE		proc

		cmp	[CompileMode], TRUE
		je	@@Compile


		; ------------
@@Execute:	; Execute mode
		; ------------

		IPUSH	TRUE			; Push value on stack
		OVSTACK				; Verify stack frame overflow
		ret				; Return to caller


		; ------------
@@Compile:	; Compile mode
		; ------------

		mov	eax, TRUE
		COMP_NB
		ret				; Return to caller
		
E_VTRUE:
VTRUE		endp


; ===============================================================
; = EQUAL							=
; =                                                 		=
; = Code for the forth word:	==				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values.				=
; ===============================================================

		align	4
EQUAL		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jne	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JNE_rel8	1				;     JNE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		cmpz	eax, 0
		jne	@@NotZero
@@Zero:		OR_EAX_EAX					;     OR    EAX, EAX
		jmp	@@ZeroDone
@@NotZero:	CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JNE_rel8	1				;     JNE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_EQUAL:
EQUAL		endp


; ===============================================================
; = NEQUAL							=
; =                                                 		=
; = Code for the forth word:	!=				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values.				=
; ===============================================================

		align	4
NEQUAL		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		je	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JE_rel8	1					;     JE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		cmpz	eax, 0
		jne	@@NotZero
@@Zero:		OR_EAX_EAX					;     OR    EAX, EAX
		jmp	@@ZeroDone
@@NotZero:	CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JE_rel8	1					;     JE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_NEQUAL:
NEQUAL		endp


; ===============================================================
; = LESS							=
; =                                                 		=
; = Code for the forth word:	<				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
LESS		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jge	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JGE_rel8 1					;     JGE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		cmpz	eax, 0
		jne	@@NotZero
@@Zero:		OR_EAX_EAX					;     OR    EAX, EAX
		jmp	@@ZeroDone
@@NotZero:	CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JGE_rel8 1					;     JGE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_LESS:
LESS		endp


; ===============================================================
; = LESSEQ							=
; =                                                 		=
; = Code for the forth word:	<=				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
LESSEQ		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jg	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JG_rel8 1					;     JG   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		cmpz	eax, 0
		jne	@@NotZero
@@Zero:		OR_EAX_EAX					;     OR    EAX, EAX
		jmp	@@ZeroDone
@@NotZero:	CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JG_rel8 1					;     JG   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_LESSEQ:
LESSEQ		endp


; ===============================================================
; = GREAT							=
; =                                                 		=
; = Code for the forth word:	>				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
GREAT		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jle	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JLE_rel8 1					;     JLE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		cmpz	eax, 0
		jne	@@NotZero
@@Zero:		OR_EAX_EAX					;     OR    EAX, EAX
		jmp	@@ZeroDone
@@NotZero:	CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JLE_rel8 1					;     JLE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_GREAT:
GREAT		endp


; ===============================================================
; = GREATEQ							=
; =                                                 		=
; = Code for the forth word:	>=				=
; =                     	                            	=
; = Stack:	( n1 n2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values signed.			=
; ===============================================================

		align	4
GREATEQ		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jl	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JL_rel8 1					;     JL   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		cmpz	eax, 0
		jne	@@NotZero
@@Zero:		OR_EAX_EAX					;     OR    EAX, EAX
		jmp	@@ZeroDone
@@NotZero:	CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JL_rel8 1					;     JL   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_GREATEQ:
GREATEQ		endp


ifdef NEVER
; ===============================================================
; = GREATLESS							=
; =                                                 		=
; = Code for the forth word:	><				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 < n3				=
; ===============================================================

		align	4
GREATLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jng	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnl	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATLESS:
GREATLESS	endp


; ===============================================================
; = GREATEQLESS							=
; =                                                 		=
; = Code for the forth word:	>=<				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 < n3				=
; ===============================================================

		align	4
GREATEQLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnge	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnl	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATEQLESS:
GREATEQLESS	endp


; ===============================================================
; = GREATLESSEQ							=
; =                                                 		=
; = Code for the forth word:	><=				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 <= n3				=
; ===============================================================

		align	4
GREATLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jng	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnle	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATLESSEQ:
GREATLESSEQ	endp


; ===============================================================
; = GREATEQLESSEQ						=
; =                                                 		=
; = Code for the forth word:	>=<=				=
; =                     	                            	=
; = Stack:	( n1 n2 n3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 <= n3				=
; ===============================================================

		align	4
GREATEQLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnge	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnle	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_GREATEQLESSEQ:
GREATEQLESSEQ	endp
endif


; ===============================================================
; = ULESS							=
; =                                                 		=
; = Code for the forth word:	U<				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
ULESS		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jae	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JAE_rel8 1					;     JAE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JAE_rel8 1					;     JAE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_ULESS:
ULESS		endp


; ===============================================================
; = ULESSEQ							=
; =                                                 		=
; = Code for the forth word:	U<=				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
ULESSEQ		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		ja	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JA_rel8 1					;     JA    LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JA_rel8 1					;     JA   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_ULESSEQ:
ULESSEQ		endp


; ===============================================================
; = UGREAT							=
; =                                                 		=
; = Code for the forth word:	U>				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
UGREAT		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jbe	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JBE_rel8 1					;     JBE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JBE_rel8 1					;     JBE   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_UGREAT:
UGREAT		endp


; ===============================================================
; = UGREATEQ							=
; =                                                 		=
; = Code for the forth word:	U>=				=
; =                     	                            	=
; = Stack:	( u1 u2 --- f )					=
; =                     	                            	=
; = Usage:	Compare two values unsigned.			=
; ===============================================================

		align	4
UGREATEQ	proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		LODSD						; Pop <n2>
		clr	ebx
		cmp	PARAM0, eax				; n1 == n2 ?
		jb	@@store
		dec	ebx
@@store:	mov	PARAM0, ebx
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL  VStack2
@@NoOpt2:	_LODSD_						;     LODSD
		CLR_EBX						;     CLR   EBX
		CMP_xESIi8_EAX					;     CMP   [ESI], EAX
		JB_rel8 1					;     JB    LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	MOV_EAX_xESIi8					;     MOV   EAX, [ESI]
		CLR_EBX						;     CLR   EBX
		CMP_EAX_i32					;     CMP   EAX, imm32
		call		CodeComma			;     Code,
@@ZeroDone:	JB_rel8 1					;     JB   LL
		DEC_EBX						;     DEC   EBX
		MOV_xESIi8_EBX					; LL: MOV [ESI], EBX

@@exit:		ret

E_UGREATEQ:
UGREATEQ	endp


ifdef NEVER
; ===============================================================
; = UGREATLESS							=
; =                                                 		=
; = Code for the forth word:	U><				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 < n3				=
; ===============================================================

		align	4
UGREATLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jna	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnb	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATLESS:
UGREATLESS	endp


; ===============================================================
; = UGREATEQLESS						=
; =                                                 		=
; = Code for the forth word:	U>=<				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 < n3				=
; ===============================================================

		align	4
UGREATEQLESS	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnae	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnb	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATEQLESS:
UGREATEQLESS	endp


; ===============================================================
; = UGREATLESSEQ						=
; =                                                 		=
; = Code for the forth word:	U><=				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 < n1 <= n3				=
; ===============================================================

		align	4
UGREATLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jna	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnbe	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATLESSEQ:
UGREATLESSEQ	endp


; ===============================================================
; = UGREATEQLESSEQ						=
; =                                                 		=
; = Code for the forth word:	U>=<=				=
; =                     	                            	=
; = Stack:	( u1 u2 u3 --- f )				=
; =                     	                            	=
; = Usage:	tf if n2 <= n1 <= n3				=
; ===============================================================

		align	4
UGREATEQLESSEQ	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; Make the test
		; -------------
		
		mov	eax, PARAM2				; Get n1
		cmp	eax, PARAM1				; n1 ? n2
		jnae	@@False
		cmp	eax, PARAM0				; n1 ? n3
		jnbe	@@False
			
@@True:		DISCARD	2					; drop drop
		mov	PARAM0, TRUE
		ret						; jmp	@@Exit
		
@@False:	DISCARD	2					; drop drop
		mov	PARAM0, FALSE


		; Return to caller
		; ----------------
		
@@Exit:		ret

E_UGREATEQLESSEQ:
UGREATEQLESSEQ	endp
endif


; ===============================================================
; = IAND							=
; =                                                 		=
; = Code for the forth word:	AND				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 AND n2					=
; ===============================================================

		align	4
IAND		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; get <n2>
		and	PARAM0, eax				; n1 and n2
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack1
@@NoOpt2:	_LODSD_						; LODSD
		AND_xESIi8_EAX					; AND    [ESI], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	and	eax, ebx
		sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		AND_xESIi8_i8	0
		jmp	@@exit
@@i32:		AND_xESIi8_i32	0

@@exit:		ret

E_IAND:
IAND		endp


; ===============================================================
; = IOR								=
; =                                                 		=
; = Code for the forth word:	OR				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 OR n2					=
; ===============================================================

		align	4
IOR		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; get <n2>
		or	PARAM0, eax				; n1 or n2
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack1
@@NoOpt2:	_LODSD_						; LODSD
		OR_xESIi8_EAX					; OR     [ESI], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	or	eax, ebx
		sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		OR_xESIi8_i8	0
		jmp	@@exit
@@i32:		OR_xESIi8_i32	0

@@exit:		ret

E_IOR:
IOR		endp


; ===============================================================
; = XOR								=
; =                                                 		=
; = Code for the forth word:	XOR				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 XOR n2					=
; ===============================================================

		align	4
IXOR		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; get <n2>
		xor	PARAM0, eax				; n1 xor n2
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack1
@@NoOpt2:	_LODSD_						; LODSD
		XOR_xESIi8_EAX					; XOR    [ESI], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize
		mov	ebx, eax				; Save 1st literal
		call	QPrev_WasNb	
		jnc	@@OneNb

@@TwoNb:	xor	eax, ebx
		sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		COMP_NB
		jmp	@@exit

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	mov	eax, ebx
		cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		XOR_xESIi8_i8	0
		jmp	@@exit
@@i32:		XOR_xESIi8_i32	0

@@exit:		ret

E_IXOR:
IXOR		endp


; ===============================================================
; = INVERT							=
; =                                                 		=
; = Code for the forth word:	INVERT				=
; =                     	                            	=
; = Stack:	( n1 --- n2 )					=
; =                     	                            	=
; = Usage:	n2 = ~ n1					=
; ===============================================================

		align	4
INVERT		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	1
		not	PARAM0					; not n
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack1				; CALL   VStack1
@@NoOpt2:	NOT_xESIi8					; NOT    dword ptr [ESI]
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		not	eax
		COMP_NB

@@exit:		ret

E_INVERT:
INVERT		endp


; ===============================================================
; = ISHL							=
; =                                                 		=
; = Code for the forth word:	shl				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 SHL n2					=
; ===============================================================

		align	4
ISHL		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		lodsd						; Pop <n1>
		mov	cl, al
		shl	PARAM0, cl				; n1 shl n2
		into						; Integer arithmetic overflow test
		ret

E_ISHL:
ISHL		endp


; ===============================================================
; = ISHR							=
; =                                                 		=
; = Code for the forth word:	shr				=
; =                     	                            	=
; = Stack:	( n1 n2 --- n3 )				=
; =                     	                            	=
; = Usage:	n3 = n1 SHR n2					=
; ===============================================================

		align	4
ISHR		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; Make the logical operation
		; --------------------------
		
		lodsd						; Pop <n1>
		mov	cl, al
		shr	PARAM0, cl				; n1 shr n2
		ret

E_ISHR:
ISHR		endp


; ===============================================================
; = SetBit							=
; =                                                 		=
; = Code for the forth word:	SetBit				=
; =                     	                            	=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
SETBIT		proc

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		lodsd						; Pop <n>
		bts	[ebx], eax				; Set bit
		ret

E_SETBIT:
SETBIT		endp


; ===============================================================
; = ResetBit							=
; =                                                 		=
; = Code for the forth word:	ResetBit			=
; =                     	                            	=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
RESETBIT	proc

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		lodsd						; Pop <n>
		btr	[ebx], eax				; Set bit
		ret

E_RESETBIT:
RESETBIT	endp


; ===============================================================
; = ComplBit							=
; =                                                 		=
; = Code for the forth word:	ComplBit			=
; =                     	                            	=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
COMPLBIT	proc

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		lodsd						; Pop <n>
		btc	[ebx], eax				; Set bit
		ret

E_COMPLBIT:
COMPLBIT	endp


; ===============================================================
; = GetBit							=
; =                                                 		=
; = Code for the forth word:	GetBit				=
; =                     	                            	=
; = Stack:	( n addr --- f )				=
; ===============================================================

		align	4
GETBIT		proc

		VSTACK	2
		lodsd						; Pop <addr>		
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		mov	eax, PARAM0				; Get <n>
		clr	ecx
		bt	[ebx], eax				; Set bit
		jnc	@@store
@@on:		dec	ecx
@@store:	mov	PARAM0, ecx
		ret

E_GETBIT:
GETBIT		endp


code32		ends
		end

