; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	MEMORY.ASM						!
; 									!
; Description:	Definitions for playing with the memory.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	MEMFREE		MEMFREE						!
;	[@]		RAT						!
;	(@)		PAT						!
;	@		AT						!
;	[C@]		RCAT						!
;	(C@)		PCAT						!
;	C8@		CAT						!
;	[W@]		RWAT						!
;	(W@)		PWAT						!
;	W@		WAT						!
;	!		STORE						!
;	(!)		PSTORE						!
;	C!		CSTORE						!
;	(C!)		PCSTORE						!
;	W!		WSTORE						!
;	(W!)		PWSTORE						!
;	C@SWAP		CATWAP						!
;	W@SWAP		WATSWAP						!
;	@SWAP		ATSWAP						!
;	+!		PLUSSTORE					!
;	-!		MINUSSTORE					!
;	*!		STARSTORE					!
;	/!		SLASHSTORE					!
;	?		QUEST						!
;	C?		CQUEST						!
;	W?		WQUEST						!
;	DYNAMIC		DYNAMIC						!
;	ALLOC		ALLOCATE					!
;	FREE		FREE						!
;	?AllotFail	QALLOTFAIL					!
;	ALLOT		ALLOT						!
;	AllotMore	ALLOTMORE					!
;	RT_DOES		PDOES						!
;	DOES>		DOES						!
;	DataSalvage	DATASALVAGE					!
;	DUST		DUST						!
;	DUSTTHRESHOLD	DUSTTHRESHOLD					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions

		
		; Used in: ALLOT and DOES
		extrn	InterpretAddr:	proc		; See INTERPRET


; ===============================================================
; = MEMFREE							=
; =                                                 		=
; = Code for the forth word:	MemFree				=
; =                     	                            	=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
MEMFREE		proc

		; Compute the free memory available into (code + data) dictionnary
		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		mov	ebx, [DDA_START]
		sub	ebx, [DDA_SIZE]
		sub	ebx, eax
		IPUSH	ebx

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E_MEMFREE:
MEMFREE		endp


; ===============================================================
; = RAT								=
; =                                                 		=
; = Code for the forth word:	[@]				=
; =								=
; = Stack:	( addr --- n )					=
; ===============================================================

		align	4
RAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		push	fs
		mov	fs, [_selzero]
		mov	eax, fs:[eax]
		pop	fs
		mov	PARAM0, eax

		; Return to caller
		ret
E_RAT:
RAT		endp


; ===============================================================
; = PAT								=
; =                                                 		=
; = Code for the forth word:	(@)				=
; =								=
; = Stack:	( addr --- n )					=
; ===============================================================

		align	4
PAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		mov	eax, [eax]
		mov	PARAM0, eax

		; Return to caller
		ret
E_PAT:
PAT		endp


; ===============================================================
; = AT								=
; =                                                 		=
; = Code for the forth word:	@				=
; =								=
; = Stack:	( addr --- n )					=
; ===============================================================

		align	4
AT		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	1
		mov	eax, PARAM0
		bound	eax, qword ptr [DDA_BOUND]
		mov	eax, [eax]
		mov	PARAM0, eax
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack1				; CALL   VStack1
@@NoOpt2:	MOV_EAX_xESIi8					; MOV    EAX, [ESI]
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOV_EAX_xEAXi8					; MOV    EAX, [EAX]
		MOV_xESIi8_EAX					; MOV    [ESI], EAX

		ret

E_AT:
AT		endp


; ===============================================================
; = RCAT							=
; =                                                 		=
; = Code for the forth word:	[C@]				=
; =								=
; = Stack:	( adr --- b )					=
; ===============================================================

		align	4
RCAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push		
		mov	eax, PARAM0
		push	fs
		mov	fs, [_selzero]
		movzx	eax, byte ptr fs:[eax]
		pop	fs
		mov	PARAM0, eax

		; Return to caller
		ret

E_RCAT:
RCAT		endp


; ===============================================================
; = PCAT							=
; =                                                 		=
; = Code for the forth word:	(C@)				=
; =								=
; = Stack:	( adr --- b )					=
; ===============================================================

		align	4
PCAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push		
		mov	eax, PARAM0
		movzx	eax, byte ptr [eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_PCAT:
PCAT		endp


; ===============================================================
; = CAT								=
; =                                                 		=
; = Code for the forth word:	C@				=
; =								=
; = Stack:	( addr --- b )					=
; ===============================================================

		align	4
CAT		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	1
		mov	eax, PARAM0
		bound	eax, qword ptr [DDA_BOUND]
		movzx	eax, byte ptr [eax]
		mov	PARAM0, eax
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack1				; CALL   VStack1
@@NoOpt2:	MOV_EAX_xESIi8					; MOV    EAX, [ESI]
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOVZX_EAX_bxEAXi8				; MOVZX  EAX, byte ptr [EAX]
		MOV_xESIi8_EAX					; MOV    [ESI], EAX

		ret

E_CAT:
CAT		endp


; ===============================================================
; = RWAT							=
; =                                                 		=
; = Code for the forth word:	[W@]				=
; =								=
; = Stack:	( adr --- w )					=
; ===============================================================

		align	4
RWAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		push	fs
		mov	fs, [_selzero]
		movzx	eax, word ptr fs:[eax]
		pop	fs
		mov	PARAM0, eax

		; Return to caller
		ret

E_RWAT:
RWAT		endp


; ===============================================================
; = PWAT							=
; =                                                 		=
; = Code for the forth word:	(W@)				=
; =								=
; = Stack:	( adr --- w )					=
; ===============================================================

		align	4
PWAT		proc

		; Verify stack frame
		VSTACK	1

		; Read value then push
		mov	eax, PARAM0
		movzx	eax, word ptr [eax]
		mov	PARAM0, eax

		; Return to caller
		ret

E_PWAT:
PWAT		endp


; ===============================================================
; = WAT								=
; =                                                 		=
; = Code for the forth word:	W@				=
; =								=
; = Stack:	( addr --- w )					=
; ===============================================================

		align	4
WAT		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	1
		mov	eax, PARAM0
		bound	eax, qword ptr [DDA_BOUND]
		movzx	eax, word ptr [eax]
		mov	PARAM0, eax
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack1				; CALL   VStack1
@@NoOpt2:	MOV_EAX_xESIi8					; MOV    EAX, [ESI]
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOVZX_EAX_wxEAXi8				; MOVZX  EAX, word ptr [EAX]
		MOV_xESIi8_EAX					; MOV    [ESI], EAX

		ret

E_WAT:
WAT		endp


; ===============================================================
; = STORE							=
; =                                                 		=
; = Code for the forth word:	!				=
; =								=
; = Stack:	( n addr --- )					=
; =								=
; = ALL REGISTERS ARE PRESERVED.				=
; ===============================================================

		align	4
STORE		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr [DDA_BOUND]
		mov	ebx, eax
		lodsd						; Pop <n>
		mov	[ebx], eax
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOV_EBX_EAX					; MOV    EBX, EAX
		_LODSD_						; LODSD
		MOV_xEBXi8_EAX					; MOV    [EBX], EAX
		ret


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@OneNb3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@OneNb3:	MOV_xEAXi8_i32	0				; MOV   dword ptr [EAX], i32
		call	CodeComma               		; Code,
		ret

E_STORE:
STORE		endp


; ===============================================================
; = PSTORE							=
; =                                                 		=
; = Code for the forth word:	(!)				=
; =								=
; = Stack:	( n adr --- )					=
; ===============================================================

		align	4
PSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Store value
		mov	eax, PARAM0				; Pop adr
		mov	ebx, PARAM1				; Pop n
		mov	[eax], ebx
		DISCARD	2

		; Return to caller
		ret

E_PSTORE:
PSTORE		endp


; ===============================================================
; = PCSTORE							=
; =                                                 		=
; = Code for the forth word:	(C!)				=
; =								=
; = Stack:	( b adr --- )					=
; ===============================================================

		align	4
PCSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Store value
		mov	eax, PARAM0				; Pop adr
		mov	ebx, PARAM1				; Pop n
		mov	[eax], bl
		DISCARD	2

		; Return to caller
		ret

E_PCSTORE:
PCSTORE		endp


; ===============================================================
; = CSTORE							=
; =                                                 		=
; = Code for the forth word:	C!				=
; =								=
; = Stack:	( b adr --- )					=
; ===============================================================

		align	4
CSTORE		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr [DDA_BOUND]
		mov	ebx, eax
		lodsd						; Pop <n>
		mov	[ebx], al
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOV_EBX_EAX					; MOV    EBX, EAX
		_LODSD_						; LODSD
		MOV_xEBXi8_AL					; MOV    [EBX], AL
		ret


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@OneNb3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@OneNb3:	MOV_bxEAXi8_i8	0				; MOV   word ptr [EAX], i8
		call	CodeCComma               		; CodeC,
		ret

E_CSTORE:
CSTORE		endp


; ===============================================================
; = WSTORE							=
; =                                                 		=
; = Code for the forth word:	W!				=
; =								=
; = Stack:	( w addr --- )					=
; ===============================================================

		align	4
WSTORE		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr [DDA_BOUND]
		mov	ebx, eax
		lodsd						; Pop <n>
		mov	[ebx], ax
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOV_EBX_EAX					; MOV    EBX, EAX
		_LODSD_						; LODSD
		MOV_xEBXi8_AX					; MOV    [EBX], AX
		ret


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@OneNb3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@OneNb3:	MOV_wxEAXi8_i16	0				; MOV   word ptr [EAX], i16
		call	CodeWComma               		; CodeW,
		ret

E_WSTORE:
WSTORE		endp


; ===============================================================
; = PWSTORE							=
; =                                                 		=
; = Code for the forth word:	(W!)				=
; =								=
; = Stack:	( w adr --- )					=
; ===============================================================

		align	4
PWSTORE		proc

		; Verify stack frame
		VSTACK	2

		; Store value
		mov	eax, PARAM0				; Pop adr
		mov	ebx, PARAM1				; Pop w
		mov	[eax], bx
		DISCARD	2

		; Return to caller
		ret

E_PWSTORE:
PWSTORE		endp


; ===============================================================
; = PLUSSTORE							=
; =                                                 		=
; = Code for the forth word:	+!				=
; =								=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
PLUSSTORE	proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		lodsd						; Pop <n>
		add	[ebx], eax
		into						; Integer arithmetic overflow test
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOV_EBX_EAX					; MOV    EBX, EAX
		_LODSD_						; LODSD
		ADD_xEBXi8_EAX					; ADD    [EBX], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@OneNb3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@OneNb3:	cmp	eax, 1
		jne	@@NotInc
@@Inc:		INC_xEAXi8
		jmp	@@exit
@@NotInc:	cmp	eax, -1
		jne	@@NotDec
@@Dec:		DEC_xEAXi8
		jmp	@@exit
@@NotDec:	cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		ADD_xEAXi8_i8	0
		jmp	@@exit
@@i32:		ADD_xEAXi8_i32	0

@@exit:		call	isQIArithOverfl				; ?iarithoverflow
		jnc	@@exit2
		_INTO_						; Integer arithmetic overflow test
@@exit2:	ret

E_PLUSSTORE:
PLUSSTORE	endp


; ===============================================================
; = MINUSSTORE							=
; =                                                 		=
; = Code for the forth word:	-!				=
; =								=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
MINUSSTORE	proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		lodsd						; Pop <n>
		sub	[ebx], eax
		into						; Integer arithmetic overflow test
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		call	isOptimize
		jc	@@Optimize


		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL VStack2
@@NoOpt2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@NoOpt3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@NoOpt3:	MOV_EBX_EAX					; MOV    EBX, EAX
		_LODSD_						; LODSD
		SUB_xEBXi8_EAX					; SUB    [EBX], EAX
		jmp	@@exit


		; Optimization
		; ------------

@@Optimize:	mov	edx, [CDA_SIZE]
		add	edx, [CDA_START]
		call	QPrev_WasNb	
		jnc	@@NoOptimize

@@OneNb:	sub	edx, [CDA_START]
		mov	[CDA_SIZE], edx
		call	isSafe
		jnc	@@OneNb2
		compCALL_rel32	VStack1				; CALL VStack1
@@OneNb2:	_LODSD_						; LODSD
		call	isSafe
		jnc	@@OneNb3
		BOUND_EAX_mem	DDA_BOUND			; BOUND  EAX, qword ptr [DDA_BOUND]
@@OneNb3:	cmp	eax, -1
		jne	@@NotInc
@@Inc:		INC_xEAXi8
		jmp	@@exit
@@NotInc:	cmp	eax, 1
		jne	@@NotDec
@@Dec:		DEC_xEAXi8
		jmp	@@exit
@@NotDec:	cmp	eax, -128
		jl	@@i32
		cmp	eax, 127
		jg	@@i32
@@i8:		SUB_xEAXi8_i8	0
		jmp	@@exit
@@i32:		SUB_xEAXi8_i32	0

@@exit:		call	isQIArithOverfl				; ?iarithoverflow
		jnc	@@exit2
		_INTO_						; Integer arithmetic overflow test
@@exit2:	ret

E_MINUSSTORE:
MINUSSTORE	endp


; ===============================================================
; = STARSTORE							=
; =                                                 		=
; = Code for the forth word:	*!				=
; =								=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
STARSTORE	proc

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		lodsd						; Pop <n>
		imul	dword ptr [ebx]				; EAX *= [addr]
		mov	[ebx], eax
		into						; Integer arithmetic overflow test
		ret

E_STARSTORE:
STARSTORE	endp


; ===============================================================
; = SLASHSTORE							=
; =                                                 		=
; = Code for the forth word:	/!				=
; =								=
; = Stack:	( n addr --- )					=
; ===============================================================

		align	4
SLASHSTORE	proc

		VSTACK	2
		lodsd						; Pop <addr>
		bound	eax, qword ptr DDA_BOUND		; Check addr against bound
		mov	ebx, eax
		lodsd						; Pop <n>
		clr	edx
		idiv	dword ptr [ebx]				; EAX *= [adr]
		mov	[ebx], eax
		ret

E_SLASHSTORE:
SLASHSTORE	endp


; ===============================================================
; = QUEST							=
; =                                                 		=
; = Code for the forth word:	?				=
; =								=
; = Stack:	( addr --- )					=
; ===============================================================

		align	4
QUEST		proc

		; Verify stack frame
		VSTACK	1

		; Do action
		DO_AT						; @
		call	DOT					; .

		; Return to caller
		ret

E_QUEST:
QUEST		endp


; ===============================================================
; = CQUEST							=
; =                                                 		=
; = Code for the forth word:	C?				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
CQUEST		proc

		; Verify stack frame
		VSTACK	1

		; Do action
		mov	eax, PARAM0
		movzx	eax, byte ptr [eax]
		mov	PARAM0, eax				; C@
		call	DOT					; .

		; Return to caller
		ret

E_CQUEST:
CQUEST		endp


; ===============================================================
; = WQUEST							=
; =                                                 		=
; = Code for the forth word:	W?				=
; =								=
; = Stack:	( adr --- )					=
; ===============================================================

		align	4
WQUEST		proc

		; Verify stack frame
		VSTACK	1

		; Do action
		mov	eax, PARAM0
		movzx	eax, word ptr [eax]
		mov	PARAM0, eax				; W@
		call	DOT					; .

		; Return to caller
		ret

E_WQUEST:
WQUEST		endp


; ===============================================================
; = DYNAMIC							=
; =                                                 		=
; = Code for the forth word:	DYNAMIC				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
DYNAMIC		proc

		; ----------------------------------------
		; Verify if the system is is execute state
		; ----------------------------------------

		call	QEXECUTE
		
		
		; ------------------------------------
		; Allready called and still not used ?
		; ------------------------------------
		
		cmp	[DynamicFlag], FALSE
		je	@@Ok
		
		; "DYNAMIC word already called"
		mov	ah, ERR_DYNAALCALL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError
		
@@Ok:

		; -----------
		; Update flag
		; -----------
		
		inc	[DynamicFlag]


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_DYNAMIC:
DYNAMIC		endp


; ===============================================================
; = ALLOCATE							=
; =                                                 		=
; = Code for the forth word:	ALLOCATE			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
ALLOCATE		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; ----------------------------------------
		; Store operation: 'MOV [AllocFlag], TRUE'
		; ----------------------------------------

		; Code operation		
		MOV_xMEM_i8
		
		; Memory relative address (rel32)
		lea	eax, AllocFlag
		call	CodeComma			; Code,
		
		; Value to store
		mov	al, TRUE
		call	CodeCComma			; CodeC,


		; ----------------------------------------
		; The word ALLOC was found while compiling
		; ----------------------------------------
		
		mov	[AllocFound], TRUE


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ALLOCATE:
ALLOCATE		endp


; ===============================================================
; = FREE							=
; =                                                 		=
; = Code for the forth word:	FREE				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
FREE		proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; -------------------------------------
		; Store operation: MOV [FreeFlag], TRUE
		; -------------------------------------

		; Code operation		
		MOV_xMEM_i8
		
		; Memory relative address (rel32)
		lea	eax, FreeFlag
		call	CodeComma			; Code,
		
		; Value to store
		mov	al, TRUE
		call	CodeCComma			; CodeC,


		; ---------------------------------------
		; The word FREE was found while compiling
		; ---------------------------------------
		
		mov	[FreeFound], TRUE


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_FREE:
FREE		endp


; ===============================================================
; = CodeComma							=
; =                                     	            	=
; = Code,							=
; =	Store a dword into CDA (Code Dictionnary Area)		=
; =	(it's a compiler action)				=
; =                       		                       	=
; = I) 	EAX value ro store					=
; = O)	none 							=
; = M) 	none							=
; ===============================================================

		align	4
CodeComma	proc

		; Dynamic allocation ?
		; --------------------
		
		cmp	[DynVarAddr], 0
		jne	@@DynaAlloc


		; NOT dynamic allocation
		; ----------------------

		; Store value
		push	ebx
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		mov	[ebx], eax
		pop	ebx
		
		; Adjust CDA size
		add	[CDA_SIZE], @WordSize
		
		; Done
		ret					; jmp	@@Exit
		
	
		; Dynamic allocation
		; ------------------
		
@@DynaAlloc:
		
		; Store value
		push	ebx
		mov	ebx, [DynVarAddr]
		mov	[ebx], eax
		pop	ebx
		
		; Ajust new address
		add	[DynVarAddr], @WordSize
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_CodeComma:
CodeComma	endp


; ===============================================================
; = CodeWComma							=
; =                                     	            	=
; = Code,							=
; =	Store a word into CDA (Code Dictionnary Area)		=
; =	(it's a compiler action)				=
; =                       		                       	=
; = I) 	AX value ro store					=
; = O)	none 							=
; = M) 	none							=
; ===============================================================

		align	4
CodeWComma	proc

		; Dynamic allocation ?
		; --------------------
		
		cmp	[DynVarAddr], 0
		jne	@@DynaAlloc


		; NOT dynamic allocation
		; ----------------------

		; Store value
		push	ebx
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		mov	[ebx], ax
		pop	ebx
		
		; Adjust CDA size
		add	[CDA_SIZE], 2
		
		; Done
		ret					; jmp	@@Exit
		
	
		; Dynamic allocation
		; ------------------
		
@@DynaAlloc:
		
		; Store value
		push	ebx
		mov	ebx, [DynVarAddr]
		mov	[ebx], ax
		pop	ebx
		
		; Ajust new address
		add	[DynVarAddr], 2
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_CodeWComma:
CodeWComma	endp


; ===============================================================
; = CodeCComma							=
; =                                     	            	=
; = Code,							=
; =	Store a byte into CDA (Code Dictionnary Area)		=
; =	(it's a compiler action)				=
; =                       		                       	=
; = I) 	AL value ro store					=
; = O)	none 							=
; = M) 	none							=
; ===============================================================

		align	4
CodeCComma	proc

		; Dynamic allocation ?
		; --------------------
		
		cmp	[DynVarAddr], 0
		jne	@@DynaAlloc


		; NOT dynamic allocation
		; ----------------------

		; Store value
		push	ebx
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		mov	[ebx], al
		pop	ebx
		
		; Adjust CDA size
		inc	[CDA_SIZE]
		
		; Done
		ret					; jmp	@@Exit
		
	
		; Dynamic allocation
		; ------------------
		
@@DynaAlloc:
		
		; Store value
		push	ebx
		mov	ebx, [DynVarAddr]
		mov	[ebx], al
		pop	ebx
		
		; Ajust new address
		inc	[DynVarAddr]
		

		; Return to caller
		; ----------------
		
@@Exit:		ret

E_CodeCComma:
CodeCComma	endp


; ===============================================================
; = QALLOTFAIL                      				=
; =                                                         	=
; = Code for the forth word:    ?AllotFail			=
; =                                                         	=
; = Stack:  ( --- addr ) 	                		=
; ===============================================================

QALLOTFAIL	proc

		; Integer run-time
		DOINT	D_QALLOTFAIL
		
		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_QALLOTFAIL:
QALLOTFAIL	endp


; ===============================================================
; = PDOES							=
; =								=
; = Code for the forth word:	RT_DOES>			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
PDOES		proc

		; ------------------------------------
		; Dynamic allocation (with ALLOCATE) ?
		; ------------------------------------
		
		cmp	[AllocFlag], FALSE
		je	@@NotDynaAlloc

		; Reset flag
		mov	[AllocFlag], FALSE
		mov	[DynVarAddr], 0

		; Done
		jmp	@@Exit
		
@@NotDynaAlloc:	


		; ------------------------------------------------------------
		; Local allocation ? No if Compile Actions called by INTERPRET
		; ------------------------------------------------------------

		lea	eax, InterpretAddr
		add	eax, [OfsDictReloc]
		cmp	dword ptr [esp + 4], eax
		je	@@StaticAll
		
		; Skip a level return
		add	esp, @WordSize
		mov	eax, [esp]
		add	eax, @WordSize				; Skip cell where is stored DataBlockAddr
		mov	[esp], eax
		ret
		
@@StaticAll:

		; -------------------------------------
		; Put 'JMP rel32' to run-time procedure
		; -------------------------------------

		; Store codop
		CALL_rel32		
		mov	eax, [esp]				; Cell where is stored where to jump
		mov	eax, [eax]
		add	eax, [NIA_START]
		mov	eax, [eax]
		sub	eax, [CDA_START]
		sub	eax, [CDA_SIZE]
		sub	eax, 4					; 4 for size REL32
		call	CodeComma				; Code,

		; ret
		_RET_

		; align 4
@@Align4:	mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		and	al, 3
		jz	@@Align4Done
		_NOP_
		jmp	@@Align4
@@Align4Done:


		; -----------------------------------------------------
		; Update the end code address of the word being defined
		; -----------------------------------------------------

		; Pointer on the word NIA address		
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		add	eax, [NIA_START]

		; Compute the size of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		sub	ebx, [eax.word_realaddr]
		
		; Store the code size
		mov	[eax.word_CodeSize], bx
		
		; Update the end of the code
		mov	ebx, [CDA_START]
		add	ebx, [CDA_SIZE]
		dec	ebx
		mov	[eax.word_EndCode], ebx


		; -----------------------------------------------
		; Terminate caller precedure (and not this one !)
		; -----------------------------------------------

@@Exit:		add	esp, 4					; Skip a level return !
		ret

E_PDOES:
PDOES		endp


; ===============================================================
; = DOES							=
; =                                                 		=
; = Code for the forth word:	DOES>				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
DOES		proc

		; 'BUILD>' not called before ?
		cmp	[CDA_BUILD_CALLED], FALSE
		jne	@@Ok

		; 'BUILD>' not called beforehand.
		mov	ah, ERR_BUILDNOTCALLED			; "'BUILD>' not called beforehand."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok:

		; DOES> already called ?
		cmp	[CDA_DOES_CALLED], FALSE
		je	@@Ok2
		
		; 'DOES>' already called
		mov	ah, ERR_DOESALCALLED			; "'DOES>' allready called."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok2:		mov	[CDA_DOES_CALLED], TRUE


		; Put 'CALL_rel32 rt_does'
		compCALL_rel32	PDOES
		
		; Store run-time call nb
		mov	ebx, [LastWordDefWColon]		; Last word (NIA) defined with : (colon)
		add	ebx, [NIA_START]
		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		add	eax, @WordSize
		mov	[ebx.word_runtaddr], eax
		
		; Store run-time call nb
		mov	eax, ebx
		sub	eax, [NIA_START]
		add	eax, word_runtaddr
		call	CodeComma				; Code,
		
		; Return to caller
		ret

E_DOES:
DOES		endp


; ===============================================================
; = LocalVar							=
; =                                                 		=
; = Code for the forth word:	LocalVar			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
LocalVar	proc

		; Save used registers
		push	eax ebx edx

		; Store into EDI the cell address
		mov	ebx, esp
		add	ebx, 4 * 3				; 12 = 4 (eax) + 4 (ebx) + 4 (edx)
		mov	edx, [ebx]

		; Get data reference (in EAX)
		mov	eax, [edx]				; Offset
		mov	eax, [edx + eax]			; Address of Data Block Nb (see .ddir)

		; Move the IP after the data reference
		add	edx, @WordSize
		mov	[ebx], edx				; 12 = 4 (eax) + 4 (ebx) + 4 (edx)

		; Push the data address
		mov	eax, [edi + eax.ddir_blkaddr]
		IPUSH	eax

		; Return to caller
		pop	edx ebx eax
		ret

E_LocalVar:
LocalVar	endp


; ===============================================================
; = DATASALVAGE							=
; =                                                 		=
; = Code for the forth word:	DataSalvage			=
; =								=
; = Stack:	( --- n )					=
; =								=
; = Usage							=
; =	Memory which can recovered by GARBAGE collector.	=
; ===============================================================

		align	4
DATASALVAGE	proc

		; Push value on stack
		mov	eax, [DDIR_MEMSALV]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_DATASALVAGE:
DATASALVAGE	endp


; ===============================================================
; = DustData							=
; =                                                 		=
; = I) EDX : pointer to record to delete.			=
; ===============================================================

		align	4
DustData	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	esi edi eax ebx ecx
		
		
		; -----------------------
		; Is-it the last record ?
		; -----------------------
		
		mov	eax, [DDIR_START]
		add	eax, [DDIR_SIZE]
		sub	eax, size ddirentry
		cmp	edx, eax
		jne	@@NotLastRec
		
		; Update the memory used
		mov	eax, [edx.ddir_blksize]
		sub	[DDA_SIZE], eax
		
		; Update the amount of memory which can be salvaged by GARBAGE collector
		sub	[DDIR_MEMSALV], eax

		; Update the size of the table
		sub	[DDIR_SIZE], size ddirentry
		
		; Update memory allocated bounds
		mov	ebx, [DDA_START]
		sub	ebx, [DDA_SIZE]
		mov	[DDA_BOUND], ebx

		; Done
		jmp	@@Exit
		
@@NotLastRec:

		; ------------------------
		; It's not the last record
		; ------------------------
		
		; Move memory
		; -----------
		
		; Source address
		mov	esi, [edx.ddir_blkaddr]
		dec	esi
		
		; Destination address
		mov	edi, esi
		add	edi, [edx.ddir_blksize]

		; Size to move
		lea	ecx, [esi + 1]
		add	ecx, [DDA_SIZE]
		sub	ecx, [DDA_START]
		js	@@DontMove
		jz	@@DontMove

		; Move block
		std                     				; Dec ESI and EDI
		rep	movsb  		            	    		; Move the block
		cld

		
		; Update each block addr 
		; ----------------------
		
		mov	ecx, [DDIR_START]
		mov	ebx, ecx
		add	ecx, [DDIR_SIZE]
		add	ebx, [DDIR_SIZE_INIT]
		mov	eax, [edx.ddir_blksize]				; Size of the block deleted
		mov	esi, [edx.ddir_blkaddr]				; Addr block deleted
		sub	ebx, size ddirentry
@@UpdtBlkAddr:	

		add	ebx, size ddirentry
		cmp	ebx, ecx
		je	@@UpdtBA2

		cmp	[ebx.ddir_blkaddr], esi
		jae	@@UpdtBlkAddr

		add	[ebx.ddir_blkaddr], eax
		jmp	@@UpdtBlkAddr
@@UpdtBA2:

@@DontMove:	
		mov	eax, [edx.ddir_blksize]				; Size of the block deleted


		; Mark the record as reusable
		; ---------------------------
		
		mov	eax, [DDIR_FFREED]
		mov	ebx, edx
		sub	ebx, [DDIR_START]
		mov	[edx.ddir_NextFreed], eax
		mov	[DDIR_FFREED], ebx


		; Update the amount of memory which can be salvaged by GARBAGE collector
		; ----------------------------------------------------------------------
		
		mov	eax, [edx.ddir_blksize]
		sub	[DDIR_MEMSALV], eax

		
		; Block is no more marked to delete
		; ---------------------------------

		and	[edx.ddir_flags], (not QDATA_MARKDEL)
		or	[edx.ddir_flags], QDATA_DELDONE
		
		
		; Update the amount of memory used
		; --------------------------------
		
		sub	[DDA_SIZE], eax


		; Update memory allocated bounds
		; ------------------------------
		
		mov	ebx, [DDA_START]
		sub	ebx, [DDA_SIZE]
		mov	[DDA_BOUND], ebx


		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	ecx ebx eax edi esi
		ret

DustData	endp

	
; ===============================================================
; = DUST							=
; =                                                 		=
; = Code for the forth word:	DUST				=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
DUST		proc
	
		; ---------
		; Dust data
		; ---------
		
		cmp	[DDIR_MEMSALV], 0			; Memory which can be recovered with GARBAGE (see: DataSalvage)
		je	@@DataDone
		
		; Scan each record
		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE]
@@Data1:	
		mov	eax, [DDIR_START]
		add	eax, [DDIR_SIZE_INIT]
		cmp	edx, eax
		je	@@DataDone
		sub	edx, size ddirentry
		test	[edx.ddir_flags], QDATA_MARKDEL
		jz	@@Data1
		call	DustData
		jne	@@Data1
@@DataDone:


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_DUST:
DUST		endp


; ===============================================================
; = DUSTTHRESHOLD                     				=
; =                                                         	=
; = Code for the forth word:    DustThreshold			=
; =                                                         	=
; = Stack:  ( --- adr )                 			=
; ===============================================================

		align	4
DUSTTHRESHOLD	proc

		; Integer run-time
		DOVAR  	D_DUSTTHRESHOLD

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_DUSTTHRESHOLD:
DUSTTHRESHOLD	endp


; ===============================================================
; = UnAllot							=
; =                                                 		=
; = Usage							=
; =	Free memory allocated for a word with ALLOT.		=
; =								=
; = I) EDX : memory block nb					=
; =								=
; = M) none							=
; ===============================================================

		align	4
UnAllot		proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax


		; ----------------------------------------
		; Is-it the last block ? If yes, delete it
		; ----------------------------------------

		mov	eax, [DDIR_SIZE]
		sub	eax, size ddirentry
		cmp	eax, edx
		jne	@@NotLast
		
		; Update the memory used
		mov	eax, [edi + edx.ddir_blksize]
		sub	[DDA_SIZE], eax
		
		; Update the size of the table
		sub	[DDIR_SIZE], size ddirentry
		
		; Update memory allocated bounds
		mov	eax, [DDA_START]
		sub	eax, [DDA_SIZE]
		mov	[DDA_BOUND], eax

		; Done
		jmp	@@Exit
		
@@NotLast:


		; ----------------------------------------------------------------------
		; Update the amount of memory which can be salvaged by GARBAGE collector
		; ----------------------------------------------------------------------
		
		test	[edi + edx.ddir_flags], QDATA_MARKDEL
		jnz	@@UpdtSalvag
		mov	eax, [edi + edx.ddir_blksize]
		add	[DDIR_MEMSALV], eax
@@UpdtSalvag:	
		

		; ---------------------------------------------------------------------
		; Mark this block as free, but not deleted (it will be deleted by DUST)
		; ---------------------------------------------------------------------

		; Mark the block as to be deleted
		or	[edi + edx.ddir_flags], QDATA_MARKDEL

		
		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	eax
		ret

E_UnAllot:
UnAllot		endp


code32		ends
		end
