; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	MEMORY.ASM						!
; 									!
; Description:	Definitions for playing with the memory.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	BMOVE		BMOVE						!
;	PBMOVE		(BMOVE)						!
;	FILL		FILL						!
;	LOGICAL		LOGICAL						!
;	ILOGICAL	ILOGICAL					!
;	BOOL_FALSESTORE	FALSE!						!
;	BOOL_TRUESTORE	TRUE!						!
;	BOOL_QFALSE	FALSE?						!
;	BOOL_QTRUE	TRUE?						!
;	INTEGER		INT		INTEGER		VARIABLE	!
;	IINTEGER	IINT		IINTEGER			!
;	CONST		CONST		CONSTANT			!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions

		
		; Used in: ALLOT and DOES
		extrn	InterpretAddr:	proc		; See INTERPRET


; ===============================================================
; = Malloc							=
; =                                     	            	=
; = Usage:	Allocate memory into the Data Dictionnary,	=
; =		via DDIR indirection (Data Directory Area).	=
; =                       		                       	=
; = I)		EAX	memory amount asked (in bytes)		=
; =		EBX	0: static memory			=
; =			1: local memory				=
; =			2: dynamic memory			=
; =                                 	                	=
; = O)		EAX	the data block reference		=
; =		EBX	the real data address			=
; =                                         	        	=
; = M)		ECX EDX						=
; ===============================================================

		align	4
Malloc		proc

		; Is there enought memory ?
		; -------------------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax
		cmp	ecx, edx
		jge	@@MemOk

		; Out of memory !
		mov	ah, ERR_NOMEMORY		; Out of memory.
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError
		
@@MemOk:

		; Add a new record into DDIR
		; --------------------------

		push	ebx
		mov	ebx, [DDIR_FFREED]
		cmp	ebx, -1
		je	@@Rec2

		; A record is freed, take it
		add	ebx, [DDIR_START]
		mov	ecx, [ebx.ddir_NextFreed]
		mov	[DDIR_FFREED], ecx
		jmp	@@RecDone
		
		; No record freed, add one (if possible...)
@@Rec2:		mov	ebx, [DDIR_SIZE]
		cmp	ebx, DDIR_WSIZE
		jl	@@RecNbOk

		; DDIR table full !
		pop	ebx				; Clean stack
		mov	ah, ERR_DDIRFULL		; "Data allocation table full."
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError
		
@@RecNbOk:	add	ebx, [DDIR_START]		; EBX points to the new record
		
		; Update DDIR size
		add	[DDIR_SIZE], size ddirentry

@@RecDone:

		
		; Get asked memory
		; ----------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax

		; Update memory pointers
		add	[DDA_SIZE], eax

		; Update memory allocated bounds
		mov	edx, [DDA_START]
		sub	edx, [DDA_SIZE]
		mov	[DDA_BOUND], edx


		; Fill the DDIR entry
		; -------------------
		
		; The record is not free
		mov	[ebx.ddir_NextFreed], -1
		
		; Block size
		mov	[ebx.ddir_blksize], eax
		
		; Memory address
		mov	[ebx.ddir_blkaddr], ecx

		; Reference to the word which owns this data
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		push	ecx
		mov	ecx, size wordentry
		clr	edx
		idiv	ecx					; eax /= size wordentry
		pop	ecx
		mov	[ebx.ddir_wordnb], ax

		; Thread nb
		mov	eax, [CurrActThread]
		mov	[ebx.ddir_threadnb], al
		
		; Thread nb, local? dynamic?
		clr	al					; Initialize flags
		pop	edx
		cmp	dl, 1
		jne	@@Not1
		or	al, QDATA_AUTO				; Automatic variable
		jmp	@@12
@@Not1:		cmp	dl, 2
		jne	@@12
		or	al, QDATA_DYNA				; Dynamic variable
@@12:		mov	[ebx.ddir_flags], al			; Set flags
		

		; Return to caller
		; ----------------

		; EAX = data block reference
		sub	ebx, [DDIR_START]
		mov	eax, ebx
		
		; EBX = real data address
		mov	ebx, ecx

		; Done				
		ret

Malloc		endp


; ===============================================================
; = ReAlloc							=
; =                                     	            	=
; = Usage:	Re-Allocate memory into the Data Dictionnary	=
; =		via DDIR indirection (Data Directory Area).	=
; =                       		                       	=
; = I)		EAX	memory amount asked (in bytes)		=
; =		EDX	data block nb				=
; =                                 	                	=
; = O)		EAX	the data block reference		=
; =		EBX	the real data address			=
; =                                         	        	=
; = M)		ECX EDX						=
; ===============================================================

		align	4
ReAlloc		proc

		; Free the data used
		; ------------------

		push	edx					; Save data block nb (for future use)
		call	UnAllot
		

		; Is there enought memory ?
		; -------------------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax
		cmp	ecx, edx
		jge	@@MemOk

		; Out of memory !
		add	esp, 1 * @WordSize			; Clean stack
		mov	ah, ERR_NOMEMORY			; Out of memory.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		jmp	doError
		
@@MemOk:


		; Add a new record into DDIR
		; --------------------------

		mov	ebx, [DDIR_FFREED]
		cmp	ebx, -1
		je	@@Rec2

		; A record is freed, take it
		add	ebx, [DDIR_START]
		mov	ecx, [ebx.ddir_NextFreed]
		mov	[DDIR_FFREED], ecx
		jmp	@@RecDone
		
		; No record freed, add one (if possible...)
@@Rec2:		mov	ebx, [DDIR_SIZE]
		cmp	ebx, DDIR_WSIZE
		jl	@@RecNbOk

		; DDIR table full !
		add	esp, 1 * @WordSize		; Clean stack
		mov	ah, ERR_DDIRFULL		; "Data allocation table full."
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		jmp	doError
		
@@RecNbOk:	add	ebx, [DDIR_START]		; EBX points to the new record
		
		; Update DDIR size
		add	[DDIR_SIZE], size ddirentry

@@RecDone:
		
		
		; Get asked memory
		; ----------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax

		; Update memory pointers
		add	[DDA_SIZE], eax

		; Update memory allocated bounds
		mov	edx, [DDA_START]
		sub	edx, [DDA_SIZE]
		mov	[DDA_BOUND], edx


		; Fill the DDIR entry
		; -------------------
		
		; The record is not free
		mov	[ebx.ddir_NextFreed], -1
		
		; Block size
		mov	[ebx.ddir_blksize], eax
		
		; Memory address
		mov	[ebx.ddir_blkaddr], ecx

		; Reference to the word which owns this data
		pop	edx					; Get saved previous record used
		mov	ax, [edi + edx.ddir_wordnb]
		mov	[ebx.ddir_wordnb], ax
		
		; Thread nb
		mov	eax, [CurrActThread]
		mov	[ebx.ddir_threadnb], al
		
		; Thread nb, local? dynamic?
		movzx	eax, [edi + edx.ddir_flags]		; Get previous flags
		and	al, (not QDATA_MARKDEL)			; Record not marked as deleted
		mov	[ebx.ddir_flags], al
		

		; Return to caller
		; ----------------

		; EAX = data block reference
		sub	ebx, [DDIR_START]
		mov	eax, ebx
		
		; EBX = real data address
		mov	ebx, ecx

		; Done				
		ret

ReAlloc		endp


; ===============================================================
; = BWD_ALLOT							=
; =                                                 		=
; = Code for the forth word:	RT_(BW)ALLOT			=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Allocate (n * [1,2,4]) bytes in the Data	=
; =		Dictionnary.					=
; = O) EAX : memory block nb (DDIR table).			=
; ===============================================================

BWD_ALLOT	MACRO	sz

		align	4

		if	sz eq 'B'
RT_BALLOT	proc
		elseif	sz eq 'W'
RT_WALLOT	proc
		else
RT_ALLOT	proc
		endif

		; ----------------------
		; CTRL-BREAK not allowed
		; ----------------------

		inc	[NoCtrlBreak]


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ---------------------------------
		; Dynamic allocation ? (with ALLOC)
		; ---------------------------------
		
		cmp	[AllocFlag], TRUE
		je	@@StaticAll


		; ----------------------------------------------------------------
		; Automatic allocation ? No if Compile Actions called by INTERPRET
		; ----------------------------------------------------------------

		lea	eax, InterpretAddr
		add	eax, [OfsDictReloc]
		cmp	dword ptr [esp + 4], eax
		je	@@StaticAll

		; Amount of memory asked
		mov	eax, PARAM0
		if	sz eq 'W'
		shl	eax, 1					; * 2
		elseif	sz eq 'D'
		shl	eax, 2					; * 4
		endif

		; Is memory already allocated for this auto variable ? If yes, release this memory before
		mov	edx, [esp + 4]
		mov	edx, [edx]
		cmp	edx, -1
		je	@@FirstAlloc


		; Re-allocation of memory
		; -----------------------

		call	ReAlloc
		jmp	@@MemDone
		

		; Get memory asked (first allocation)
		; -----------------------------------
		
@@FirstAlloc:	mov	bl, 1					; 1 mean Local variable
		call	Malloc					; Get memory


		; Store the data block address
@@MemDone:	mov	PARAM0, ebx 				; Push block address on stack
		mov	ecx, [esp + 4]				; Addr of cell where we have to store DataBlockAddr
		mov	[ecx], eax

		; Done for dynamic allocation
		jmp	@@Exit					; ret

		
		; -------------------------------------------
		; Static allocation (not inside a definition)
		; -------------------------------------------
		
@@StaticAll:

		; Compile: CALL $
		CALL_rel32
		clr	eax
		call	CodeComma               		; Code,

		; Compile: ADD DWORD PTR [ESP], 4+6+3+2+1
		ADD_xESPi8_i8	0, 4+6+3+2+1

		; Get memory asked (returned: <n>, block memory nb)
		; then compile: MOV EAX, [EDI.ddir_blkaddr + <n>]
		MOV_EAX_xEDIi32
		mov	eax, PARAM0
		if	sz eq 'W'
		shl	eax, 1					; * 2
		elseif	sz eq 'D'
		shl	eax, 2					; * 4
		endif
		cmp	[AllocFlag], TRUE
		je	@@DynaVar
		clr	bl					; 0 means static  variable
		jmp	@@malloc
@@DynaVar:	mov	bl, 2					; 2 means dynamic variable
@@malloc:	call	Malloc					; Get memory
		mov	PARAM0, ebx 				; Push block address on stack
		add	eax, ddir_blkaddr
		call	CodeComma				; Code,
		
		; Compile: SUB ESI, 4
		SUB_ESI_i8	@WordSize
		
		; Compile: MOV [ESI], eax
		MOV_xESIi8_EAX
		
		; Compile: RET
		_RET_

		; Record the addr of the block, which can be used by ALLOTMORE (if static allocation)
		cmp	[AllocFlag], TRUE
		mov	[REF_BLK_ALLCTED], eax

		; On output, EAX is block nb
		sub	eax, ddir_blkaddr
		

		; ------------------------------------------------------
		; Return to caller, with real block address on top stack
		; ------------------------------------------------------
		
@@Exit:		
		; CTRL-BREAK allowed
		call	CTRLBREAK_REST

		ret

		if	sz eq 'B'
E_RT_BALLOT:
		elseif	sz eq 'W'
E_RT_WALLOT:
		else
E_RT_ALLOT:
		endif
		endp


; ===============================================================
; = ALLOT							=
; =                                                 		=
; = Code for the forth word:	ALLOT				=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Allocate (n * [1, 2, 4]) bytes in the Data 	=
; =		Dictionnary.					=
; ===============================================================

		align	4

		if	sz eq 'B'
BALLOT		proc
		elseif	sz eq 'W'
WALLOT		proc
		else
ALLOT		proc
		endif

		; 'BUILD>' not called before ?
		cmp	[CDA_BUILD_CALLED], FALSE
		jne	@@Ok

		; 'BUILD>' not called beforehand.
		mov	ah, ERR_BUILDNOTCALLED			; "'BUILD>' not called beforehand."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok:

		; DOES> already called ?
		cmp	[CDA_DOES_CALLED], FALSE
		je	@@Ok2
		
		; 'DOES>' already called
		mov	ah, ERR_DOESALCALLED			; "'DOES>' already called."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok2:		

		; ALLOT already called ?
		cmp	[CDA_ALLOT_CALLED], FALSE
		je	@@Ok3
		
		; 'ALLOT' already called
		mov	ah, ERR_ALLOTALCALLED			; "'ALLOT' already called."
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok3:		mov	[CDA_ALLOT_CALLED], TRUE

		; Put 'CALL_rel32'
		if	sz eq 'B'
			compCALL_rel32	RT_BALLOT
		elseif	sz eq 'W'
			compCALL_rel32	RT_WALLOT
		else
			compCALL_rel32	RT_ALLOT
		endif
		
		; Return to caller
		ret

		if	sz eq 'B'
E_BALLOT:
		elseif	sz eq 'W'
E_WALLOT:
		else
E_ALLOT:
		endif
		endp

		endm


BWD_ALLOT	'B'
BWD_ALLOT	'W'
BWD_ALLOT	'D'


; ===============================================================
; = BWD_ALLOTMORE						=
; =                                                 		=
; = Code for the forth words:	ALLOTMORE			=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Re-Allocate (n * [1,2,4]) bytes in the Data 	=
; =		Dictionnary.					=
; ===============================================================

BWD_ALLOTMORE	MACRO	sz

		align	4

		if	sz eq 'B'
RT_BALLOTMORE	proc
		elseif	sz eq 'W'
RT_WALLOTMORE	proc
		else
RT_ALLOTMORE	proc
		endif

		; CTRL-BREAK not allowed
		; ----------------------
		inc	[NoCtrlBreak]


		; Get last block allocated by ALLOT
		cmp	[AllocFlag], TRUE
		je	@@NOK
		mov	ebx, [REF_BLK_ALLCTED]
		cmp	ebx, -1
		jne	@@OK

		; "'ALLOTMORE' can be used only after ALLOT (non dynamic)."
@@NOK:		mov	ah, ERR_ALLOTMOREBAD			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@OK:		sub	ebx, ddir_blkaddr
		add	ebx, [DDIR_START]

		; Get size to add
		mov	eax, PARAM0
		if	sz eq 'W'
		shl	eax, 1					; * 2
		elseif	sz eq 'D'
		shl	eax, 2					; * 4
		endif


		; Get asked memory
		; ----------------

		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, eax
		cmp	ecx, edx
		jge	@@MemOk

		; Out of memory !
		mov	ah, ERR_NOMEMORY			; Out of memory.
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@MemOk:

		; Update memory pointers
		add	[DDA_SIZE], eax

		; Update memory allocated bounds
		mov	edx, [DDA_START]
		sub	edx, [DDA_SIZE]
		mov	[DDA_BOUND], edx

		; Move data
		push	esi edi
		mov	edi, PARAM0				; Size of mem asked
		if	sz eq 'W'
		shl	edi, 1					; * 2
		elseif	sz eq 'D'
		shl	edi, 2					; * 4
		endif
		mov	esi, [ebx.ddir_blkaddr]			; Source address
		neg	edi
		add	edi, esi
		mov	ecx, [ebx.ddir_blksize]			; Size of block to move
		cld                    				; Inc ESI and EDI
		rep	movsb  		            		; Move the block
		pop	edi esi

		
		; Update the DDIR entry
		; ---------------------
		
		sub	[ebx.ddir_blkaddr], eax
		add	[ebx.ddir_blksize], eax

		
		; Push on stack begin of new block allocated
		; ------------------------------------------
		
		mov	ecx, [ebx.ddir_blkaddr]
		add	ecx, [ebx.ddir_blksize]
		sub	ecx, eax
		mov	PARAM0, ecx
	

		; CTRL-BREAK allowed
		; ------------------
		
		call	CTRLBREAK_REST
		

		; Return to caller
		; ----------------
		
		ret

		if	sz eq 'B'
E_RT_BALLOTMORE:
		elseif	sz eq 'W'
E_RT_WALLOTMORE:
		else
E_RT_ALLOTMORE:
		endif
		endp


; ===============================================================
; = BWD_ALLOTMORE						=
; =                                                 		=
; = Code for the forth word:	(BW)ALLOTMORE			=
; =								=
; = Stack:	( n --- addr )					=
; =								=
; = Usage:	Re-Allocate (n * [1,2,4]) bytes in the Data 	=
; =		Dictionnary.					=
; ===============================================================

		align	4

		if	sz eq 'B'
BALLOTMORE	proc
		elseif	sz eq 'W'
WALLOTMORE	proc
		else
ALLOTMORE	proc
		endif


		; Verify if the system is in compile state
		call	QCOMPILE


		; Put 'CALL_rel32'
		if	sz eq 'B'
			compCALL_rel32	RT_BALLOTMORE
		elseif	sz eq 'W'
			compCALL_rel32	RT_WALLOTMORE
		else
			compCALL_rel32	RT_ALLOTMORE
		endif
		
		
		; Return to caller
		ret

		if	sz eq 'B'
E_BALLOTMORE:
		elseif	sz eq 'W'
E_WALLOTMORE:
		else
E_ALLOTMORE:
		endif
		endp


		endm


BWD_ALLOTMORE	'B'
BWD_ALLOTMORE	'W'
BWD_ALLOTMORE	'D'


; ===============================================================
; = BMOVE							=
; =								=
; = Code for the forth word:	BMOVE				=
; =                     	                            	=
; = Stack:	( dest src size --- )				=
; ===============================================================

		align	4
BMOVE		proc

		; Verify stack frame
		VSTACK	3

		; Save regs used
		push	esi edi

		; Get data
		mov	edi, PARAM2			; Destination address
		mov	ecx, PARAM0			; Size of the block to move
		mov	esi, PARAM1			; Source address

		; Check dest addr against bound
		bound	edi, qword ptr DDA_BOUND

		; Choose direction
		cmp	esi, edi
		ja	@@great
		std
;		mov	eax, 1
		lea	ebx, [edi - ecx + 1]		
		lea	esi, [esi + ecx - 1]
		lea	edi, [edi + ecx - 1]
		jmp	@@greatlow
@@great:		cld
;		clr	eax
		lea	ebx, [edi + ecx - 1]
@@greatlow:	bound	ebx, qword ptr DDA_BOUND

;		; Align on multiple of 4
;@@align4:	test	edi, 3
;		jz	@@isAlign4
;		jecxz	@@done
;		movsb
;		dec	ecx
;		jmp	@@align4
;@@isAlign4:

		; Move data
;		shr	ecx, 2
		rep	movsb				; Move the block

		; Restore regs used
@@done:		pop	edi esi

		; Return to caller
		cld
		add	esi, 3 * @WordSize		; Free parameters from the stack
		ret

E_BMOVE:
BMOVE		endp


; ===============================================================
; = PBMOVE							=
; =								=
; = Code for the forth word:	(BMOVE)				=
; =                     	                            	=
; = Stack:	( dest src size --- )				=
; ===============================================================

		align	4
PBMOVE		proc

		; Verify stack frame
		VSTACK	3

		; Save regs used
		push	esi edi

		; Get data
		mov	edi, PARAM2			; Destination address
		mov	ecx, PARAM0			; Size of the block to move
		mov	esi, PARAM1			; Source address

		; Choose direction
		cmp	esi, edi
		ja	@@great
		std
;		mov	eax, 1
		lea	ebx, [edi - ecx + 1]		
		lea	esi, [esi + ecx - 1]
		lea	edi, [edi + ecx - 1]
		jmp	@@greatlow
@@great:		cld
;		clr	eax
		lea	ebx, [edi + ecx - 1]
@@greatlow:	

;		; Align on multiple of 4
;@@align4:	test	edi, 3
;		jz	@@isAlign4
;		jecxz	@@done
;		movsb
;		dec	ecx
;		jmp	@@align4
;@@isAlign4:

		; Move data
;		shr	ecx, 2
		rep	movsb				; Move the block

		; Restore regs used
@@done:		pop	edi esi

		; Return to caller
		cld
		add	esi, 3 * @WordSize		; Free parameters from the stack
		ret

E_PBMOVE:
PBMOVE		endp


; ===============================================================
; = FILL							=
; =								=
; = Code for the forth word:	FILL				=
; =                     	                            	=
; = Stack:	( addr u c --- )				=
; ===============================================================

		align	4
FILL		proc

		; Verify stack frame
		VSTACK	3

		; Get data
		mov	edx, PARAM2			; Destination address
		mov	ecx, PARAM1			; Size of the block to fill
		mov	eax, PARAM0			; Value to store

		; Check dest addr against bound (start)
		bound	edx, qword ptr DDA_BOUND

		; Check dest addr against bound (start)
		lea	ebx, [edx + ecx - 1]
		bound	ebx, qword ptr DDA_BOUND

		; Fill data
		push	edi
		mov	edi, edx
		cld
		rep	stosb				; Fill the block (byte)
		pop	edi

		; Return to caller
		DISCARD	3				; Free parameters from the stack
		ret

E_FILL:
FILL		endp


; ===============================================================
; = LOGICAL							=
; =                                                 		=
; = Code for the forth word:	LOGICAL				=
; =								=
; = Stack:	( --- )						=
; ==============================================================

		align	4
BOOLEAN		proc

		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------
		
		; 1 ALLOT DROP
		IPUSH	1
		call	RT_ALLOT
		DISCARD	1				; DROP

		; DOES>		(this ends the compilation sequence)
		call	PDOES
		dd	R_BOOLEAN

		
		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_BOOLEAN:

		; Return to caller
		ret

E_BOOLEAN:
BOOLEAN		endp


; ===============================================================
; = ILOGICAL							=
; =                                                 		=
; = Code for the forth word:	ILOGICAL			=
; =								=
; = Stack:	( f --- )					=
; ===============================================================

		align	4
IBOOLEAN	proc

		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------
		
		; 1 ALLOT !
		IPUSH	1
		call	RT_ALLOT		; 1 ALLOT
		DO_STORE			; !

		; DOES>		(this ends the compilation sequence)
		call	PDOES
		dd	R_IBOOLEAN

		
		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_IBOOLEAN:

		; Return to caller
		ret

E_IBOOLEAN:
IBOOLEAN	endp


; ===============================================================
; = BOOL_FALSESTORE						=
; =                                                 		=
; = Code for the forth word:	FALSE!				=
; =								=
; = Stack:	( addr --- )					=
; ===============================================================

		align	4
BOOL_FALSESTORE	proc

		; Verify stack frame
		VSTACK	1

		; Get addr
		mov	eax, PARAM0				; Pop adr

		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND

		; Store value
		mov	byte ptr [eax], FALSE
		DISCARD	1

		; Return to caller
		ret

E_BOOL_FALSESTORE:
BOOL_FALSESTORE	endp


; ===============================================================
; = BOOL_TRUESTORE						=
; =                                                 		=
; = Code for the forth word:	TRUE!				=
; =								=
; = Stack:	( addr --- )					=
; ===============================================================

		align	4
BOOL_TRUESTORE	proc

		; Verify stack frame
		VSTACK	1

		; Get addr
		mov	eax, PARAM0				; Pop adr

		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND

		; Store value
		mov	byte ptr [eax], TRUE
		DISCARD	1

		; Return to caller
		ret

E_BOOL_TRUESTORE:
BOOL_TRUESTORE	endp


; ===============================================================
; = BOOL_QFALSE							=
; =                                                 		=
; = Code for the forth word:	FALSE?				=
; =								=
; = Stack:	( addr --- f )					=
; ===============================================================

		align	4
BOOL_QFALSE	proc

		; Verify stack frame
		VSTACK	1

		; Get addr
		mov	eax, PARAM0
		
		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND
		
		; Read value then push
		cmp	byte ptr [eax], FALSE
		je	@@false
@@true:		mov	PARAM0, FALSE
		ret
@@false:	mov	PARAM0, TRUE

		; Return to caller
		ret

E_BOOL_QFALSE:
BOOL_QFALSE	endp


; ===============================================================
; = BOOL_QTRUE							=
; =                                                 		=
; = Code for the forth word:	TRUE?				=
; =								=
; = Stack:	( addr --- f )					=
; ===============================================================

		align	4
BOOL_QTRUE	proc

		; Verify stack frame
		VSTACK	1

		; Get addr
		mov	eax, PARAM0
		
		; Check addr against bound
		bound	eax, qword ptr DDA_BOUND
		
		; Read value then push
		cmp	byte ptr [eax], FALSE
		je	@@false
@@true:		mov	PARAM0, TRUE
		ret
@@false:	mov	PARAM0, FALSE

		; Return to caller
		ret

E_BOOL_QTRUE:
BOOL_QTRUE	endp


; ===============================================================
; = INTEGER							=
; =                                                 		=
; = Code for the forth word:	INT				=
; =								=
; = Stack:	( --- )						=
; =								=
; = Input stream:	INTEGER var				=
; ===============================================================

		align	4
INTEGER		proc

		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------

		; 1 ALLOT DROP
		IPUSH	1
		call	RT_ALLOT
		DISCARD	1

		; DOES>		(this ends the compilation sequence)
		call	PDOES
		dd	R_INTEGER

		
		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_INTEGER:	

		; Return to caller
		ret

E_INTEGER:
INTEGER		endp


; ===============================================================
; = IINTEGER							=
; =                                                 		=
; = Code for the forth word:	IINT				=
; =								=
; = Stack:	( n --- )					=
; =								=
; = Input stream:	n IINT var				=
; ===============================================================

		align	4
IINTEGER	proc

		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------

		; Verify stack frame
		VSTACK	1

		; 1 ALLOT !
		IPUSH	1
		call	RT_ALLOT
		DO_STORE					; !
		
		; DOES>		(this ends the compilation sequence)
		call	PDOES
		dd	R_IINTEGER

		
		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_IINTEGER:	

		; Return to caller
		ret


E_IINTEGER:
IINTEGER	endp


; ===============================================================
; = CONST							=
; =                                                 		=
; = Code for the forth word:	CONST				=
; =								=
; = Stack:	( n --- )					=
; =								=
; = Input stream:	n CONST var				=
; ===============================================================

		align	4
CONSTANT	proc

		; ---------------------------------------
		; Define the compilation actions sequence
		; ---------------------------------------

		; Verify stack frame
		VSTACK	1

		; 1 ALLOT !
		IPUSH	1
		call	RT_ALLOT
		DO_STORE					; !
		
		; DOES>		(this ends the compilation sequence)
		call	PDOES
		dd	R_CONSTANT


		; ------------------------------------
		; Define the run-time actions sequence
		; ------------------------------------

RT_CONSTANT:

		; @
		mov	eax, PARAM0
		mov	eax, [eax]
		mov	PARAM0, eax
		
		; Return to caller
		ret

E_CONSTANT:
CONSTANT	endp


; ===============================================================
; = FreeLocalVar						=
; =                                                 		=
; = Usage							=
; =	Free memory allocated for a word with ALLOT.		=
; ===============================================================

		align	4
FreeLocalVar	proc

		; --------------------------
		; Free each dynamic variable
		; --------------------------

		push	esi		
		mov	eax, [esp + 1*4]		; Return addr is addr of cell where is stored nb of dyn var
		mov	ecx, eax
		add	ecx, @WordSize			; Ptr do start of list
		mov	eax, [eax]			; Nb of dynamic variables to free
		clr	ebx				; Dyn var counter
@@Free:		mov	esi, [ecx]			; Memory block nb
		mov	edx, [esi + ecx]
		cmp	edx, -1
		je	@@Free2				; Memory never allocated ('integer xxx' not executed for ex.)
		call	UnAllot
		mov	dword ptr [esi + ecx], -1	; Memory is freed
@@Free2:	add	ecx, @WordSize			; Next cell
		inc	ebx
		cmp	bl, al
		jne	@@Free
		pop	esi
		
		
		; ----------------------
		; Skip data in code area
		; ----------------------
		
		mov	[esp], ecx


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_FreeLocalVar:
FreeLocalVar	endp


; ===============================================================
; = FreeData							=
; =                                                 		=
; = Usage							=
; =	Free memory allocated for a word with ALLOT. Is called	=
; =	only when the word is forgetten.			=
; =								=
; = I) EDX : word NIA address					=
; =								=
; = M) none							=
; ===============================================================

		align	4
FreeData	proc

		; -------------------
		; Save used registers
		; -------------------
		
		push	eax ebx ecx edx esi
		

		; -----------
		; Get word nb
		; -----------
		
		sub	edx, [NIA_START]

		
		; -----------------------------------------------
		; Found each memory block allocated for this word
		; -----------------------------------------------

		; Find last record freed (linked list)
		mov	esi, [DDIR_FFREED]
		cmp	esi, -1
		je	@@LL2
@@LL1:		add	esi, [DDIR_START]
		cmp	[esi.ddir_NextFreed], -1
		je	@@LL3
		mov	esi, [esi.ddir_NextFreed]
		jmp	@@LL1
@@LL2:		clr	esi
@@LL3:		

		; Initializations		
		mov	ecx, edi				; EDI = [DDIR_START]
		mov	ebx, ecx
		add	ebx, [DDIR_SIZE]
		
		; End of table ?
@@TestBlock:	cmp	ecx, ebx
		je	@@Done
		
		; First record freed ?
		cmp	ecx, esi
		je	@@NextBlock

		; Record freed ?
		cmp	[ecx.ddir_NextFreed], -1
		jne	@@NextBlock
		
		; Test this memory block
		movzx	eax, [ecx.ddir_wordnb]
		imul	eax, size wordentry
		cmp	eax, edx
		jne	@@NextBlock
	
		; Free this block
		push	edx
		mov	edx, ecx
		sub	edx, edi				; EDI = [DDIR_START]
		call	UnAllot
		pop	edx
		
		; Next ddir block to test
@@NextBlock:	add	ecx, size ddirentry
		jmp	@@TestBlock
		
@@Done:

		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	esi edx ecx ebx eax
		ret

FreeData	endp


; ===============================================================
; = DiscardLocalVars						=
; =                                                 		=
; = Usage							=
; =	Discard all local variables. Called by 'doError'.	=
; =								=
; = I) none							=
; = O) none							=
; = M) none							=
; ===============================================================

		align	4
DiscardLocalVars proc

		; --------------------------------
		; Scan each memory block allocated
		; --------------------------------

		; Find last record freed (linked list)
		push	ebx ecx edx
		mov	ebx, [DDIR_FFREED]
		cmp	ebx, -1
		je	@@LL2
@@LL1:		add	ebx, [DDIR_START]
		cmp	[ebx.ddir_NextFreed], -1
		je	@@LL3
		mov	ebx, [ebx.ddir_NextFreed]
		jmp	@@LL1
@@LL2:		clr	ebx
@@LL3:		

		; Initializations		
		mov	ecx, [DDIR_START]
		add	ecx, [DDIR_SIZE]
		
		; Next ddir block to test
		; End of table ?
@@NextBlock:	
		sub	ecx, size ddirentry
@@NextBlock2:	cmp	ecx, [DDIR_START]
		je	@@Done

		; First record freed ?
		cmp	ecx, ebx
		je	@@NextBlock

		; Record freed ?
		cmp	[ecx.ddir_NextFreed], -1
		jne	@@NextBlock
		
		; Test this memory block: is-it an automatic variable ?
		test	[ecx.ddir_flags], QDATA_AUTO
		jz	@@NextBlock

		; Free this block
		mov	edx, ecx
		sub	edx, [DDIR_START]
		call	UnAllot
		
		; Next ddir block to test
		jmp	@@NextBlock
		
@@Done:

		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	edx ecx ebx
		ret

DiscardLocalVars endp


; ===============================================================
; = DiscardThreadLVars						=
; =                                                 		=
; = Usage							=
; =	Discard all local variables, for a given thread.	=
; =	Called by 'ENDTHREAD'					=
; =								=
; = I) EAX : thread nb						=
; = O) none							=
; = M) EBX ECX EDX						=
; ===============================================================

		align	4
DiscardThreadLVar proc

		; --------------------------------
		; Scan each memory block allocated
		; --------------------------------

		; Find last record freed (linked list)
		push	ebx ecx edx
		mov	ebx, [DDIR_FFREED]
		cmp	ebx, -1
		je	@@LL2
@@LL1:		add	ebx, [DDIR_START]
		cmp	[ebx.ddir_NextFreed], -1
		je	@@LL3
		mov	ebx, [ebx.ddir_NextFreed]
		jmp	@@LL1
@@LL2:		clr	ebx
@@LL3:		

		; Initializations		
		mov	ecx, edi					; EDI = [DDIR_START]
		add	ecx, [DDIR_SIZE]
		
		; End of table ?
@@TestBlock:	cmp	ecx, edi
		je	@@Done
		sub	ecx, size ddirentry

		; First record freed ?
		cmp	ecx, ebx
		je	@@NextBlock

		; Record freed ?
		cmp	[ecx.ddir_NextFreed], -1
		jne	@@NextBlock
		
		; Test this memory block
		test	[ecx.ddir_flags], QDATA_AUTO
		jz	@@NextBlock				; No automatic variable
		cmp	al, [ecx.ddir_threadnb]			; Thread nb
		jne	@@NextBlock
	
		; Free this block
		mov	edx, ecx
		sub	edx, edi					; EDI = [DDIR_START]
		call	UnAllot
		
		; Next ddir block to test
@@NextBlock:	jmp	@@TestBlock
		
@@Done:

		; ----------------
		; Return to caller
		; ----------------

@@Exit:		pop	edx ecx ebx
		ret

DiscardThreadLVar endp


code32		ends
		end
