; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	SAVEEXE.ASM						!
; 									!
; Description:	Definitions for playing with building turkey		!
;		application.						!
; 									!
; List of FROTH definitions defined in this module:			!
;	SAVE_EXE	SAVE.EXE					!
;	QTURNKEY	?TURNKEY					!
;	QFROTHTURNKEY	?FROTHTURNKEY					!
;	NBCMDARGS	#CMDARGS					!
;	CMDARGS		CMDARGS$					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc					; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; ===============================================================
; = SAVE_EXE							=
; =								=
; = Code for the forth word:	SAVE_EXE			=
; =								=
; = Stack:	( addrz --- n )					=
; ===============================================================

		align	4
SAVE_EXE	proc

		; -----------------------------------------
		; Save input parameters into Temp variables
		; -----------------------------------------

t_FileHdl	equ	word ptr [ebp - 4]
t_FrothExeHdl	equ	word ptr [ebp - 8]
t_FrothExeSize	equ	[ebp - 12]
t_BuffAddr	equ	[ebp - 16]
t_nb		equ	dword ptr [ebp - 20]
		enter	5 * @WordSize, 0


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1

		; Verify if a valid string
		call	QSTRING					; ?String

		
		; -----------------
		; Nothing to save ?
		; -----------------
		
		mov	eax, [CDA_SIZE]
		cmp	eax, [CDA_SIZE_INIT]
		jne	@@Saveable

		; "Nothing to save !"
		mov	ah, ERR_FROTHEXENOUWORDS		; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Saveable:	

		; -----------------------------
		; Garbage collector (NO hole !)
		; -----------------------------
		
		call	DUST


		; ----------------------
		; Try to delete the file
		; ----------------------
		
		mov	ebx, PARAM0		; Get filename
		call	doDeleteFile		; Delete the file

		
		; --------------------
		; Try to open the file
		; --------------------
		
		; Pointer to filename FROTH.EXE
		mov	ebx, PARAM0		; Get filename
		
		; Access mode
		mov	eax, 2091h		; W/O  b13: no int24h   b7: don't pass handle to child   b4: no sharing

		; Authorize new file creation ? (update ZF flag)
		mov	cl, 1
		
		; Call BIOS
		call	doWideOpenFile
		jnc	@@OpenOk
		
		mov	PARAM0, -1		; Error
		jmp	@@CloseFile
		
@@OpenOk:	
		; Save file handle
		mov	[t_FileHdl], bx
		

		; ------------------------
		; Read FROTH.EXE in memory
		; ------------------------

		; Open FROTH.EXE file
		mov	edx, [QuoteStrBuff]
		add	edx, size strentry
		call	GetProgName
		mov	ebx, edx
		mov	eax, 2090h		; R/O  b13: no int24h   b7: don't pass handle to child   b4: no sharing
		clr	cl
		call	doWideOpenFile
		jc	@@CloseFile
		mov	[t_FrothExeHdl], bx

		; Size of FROTH.EXE
		call	doFileSize
		mov	[t_FrothExeSize], ebx

		; Try to find memory for a buffer enought big to hold FROTH.EXE
		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE]
		mov	ecx, [DDA_START]
		sub	ecx, [DDA_SIZE]
		sub	ecx, [t_FrothExeSize]
		cmp	ecx, edx
		jl	@@CloseFile
		mov	[t_BuffAddr], ecx

		; Read entire FROTH.EXE file
		mov	bx, [t_FrothExeHdl]			; File handle
		mov	edx, [t_BuffAddr]			; Buffer address
		mov	ecx, [t_FrothExeSize]			; Bytes to read
		call	doReadFile				; in EBX: bytes read from the file

		; Close FROTH.EXE file
		mov	bx, [t_FrothExeHdl]
		call	doCloseFile


		; ----------------------------------
		; Find mark into buffer, then update
		; ----------------------------------

		; Find mark
		mov	ebx, [t_BuffAddr]
		clr	edx
@@FindMark:	cmp	dword ptr [ebx + edx], 'iSlO'
		jne	@@FindMark2
		cmp	dword ptr [ebx + edx + 4], 'CHPC'
		je	@@FindMark3
@@FindMark2:	inc	edx
		jmp	@@FindMark
@@FindMark3:	

		; Now modify variables
		lea	ebx, [ebx + edx + 8]
		mov	byte ptr [ebx], TRUE			; ExeFile
		
		; Word to execute relative address
		mov	eax, [NIA_START]
		add	eax, [NIA_SIZE]
		sub	eax, size wordentry
		mov	eax, [eax.word_realaddr]
		sub	eax, [CDA_START]
		mov	dword ptr [ebx + 1*4], eax		; ExeWordAddr

		; Adress and Size of user's code into .EXE file
		mov	eax, [t_FrothExeSize]
		mov	dword ptr [ebx + 2*4], eax		; ExeCdaBegin
		mov	eax, [CDA_SIZE]
		sub	eax, [CDA_SIZE_INIT]			; Bytes to write
		mov	dword ptr [ebx + 3*4], eax		; ExeCdaSize

		; Adress of user's data into .EXE file
		add	eax, [t_FrothExeSize]			
		mov	dword ptr [ebx + 4*4], eax		; ExeDdaBegin
		mov	eax, [DDIR_SIZE]
		sub	eax, [DDIR_SIZE_INIT]			; Bytes to write
		mov	dword ptr [ebx + 5*4], eax		; ExeDdaSize


		; -----------------------------		
		; Save FROTH.EXE into .EXE file
		; -----------------------------		
		
		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, [t_FrothExeSize]			; Bytes to write
		mov	edx, [t_BuffAddr]			; Buffer address
		call	doWriteFile
		jc	@@CloseFile


		; ------------------------------
		; Save user defined words (code)
		; ------------------------------

		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, [CDA_SIZE]
		sub	ecx, [CDA_SIZE_INIT]			; Bytes to write
		mov	edx, [CDA_START]
		add	edx, [CDA_SIZE_INIT]			; Buffer address
		call	doWriteFile
		jc	@@CloseFile


		; ------------------------------
		; Save user defined words (data)
		; ------------------------------

		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, [DDIR_SIZE]
		sub	ecx, [DDIR_SIZE_INIT]			; Bytes to write
		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE_INIT]			; Buffer address
		call	doWriteFile
		jc	@@CloseFile


		; ----------------
		; Save static data
		; ----------------

		mov	ebx, [DDIR_START]
		add	ebx, [DDIR_SIZE_INIT]

		mov	edx, [DDIR_START]
		add	edx, [DDIR_SIZE]

@@SaveDDA:	
		; End reached ?
		cmp	ebx, edx
		je	@@SaveDDA3

		; Save only static variables (not auto, not dynamic)
		test	[ebx.ddir_flags], (QDATA_AUTO or QDATA_DYNA)		; Automatic or Dynamic variable ?
		jnz	@@SaveDDA2

		; Save DDIR reference
		pushad
		mov	eax, ebx
		sub	eax, [DDIR_START]
		mov	[t_nb], eax
		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, @WordSize				; Bytes to write
		lea	edx, t_nb				; Buffer address
		call	doWriteFile
		popad

		; Save DDA
		pushad
		mov	ecx, [ebx.ddir_blksize]			; Bytes to write
		mov	edx, [ebx.ddir_blkaddr]			; Buffer address
		mov	bx, [t_FileHdl]				; File handle
		call	doWriteFile
		popad
		jmp	@@SaveDDA2

		; Next variable, if any
@@SaveDDA2:	add	ebx, size ddirentry
		jmp	@@SaveDDA

@@SaveDDA3:	

		; Write final -1
		mov	[t_nb], -1
		mov	bx, [t_FileHdl]				; File handle
		mov	ecx, @WordSize				; Bytes to write
		lea	edx, t_nb				; Buffer address
		call	doWriteFile


		; ---------------------------------
		; Push size of .EXE file onto stack
		; ---------------------------------

		mov	bx, [t_FileHdl]				; File handle
		call	doFilePos
		mov	PARAM0, ebx				; Store file size


		; --------------
		; Close the file
		; --------------
		
@@CloseFile:	mov	bx, [t_FileHdl]
		call	doCloseFile

		
		; ----------------
		; Return to caller
		; ----------------

		leave		
		ret
		
E_SAVE_EXE:
SAVE_EXE	endp


; ===============================================================
; = QTURNKEY							=
; =								=
; = Code for the forth word:	?turnkey			=
; =								=
; = Stack:	( --- f )					=
; ===============================================================

		align	4
QTURNKEY	proc

		; Is-it a turnkey application ?
		DISPOSE	1
		cmp	[ExeFile], TRUE				; Is-ir a turnkey application ?
		je	@@turnkey

		; This is not a turnkey application
		mov	PARAM0, FALSE				; push ff
		jmp	@@exit
		
		; This is a turnkey application
@@turnkey:	mov	PARAM0, TRUE				; push tf

		; Verify stack frame overflow, then return to caller
@@exit:		OVSTACK
		ret

E_QTURNKEY:
QTURNKEY	endp


; ===============================================================
; = QFROTHTURNKEY						=
; =								=
; = Code for the forth word:	?FrothTurnkey			=
; =								=
; = Stack:	( --- f )					=
; ===============================================================

		align	4
QFROTHTURNKEY	proc

		; Push flag
if TURNKEYED eq 0
		IPUSH	FALSE
else
		IPUSH	TRUE
endif

		; Verify stack frame overflow, then return to caller
		OVSTACK
		ret

E_QFROTHTURNKEY:
QFROTHTURNKEY	endp


; ===============================================================
; = NBCMDARGS							=
; =								=
; = Code for the forth word:	NBCMDARGS			=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
NBCMDARGS	proc

		; ----------------------------
		; Find the number of arguments
		; ----------------------------

		; Save vital registers
		push	es

		; Length of the cmd line
		mov	ax, _selzero
		mov	es, ax
		mov	edx, _pspa
		add	edx, 81h
		movzx	ebx, byte ptr es:[edx - 1]		; nb of chars of the cmd line

		; Count the number of arguments
		clr	ecx

		; skip blanks, i.e. goto 1st non blank char
		clr	eax
@@nextword:	dec	eax
@@blank:	inc	eax
		cmp	al, bl
		je	@@finish
		cmp	byte ptr es:[edx + eax], ' '
		je	@@blank

		; one more argument
		inc	ecx

		; skip non blanks, i.e. goto 1st blank char
		dec	eax
@@nonblank:	inc	eax
		cmp	al, bl
		je	@@finish
		cmp	byte ptr es:[edx + eax], ' '
		jne	@@nonblank
		jmp	@@nextword

		; Restore vital registers
@@finish:	pop	es

		; Number of cmd line arguments
		IPUSH	ecx

		; Verify stack frame overflow, then return to caller
@@exit:		OVSTACK
		ret

E_NBCMDARGS:
NBCMDARGS	endp


; ===============================================================
; = CMDARGS							=
; =								=
; = Code for the forth word:	CMDARGS$			=
; =								=
; = Stack:	( n --- addrz )					=
; ===============================================================

		align	4
CMDARGS		proc

		; Verify stack frame
		VSTACK	1

		; arg # is correct ?
		call	NBCMDARGS			; #CMDARGS	( n1 --- n1 n2 )
		call	OVER				; OVER		( n1 --- n1 n2 n1 )
		call	GREAT				; >		( n --- n f )
		mov	eax, PARAM0
		call	DROP				; DROP		( n f --- n )
		cmpz	eax, FALSE
		jne	@@ok

		; incorrect arg #
		call	DROP				; DROP		( n --- )
		call	NULLSTR				; null$		( --- addrz )
		jmp	@@exit

		; --------------------
		; Locate the given arg
		; --------------------

@@ok:		
		; Save vital registers
		mov	edx, PARAM0				; DL is the arg #		
		push	edi es

		; Length of the cmd line
		mov	ax, _selzero
		mov	es, ax
		mov	edi, _pspa
		add	edi, 81h
		movzx	ebx, byte ptr es:[edi - 1]		; nb of chars of the cmd line

		; Count the number of arguments
		clr	ecx

		; skip blanks, i.e. goto 1st non blank char
		clr	eax
@@nextword:	dec	eax
@@blank:	inc	eax
		cmp	al, bl
		je	@@finish
		cmp	byte ptr es:[edi + eax], ' '
		je	@@blank

		; one more argument
		cmp	cl, dl
		je	@@finish
		inc	ecx

		; skip non blanks, i.e. goto 1st blank char
		dec	eax
@@nonblank:	inc	eax
		cmp	al, bl
		je	@@finish
		cmp	byte ptr es:[edi + eax], ' '
		jne	@@nonblank
		jmp	@@nextword

		; Restore vital registers
@@finish:	mov	edx, edi
		add	edi, eax
		mov	edx, offset @@arg
		clr	ebx
		dec	ebx
@@copy:		inc	ebx
		mov	al, es:[edi + ebx]
		mov	[@@arg + ebx], al
		cmp	al, 0Dh
		je	@@endcopy
		cmp	al, ' '
		jne	@@copy
@@endcopy:	mov	[@@arg + ebx], 0
		pop	es edi

		; Update string header
		mov	PARAM0, edx
		sub	edx, size strentry
		mov	[edx.str_currlen], ebx			; Current length
		mov	[edx.str_currsize], ebx			; Current size allocated
		mov	[edx.str_incsize], 0			; 0 (if fixed str)
		mov	[edx.str_memblknb], -1			; Memory block nb (no ref to DDIR)
	
		; Return to caller
@@exit:		ret

		; local variables
		; ---------------

		strentry	<?>
@@arg		db		128 dup (?)

E_CMDARGS:
CMDARGS		endp


code32		ends
		end
