; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	SERIAL.ASM						!
; 									!
; Description:	Definitions for playing with serial I/O.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	SERIALOPEN	SerialOpen					!
;	NONEPARITY	NoneParity					!
;	ODDPARITY	OddParity					!
;	EVENPARITY	EvenParity					!
;	SERIALDEFPCK	SerialDefPacket					!
;	SERIALPCKQ	SerialPackets?					!
;	SERIALPCKREAD	SerialPacketRead				!
;	SERIALPARAM	SerialParam					!
;	SERIALCLOSE	SerialClose					!
;	SERIALIN	SerialIn					!
;	SERIALINQ	SerialIn?					!
;	SERIALREAD	SerialRead					!
;	SERIALFLUSHIN	SerialFlushIn					!
;	SERIALOUT	SerialOut					!
;	SERIALWRITE	SerialWrite					!
;	SERIALFLUSHOUT	SerialFlushOut					!
;	SERIALABORT	SerialAbort					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32 		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc


; -----------------------------------------------
; ITENTRY : general structure, for each interrupt
; -----------------------------------------------

itentry		struc

		; General parameters
it_active	db	?			; Interrupt actived ?	
it_ioaddr	dw	?			; Base IO address
it_irq		db	?			; Irq nb

		; Input processing
it_in_buffaddr	dd	?			; Buffer address for receive
it_in_buffsize	dd	?			; Size of the buffer
it_in_head	dd	?			; FIFO stack: head pointer
it_in_tail	dd	?			; FIFO stack: tail pointer
it_in_nbchars	dd	?			; Nb of chars available
it_in_nbpackets	dd	?			; Nb of packets available

		; Output processing
it_out_buffaddr	dd	?			; Buffer address for receive
it_out_buffsize	dd	?			; Size of the buffer
it_out_head	dd	?			; FIFO stack: head pointer
it_out_tail	dd	?			; FIFO stack: tail pointer
it_out_nbchars	dd	?			; Nb of chars available

		; Packet definitions (for input processing)
it_pck_nb	db	?						; Nb of definitions
it_pck_def	db	(MAXNB_DEF_ITPCK * MAXNB_CHARS_ITPCK) dup (?)	; Each packet definition

		ends


; ---> LOCAL variables to this module ---> ---------------------------------------------------------------------------------------

		align	4
itstable	itentry	MAXNB_ITS dup (<>)			; Its table

; <--- LOCAL variables to this module <--- ---------------------------------------------------------------------------------------


; ===============================================================
; = SERIALOPEN							=
; =								=
; = Code for the froth word:	SerialOpen			=
; =								=
; = Stack:  ( n1 n2 addr1 n3 addr2 n4 --- n5 )			=
; = 	n1	IO address					=
; =	n2	IRQ						=
; =	addr1	Buffer IN address				=
; =	n3	Buffer IN size					=
; =	addr2	Buffer OUT address				=
; =	n4	Buffer OUT size					=
; =	---							=
; =	n5	Serial channel (or -1 if error)			=
; ===============================================================

		align	4
SERIALOPEN   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	6


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALOPEN:
SERIALOPEN	endp


; ===============================================================
; = NONEPARITY							=
; =								=
; = Code for the froth word:	NONEPARITY			=
; =								=
; = Stack:  ( --- 2 )						=
; ===============================================================

		align	4
NONEPARITY   	proc

		; -----------------------------------
		; Push constant onto parameters stack
		; -----------------------------------
		
		IPUSH	0


		; ------------------------------------------
		; Test stack overflow, then Return to caller
		; ------------------------------------------
		
		OVSTACK
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_NONEPARITY:
NONEPARITY	endp


; ===============================================================
; = ODDPARITY							=
; =								=
; = Code for the froth word:	ODDPARITY			=
; =								=
; = Stack:  ( --- 1 )						=
; ===============================================================

		align	4
ODDPARITY   	proc

		; -----------------------------------
		; Push constant onto parameters stack
		; -----------------------------------
		
		IPUSH	1


		; ------------------------------------------
		; Test stack overflow, then Return to caller
		; ------------------------------------------
		
		OVSTACK


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_ODDPARITY:
ODDPARITY	endp


; ===============================================================
; = EVENPARITY							=
; =								=
; = Code for the froth word:	EVENPARITY			=
; =								=
; = Stack:  ( --- 2 )						=
; ===============================================================

		align	4
EVENPARITY   	proc

		; -----------------------------------
		; Push constant onto parameters stack
		; -----------------------------------
		
		IPUSH	2


		; ------------------------------------------
		; Test stack overflow, then Return to caller
		; ------------------------------------------
		
		OVSTACK


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_EVENPARITY:
EVENPARITY	endp


; ===============================================================
; = SERIALDEFPCK						=
; =								=
; = Code for the froth word:	SerialDefPacket			=
; =								=
; = Stack:  ( n1 n2 n3 n4 n5 n6 --- )				=
; =	n1	   serial channel				=
; =	n2	   packet definition #n2			=
; =	n3 .. n6   chars which define a packet (0 -> not used)	=
; ===============================================================

		align	4
SERIALDEFPCK	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	6


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALDEFPCK:
SERIALDEFPCK	endp


; ===============================================================
; = SERIALPCKQ							=
; =								=
; = Code for the froth word:	SerialPackets?			=
; =								=
; = Stack:  ( n1 --- n2 )					=
; =	n1	serial channel					=
; =	n2	nb of available packets 			=
; ===============================================================

		align	4
SERIALPCKQ	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALPCKQ:
SERIALPCKQ	endp


; ===============================================================
; = SERIALPCKREAD						=
; =								=
; = Code for the froth word:	SerialPacketRead		=
; =								=
; = Stack:  ( n1 addr n2 --- n3 )				=
; =	n1	serial channel					=
; =	n2	nb of available packets 			=
; ===============================================================

		align	4
SERIALPCKREAD	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALPCKREAD:
SERIALPCKREAD	endp


; ===============================================================
; = SERIALPARAM							=
; =								=
; = Code for the froth word:	SerialParam			=
; =								=
; = Stack:  ( n1 n2 n3 n4 n5 n6 --- )				=
; =	n1	serial channel					=
; =	n2	speed (1200, 2400, ... )			=
; =	n3	EVENPARITY ODDPARITY NONEPARITY			=
; =	n4	Data bits (7 or 8)				=
; =	n5	Start bits nb (0, 1 or 2)			=
; =	n6	Stop  bits nb (0, 1 or 2)			=
; ===============================================================

		align	4
SERIALPARAM   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	6


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALPARAM:
SERIALPARAM	endp


; ===============================================================
; = SERIALCLOSE							=
; =								=
; = Code for the froth word:	SerialClose			=
; =								=
; = Stack:  ( n --- )						=
; =	n	serial channel					=
; ===============================================================

		align	4
SERIALCLOSE	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALCLOSE:
SERIALCLOSE	endp


; ===============================================================
; = SERIALIN							=
; =								=
; = Code for the froth word:	SerialIn			=
; =								=
; = Stack:  ( n --- c )						=
; =	n	serial channel					=
; =	c	char read					=
; ===============================================================

		align	4
SERIALIN		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALIN:
SERIALIN		endp


; ===============================================================
; = SERIALINQ							=
; =								=
; = Code for the froth word:	SerialIn?			=
; =								=
; = Stack:  ( n1 --- n2 )					=
; =	n1	serial channel					=
; =	n2	nb of available chars				=
; ===============================================================

		align	4
SERIALINQ	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALINQ:
SERIALINQ	endp


; ===============================================================
; = SERIALREAD							=
; =								=
; = Code for the froth word:	SerialRead			=
; =								=
; = Stack:  ( n1 addr n2 --- n3 )				=
; =	n1	serial channel					=
; =	addr	buffer to store read characters			=
; =	n2	nb of chars requested				=
; =	---							=
; =	n3	nb of read chars (or -1) if error		=
; ===============================================================

		align	4
SERIALREAD	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALREAD:
SERIALREAD	endp


; ===============================================================
; = SERIALFLUSHIN						=
; =								=
; = Code for the froth word:	SerialFlushIn			=
; =								=
; = Stack:  ( n --- )						=
; =	n	serial channel					=
; ===============================================================

		align	4
SERIALFLUSHIN	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALFLUSHIN:
SERIALFLUSHIN	endp


; ===============================================================
; = SERIALOUT							=
; =								=
; = Code for the froth word:	SerialOut			=
; =								=
; = Stack:  ( n c --- )						=
; =	n	serial channel					=
; =	c	char to write					=
; ===============================================================

		align	4
SERIALOUT	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	2


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALOUT:
SERIALOUT	endp


; ===============================================================
; = SERIALWRITE							=
; =								=
; = Code for the froth word:	SerialWrite			=
; =								=
; = Stack:  ( n1 addr n2 --- )					=
; =	n1	serial channel					=
; =	addr	buffer address					=
; =	n2	nb of chars to write				=
; ===============================================================

		align	4
SERIALWRITE	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	3


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALWRITE:
SERIALWRITE	endp


; ===============================================================
; = SERIALFLUSHOUT						=
; =								=
; = Code for the froth word:	SerialFlushOut			=
; =								=
; = Stack:  ( n --- )						=
; =	n	serial channel					=
; ===============================================================

		align	4
SERIALFLUSHOUT	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALFLUSHOUT:
SERIALFLUSHOUT	endp


; ===============================================================
; = SERIALABORT							=
; =								=
; = Code for the froth word:	SerialAbort			=
; =								=
; = Stack:  ( n --- )						=
; =	n	irq						=
; ===============================================================

		align	4
SERIALABORT   	proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_SERIALABORT:
SERIALABORT	endp


code32		ends
		end
