; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	STACK.ASM						!
; 									!
; Description:	Definitions for playing with the stack.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	DEPTH			DEPTH					!
; 	RT_ENOUGHT?		RT_ENOUGHT				!
; 	?ENOUGHT		QENOUGHT				!
; 	DROP			DROP					!
; 	DROPS			DROPS					!
; 	DDROP			DDROP					!
; 	DUP			_DUP					!
; 	DDUP			DDUP					!
; 	PICK			PICK					!
; 	SWAP			SWAP					!
; 	DSWAP			DSWAP					!
; 	OVER			OVER					!
; 	ABOVE			ABOVE					!
; 	UNDER			UNDER					!
; 	BELOW			BELOW					!
; 	TUCK			TUCK					!
; 	ROT			ROT					!
; 	-ROT			MROT					!
; 	ROLL			ROLL					!
; 	>R			TOR					!
; 	R>			FROMR					!
; 	R			R					!
; 	RDUP			RDUP					!
; 	RSWAP			RSWAP					!
; 	RDROP			RDROP					!
; 	RUNDER			RUNDER					!
; 	RBELOW			RBELOW					!
; 	ROVER			ROVER					!
; 	RABOVE			RABOVE					!
; 	RROT			RROT					!
; 	R-ROT			RMROT					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions


; ===============================================================
; = isSpeed							=
; = 	*** This is a local function ***			=
; =                                                 		=
; = Usage:							=
; =    Return if ?SPEED is TRUE (cf=1), else cf=0.		=
; =								=
; = I) none							=
; = O) CF							=
; = M) none							=
; ===============================================================

		align	4
isSpeed		proc

		; Test flag
		push	eax
		mov	eax, [edi + ddir_blkaddr + D_QSPEED]
		cmp	byte ptr [eax], FALSE
		jne	@@On
		
@@Off:		pop	eax
		clc						; CF = 0
		ret						; jmp	@@Exit

@@On:		pop	eax
		stc						; CF = 1
		ret

isSpeed		endp


; ===============================================================
; = isSafe							=
; = 	*** This is a service function ***			=
; =                                                 		=
; = Usage:							=
; =    Return if ?SAFE is TRUE (cf=1), else cf=0.		=
; =								=
; = I) none							=
; = O) CF							=
; = M) none							=
; ===============================================================

		align	4
isSafe		proc

		; Test flag
		push	eax
		mov	eax, [edi + ddir_blkaddr + D_QSAFE]
		cmp	byte ptr [eax], FALSE
		jne	@@On
		
@@Off:		pop	eax
		clc						; CF = 0
		ret						; jmp	@@Exit

@@On:		pop	eax
		stc						; CF = 1
		ret

isSafe		endp


; ===============================================================
; = DEPTH							=
; =                                                 		=
; = Code for the forth word:	DEPTH				=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
DEPTH		proc

		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * @WordSize]
		sub	eax, esi
		shr	eax, 2
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret
		
E_DEPTH:
DEPTH		endp


; ===============================================================
; = RT_QENOUGHT							=
; =                                                 		=
; = Code for the forth word:	RT_?ENOUGHT			=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
RT_QENOUGHT	proc

		; Verify stack frame
		VSTACK	1

		; Get <n>
		mov	ebx, PARAM0
		DISCARD	1					; drop

		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * @WordSize]

		; Stack overflow ?
		mov	ecx, esi
		add	ecx, PSTACKSIZE
		cmp	eax, ecx
		jne	@@NotOverflow

		; Error: stack overflow
		mov	ah, ERR_STACKOVERFLOW			; Stack overflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

		; Verify stack frame
@@NotOverflow:	sub	eax, esi
		shr	eax, 2					; EAX is the nb of words on the stack
		cmp	ax, bx
		jae	@@Ok

		; Error: stack underflow
		mov	ah, ERR_STACKUNDERFLOW			; Stack underflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@Ok:

		; Return to caller
		ret

E_RT_QENOUGHT:
RT_QENOUGHT	endp


; ===============================================================
; = QENOUGHT							=
; =                                                 		=
; = Code for the forth word:	?ENOUGHT			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
QENOUGHT	proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE


		; ---------------------------
		; Ask for the value of ?SAFE
		; ---------------------------
		
		call	isSafe
		jc	@@Safe
		

		; -------------------------------------
		; Not safe, remove previous nb compiled
		; -------------------------------------

		; 1st address of the code of the current word being defined
		mov	edx, [VOC_CURRENT]
		add	edx, [VOC_START]
		mov	edx, [edx.voc_lastword]
		add	edx, [NIA_START]
		mov	edx, [edx.word_realaddr]

		; Current position into code
		mov	ecx, [CDA_START]
		add	ecx, [CDA_SIZE]
		sub	ecx, 9					; 9: 3 = SUB ESI,4   6 = MOV DWORD PTR [ESI], i32

		; Not a constant given ?
		cmp	ecx, edx
		jae	@@OK
		
		; "An integer constant must be compiled before ?ENOUGHT"
@@NOK:		mov	ah, ERR_QENOUGHTNOTCSTG			; Error #
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@OK:		cmp	dword ptr [ecx], 0C704EE83h		; 83 EE 04 -> sub esi, 4
		jne	@@NOK
		cmp	byte ptr [ecx + 4], 06h			; C7 06 -> mov dword ptr [esi], i32
		jne	@@NOK
		
		; Ok, remove the constant
		sub	[CDA_SIZE], 9
		
		; Done
		ret						; jmp	@@Exit

		
		; ---------------------------------------
		; Put 'SUMMON RT_?ENOUGHT' code operation
		; ---------------------------------------
		
@@Safe:		compCALL_rel32	RT_QENOUGHT
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_QENOUGHT:
QENOUGHT	endp


; ===============================================================
; = DROP							=
; =                                                 		=
; = Run-time code for the forth word:	DROP			=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
DROP		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	1
		lodsd
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack1				; CALL   VStack1
@@NoOpt2:	_LODSD_
		ret

E_DROP:
DROP		endp


; ===============================================================
; = DROPS							=
; =                                                 		=
; = Run-time code for the forth word:	DROPS			=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
DROPS		proc

		; ------------------------------
		; Get <n> and verify stack frame
		; ------------------------------

		; Verify stack frame
		VSTACK	1

		; Get <n>		
		mov	ebx, PARAM0

		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * @WordSize]
		sub	eax, esi
		shr	eax, 2
		cmp	eax, ebx
		ja	@@Ok

		; Error: stack underflow
		mov	ah, ERR_STACKUNDERFLOW			; Stack underflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@Ok:


		; ----------------------------------
		; Drop the <n> top values from stack
		; ----------------------------------

		inc	ebx
		lea	esi, [esi + ebx * @WordSize]


		; ----------------
		; Return to caller
		; ----------------

		ret

E_DROPS:
DROPS		endp


; ===============================================================
; = DDROP							=
; =                                                 		=
; = Run-time code for the forth word:	DDROP			=
; =								=
; = Stack:	( n1 n2 --- )					=
; ===============================================================

		align	4
DDROP		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		DISCARD	2
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL   VStack2
@@NoOpt2:	ADD_ESI_i8	2*@WordSize
		ret

E_DDROP:
DDROP		endp


; ===============================================================
; = DUP								=
; =                                                 		=
; = Run time code for the forth word:	DUP			=
; =								=
; = Stack:	( n --- n n )					=
; ===============================================================

		align	4
_DUP		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	1
		mov	eax, PARAM0
		sub	esi, @WordSize
		mov	PARAM0, eax
		call	OStack
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack1				; CALL   VStack1
@@NoOpt2:	MOV_EAX_xESIi8
		SUB_ESI_i8	@WordSize
		call	isSafe
		jnc	@@NoOpt3
		compCALL_rel32	OStack				; CALL   OStack
@@NoOpt3:	MOV_xESIi8_EAX
		ret

E__DUP:
_DUP		endp


; ===============================================================
; = DDUP							=
; =                                                 		=
; = Code for the forth word:	DDUP				=
; =								=
; = Stack:	( n1 n2 --- n1 n2 n1 n2 )			=
; ===============================================================

		align	4
DDUP		proc

		; Verify stack frame
		VSTACK	2

		; Duplicate values
		DISPOSE	2
		mov	eax, PARAM3					; n1
		mov	PARAM1, eax
		mov	eax, PARAM2					; n2
		mov	PARAM0, eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_DDUP:
DDUP		endp


; ===============================================================
; = PICK							=
; =                                                 		=
; = Code for the forth word:	PICK				=
; =								=
; = Stack:	( n --- n n )					=
; ===============================================================

		align	4
PICK		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------------------------
		; Get <n> and verify stack frame
		; ------------------------------

		; Get <n>		
		mov	ebx, PARAM0

		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * @WordSize]
		sub	eax, esi
		shr	eax, 2
		cmp	eax, ebx
		ja	@@Ok

		; Error: stack underflow
		mov	ah, ERR_STACKUNDERFLOW			; Stack underflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError


		; -----------------------------------
		; Duplicate the n-th value on the top
		; -----------------------------------
		
@@Ok:		mov	eax, [esi + ebx * @WordSize]
		mov	PARAM0, eax


		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK

		; Return to caller
		ret
		

E_PICK:
PICK		endp


; ===============================================================
; = SWAP							=
; =                                                 		=
; = Run-time code for the forth word:	SWAP			=
; =								=
; = Stack:	( n1 n2 --- n2 n1 )				=
; ===============================================================

		align	4
SWAP		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		mov	eax, PARAM0				; pop <n2>
		xchg	PARAM1, eax				; Echg <n1> <n2>
		mov	PARAM0, eax
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL   VStack2
@@NoOpt2:	MOV_EAX_xESIi8
		XCHG_EAX_xESIi8	@WordSize
		MOV_xESIi8_EAX
		ret

E_SWAP:
SWAP		endp


; ===============================================================
; = DSWAP							=
; =                                                 		=
; = Code for the forth word:	DSWAP				=
; =								=
; = Stack:	( n1 n2 n3 n4 --- n3 n4 n1 n2 )			=
; ===============================================================

		align	4
DSWAP		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	4


		; ------------
		; Do operation
		; ------------
		
		mov	eax, PARAM0			; pop <n4>
		mov	ebx, PARAM1			; pop <n3>
		mov	ecx, PARAM2			; pop <n2>
		mov	edx, PARAM3			; pop <n1>
		mov	PARAM0, ecx
		mov	PARAM1, edx
		mov	PARAM2, eax
		mov	PARAM3, ebx


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
E_DSWAP:
DSWAP		endp


; ===============================================================
; = OVER							=
; =                                                 		=
; = Run-time code for the forth word:	OVER			=
; =								=
; = Stack:	( n1 n2 --- n1 n2 n1 )				=
; ===============================================================

		align	4
OVER		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		mov	eax, PARAM1
		sub	esi, @WordSize
		mov	PARAM0, eax
		call	OStack
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL   VStack2
@@NoOpt2:	MOV_EAX_xESIi8	@WordSize
		SUB_ESI_i8	@WordSize
		call	isSafe
		jnc	@@NoOpt3
		compCALL_rel32	OStack				; CALL   OStack
@@NoOpt3:	MOV_xESIi8_EAX
		ret

E_OVER:
OVER		endp


; ===============================================================
; = ABOVE							=
; =                                                 		=
; = Run-time code for the forth word:	ABOVE			=
; =								=
; = Stack:	( n1 n2 n3 --- n1 n2 n3 n1 )			=
; =								=
; ===============================================================

		align	4
ABOVE		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	3
		mov	eax, PARAM2
		sub	esi, @WordSize
		mov	PARAM0, eax
		call	OStack
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack3				; CALL   VStack3
@@NoOpt2:	MOV_EAX_xESIi8	2*@WordSize
		SUB_ESI_i8	@WordSize
		call	isSafe
		jnc	@@NoOpt3
		compCALL_rel32	OStack				; CALL   OStack
@@NoOpt3:	MOV_xESIi8_EAX
		ret

E_ABOVE:
ABOVE		endp


; ===============================================================
; = UNDER							=
; =                                                 		=
; = Run-time code for the forth word:	UNDER			=
; =								=
; = Stack:	( n1 n2 --- n2 )				=
; =								=
; ===============================================================

		align	4
UNDER		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	2
		lodsd
		mov	PARAM0, eax
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack2				; CALL   VStack2
@@NoOpt2:	_LODSD_
		MOV_xESIi8_EAX
		ret

E_UNDER:
UNDER		endp


; ===============================================================
; = BELOW							=
; =                                                 		=
; = Code for the forth word:	BELOW				=
; =								=
; = Stack:	( n1 n2 n3 --- n2 n3 )				=
; ===============================================================

		align	4
BELOW		proc

		cmp	[CompileMode], TRUE
		je	@@Compile

		; ------------
@@Execute:	; Execute mode
		; ------------

		VSTACK	3
		mov	ebx, PARAM1				; Pop <n2>
		lodsd						; Pop <n3>
		mov	PARAM0, eax				; push <n3>
		mov	PARAM1, ebx				; push <n2>
		ret


		; ------------
@@Compile:	; Compile mode
		; ------------

		; No optimization
		; ---------------

@@NoOptimize:	call	isSafe
		jnc	@@NoOpt2
		compCALL_rel32	VStack3				; CALL   VStack3
@@NoOpt2:	MOV_EBX_xESIi8	@WordSize
		_LODSD_
		MOV_xESIi8_EAX
		MOV_xESIi8_EBX	@WordSize
		ret

E_BELOW:
BELOW		endp


; ===============================================================
; = TUCK							=
; =                                                 		=
; = Code for the forth word:	TUCK				=
; =								=
; = Stack:	( n1 n2 --- n2 n1 n2 )				=
; ===============================================================

		align	4
TUCK		proc

		VSTACK	2
		mov	eax, PARAM0				; pop <n2>
		mov	ebx, PARAM1				; pop <n1>
		DISPOSE	1		
		mov	PARAM0, eax				; push <n2>
		mov	PARAM1, ebx				; push <n1>
		mov	PARAM2, eax				; push <n1>
		OVSTACK
		ret
		
E_TUCK:
TUCK		endp


; ===============================================================
; = ROT								=
; =                                                 		=
; = Code for the forth word:	ROT				=
; =								=
; = Stack:	( n1 n2 n3 --- n2 n3 n1)			=
; ===============================================================

		align	4
ROT		proc

		; Verify stack frame
		VSTACK	3

		; Rotate the three values on the stack
		mov	eax, PARAM0			; pop <n3>
		mov	ebx, PARAM1			; pop <n2>
		mov	ecx, PARAM2			; pop <n1>
		mov	PARAM2, ebx			; <n2>
		mov	PARAM1, eax			; <n3>
		mov	PARAM0, ecx			; <n1>

		; Return to caller
		ret

E_ROT:
ROT		endp


; ===============================================================
; = MROT							=
; =                                                 		=
; = Code for the forth word:	-ROT				=
; =								=
; = Stack:	( n1 n2 n3 --- n2 n3 n1)			=
; ===============================================================

		align	4
MROT		proc

		; Verify stack frame
		VSTACK	3

		; Rotate the three values on the stack
		mov	eax, PARAM0			; pop <n3>
		mov	ebx, PARAM1			; pop <n2>
		mov	ecx, PARAM2			; pop <n1>
		mov	PARAM2, eax			; <n3>
		mov	PARAM1, ecx			; <n1>
		mov	PARAM0, ebx			; <n2>

		; Return to caller
		ret

E_MROT:
MROT		endp


; ===============================================================
; = ROLL							=
; =                                                 		=
; = Code for the forth word:	ROLL				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
ROLL		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------------------------
		; Get <n> and verify stack frame
		; ------------------------------

		; Get <n>		
		mov	ebx, PARAM0

		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * @WordSize]
		sub	eax, esi
		shr	eax, 2
		cmp	eax, ebx
		ja	@@Ok

		; Error: stack underflow
		mov	ah, ERR_STACKUNDERFLOW			; Stack underflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@Ok:		DISCARD	1					; drop


		; --------------
		; Stack rotation
		; --------------
		
		; Do nothing if n < 3
		cmp	bl, 3
		jl	@@Exit

		; Save EDI
		push	esi edi
		
		; Save bottom value
		dec	ebx
		mov	eax, [esi + ebx * @WordSize]
		
		; Nb of dword to move
		mov	ecx, ebx

		; Destination address
		lea	edi, [esi + ebx * @WordSize]

		; Source address to move
		mov	esi, edi
		sub	esi, @WordSize
		
		; Move block
		std  	                   			; Dec ESI and EDI
		rep	movsd 	 		            	; Move the block
		cld
		
		; Restore EDI
		pop	edi esi
		
		; Store value on top
		mov	PARAM0, eax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_ROLL:
ROLL		endp


; ===============================================================
; = MROLL							=
; =                                                 		=
; = Code for the forth word:	-ROLL				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
MROLL		proc

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------------------------
		; Get <n> and verify stack frame
		; ------------------------------

		; Get <n>		
		mov	ebx, PARAM0

		; The number of words on the stack
		mov	eax, [CurrActThread]
		mov	eax, [PStackStart + eax * @WordSize]
		sub	eax, esi
		shr	eax, 2
		cmp	eax, ebx
		ja	@@Ok

		; Error: stack underflow
		mov	ah, ERR_STACKUNDERFLOW			; Stack underflow
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
		
@@Ok:		DISCARD	1					; drop


		; --------------
		; Stack rotation
		; --------------
		
		; Do nothing if n < 3
		cmp	bl, 3
		jl	@@Exit

		; Save EDI
		push	esi edi
		
		; Save top value
		mov	eax, PARAM0
		
		; Nb of dword to move
		mov	ecx, ebx
		dec	ecx
		
		; Destination address
		mov	edi, esi

		; Source address to move
		add	esi, @WordSize
		
		; Move block
		cld  	                   			; Inc ESI and EDI
		rep	movsd 	 		            	; Move the block
		
		; Restore EDI
		pop	edi esi
		
		; Store value on bottom
		dec	ebx
		mov	[esi + ebx * @WordSize], eax


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

E_MROLL:
MROLL		endp


; ===============================================================
; = TOR								=
; =                                                 		=
; = Code for the forth word:	>R				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
TOR		proc

		; Verify stack frame
		VSTACK	1

		; Store top stack value to R-stack
		lodsd
		mov	ebx, [CurrActThread]
		lea	ebx, RStackPtr + ebx * @WordSize
		sub	dword ptr [ebx], @WordSize
		mov	ebx, [ebx]
		mov	dword ptr [ebx], eax

		; Return to caller
		ret

E_TOR:
TOR		endp


; ===============================================================
; = FROMR							=
; =                                                 		=
; = Code for the forth word:	R>				=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
FROMR		proc

		; Verify second stack frame
		VRSTACK	1

		; Store top 2nd stack value to stack
		mov	ebx, [CurrActThread]
		lea	ebx, RStackPtr + ebx * @WordSize
		mov	ecx, [ebx]
		mov	eax, [ecx]
		add	dword ptr [ebx], @WordSize
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_FROMR:
FROMR		endp


; ===============================================================
; = R								=
; =                                                 		=
; = Code for the forth word:	R				=
; =								=
; = Stack:	( --- n )					=
; ===============================================================

		align	4
R		proc

		; Verify second stack frame
		VRSTACK	1

		; Store top 2nd stack value to stack
		mov	ebx, [CurrActThread]
		mov	ecx, [RStackPtr + ebx * @WordSize]
		mov	eax, [ecx]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_R:
R		endp


; ===============================================================
; = RDUP							=
; =                                                 		=
; = Code for the forth word:	RDUP				=
; =								=
; = 2nd stack:	( n --- n n )					=
; ===============================================================

		align	4
RDUP		proc

		; Verify second stack frame
		VRSTACK	1
	
		; Get value
		mov	ebx, [CurrActThread]
		lea	ebx, RStackPtr + ebx * @WordSize
		mov	ecx, [ebx]
		mov	eax, [ecx]				; Get <n>

		; Duplicate top 2nd stack value
		sub	dword ptr [ebx], @WordSize
		mov	[ecx - @WordSize], eax

		; Return to caller
		ret

E_RDUP:
RDUP		endp


; ===============================================================
; = RSWAP							=
; =                                                 		=
; = Code for the forth word:	RSWAP				=
; =								=
; = 2nd stack:	( n1 n2 --- n2 n1 )				=
; ===============================================================

		align	4
RSWAP		proc

		; Verify second stack frame
		VRSTACK	2
	
		; Do SWAP
		mov	ebx, [CurrActThread]
		mov	ecx, [RStackPtr + ebx * @WordSize]
		mov	eax, [ecx + @WordSize]				; Get n1
		mov	ebx, [ecx]					; Get n2
		mov	[ecx + @WordSize], ebx				; swap
		mov	[ecx], eax					; swap

		; Return to caller
		ret

E_RSWAP:
RSWAP		endp


; ===============================================================
; = RDROP							=
; =                                                 		=
; = Code for the forth word:	RDROP				=
; =								=
; = 2nd stack:	( n --- )					=
; =								=
; ===============================================================

		align	4
RDROP		proc

		; Verify second stack frame
		VRSTACK	1
	
		; Drop value
		mov	ebx, [CurrActThread]
		add	dword ptr [RStackPtr + ebx * @WordSize], @WordSize

		; Return to caller
		ret

E_RDROP:
RDROP		endp


; ===============================================================
; = RUNDER							=
; =                                                 		=
; = Code for the forth word:	RUNDER				=
; =								=
; = 2nd stack:	( n1 n2 --- n2 )				=
; ===============================================================

		align	4
RUNDER		proc

		; Verify second stack frame
		VRSTACK	2
	
		; Drop value
		mov	ebx, [CurrActThread]
		lea	ebx, RStackPtr + ebx * @WordSize
		mov	ecx, [ebx]
		mov	eax, [ecx]					; Get n2
		add	dword ptr [ebx], @WordSize			; DROP
		mov	[ecx + @WordSize], eax

		; Return to caller
		ret

E_RUNDER:
RUNDER		endp


; ===============================================================
; = RBELOW							=
; =                                                 		=
; = Code for the forth word:	RBELOW				=
; =								=
; = 2nd stack:	( n1 n2 n3 --- n2 n3 )				=
; ===============================================================

		align	4
RBELOW		proc

		; Verify second stack frame
		VRSTACK	3
	
		; Do operation
		mov	ebx, [CurrActThread]
		lea	ebx, RStackPtr + ebx * @WordSize
		mov	ecx, [ebx]
		mov	eax, [ecx + 1 * @WordSize]			; Get <n2>
		mov	edx, [ecx + 0 * @WordSize]			; Get <n3>
		add	dword ptr [ebx], @WordSize			; DROP
		mov	[ecx + 1 * @WordSize], eax			; Put <n2>
		mov	[ecx + 0 * @WordSize], edx			; Put <n3>

		; Return to caller
		ret

E_RBELOW:
RBELOW		endp


; ===============================================================
; = ROVER							=
; =                                                 		=
; = Code for the forth word:	ROVER				=
; =								=
; = 2nd stack:	( n1 n2 --- n1 n2 n1 )				=
; ===============================================================

		align	4
ROVER		proc

		; Verify second stack frame
		VRSTACK	2
	
		; Do operation
		mov	ebx, [CurrActThread]
		lea	ebx, RStackPtr + ebx * @WordSize
		mov	ecx, [ebx]
		mov	eax, [ecx + @WordSize]			; Get n1
		sub	dword ptr [ebx], @WordSize
		mov	[ecx - @WordSize], eax

		; Return to caller
		ret

E_ROVER:
ROVER		endp


; ===============================================================
; = RABOVE							=
; =                                                 		=
; = Code for the forth word:	RABOVE				=
; =								=
; = 2nd stack:	( n1 n2 n3 --- n1 n2 n3 n1 )			=
; ===============================================================

		align	4
RABOVE		proc

		; Verify second stack frame
		VRSTACK	3
	
		; Do operation
		mov	ebx, [CurrActThread]
		lea	ebx, RStackPtr + ebx * @WordSize
		mov	ecx, [ebx]
		mov	eax, [ecx + 2 * @WordSize]			; Get n1
		sub	dword ptr [ebx], @WordSize
		mov	[ecx - @WordSize], eax

		; Return to caller
		ret

E_RABOVE:
RABOVE		endp


; ===============================================================
; = RROT							=
; =                                                 		=
; = Code for the forth word:	RROT				=
; =								=
; = 2nd stack:	( n1 n2 n3 --- n2 n3 n1 )			=
; ===============================================================

		align	4
RROT		proc

		; Verify second stack frame
		VRSTACK	3
	
		; Do operation
		mov	ebx, [CurrActThread]
		mov	ecx, [RStackPtr + ebx * @WordSize]
		mov	eax, [ecx + 2 * @WordSize]			; Get n1
		mov	ebx, [ecx + 1 * @WordSize]			; Get n2
		mov	edx, [ecx + 0 * @WordSize]			; Get n3
		mov	[ecx + 2 * @WordSize], ebx
		mov	[ecx + 1 * @WordSize], edx
		mov	[ecx + 0 * @WordSize], eax

		; Return to caller
		ret

E_RROT:
RROT		endp


; ===============================================================
; = RMROT							=
; =                                                 		=
; = Code for the forth word:	R-ROT				=
; =								=
; = 2nd stack:	( n1 n2 n3 --- n3 n1 n2 )			=
; ===============================================================

		align	4
RMROT		proc

		; Verify second stack frame
		VRSTACK	3
	
		; Do operation
		mov	ebx, [CurrActThread]
		mov	ecx, [RStackPtr + ebx * @WordSize]
		mov	eax, [ecx + 2 * @WordSize]			; Get n1
		mov	ebx, [ecx + 1 * @WordSize]			; Get n2
		mov	edx, [ecx + 0 * @WordSize]			; Get n3
		mov	[ecx + 2 * @WordSize], edx
		mov	[ecx + 1 * @WordSize], eax
		mov	[ecx + 0 * @WordSize], ebx

		; Return to caller
		ret

E_RMROT:
RMROT		endp


code32		ends

		end

