; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	TRACE.ASM						!
; 									!
; Description:	Definitions for playing with tracing msg.		!
; 									!
; List of FROTH definitions defined in this module:			!
;	TRACE_PAGE	TRACE_PAGE					!
;	TRACE_CLEAR	TRACE_CLEAR					!
;	TRACE_BEGIN	TRACE_BEGIN					!
;	TRACE_END	TRACE_END					!
;	TRACE_DOT	TRACE_.						!
;	TRACE_DOTSTACK	TRACE_.S					!
;	TRACE_TYPE	TRACE_TYPE					!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc
		
if TURNKEYED eq 0
		
; ---> Local variables ---> ------------------------------------------------------------------------------------------------------

		align	4
InitDone	db	?				; Is TRACE initialized ?
		align	4
RecordsNb	dd	?				; Nb of records stored
SizeUsed	dd	?				; Size currently in use
TopRecNb	dd	?				; 1st record shown on top
		align	4
CurrMsg		db	256 dup (?)			; Current msg
		align	4
CurrMsgLen	dd	?				; Current msg length
		align	4

; ---> End of local variables ---> -----------------------------------------------------------------------------------------------


; ===============================================================
; = DrawMsg							=
; =								=
; = Usage:							=
; =	Draw msg, starting at [TopRecNb].			=
; =								=
; = I) none							=
; = O) none							=
; = M) undetermined.						=
; ===============================================================

		align	4
DrawMsg		proc

		; -----------------------------
		; Compute address of top record
		; -----------------------------
		
		mov	edx, [TraceBuffer]
		clr	ecx
		mov	ebx, [TopRecNb]
@@FindAddr:	cmp	ecx, ebx
		je	@@FindAddr2
		movzx	eax, byte ptr [edx]
		add	edx, eax
		add	edx, 2
		inc	ecx
		jmp	@@FindAddr
@@FindAddr2:


		; -------------
		; Print the msg
		; -------------
		
		clr	ecx					; Line counter
		
@@PrMsg:	
		; Move cursor
		push	ecx edx
		sub	esi, @WordSize
		mov	dword ptr [esi], 0			; col
		sub	esi, @WordSize
		mov	[esi], ecx				; row
		call	GOTOXY
		pop	edx ecx
		
		; Print msg
		push	ecx edx
		sub	esi, @WordSize
		inc	edx

		ifdef	DOSREAL
		
		; Copy the string below 1 MB ?
		cmp	edx, 10000h
		jb	@@Below1MB
		
		; Move string
		push	eax ecx
		mov	ecx, [QuoteStrBuff]
		add	ecx, size strentry
@@MovStr1:	mov	al, [edx]
		mov	[ecx], al
		cmp	al, 0
		je	@@MovStr2
		inc	edx
		inc	ecx
		jmp	@@MovStr1
@@MovStr2:	mov	edx, [QuoteStrBuff]
		add	edx, size strentry
		pop	ecx eax
		
@@Below1MB:

		endif

		mov	[esi], edx				; String address
		call	_TYPE					; TYPE
		call	CLREOL					; ClrEOL
		pop	edx ecx

		; All the screen is filled ?
		call	NBROWS					; #Rows
		mov	eax, [esi]
		add	esi, @WordSize
		sub	eax, 2
		cmp	ecx, eax
		je	@@Exit
		
		; Next line, if any
		inc	ecx
		mov	eax, [TopRecNb]
		add	eax, ecx
		cmp	eax, [RecordsNb]
		je	@@Stop
		
		; Skip to next record
		movzx	eax, byte ptr [edx]
		add	edx, eax
		add	edx, 2
		jmp	@@PrMsg

		; Blank the rest of the screen
@@Stop:
@@Blank:	
				
		; Clear line
		pushad
		sub	esi, @WordSize
		mov	[esi], ecx				; row
		call	CLRLINE
		popad
		
		; All the screen is filled ?
		call	NBROWS					; #Rows
		mov	eax, [esi]
		add	esi, @WordSize
		sub	eax, 2
		cmp	ecx, eax
		je	@@BlankDone
		inc	ecx
		jmp	@@Blank
@@BlankDone:
		

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

DrawMsg		endp


; ===============================================================
; = TraceUser							=
; =								=
; = Usage:							=
; =	User actions within trace.				=
; =								=
; = I) none							=
; = O) none							=
; = M) undetermined.						=
; ===============================================================

		align	4
TraceUser	proc

		; -----------------
		; Print a msg below
		; -----------------

		; Move cursor
		sub	esi, @WordSize
		mov	dword ptr [esi], 0			; col
		call	NBROWS
		dec	dword ptr [esi]				; row
		call	GOTOXY
		
		; Print a msg
		call	PDOTQ
		db	"< > Move by one line  <PgUp PgDn> Move by one page   <ESC> abort"
		db	0
		
		; Draw msg, if any
		cmp	[RecordsNb], 0
		je	@@NoRec
		call	DrawMsg
@@NoRec:


		; ----------
		; Wait a key
		; ----------
		
@@WaitKey:	

		; Print position (based on top)
		sub	esi, @WordSize
		mov	dword ptr [esi], 70			; col
		call	NBROWS
		dec	dword ptr [esi]				; row
		call	GOTOXY
		mov	eax, [TopRecNb]
		IPUSH	eax
		IPUSH	4
		call	DOTR
		mov	al, '/'
		call	COUT
		mov	eax, [RecordsNb]
		IPUSH	eax
		IPUSH	4
		call	DOTR
		

		; Wait a key
		call	CharIn

		
		; Is-it <ESC> ? -> exit trace
		cmp	ax, EKEY_ESC
		je	@@Exit
		
		; Error, unknow action
		jmp	@@WaitKey


		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		ret

TraceUser	endp


; ===============================================================
; = AddRecord							=
; =								=
; = Usage:							=
; =	Add a new record into trace buffer.			=
; =								=
; = I) EDX : pointer to a string endded with '\0'.		=
; = O) none							=
; = M) undetermined.						=
; ===============================================================

		align	4
AddRecord	proc

		; ------------------------------
		; Compute the size of the string
		; ------------------------------
		
		clr	ecx
@@StrLen:	cmp	byte ptr [edx + ecx], 0
		je	@@StrLen2
		inc	ecx
		jmp	@@StrLen
@@StrLen2:


		; ---------------------------
		; Enought space into buffer ?
		; ---------------------------

		mov	eax, [SizeUsed]
		add	eax, ecx
		add	eax, 2					; EAX = New buffer size
		cmp	eax, [TraceBufferMaxSz]
		jbe	@@SpaceOk


		; ----------------------------------
		; Delete the 1st record to get space
		; ----------------------------------

		
		
@@SpaceOk:
		

		; ---------------------
		; Record the new string
		; ---------------------
		
		; Start address
		mov	ebx, [TraceBuffer]
		add	ebx, [SizeUsed]
		
		; Store the length of the string
		mov	[ebx], cl
		inc	ebx

		; Now Store the string
		push	esi edi ecx
		mov	esi, edx				; Source address
		mov	edi, ebx				; Destination address
		inc	ecx					; Size to move (String length + 1)
		cld                     			; Inc SI and DI
		rep	movsb  		            	    	; Move the block
		pop	ecx edi esi
		
		; Update the nb of records (a record = a line)
		inc	[RecordsNb]
		
		; Update the size used into the buffer
		add	ecx, 2
		add	[SizeUsed], ecx


		; ----------------
		; Return to caller
		; ----------------
		
		ret

AddRecord	endp


; ===============================================================
; = AddMsg							=
; =								=
; = Usage:							=
; =	Add a new msg to the current msg.			=
; =								=
; = I) EDX : pointer to a string endded with '\0'.		=
; = O) none							=
; = M) undetermined.						=
; ===============================================================

		align	4
AddMsg		proc

		; ------------------------------
		; Compute the size of the string
		; ------------------------------
		
		clr	ecx
@@StrLen:	cmp	byte ptr [edx + ecx], 0
		je	@@StrLen2
		inc	ecx
		jmp	@@StrLen
@@StrLen2:


		; -------------------------
		; New msg will be to long ?
		; -------------------------
		
		call	NBCOLS
		mov	eax, [esi]
		add	esi, @WordSize
		mov	ebx, [CurrMsgLen]
		inc	ebx
		cmp	ebx, eax
		jae	@@Exit


		; -----------
		; Add the msg
		; -----------
		
		push	esi edi ecx
		mov	esi, edx				; Source address
		lea	edi, CurrMsg
		add	edi, [CurrMsgLen]			; Destination address
		inc	ecx					; Nb of bytes to move
		cld                     			; Inc SI and DI
		rep	movsb  		            	    	; Move the block
		pop	ecx edi esi
		
		; Update ths current msg length
		add	[CurrMsgLen], ecx


@@Exit:		ret

AddMsg		endp


; ===============================================================
; = doTraceInit							=
; =								=
; = Initialize things for trace manager.			=
; = Called by: INTERPRET					=
; ===============================================================

		align	4
doTraceInit	proc

		; Already initialized ?
		cmp	[InitDone], FALSE
		jne	@@Exit

		; Initial nb of records
		mov	[RecordsNb], 0
		
		; Size currently used
		mov	[SizeUsed], 0

		; Update flag
		mov	[InitDone], TRUE

		; Return to caller
@@Exit:		ret

doTraceInit	endp


; ===============================================================
; = TRACE_PAGE							=
; =								=
; = Code for the forth word:	TRACE_PAGE			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
TRACE_PAGE	proc

		; Get the nb of video page available (two video pages at least)
		call	NBPAGES					; #pages
		mov	eax, [esi]
		add	esi, @WordSize
		cmp	al, 2
		jl	@@Exit

		; Get current video page
		call	PAGENB					; page#
		DO_AT						; @
		mov	eax, [esi]
		add	esi, @WordSize
		mov	@@PageNb, eax

		; Commut on last video page available
		call	NBPAGES					; #pages
		mov	eax, [esi]
		dec	eax
		mov	[esi], eax
		call	CHANGEPAGE				; SelectPage
		
		; 1st record shown on top
		call	NBROWS					; #rows
		mov	eax, [esi]
		add	esi, @WordSize
		dec	eax
		mov	ebx, [RecordsNb]
		sub	ebx, eax
		cmp	ebx, 0
		jge	@@1stRecOk
		clr	ebx
@@1stRecOk:	mov	[TopRecNb], ebx

		; Now wait user actions
		call	TraceUser
		
		; Restore previous video page
		mov	eax, [@@PageNb]
		sub	esi, @WordSize
		mov	[esi], eax
		call	CHANGEPAGE				; SelectPage

		; Return to caller
@@Exit:		ret

; Local variables
; ---------------

		align	4
@@PageNb	dd	?					; Current video page (before commut)

E_TRACE_PAGE:
TRACE_PAGE	endp


; ===============================================================
; = TRACE_CLEAR							=
; =								=
; = Code for the forth word:	TRACE_CLEAR			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
TRACE_CLEAR	proc

		; Initial nb of records
		mov	[RecordsNb], 0
		
		; Size currently used
		mov	[SizeUsed], 0

		; Return to caller
		ret

E_TRACE_CLEAR:
TRACE_CLEAR	endp


; ===============================================================
; = TRACE_BEGIN							=
; =								=
; = Code for the forth word:	TRACE_BEGIN			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
TRACE_BEGIN	proc

		; Initialize current msg
		mov	[CurrMsgLen], 0

		; Return to caller
		ret

E_TRACE_BEGIN:
TRACE_BEGIN	endp


; ===============================================================
; = TRACE_END							=
; =								=
; = Code for the forth word:	TRACE_END			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
TRACE_END	proc

		; Add record
		lea	edx, CurrMsg
		call	AddRecord

		; Re-Initialize current msg
		mov	[CurrMsgLen], 0

		; Return to caller
		ret

E_TRACE_END:
TRACE_END	endp


; ===============================================================
; = TRACE_DOT							=
; =								=
; = Code for the forth word:	TRACE_.				=
; =								=
; = Stack:	( n --- )					=
; ===============================================================

		align	4
TRACE_DOT	proc

		; Verify stack frame
		VSTACK	1

		; Get value to print
		call	BSHARP					; <#	( n --- n )
		call	SHARPS					; #S	( n --- 0 )
		IPUSH	' '					; BL
		call	HOLD					; HOLD
		call	ESHARP					; #>	( 0 --- addr )

		; Add msg
		mov	edx, [esi]
		add	esi, @WordSize
		call	AddMsg

		; Return to caller
		ret

E_TRACE_DOT:
TRACE_DOT	endp


; ===============================================================
; = TRACE_DOTSTACK						=
; =								=
; = Code for the forth word:	TRACE_.S			=
; =								=
; = Stack:	( --- )						=
; ===============================================================

		align	4
TRACE_DOTSTACK	proc

		; "Sorry, this function is not yet implemented. Next release..."
		mov	ah, ERR_NOTYETAVAIL		; Error #
		clr	edx				; No additionnal msg
		clr	ecx				; No line #
		call	doError

		; Return to caller
		ret

E_TRACE_DOTSTACK:
TRACE_DOTSTACK	endp


; ===============================================================
; = TRACE_TYPE							=
; =								=
; = Code for the forth word:	TRACE_TYPE			=
; =								=
; = Stack:	( addrz --- )					=
; ===============================================================

		align	4
TRACE_TYPE	proc

		; Verify stack frame
		VSTACK	1

		; Verify if a valid string
		call	QSTRING					; ?String

		; Add msg
		mov	edx, [esi]
		add	esi, @WordSize
		call	AddMsg

		; Return to caller
		ret

E_TRACE_TYPE:
TRACE_TYPE	endp

endif

code32		ends
		end
