; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	VARIED.ASM						!
; 									!
; Description:	Varied definitions.					!
; 									!
; List of FROTH definitions defined in this module:			!
;	WORDSIZE		WORDSIZE				!
;	QOK			?OK					!
;	QSAFE			?SAFE					!
;	QSPEED			?SPEED					!
;	QOPTIMIZE		?OPTIMIZE				!
;	QARITHOVERFLOW		?ArithOverflow				!
;	QCHAT			?CHAT					!
;	QREDEF			?REDEFINITIONS				!
;	BYE			BYE					!
;	PBYE			(BYE)					!
;	WINFOS			WINF					!
;	QUOTELOAD		"LOAD					!
;	LOAD			LOAD					!
;	ED			ED					!
;	RND			RND					!
;	MEM			MEM					!
;	INFOS			INFOS					!
;	LIMITS			LIMITS					!
;	PMODE			PMODE					!
;	CTRLBREAK_OFF		CtrlBreak_Off				!
;	CTRLBREAK_REST		CtrlBreak_Restore			!
;	GETTIME			GetTime					!
;	GETDATE			GetDate					!
;	TIMEANDDATE		Time&Date				!
;	GETTICKS		GetTicks				!
;	DELAY			DELAY					!
;	FROTHVERSION		FrothVersion				!
;	LANGUAGE		LANGUAGE				!
;	ENGLISH			ENGLISH					!
;	FRENCH			FRENCH					!
;	FROTHDATE		FrothDate				!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32  	segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions
include		Ed.inc                      		; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = GetArgc							=
; =                                                         	=
; = Usage:							=
; =	Search the string for a given switch letter.		=
; =                                                         	=
; = I) AL  : switch letter					=
; =    EDX : buffer for string.					=
; =                                                         	=
; = O) CF  = 1 -> switch not found				=
; =    CF  = 0 -> switch found, and EDX is the string.		=
; =                                                         	=
; = M) EDX							=
; ===============================================================

		align	4
; Set up ES, EDI, ECX and chek length
Argc_SetAndCheck proc
		mov	bx, _selzero
		mov	es, bx
		mov	edi, _pspa
		add	edi, 81h
		movzx	ecx, byte ptr es:[edi - 1]
		cmp	cl, ah
		jnb	@@Exit
		add	esp, 4
		pop	es
		popad
		stc
@@Exit:		ret
Argc_SetAndCheck endp

; Find switch AL on command line
Argc_FindSwitch	proc
		mov	ah, al
		mov	ebx, edi
		mov	ebp, ecx
@@Find1:	mov	al, '/'
		repnz	scasb
		jecxz	@@Find2
		mov	al, ah
		dec	ecx
		scasb
		jne	@@Find1
		ret
@@Find2:	mov	edi, ebx
		mov	ecx, ebp
@@Find3:	mov	al, '-'
		repnz	scasb
		jecxz	@@Find4
		mov	al, ah
		dec	ecx
		scasb
		jne	@@Find3
		ret
@@Find4:							;  command line too short
		add	esp, 4
		pop	es
		popad
		stc
		ret
Argc_FindSwitch	endp

GetArgc		proc
		pushad
		push	es
		mov	ah, 3					; Minimal size of the cmd line is 2 (-?)
		call	Argc_SetAndCheck
		call	Argc_FindSwitch
;		jecxz	@@NotFound
;		cmp	byte ptr es:[edi], ' '
;		je	@@NotFound
		push	ds
		mov	ax, es
		mov	bx, ds
		mov	es, bx
		mov	ds, ax
		mov	esi, edi
		mov	edi, edx
@@Loop:		lodsb
		stosb
		cmp	al, ' '
		loopnz	@@Loop
		jnz	@@Loop2
		dec	edi
@@Loop2:	clr	al
		stosb
		db	01Fh					; pop ds
		pop	es
		popad
		clc
		ret
@@NotFound:	pop	es
		popad
		stc
		ret
GetArgc		endp


; ===============================================================
; = GetProgName							=
; =                                                         	=
; = Usage:							=
; =	Return Path+Name of current program.			=
; =                                                         	=
; = I) EDX : buffer						=
; =                                                         	=
; = O) EDX : buffer						=
; =                                                         	=
; = M) EDX							=
; ===============================================================

		align	4
GetProgName	proc

		; --------------
		; Save used regs
		; --------------
		
		push	esi ax ecx
		
		
		; ---------------------
		; Get program path+name
		; ---------------------

		mov	esi, [_pspa]
		sub	esi, _code32a
		movzx	esi, word ptr [esi + 02Ch]
		shl	esi, 4
		sub	esi, _code32a
@@xx:		mov	al, [esi]
		cmpz	al, 0
		je	@@yy
		inc	esi
		jmp	@@xx
@@yy:		inc	esi
		mov	al, [esi]
		cmpz	al, 0
		je	@@zz
		inc	esi
		jmp	@@xx
@@zz:	

		clr	ecx
		dec	ecx
@@aa:		inc	ecx
		mov	al, [esi + 3 + ecx]
		mov	[edx + ecx], al
		cmpz	al, 0
		jne	@@aa


		; ----------------
		; Return to caller
		; ----------------

		pop	ecx ax esi		
		ret

GetProgName	endp


; ===============================================================
; = prGreatingMsg						=
; =                                                         	=
; = Usage:							=
; =	Print a greating message at starting.			=
; =                                                     	=
; = I) none							=
; =                                                         	=
; = O) none							=
; =                                                         	=
; = M) none							=
; ===============================================================

		align	4
prGreatingMsg	proc

		; -------------------
		; Print a version msg
		; -------------------

		call	DECIMAL					; decimal		
		call	PDOTQ					; rt."
		db 	13, 10
		db 	"FROTH v"
		db	FROTH_VERSION + '0', '.', (FROTH_RELEASE/10) + '0', (FROTH_RELEASE MOD 10) + '0', '.', FROTH_UNDERREL + 'a', " "
		db 	0

		IPUSH	FROTH_DATE_MONTH-1
		call	STR_MONTH				; $MONTH
		call	_TYPE					; TYPE

		call	PDOTQ					; rt."
		db	'-'
		db	(FROTH_DATE_DAY   / 10) + '0', (FROTH_DATE_DAY   mod 10) + '0', '-'
		db	(FROTH_DATE_YEAR / 1000) + '0', ((FROTH_DATE_YEAR  mod 1000) / 100) + '0'
		db	((FROTH_DATE_YEAR mod 100 ) / 10) + '0', (FROTH_DATE_YEAR  mod 10) + '0'
		db 	"  [OS]", 13, 10
if TURNKEYED eq 1
		db	"   Turnkey edition (no editor, no on-line help, ...)", 13, 10
endif
		db 	0

		; ---------------------------
		; Print amount of free memory
		; ---------------------------

		call	PDOTQ					; rt."
		db 	"Memory free: ", 0
		call	MEMFREE					; MemFree
		call	DOT					; .
		call	PDOTQ					; rt."
		db	" bytes", 0

	ifdef	UNDERDEV
		call	PDOTQ					; rt."
		db	13, 10
		db	"(Lomem free: ", 0
		lea	eax, _lomemsize
		call	eax
		sub	esi, @WordSize
		mov	[esi], eax
		call	DOT					; .
		mov	al, ')'
		call	COUT
	endif
		
			
		call	CR					; CR
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret
@@flag		db	0

prGreatingMsg	endp


; ===============================================================
; = WORDSIZE							=
; =                                                         	=
; = Code for the froth word:	WordSize			=
; =                                                     	=
; = Stack: ( --- n )						=
; ===============================================================

		align	4
WORDSIZE	proc

		cmp	[CompileMode], TRUE
		je	@@Compile


		; ------------
@@Execute:	; Execute mode
		; ------------

		IPUSH	@WordSize		; Push value on stack
		OVSTACK				; Verify stack frame overflow
		ret				; Return to caller


		; ------------
@@Compile:	; Compile mode
		; ------------

		mov	eax, @WordSize
		COMP_NB
		ret				; Return to caller
		
E_WORDSIZE:
WORDSIZE	endp


; ===============================================================
; = QOK	                            				=
; =                                                         	=
; = Code for the forth word:    ?OK                		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
QOK		proc

		; Get system integer var address
		DOVAR	D_QOK

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QOK:
QOK		endp


; ===============================================================
; = QSAFE                            				=
; =                                                         	=
; = Code for the forth word:    ?SAFE                		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
QSAFE		proc

		; Get system integer var address
		DOVAR	D_QSAFE

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QSAFE:
QSAFE		endp


; ===============================================================
; = QSPEED                            				=
; =                                                         	=
; = Code for the forth word:    ?SPEED               		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
QSPEED		proc

		; Get system integer var address
		DOVAR	D_QSPEED

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QSPEED:
QSPEED		endp


; ===============================================================
; = QOPTIMIZE                          				=
; =                                                         	=
; = Code for the forth word:    ?OPTIMIZE              		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
QOPTIMIZE	proc

		; Get system integer var address
		DOVAR	D_QOPTIMIZE

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QOPTIMIZE:
QOPTIMIZE	endp


; ===============================================================
; = QIARITHOVERFLOW                         			=
; =                                                         	=
; = Code for the forth word:    ?IARITHOVERFLOW        		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
QIARITHOVERFLOW	proc

		; Get system integer var address
		DOVAR	D_QIARITHOVERFLOW

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QIARITHOVERFLOW:
QIARITHOVERFLOW	endp


; ===============================================================
; = QCHAT                             				=
; =                                                         	=
; = Code for the forth word:    ?CHAT                		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
QCHAT		proc

		; Get system integer var address
		DOVAR	D_QCHAT

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QCHAT:
QCHAT		endp


; ===============================================================
; = QREDEF                             				=
; =                                                         	=
; = Code for the forth word:    ?redefinitions			=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
QREDEF		proc

		; Get system integer var address
		DOVAR	D_QREDEF

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_QREDEF:
QREDEF		endp


; ===============================================================
; = BYE                             				=
; =                                                         	=
; = Code for the forth word:    BYE             		=
; =                                                         	=
; = Stack:  ( --- )                                         	=
; ===============================================================

		align	4
doBye		proc

		; Load the STARTUP.FTH file ?
		cmp	[ExeFile], TRUE
		je	@@noload

		; Load the file 'SHUTDOWN.FTH'
		mov	[ByeRequest], TRUE			; INTERPRET will end FROTH when LOAD will be finished
		DISPOSE	1
		mov	PARAM0, offset @@FileName
		jmp	QUOTELOAD0				; "LOAD

@@noload:
	
		; Desinstall the keyboard interrupt handler (only if installed)
		call	DesinstallKbIt

		; Return to host OS
		lea	eax, _exit		
		jmp	eax

		align	4
@@FileName	db	SHUTDOWN_FILENAME			; File loaded when FROTH is endded

doBye		endp

		
		align	4
BYE     	proc

		; --------------------
		; Release msec counter
		; --------------------

		push	fs
		mov	fs, [_selzero]
		mov	dword ptr fs:[049Ch], 100000
@@wait:		cmp	dword ptr fs:[049Ch], 0
		jg	@@wait
		pop	fs


		; -------------------------------------------------------------
		; Desinstall the keyboard interrupt handler (only if installed)
		; CR
		; Return to host OS
		; -------------------------------------------------------------

		jmp	doBye

E_BYE:
BYE     	endp


; ===============================================================
; = PBYE                             				=
; =                                                         	=
; = Code for the forth word:    (BYE)             		=
; =                                                         	=
; = Stack:  ( --- )                                         	=
; ===============================================================

		align	4
PBYE     	proc

		; -------------------------------------------------------------
		; Desinstall the keyboard interrupt handler (only if installed)
		; CR
		; Return to host OS
		; -------------------------------------------------------------

		jmp	doBye

E_PBYE:
PBYE     	endp


if TURNKEYED eq 0
; ===============================================================
; = WINFOS                             				=
; =                                                         	=
; = Code for the forth word:    WINFOS              		=
; =                                                         	=
; = Stack:  ( --- )                     			=
; =                                                         	=
; = Input stream: WINFOS name					=
; ===============================================================

		align	4
WINFOS		proc

t_bmsgprinted	equ	byte ptr [ebp - 4]			; Msg printed ?

		enter	4, 0


		; --------------------
		; Go into decimal mode
		; --------------------
		
		call	_BASE					; BASE
		DO_AT						; @
		call	TOR					; >R
		call	DECIMAL					; DECIMAL
		

		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; NIA word in EDX
		call	IsolateWordEx


		; -----------------------
		; Oliver's specific infos
		; -----------------------

		; NIA #
		call	PDOTQ
		db	13, 10
		db	"       NIA:  Nb = ", 0
		mov	eax, edx
		sub	eax, [NIA_START]
		add	eax, word_realaddr
		IPUSH	eax
		call	HDOT					; H.
		push	edx
		call	PDOTQ
		db	" (#", 0
		clr	edx
		sub	eax, word_realaddr
		mov	ebx, size wordentry
		idiv	ebx
		IPUSH	eax
		call	DOT					; .
		call	PDOTQ
		db	")", 0
		pop	edx
		
		; NIA address
		call	PDOTQ
		db	"   Addr = ", 0
		mov	eax, edx
		IPUSH	eax
		call	HDOT					; H.
		
		; Code address (begin)
		call	PDOTQ
		db	13, 10
		db	"      Code:  StartAddr = ", 0
		mov	eax, [edx.word_realaddr]
		IPUSH	eax
		call	HDOT					; H.
		
		; Code address (end)
		call	PDOTQ
		db	"h   EndAddr = ", 0
		mov	eax, [edx.word_EndCode]
		IPUSH	eax
		call	HDOT					; H.
		
		; Code size
		call	PDOTQ					; (.")
		db	"h   Size = ", 0
		movzx	eax, [edx.word_CodeSize]
		IPUSH	eax
		call	DOT					; .

		; Run-time code nb
		mov	eax, [edx.word_runtaddr]
		cmpz	eax, 0
		je	@@rtdone

		call	PDOTQ
		db	13, 10
		db	"  Run-time:  Nb = ", 0
		mov	eax, edx
		sub	eax, [NIA_START]
		add	eax, word_runtaddr
		IPUSH	eax
		call	HDOT					; H.

		call	PDOTQ
		db	"   Addr = ", 0
		mov	eax, [edx.word_runtaddr]
		IPUSH	eax
		call	HDOT					; H.
		
@@rtdone:

		; ---------------
		; Vocabulary name
		; ---------------

				
		call	PDOTQ
		db	13, 10
		db	"       Voc:  ", 0
		movzx	eax, [edx.word_vocptr]
		add	eax, [VOC_START]
		add	eax, voc_name
		sub	esi, @WordSize
		mov	[esi], eax
		call	_TYPE

		
		; --------------------
		; List word attributes
		; --------------------
		
		call	PDOTQ
		db	13, 10
		db	"Attributes:", 0
		
		; System word ?
		test	[edx.word_flags], FSYSTEM
		jz	@@NotSys
		call	PDOTQ
		db	"  System", 0
@@NotSys:
		
		; Immediate word ?
		test	[edx.word_flags], FIMMEDIATE
		jz	@@NotImm
		call	PDOTQ
		db	"  Immediate", 0
@@NotImm:
		
		; Macro word ?
		test	[edx.word_flags], FMACRO
		jz	@@NotMacro
		call	PDOTQ
		db	"  Macro", 0
@@NotMacro:
		
		; Run-time word ?
		test	[edx.word_flags], FRUNTIME
		jz	@@NotRT
		call	PDOTQ
		db	"  Run-time", 0
@@NotRT:
		
		; No compile word ?
		test	[edx.word_flags], FNOCOMPILE
		jz	@@NotNC
		call	PDOTQ
		db	"  Not compilable", 0
@@NotNC:

		; Colon-definition ?
		test	[edx.word_flags], FCOLONDEF
		jz	@@NotCDEF
		PRSTR	"  Colon definition"
@@NotCDEF:

		; Sumdged definition ?
		test	[edx.word_flags], FSMUDGE
		jz	@@NotSmudge
		PRSTR	"  Smudge"
@@NotSmudge:

		; None attribute ?
		cmp	[edx.word_flags], 0
		jne	@@NoNone
		call	PDOTQ
		db	"  None", 0
@@NoNone:


		; ----------------------------------------
		; Is-it a word defined by a definig-word ?
		; ----------------------------------------
		
		mov	eax, [edx.word_DefiningWord]
		cmpz	eax, 0
		je	@@DWdone
		call	PDOTQ					; (.")
		db	13, 10
		db	"Word build by the defining-word: ", 0
		add	eax, [NIA_START]
		IPUSH	eax
		call	_TYPE		
		
@@DWdone:	


		; --------------------------------------
		; Word 'system' ? If so, no more infos !
		; --------------------------------------

		test	[edx.word_flags], FSYSTEM
		jnz	@@Exit

		
		; -----------
		; Nb of users
		; -----------
		
		call	PDOTQ					; (.")
		db	13, 10
		db	"Users nb:  ", 0
		movzx	eax, [edx.word_NbUses]
		IPUSH	eax
		call	DOT					; .
		

if TURNKEYED eq 0
		; --------------------
		; Which words it calls
		; --------------------
		
		push	edx

		mov	[t_bmsgprinted], FALSE			; Msg printed ?

		mov	ecx, edx
		sub	ecx, [NIA_START]			; The caller word
		clr	edx					; The 1st word called
		call	CRefSrch_RD
		
		mov	eax, 3					; Char counter on the line

@@Caller1:	cmp	ebx, [CrossRefNb]
		je	@@Caller2
		mov	edx, ebx
		shl	edx, 2
		add	edx, [CRefNdxRefRD]
		mov	edx, [edx]
		imul	edx, size wrefentry
		add	edx, [CrossRef]
		cmp	[edx.wref_caller], ecx
		jne	@@Caller2

		; Msg printed ?
		cmp	[t_bmsgprinted], TRUE
		je	@@CallerMsgPr
		call	PDOTQ					; (.")
		db	13, 10
		db	"This word calls : "
		db	13, 10, "   "
		db	0
		mov	[t_bmsgprinted], TRUE
@@CallerMsgPr:	

		; Print this word name		
		push	ecx edx
		mov	edx, [edx.wref_called]
		add	edx, [NIA_START]
		add	edx, word_name
		sub	esi, 4
		mov	[esi], edx
@@Caller3:	inc	eax					; Count the 'space' bewteen each word
@@Caller4:	cmp	byte ptr [edx], 0
		je	@@Caller5
		inc	eax
		inc	edx
		jmp	@@Caller4
@@Caller5:	pop	edx
		push	eax
		call	NBCOLS					; #COLS
		pop	eax
		mov	ecx, [esi]
		add	esi, 4
		cmp	eax, ecx
		jl	@@Caller6
		call	PDOTQ					; (.")
		db	13, 10, "   ", 0
		mov	eax, 3					; Reset counter on next line
		jmp	@@Caller3
@@Caller6:	call	_TYPE					; TYPE
		movzx	eax, [edx.wref_calls]
		cmp	eax, 1
		je	@@PrNbDone
		call	PDOTQ					; (.")
		db	':', 0
		sub	esi, @WordSize
		mov	[esi], eax
		call	DOT					; .
@@PrNbDone:	call	SPACE					; SPACE
		pop	ecx
		
		; Next word called, if any
		inc	ebx
		jmp	@@Caller1
		
@@Caller2:	pop	edx
		

		; ---------------------------
		; Which words calls this word
		; ---------------------------

		mov	[t_bmsgprinted], FALSE			; Msg printed ?

		clr	ecx					; The caller word
		sub	edx, [NIA_START]			; The 1st word called
		call	CRefSrch_DR
		mov	ecx, edx

		mov	eax, 3					; Char counter on the line

@@Called1:	cmp	ebx, [CrossRefNb]
		je	@@Called2
		mov	edx, ebx
		shl	edx, 2
		add	edx, [CRefNdxRefDR]
		mov	edx, [edx]
		imul	edx, size wrefentry
		add	edx, [CrossRef]
		cmp	[edx.wref_called], ecx
		jne	@@Called2

		; Msg printed ?
		cmp	[t_bmsgprinted], TRUE
		je	@@CalledMsgPr
		call	PDOTQ					; (.")
		db	13, 10
		db	"This word is called by : "
		db	13, 10, "   "
		db	0
		mov	[t_bmsgprinted], TRUE
@@CalledMsgPr:	
		
		; Print this word name		
		push	ecx edx
		mov	edx, [edx.wref_caller]
		add	edx, [NIA_START]
		add	edx, word_name
		sub	esi, 4
		mov	[esi], edx
@@Called3:	inc	eax					; Count the 'space' bewteen each word
@@Called4:	cmp	byte ptr [edx], 0
		je	@@Called5
		inc	eax
		inc	edx
		jmp	@@Called4
@@Called5:	pop	edx
		push	eax
		call	NBCOLS					; #COLS
		pop	eax
		mov	ecx, [esi]
		add	esi, 4
		cmp	eax, ecx
		jl	@@Called6
		call	PDOTQ					; (.")
		db	13, 10, "   ", 0
		mov	eax, 3					; Reset counter on next line
		jmp	@@Called3
@@Called6:	call	_TYPE					; TYPE
		movzx	eax, [edx.wref_calls]
		cmp	eax, 1
		je	@@PrNbDone2
		call	PDOTQ					; (.")
		db	':', 0
		sub	esi, @WordSize
		mov	[esi], eax
		call	DOT					; .
@@PrNbDone2:	call	SPACE					; SPACE
		pop	ecx
		
		; Next word caller, if any
		inc	ebx
		jmp	@@Called1
		
@@Called2:
endif
		
		
		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		
		; Restore previous base
		call	FROMR					; R>
		call	_BASE					; BASE
		DO_STORE					; !

		leave
		ret

E_WINFOS:
WINFOS      	endp
endif


; ===============================================================
; = QUOTELOAD                           			=
; =                                                         	=
; = Code for the forth word:    "LOAD				=
; =                                                     	=
; = Stack:  ( addrz --- )                 			=
; ===============================================================

		align	4
QUOTELOAD   	proc

		; Verify stack frame
		; ------------------
		
		VSTACK	1
		call	QSTRING				; ?String


QUOTELOAD0:						; Enter for cold start
		; Copy the filename to temp buffer
		; --------------------------------
		
		mov	edx, PARAM0			; Pop filename addr
		DISCARD	1
		lea	ebx, @@FileName			; Local buffer
		clr	ecx				; Char counter
		dec	ecx
@@CopyFileName:	inc	ecx
		mov	al, [edx + ecx]
		mov	[ebx + ecx], al
		cmpz	al, 0
		jne	@@CopyFileName


		; Add .FTH extension if one is not given and default directory (ED_DIR) if one is not given
		; -----------------------------------------------------------------------------------------
		
		mov	edx, ebx			; FileName address
		call	EdAddExtFTH


		; Try to open the file in R/O
		; ---------------------------
		
		lea	ebx, @@FileName			; FileName address
		mov	ax, 10h                 	; 0 001 0000 -> (inherited, deny both, read)
		call	doOpenFile
		jnc	@@OpenOk


		; An error occurs
		; ---------------
		
QuoteLoadErr:	mov	edx, ebx			; Pointer on filename
		clr	ecx				; No line #
		cmp	ax, 2
		jne	@@LL1
		mov	ah, ERR_FILENF			; File not found
		call	doError
@@LL1:      	cmp	ax, 3
		jne	@@LL2
		mov	ah, ERR_PATHNF			; Path not found
		call	doError
@@LL2:      	cmp	ax, 4
		jne	@@LL3
		mov	ah, ERR_TOOMANYOF		; Too many open files
		call	doError
@@LL3:      	mov	ah, ERR_ACCDENIED		; Access denied
		call	doError
		
		
		; No error on opening the file to load
		; ------------------------------------      

@@OpenOk:   	; Save the handle
		mov	eax, [LoadFilesNb]
		imul	ecx, eax, size loadfileentry
		add	ecx, [LoadFilesInfos]
		mov	[ecx.loadfile_hdl], bx
		
		; Save the filename
		lea	ebx, @@FileName			; FileName address
		clr	edx				; Counter
@@SaveFileName:	mov	al, [ebx + edx]
		mov	[ecx.loadfile_name + edx], al
		cmpz	al, 0
		je	@@EndCopyFN
		inc	edx
		jmp	@@SaveFileName
@@EndCopyFN:


		; Print the msg "Compiling..." ont first load (not nesting)
		; ---------------------------------------------------------

		cmp	[GreatingMsgPr], FALSE			; If STARTUP.FTH is not still loaded
		je	@@NotMsg
		cmp	[LoadFilesNb], 0
		jne	@@NotMsg
		cmp	[ByeRequest], TRUE
		je	@@NotMsg
		call	CR
		call	PDOTQ					; rt_."
		db	CMSG_COMPILING
@@NotMsg:


		; Reset the line nb into the file
		; -------------------------------
		
		mov	[ecx.loadfile_linenb], 0
		
		
		; Update the 'depth' of files loading
		; -----------------------------------
		
		inc	[LoadFilesNb]
	

		; Reset infos
		; -----------
		
		mov	[ecx.loadfile_inFile], 0 		; Current position in the file (seek)
		mov	[ecx.loadfile_inBuff], 0		; Current position in the buffer


		; Read the file into buffer
		; -------------------------

		mov	eax, ecx
		mov	bx, [eax.loadfile_hdl]              	; File handle
		mov	ecx, LOADFILEBUFFSZ + IBUFFMAXS     	; Bytes to read
		mov	edx, [eax.loadfile_buff]            	; Buffer address
		call	doReadFile                 	   	; in EBX: bytes read from the file
		jc	QuoteLoadErr


		; Put a '\0' at the end of the buffer
		; -----------------------------------

		mov	ecx, [LoadFilesNb]
		dec	ecx
		imul	edx, ecx, size loadfileentry
		add	edx, [LoadFilesInfos]
		mov	ecx, [edx.loadfile_buff]		; Pointer to buffer
		mov	byte ptr [ecx + ebx], 0             	; '\0'


		; Return to caller
		; ----------------
		
		ret

		align	4
@@FileName	db	MAXSZFILENAME+1 dup (?)			; Local buffer to store modified filename

E_QUOTELOAD:
QUOTELOAD   	endp


if TURNKEYED eq 0
; ===============================================================
; = LOAD    	                       				=
; =                                                         	=
; = Code for the forth word:    LOAD				=
; =                                                     	=
; = Stack:  ( --- )    	             				=
; ===============================================================

		align	4
LOAD   		proc

		; Is the 1st file in FROSTED given ?
		; ----------------------------------

		lea	ebx, EdBuffInfos		
		cmp	byte ptr [ebx.binf_FileName], 0
		jne	@@Ok
		mov	ah, ERR_EDFILENG			; Code # error
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError
@@Ok:


		; Bring this filename to string buffer
		; ------------------------------------
		
		mov	ecx, [QuoteStrBuff]
		add	ecx, size strentry
		clr	edx
@@Loop:		mov	al, [ebx + edx]
		mov	[ecx + edx], al
		cmpz	al, 0
		je	@@LoopEnd
		inc	edx
		jmp	@@Loop
@@LoopEnd:	mov	[ecx.str_currlen - size strentry], edx		; Current length

		; Push string buffer on stack
		IPUSH	ecx


		; Continue with "Load
		; -------------------
		
		call	QUOTELOAD


		; Return to caller
		; ----------------
		
		ret

E_LOAD:
LOAD   		endp


; ===============================================================
; = ED                              				=
; =                                                         	=
; = Code for the forth word:    ED              		=
; =                                                         	=
; = Stack:  ( --- )                     			=
; ===============================================================

		align	4
ED		proc

		; ---------------------------------------------
		; Enter in editor session, then edit and return
		; ---------------------------------------------

		; Disable Ctrl-break flag
		inc	[NoCtrlBreak]
				
		; Enter into editor
		call	Editor
		
		; Restore Ctrl-break save
		dec	[NoCtrlBreak]


		; ---------------
		; LOAD is asked ?
		; ---------------
		
		jnc	@@NoLoad
		call	LOAD					; LOAD
@@NoLoad:


		; ----------------
		; Return to caller
		; ----------------
		
		ret
		
		align	4
@@edx		dd	?
@@eax		dd	?

E_ED:
ED      	endp
endif


; ===============================================================
; = RND                            				=
; =                                                         	=
; = Code for the forth word:    RND                		=
; =                                                         	=
; = Stack:  ( n1 n2 --- n3 )                 			=
; ===============================================================

		align	4
RND		proc

		; Verify stack frame
		; ------------------
		
		VSTACK	2
		
		
		; Generate a random number
		; ------------------------

		; Compute new seed
		mov	eax, [@@Seed]
		mov	ebx, 015A4E35h
		mul	ebx					; EDX:EAX = M * seed
		inc	eax
		mov	[@@Seed], eax
		
		; Compute limit
		mov	ecx, [esi]
		add	esi, 1 * 4				; drop
		mov	ebx, [esi]				; Save mini
		sub	ecx, ebx
		inc	ecx					; Limit

		; Return value
		clr	edx
		shr	eax, 16
		div	ecx					; eax(Q):edx(R) = EDX:EAX / ecx
		add	edx, ebx
		mov	[esi], edx
		

		; Return to caller
		; ----------------
		
		ret

		align	4		
@@Seed		dd	1
		
E_RND:
RND		endp


if TURNKEYED eq 0
; ===============================================================
; = MEM                            				=
; =                                                         	=
; = Code for the forth word:    MEM                		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

PRPOURC		MACRO
		PRSTR	" ("
		imul	eax, 100
		cdq						; EDX:EAX <- EAX
		idiv	ebx					; EDX:EAX <- EDX:EAX / EBX
		IPUSH	eax
		IPUSH	2
		call	DOTR					; 2 .R
		PRSTR	"%)"
		ENDM

		align	4
MEM		proc


		; ----------------
		; Vocabulary infos
		; ----------------
		
		call	PDOTQ					; (.")
		db	13, 10
		db	"            Vocabulary:"
		db	" Area = ", 0
		mov	eax, [VOC_START]
		call	PrEAX
		PRSTR	"- "
		mov	eax, [VOC_START]
		add	eax, VOC_MAXSIZE - 1
		call	PrEAX
		PRSTR	"   Used = "
		mov	eax, [VOC_NB]
		imul	eax, size vocentry
		call	PrEAX

		; %
		mov	ebx, VOC_MAXSIZE
		PRPOURC


		; ----------------------
		; Names Infos Area (NIA)
		; ----------------------

		call	PDOTQ					; (.")
		db	13, 10
		db	"      Names Infos Area:"
		db	" Area = ", 0
		mov	eax, [NIA_START]
		call	PrEAX
		PRSTR	"- "
		mov	eax, [NIA_START]
		add	eax, NIA_WSIZE - 1
		call	PrEAX
		PRSTR	"   Used = "
		mov	eax, [NIA_SIZE]
		call	PrEAX

		; %
		mov	ebx, NIA_WSIZE
		PRPOURC


if TURNKEYED eq 0

		; ----------------------
		; Words Cross References
		; ----------------------

		call	PDOTQ					; (.")
		db	13, 10
		db	"Words cross references:"
		db	" Area = ", 0
		mov	eax, [CrossRef]
		call	PrEAX
		call	PDOTQ					; (.")
		db	"- ", 0
		mov	eax, [CrossRef]
		add	eax, MAXNBWREF * size wrefentry - 1
		call	PrEAX
		call	PDOTQ					; (.")
		db	"   Used = ", 0
		clr	edx
		mov	eax, [CrossRefTotalNb]
		imul	eax, size wrefentry
		call	PrEAX

		; %
		mov	eax, [CrossRefTotalNb]
		mov	ebx, MAXNBWREF
		PRPOURC

endif
		

		; --------------------------
		; Data Directory Area (DDIR)
		; --------------------------

		call	PDOTQ					; (.")
		db	13, 10
		db	"   Data Directory Area:"
		db	" Area = ", 0
		mov	eax, [DDIR_START]
		call	PrEAX
		call	PDOTQ					; (.")
		db	"- ", 0
		mov	eax, [DDIR_START]
		add	eax, [DDIR_SIZE]
		dec	eax
		call	PrEAX
		call	PDOTQ					; (.")
		db	"   Used = ", 0
		mov	eax, [DDIR_SIZE]
		call	PrEAX
	
		; %
		mov	ebx, DDIR_WSIZE
		PRPOURC


		; --------------------
		; Code Dict Area (CDA)
		; --------------------

		call	PDOTQ					; (.")
		db	13, 10
		db	"        Code Dict Area:"
		db	" Area = ", 0
		mov	eax, [CDA_START]
		call	PrEAX
		call	PDOTQ					; (.")
		db	"- ", 0
		mov	eax, [CDA_START]
		add	eax, [CDA_SIZE]
		dec	eax
		call	PrEAX
		call	PDOTQ					; (.")
		db	"   Size = ", 0
		mov	eax, [CDA_SIZE]
		call	PrEAX

		; %
		call	MEMFREE
		mov	eax, [CDA_SIZE]
		mov	ebx, PARAM0
		DISCARD	1
		PRPOURC
		
		PRSTR	"  InitSize = "
		mov	eax, [CDA_SIZE_INIT]
		call	PrEAX
		

		; --------------------
		; Data Dict Area (DDA)
		; --------------------

		call	PDOTQ					; (.")
		db	13, 10
		db	"        Data Dict Area:"
		db	" Area = ", 0
		mov	eax, [DDA_START]
		call	PrEAX
		call	PDOTQ					; (.")
		db	"- ", 0
		mov	eax, [DDA_START]
		add	eax, [DDA_SIZE]
		dec	eax
		call	PrEAX
		call	PDOTQ					; (.")
		db	"   Size = ", 0
		mov	eax, [DDA_SIZE]
		call	PrEAX

		; %
		call	MEMFREE
		mov	eax, [DDA_SIZE]
		mov	ebx, PARAM0
		DISCARD	1
		PRPOURC
		
		PRSTR	"  InitSize = "
		mov	eax, [DDA_SIZE_INIT]
		call	PrEAX
		

		; ----------------
		; Return to caller
		; ----------------

		call	CR					; CR		
		ret
		
E_MEM:
MEM		endp
endif


; ===============================================================
; = INFOS                            				=
; =                                                         	=
; = Code for the forth word:    INFOS                		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; =                                                         	=
; ===============================================================

		align	4
INFOS		proc

		; --------------------
		; Show Code/Data infos
		; --------------------

		; Code
		call	PDOTQ					; rt."
		db	13, 10
		db	"Code size :", 0
		mov	eax, [CDA_SIZE]
		sub	eax, [CDA_SIZE_INIT]
		IPUSH	eax
		IPUSH	9
		call	DOTR					; .R
		
		; Data
		call	PDOTQ
		db	13, 10
		db	"Data size :", 0
		mov	eax, [DDA_SIZE]
		sub	eax, [DDA_SIZE_INIT]
		IPUSH	eax
		IPUSH	9
		call	DOTR					; .R

		; Recoverable
		call	DATASALVAGE
		mov	eax, PARAM0
		DISCARD	1
		cmpz	eax, 0
		je	@@NoRecov
		
		PRSTR	"  (recoverable with DUST:"
		call	DATASALVAGE
		IPUSH	9
		call	DOTR					; .R
		PRSTR	")"

@@NoRecov:	
		

		; ----------------
		; Return to caller
		; ----------------

		ret
		
E_INFOS:
INFOS		endp


if TURNKEYED eq 0
; ===============================================================
; = LIMITS                            				=
; =                                                         	=
; = Code for the forth word:    LIMITS               		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
doLimit		proc

		; Push nb on parameters stack
		DISPOSE	1
		mov	PARAM0, eax

		; Get string address
		mov	ebx, [esp]
		
		; Print the string
		PRSTR	"   "
		clr	ecx
@@Print:		mov	al, [ebx]
		cmpz	al, 0
		je	@@PrintSpaces
		call	COUT
		inc	ebx
		inc	ecx
		jmp	@@Print
		
		; Complement with spaces
@@PrintSpaces:	mov	al, ' '
		call	COUT
		inc	ecx
		cmp	cl, 54
		jl	@@PrintSpaces

		; Update return address (after the string !)
		inc	ebx
		mov	[esp], ebx

		; Print number
		IPUSH	6
		call	DOTR

		; Return to caller
		call	CR
		ret

		endp


		align	4
LIMITS		proc

		; ------------
		; FROTH system
		; ------------

		call	DECIMAL
		call	CR
		call	PDOTQ
		db	"FROTH interpret / compiler:", 13, 10, 0
		
		mov	eax, KB_FIFOSZ
		call	doLimit
		db	"Size of FIFO used by keyboard it handler", 0

		mov	eax, WORDMAXLEN
		call	doLimit
		db	"Max size of a word name", 0

		mov	eax, IBUFFMAXS
		call	doLimit
		db	"Max line length that can be treated by interpret", 0
		
		mov	eax, MAXNB_THREADS
		call	doLimit
		db	"Max nb of threads", 0
		
		mov	eax, (PSTACKSIZE / @WordSize)
		call	doLimit
		db	"Max nb of int on parameters stack", 0
		
		mov	eax, MAXNB_WORDS
		call	doLimit
		db	"Max nb of definitions (included on-line words)", 0

if TURNKEYED eq 0
		mov	eax, MAXNBWREF
		call	doLimit
		db	"Max nb of cross word references", 0
endif
		
		mov	eax, MAXNB_VOCS
		call	doLimit
		db	"Max nb of vocabularies", 0
		
		mov	eax, (DDIR_WSIZE / size ddirentry)
		call	doLimit
		db	"Max nb of global variables", 0
		
		mov	eax, LDA_MAXNB
		call	doLimit
		db	"Max nb of automatic variables", 0
		
		mov	eax, MAXSZFILENAME
		call	doLimit
		db	"Max length for a filename (path + filename)", 0
		
		mov	eax, MAXNBLOADFILES
		call	doLimit
		db	"Max nb of nested load files", 0
		
		mov	eax, CONTROLSTACKSZ
		call	doLimit
		db	"Max nb of nested control structures", 0
		
		mov	eax, LOOPSTACKMAXNB
		call	doLimit
		db	"Max nb of nested DO..ENDDO loops", 0
		
		mov	eax, CASESTACKMAXNB
		call	doLimit
		db	"Max nb of nested CASE..ENDCASE structures", 0
		
		
		; ------
		; SERIAL
		; ------
		
		call	PDOTQ
		db	"Serial:", 13, 10, 0
		
		mov	eax, MAXNB_ITS
		call	doLimit
		db	"Max nb of handled serial IT", 0
		
		mov	eax, MAXNB_DEF_ITPCK
		call	doLimit
		db	"Max nb of definable packets", 0
		
		mov	eax, MAXNB_CHARS_ITPCK
		call	doLimit
		db	"Max nb of chars to define a packet", 0
		

		; -------
		; FROSTED
		; -------
		
		call	PDOTQ
		db	"FROSTED editor:", 13, 10, 0
		
		mov	eax, EDMAXSZLINE
		call	doLimit
		db	"Max line length", 0
		
		mov	eax, EDMAXNBBUFF
		call	doLimit
		db	"Max nb open files", 0
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_LIMITS:
LIMITS		endp
endif


; ===============================================================
; = PMODE                            				=
; =                                                         	=
; = Code for the forth word:    PMODE               		=
; =                                                         	=
; = Stack:  ( --- n )                 				=
; ===============================================================

		align	4
PMODE		proc

		; ---------------------
		; Push value onto stack
		; ---------------------

		movzx	eax, [_sysbyte0]
		sub	esi, @WordSize
		mov	[esi], eax


		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK
		
		; Return to caller
		ret
		
		ret

E_PMODE:		
PMODE		endp


; ===============================================================
; = CTRLBREAK_OFF						=
; =                                                         	=
; = Code for the forth word:  CtrlBreak_Off            		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
CTRLBREAK_OFF	proc

		; -----------------------
		; Disable Ctrl-break flag
		; -----------------------

		inc	[NoCtrlBreak]

				
		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_CTRLBREAK_OFF:
CTRLBREAK_OFF	endp


; ===============================================================
; = CTRLBREAK_REST						=
; =                                                         	=
; = Code for the forth word:  CtrlBreak_Restore            	=
; =                                                         	=
; = Stack:  ( --- )                 				=
; =                                                         	=
; = FLAGS ARE PRESERVED.					==
; ==============================================================

		align	4
CTRLBREAK_REST	proc

		; -------------------------------
		; Ctrl-BREAK always not allowed ?
		; -------------------------------

		pushfd		
		dec	[NoCtrlBreak]
		cmp	[NoCtrlBreak], 0
		jne	@@Exit

		; Running a turnkey .exe ?
		cmp	[ExeFile], TRUE
		je	@@Exit
		
		; Inside FROSTED editor ?
		cmp	[EditorActive], TRUE
		je	@@Exit


		; ---------------------
		; Ctrl-BREAK occurred ?
		; ---------------------

		cmp	[CtrlBreakEvent], FALSE
		je	@@Exit
		popfd
		jmp	CtrlBreakRecov
	

		; ----------------
		; Return to caller
		; ----------------
		
@@Exit:		popfd
		ret

E_CTRLBREAK_REST:
CTRLBREAK_REST	endp


; ===============================================================
; = GETTIME							=
; =                                                         	=
; = Code for the forth word:    GetTime              		=
; =                                                         	=
; = Stack:  ( --- h mn sec msec )              			=
; ===============================================================

		align	4
GETTIME		proc

		; -------
		; Ask DOS
		; -------

		mov	[v86r_ah], 02Ch				; Asked function
		mov 	al, 21h					; Call DOS
		int 	33h
		mov	cx, [v86r_cx]
		mov	dx, [v86r_dx]
		
		
		; ------------------
		; Push data on stack
		; ------------------
		
		sub	esi, 4 * @WordSize
		call	OStack					; Stack overflow ?
		movzx	ebx, ch
		mov	PARAM0, ebx				; Hour
		movzx	ebx, cl
		mov	PARAM1, ebx				; Mn
		movzx	ebx, dh
		mov	PARAM2, ebx				; sec
		movzx	ebx, dl
		mov	PARAM3, ebx				; 1/100 sec
		

		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK
		ret
		
E_GETTIME:
GETTIME		endp


; ===============================================================
; = GETDATE							=
; =                                                         	=
; = Code for the forth word:    GetDate              		=
; =                                                         	=
; = Stack:  ( --- c y m d )                 			=
; ===============================================================

		align	4
GETDATE		proc

		; -------
		; Ask DOS
		; -------

		mov	[v86r_ah], 02Ah			; Asked function
		mov 	al, 21h				; Call DOS
		int 	33h
		mov	ax, [v86r_ax]
		mov	cx, [v86r_cx]
		mov	dx, [v86r_dx]
		
		
		; ------------------
		; Push data on stack
		; ------------------
		
		sub	esi, 4 * @WordSize
		call	OStack				; Stack overflow ?
		movzx	ebx, cx
		mov	PARAM3, ebx			; Year
		movzx	ebx, dh
		mov	PARAM2, ebx			; Month
		movzx	ebx, dl
		mov	PARAM1, ebx			; Day
		movzx	ebx, al
		mov	PARAM0, ebx			; Day of the week
		

		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK
		ret
		
E_GETDATE:
GETDATE		endp


; ===============================================================
; = TIMEANDDATE							=
; =                                                         	=
; = Code for the forth word:    Time&Date              		=
; =                                                         	=
; = Stack:  ( --- h mn sec day mon yr )            		=
; ===============================================================

		align	4
TIMEANDDATE	proc


		; -------
		; Ask DOS
		; -------

		mov	[v86r_ah], 02Ch				; Asked function
		mov 	al, 21h					; Call DOS
		int 	33h
		mov	cx, [v86r_cx]
		mov	dx, [v86r_dx]
		
		
		; ------------------
		; Push data on stack
		; ------------------
		
		DISPOSE	6
		call	OStack					; Stack overflow ?
		movzx	ebx, ch
		mov	PARAM5, ebx				; Hour
		movzx	ebx, cl
		mov	PARAM4, ebx				; Mn
		movzx	ebx, dh
		mov	PARAM3, ebx				; sec
		

		; -------
		; Ask DOS
		; -------

		mov	[v86r_ah], 02Ah			; Asked function
		mov 	al, 21h				; Call DOS
		int 	33h
		mov	ax, [v86r_ax]
		mov	cx, [v86r_cx]
		mov	dx, [v86r_dx]
		
		
		; ------------------
		; Push data on stack
		; ------------------
		
		movzx	ebx, cx
		mov	PARAM0, ebx			; Year
		movzx	ebx, dh
		mov	PARAM1, ebx			; Month
		movzx	ebx, dl
		mov	PARAM2, ebx			; Day
		

		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		OVSTACK
		ret
		
E_TIMEANDDATE:
TIMEANDDATE	endp


; ===============================================================
; = GETTICKS							=
; =                                                         	=
; = Code for the forth word:    GetTicks             		=
; =                                                         	=
; = Stack:  ( --- n ) nmsec	                		=
; ===============================================================

		align	4
GETTICKS	proc

		; ------------------
		; Push data on stack
		; ------------------
		
		sub	esi, @WordSize
		OVSTACK					; Stack overflow ?
		mov	eax, [TicksCnt]
		mov	PARAM0, eax
		

		; --------------------------------------------------
		; Verify stack frame overflow, then return to caller
		; --------------------------------------------------
		
		ret

global	TicksCnt:	dword		; Used in PMODE.ASM, by irq 08h
TicksCnt		dd	0
		
E_GETTICKS:
GETTICKS		endp


; ===============================================================
; = DELAY							=
; =                                                         	=
; = Code for the forth word:    DELAY             		=
; =                                                         	=
; = Stack:  ( n --- )                 				=
; ===============================================================

		align	4
DELAY		proc

t_yield		equ	byte ptr [ebp - 4]			; YIELD done at least one time ?

		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1
		enter	4, 0
		mov	[t_yield], FALSE


		; -------------
		; Wait <n> msec
		; -------------
		
		; Value to wait
		mov	edx, PARAM0
		DISCARD	1
		add	edx, [TicksCnt]
		
		; Delay exired ?
@@again:	cmp	[TicksCnt], edx
		jae	@@exit
		
		; Yield to other threads, then again
		call	YIELD					; Yield to another active thread
		mov	[t_yield], TRUE				; YIELD done at least one time ?
		jmp	@@again


		; ----------------
		; Return to caller
		; ----------------

@@exit:		; To be sure to yield at least one time
		cmp	[t_yield], TRUE
		je	@@exit2
		call	YIELD					; Yield to another active thread
@@exit2:	leave
		ret

E_DELAY:
DELAY		endp


; ===============================================================
; = FROTHVERSION						=
; =                                                         	=
; = Code for the forth word:    FROTHVERSION         		=
; =                                                         	=
; = Stack:  ( --- )                 				=
; ===============================================================

		align	4
FROTHVERSION	proc

		; ------------------
		; Push data on stack
		; ------------------
		
		DISPOSE	3
		call	OStack					; Stack overflow ?
		mov	PARAM0, FROTH_VERSION
		mov	PARAM1, FROTH_RELEASE
		mov	PARAM2, FROTH_UNDERREL


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_FROTHVERSION:
FROTHVERSION	endp


; ===============================================================
; = LANGUAGE							=
; =                                                         	=
; = Code for the forth word:    LANGUAGE        		=
; =                                                         	=
; = Stack:  ( --- addr )                 			=
; ===============================================================

		align	4
LANGUAGE	proc

		; Get system integer var address
		DOVAR	D_LANGUAGE

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret
		
E_LANGUAGE:
LANGUAGE	endp


; ===============================================================
; = ENGLISH							=
; =                                                 		=
; = Froth word:		ENGLISH					=
; =                     	                            	=
; = Stack:		( --- 0 )				=
; ===============================================================

		align	4
ENGLISH		proc

		; Push value on stack
		IPUSH	0

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_ENGLISH:
ENGLISH		endp


; ===============================================================
; = FRENCH							=
; =                                                 		=
; = Froth word:		FRENCH					=
; =                     	                            	=
; = Stack:		( --- 1 )				=
; ===============================================================

		align	4
FRENCH		proc

		; Push value on stack
		IPUSH	1

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_FRENCH:
FRENCH		endp


; ===============================================================
; = FROTHDATE							=
; =                                                         	=
; = Code for the forth word:    FROTHDATE         		=
; =                                                         	=
; = Stack:  ( --- y d m )                 			=
; ===============================================================

		align	4
FROTHDATE	proc


		; ------------------
		; Push data on stack
		; ------------------
		
		DISPOSE	3
		call	OStack					; Stack overflow ?
		mov	PARAM2, FROTH_DATE_YEAR
		mov	PARAM1, FROTH_DATE_DAY
		mov	PARAM0, FROTH_DATE_MONTH


		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_FROTHDATE:
FROTHDATE	endp


ifdef NEVER
global	UpdateCodeSel:	proc

; ===============================================================
; = UpdateCodeSel						=
; =                                                         	=
; = Usage							=
; =    Update the descriptor of code segment.			=
; =                                                         	=
; = I) none							=
; = O) none							=
; = I) all regs							=
; ===============================================================

XX	macro
B_SHOW	'1'
mov	ax, [edx]
SHOW	ax
mov	ax, [edx + 2]
SHOW	ax
mov	ax, [edx + 4]
SHOW	ax
mov	ax, [edx + 6]
SHOW	ax
E_SHOW
	endm

		align	4
UpdateCodeSel	proc

		; -----------------------------
		; Pointer to segment descriptor
		; -----------------------------

		movzx	eax, [_selcode]
		lea	edx, [gdt32 + eax]
XX
		
		; Base address of the segment
		mov	eax, [CDA_START]			; Start address of code
		mov	[edx + 2], ax				; Bits 0-15
		shr	eax, 16
		mov	[edx + 4], al				; Bits 16-23
		shr	ax, 8
		mov	[edx + 7], al				; Bits 24-31
XX

		; Length of the segment
		mov	eax, [CDA_SIZE]
		mov	[edx], ax				; Bits 0-15
		shr	eax, 16
		and	al, 0Fh
		mov	bl, [edx + 6]
		and	bl, 01110000b				; b7=bytes (not 4Kb)
		or	bl, al
		mov	[edx + 6], bl				; Bits 16-19
XX
	

		; ----------------
		; Return to caller
		; ----------------

		ret

UpdateCodeSel	endp
endif


; ===============================================================
; = EdNumber							=
; =								=
; = Usage:							=
; = 	Convert a stringz into a decimal number.		=
; =								=
; = I)	EDX : pointer to stringz				=
; =								=
; = O)	EAX = the number					=
;        CF  = 1 if error, CF = 0 if conversion OK		=
; =								=
; = M)	All registers						=
; ===============================================================

		align	4
EdNumber	proc

		; Pointer to the end of the string
		mov	ecx, edx
@@SearchEnd:	inc	ecx
		cmp	byte ptr [ecx], 0
		jne	@@SearchEnd
		dec	ecx

		; Conversion
		mov	esi, edx
		clr	edi
		clr	ebx					; Digit counter
@@Convert:	movzx	eax, byte ptr [ecx]
		cmp	al, '0'
		jb	@@Error
		cmp	al, '9'
		ja	@@Error
		sub	al, '0'
		imul	eax, dword ptr [@@Powers + ebx * 4]
		add	edi, eax
		inc	bx
		cmp	bx, 6
		je	@@Error
		dec	ecx
		cmp	ecx, esi
		jae	@@Convert
		mov	eax, edi

		; No error
		clc						; CF = 0
		jmp	@@Exit

		; Error
@@Error:	stc						; CF = 1

		; Return to caller
@@Exit:		ret

		align	4
@@Powers:	dd	1, 10, 100, 1000, 10000, 100000, 1000000

EdNumber	endp


; ===============================================================
; = EdPDotQ			                           	=
; =								=
; = Run time function for the forth word ."			=
; =								=
; = Regs modified:  none                  			=
; ===============================================================

		align	4
EdPDotQ		proc

		; Save used registers
		push	ax ecx

		; Store into DI the string address
		mov	ecx, [esp + 6]

@@LL1:      	mov	al, [ecx]
		inc	ecx
		cmpz	al, 0
		je	@@LL2
		call	COUT
		jmp	@@LL1
@@LL2:

		; Move the IP after the string
		mov	[esp + 6], ecx

		; Return to caller
		pop	ecx ax
		ret

EdPDotQ       	endp


; ===============================================================
; = EdAddExtFTH							=
; =								=
; = Usage:							=
; =	If no extension is given, add the standard .FTH		=
; =	extension.						=
; =								=
; = I)	EDX : filename address (zstring)			=
; =								=
; = O)	none							=
; =								=
; = M)	none							=
; ===============================================================

		align	4
EdAddExtFTH	proc

		; Save registers used
		push	eax ebx ecx edx esi
		
		; String empty ?
		cmp	byte ptr [edx], 0
		je	@@Exit
		
		; Found last char address
		mov	ecx, edx
		clr	bl					; String length
@@FLCA:		inc	ecx
		inc	bl
		cmp	byte ptr [ecx], 0
		jne	@@FLCA
		dec	ecx
		dec	bl
		
		; Extension ? Test:  XX.
		cmp	byte ptr [ecx], '.'
		je	@@AddDir
		
		; Extension ? Test:  XX.Y
		cmp	bl, 2 
		jl	@@AddExt
		cmp	byte ptr [ecx - 1], '.'
		je	@@AddDir
		
		; Extension ? Test:  XX.YY
		cmp	bl, 3
		jl	@@AddExt
		cmp	byte ptr [ecx - 2], '.'
		je	@@AddDir
		
		; Extension ? Test:  XX.YYY
		cmp	bl, 4
		jl	@@AddExt
		cmp	byte ptr [ecx - 3], '.'
		je	@@AddDir
		
		; Add extension .FTH
@@AddExt:	mov	dword ptr [ecx + 1], 'HTF.'
		mov	byte ptr [ecx + 5], 0

		; Do we add the default directory, if one is not given ?
@@AddDir:	cmp	byte ptr [edx], '\'
		je	@@Exit
		cmp	byte ptr [edx + 1], ':'
		je	@@Exit
		
		; Pointer on default directory to add
		mov	esi, D_ED_DIR
		add	esi, [DDIR_START]
		mov	esi, [esi.ddir_blkaddr]
		add	esi, size strentry
		
		; Total length will be ok ?
		clr	ebx
		dec	ebx
@@Len1:		inc	ebx
		cmp	byte ptr [edx + ebx], 0
		jne	@@Len1
		clr	ecx
		dec	ecx
@@Len2:		inc	ecx
		cmp	byte ptr [esi + ecx], 0
		jne	@@Len2
		mov	eax, ebx
		add	eax, ecx
		cmp	eax, MAXSZFILENAME - 1
		jge	@@Exit
		
		; Make a hole
		push	esi edi ecx
		mov	esi, edx					; Source address
		mov	edi, edx
		add	edi, ecx					; Destination address
		mov	ecx, ebx
		inc	ecx						; Nb of bytes to move
		cld                     				; Inc SI and DI
		rep	movsb  		            			; Move the block
		pop	ecx edi esi
		
		; Then copy the default directory
		push	esi edi ecx
								; ESI = destination address
		mov	edi, edx				; EDI = destination address
								; ECX = nb of bytes to move
		cld                     			; Inc SI and DI
		rep	movsb  		            		; Move the block
		pop	ecx edi esi
		
		; Return to caller
@@Exit:		pop	esi edx ecx ebx eax
		ret

EdAddExtFTH	endp


code32		ends
		end
